### ===========================================================================
### -package  : SimpleBase
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library base package.
# -module     : implementation
# -overview   :
#
#     This package contains the options applicable to The Simple Development
#  Library as a whole as well as a few procedures for lists handling.
#
# -usage      :
#
#     The K<::Simple::List::enumerate> procedure enumerates the elements of a
#  list while K<::Simple::List::globs> returns the elements in a list which
#  match at least one (or do not match any) pattern among those in another
#  list.
#
# -api        :
#
#  * K<::Simple::configure> ?-storemetadata extbool?
#    Configures The Simple Development Library options.
#
#  * K<::Simple::cget> ?-storemetadata?
#    Gets The Simple Development Library package options.
#
#  * K<::Simple::List::enumerate> list ?separator?
#    Enumerates the elements of a list.
#
#  * K<::Simple::List::globs> ?-negate? candidates patterns
#    Performs globbing based on candidates and pattern lists.
#
# -variable   :
#
#  {
#     ::Simple::BaseCommandTag
#
#  } -description {
#
#     The base command tag.  A procedure is considered a base
#     command procedure if its body contains this tring
#
#  } -type string -value %BASE-COMMAND%
#
# -option     :
#
#  {
#     -storemetadata
#
#  } -description {
#
#        Whether non-essential details (''metadata'') of most items declared
#     with The Simple Development Library (errors, packages, types,
#     procedures, classes, ...) are stored or not. 
#
#        While not storing the metadata saves some memory, the information
#     returned by the various K<information> commands may be incomplete. 
#     Notice that only those items created ''after'' modyfing this option are
#     affected by the change.  Notice also that items within packages declared
#     via K<declare-package> are only created upon installing the package.
#
#  } -type boolean -value 0
#
# -history    :
#
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

# Require the proper Tcl version
package require Tcl 8.1

### Create the package namespaces
namespace eval ::Simple {}
namespace eval ::Simple::Priv {}
namespace eval ::Simple::List {}

### Declare the package
#   Utility procedure to declare the package
proc ::Simple::Priv::declare-package {} {
   ::Simple::Package::declare SimpleBase 1.0 -requisites {

      {Tcl 8.1}

   } -namespaces {

      ::Simple
      ::Simple::Priv
      ::Simple::List
   }

   # Delete this procedure
   rename [info level 0] {}
}

### Create the package variables
set ::Simple::BaseCommandTag %BASE-COMMAND%

### Create the package options

# A<-storemetadata>
set ::Simple::%Options%(-storemetadata) 0

### Declare the package types

### Declare the package errors
#   Utility procedure to declare the package errors
proc ::Simple::Priv::declare-errors {} {

   ::Simple::Error::declare {
      ::Simple::NO-OPTIONS 
   } -message {
      package with no options
   } -explanation {
      This package K<configure> command is not functional because the package
      has no options
   }

   ::Simple::Error::declare {
      ::Simple::NON-EXISTING 
   } -message {
      non-existing %s "%s"
   } -explanation {
      The <item> named <name> does not exist
   }

   ::Simple::Error::declare {
      ::Simple::NO-SUBCOMMAND 
   } -message {
      wrong # args: should be "%s option ?arg ...?"
   } -explanation {
      The base command <command> has been invoked with no subcommand
   } -corrective {
      Invoke the base command giving a subcommand
   }

   ::Simple::Error::declare {
      ::Simple::BAD-SUBCOMMAND 
   } -message {
      bad option "%s": must be %s
   } -explanation {
      Subcommand <subcommand> is not in the <list of allowed subcommands> for
      the base command
   } -corrective {
      Invoke the base command giving a subcommand among those in the list
   }

   ::Simple::Error::declare {
      ::Simple::BAD-SUBCOMMAND-ARGS 
   } -message {
      wrong # args: should be "%s %s"
   } -explanation {
      Subcommand procedure <subcommand procedure> has been invoked with a
      bad <argument list>
   } -corrective {
      Invoke the subcommand procedure matching the argument list
   }

   ::Simple::Error::declare {
      ::Simple::BAD-SUBCOMMAND-ARGS-EMPTY
   } -message {
      wrong # args: should be "%s"
   } -explanation {
      Subcommand procedure <subcommand procedure> has been invoked with a
      bad <argument list>
   } -corrective {
      Invoke the subcommand procedure without arguments
   }

   ::Simple::Error::declare {
      ::Simple::FLAG-REQUIRED 
   } -message {
      flag %s required
   } -explanation {
      The flag or flags given in the <flag or flags combination> are required
   }

   ::Simple::Error::declare {
      ::Simple::AT-LEAST-ONE-FLAG 
   } -message {
      at least one flag required, flags are %s
   } -explanation {
      No flag was supplied
   } -corrective {
      Supply a flag from the <list of flags>
   } -translation {<list of flags>}

   ::Simple::Error::declare {
      ::Simple::BAD-FLAG 
   } -message {
      bad flag "%s": must be %s
   } -explanation {
      Flag <flag> is not in the <list of flags>
   } -corrective {
      Supply a flag among those in the list
   }

   ::Simple::Error::declare {
      ::Simple::DUPLICATED-FLAG-OR-OPTION 
   } -message {
      duplicated %s %s
   } -explanation {
      The <flag or option> <flag or option> has been given twice
   }

   ::Simple::Error::declare {
      ::Simple::BAD-VALUE 
   } -message {
      invalid value "%s" for %s "%s" of type "%s"%s
   } -explanation {
      A <variable, array element or argument> value did not conform to its type
   } -corrective {
      Either modify the <variable, array element or argument> value or type
   } -explained {
      invalid value "<value>" for <variable, array element or argument>
      "<variable, array element or argument name>" of type "<type>"[:
      <further explanation>]
   }

# fixme: tcl 8.4a2 gives a single error when not suplying enough arguments to
#        a command or when suplying more arguments that expected 
#           wrong # args: should be "foo arg1 arg2 ..." 
#        instead of the various different errors than previous tcl versions gave
   ::Simple::Error::declare {
      ::Simple::TOO-MANY-ARGS 
   } -message {
      called "%s" with too many arguments
   } -explanation {
      A call to the command <command> failed because
      too many arguments were supplied
   } -corrective {
      Use the correct list of arguments
   }

   ::Simple::Error::declare {
      ::Simple::TOO-MANY-ARGS-NO-CALL
   } -message {
      too many arguments
   } -explanation {
      Too many arguments were supplied
   } -corrective {
      Use the correct list of arguments
   }

   ::Simple::Error::declare {
      ::Simple::NO-VALUE-FOR-PARAM 
   } -message {
      no value given for parameter "%s" to "%s"
   } -explanation {
      A call to the procedure <procedure> failed because no value was supplied
      for argument <argument>
   } -corrective {
      Use the correct list of arguments
   } -translation {
      <argument> <procedure>
   }

   ::Simple::Error::declare {
      ::Simple::BAD-ITEM 
   } -message {
      expected %s but got "%s"
   } -explanation {
      <bad item> was found when an <item> was expected
   } -corrective {
      Provide a valid <item>
   } -translation {<item> <bad item>}

   # Delete this procedure
   rename [info level 0] {}
}

### ===========================================================================
### -command  : ::Simple::Priv::is-not-internal
# -synopsis   : Returns whether an item is internal to The Simple Development Library
# -returns    : 
#
#  Whether the given item is internal to The Simple Development Library
#
# -overview   :
#
#     This procedure returns whether the given item is internal to The Simple
#  Development Library.  These items contain a percent sign in its name.  This
#  occurs, for example, with "after flags" auxiliary commands from the
#  P<SimpleProc> package
#
# -arguments  :
#  { item             unqualifiedname
#                                    {Item name}}
#
proc ::Simple::Priv::is-not-internal {
   item
} {
   return [expr {[string first % $item] == -1}]
}

### ===========================================================================
### -command  : ::Simple::Priv::qualify-split
# -synopsis   : Splits a name.
# -overview   :
#
#     This procedure splits a name into its namespace qualifiers and
#  unqualified name components and returns the pair.  If the name is
#  unqualified, it is first qualified up the calling stack. By default,
#  two levels up the stack are used.
#
# -arguments  :
#  { name             variableorcommandname 
#                                    {Variable or procedure name}}
#  {?levels?          integer 2      {Levels up the stack in which to operate}}
#
# -returns    : The qualified name, namespace and unqualified name.
#
proc ::Simple::Priv::qualify-split {
   name
   {levels 2}
} {
   ### Qualify the name a two levels up the stack
   set name [::Simple::Priv::qualify $name $levels]

   ### Split the name
   regexp -- {^(.*::)(.*)$} $name -> namespace unqualifiedName

   ### Return the qualified name, namespace and unqualified name
   return [list $namespace$unqualifiedName $namespace $unqualifiedName]
}

### ===========================================================================
### -command  : ::Simple::Priv::qualify
# -synopsis   : Qualifies a name up the calling stack.
# -overview   :
#
#     This procedure qualifies a variable or procedure name.  It is somewhat
#  similar to the K<auto_qualify> procedure but the namespace is taken from a
#  scope up the calling stack.  By default, one level up the stack is used.
#
# -arguments  :
#  { name             name           {Variable or procedure name}}
#  {?levels?          integer 1      {Levels up the stack in which to operate}}
#
# -returns    : The qualified name.
#
# -limitations:
#
#  * This procedure must be called from another procedure.  If the A<levels>
#    optional argument is specified, the corresponding number of levels up the
#    stack are necessary.
#
# -references :
#
#  * The K<auto_qualify> procedure.
#
#  * The K<::Simple::Namespace::qualify> procedure.
#
proc ::Simple::Priv::qualify {
   name
   {levels 1}
} {
   ### Unqualified name
   if {![string match ::* $name]} {

      ### Take the namespace from a certain level up the stack
      #   This throws a "bad level <level>" error is this procedure is not
      #   called from within another one.
      set namespace [uplevel [incr levels] namespace current]

      ### Namespace is not "::"
      if {[string compare $namespace ::]} {

         ### namespace::name
         set answer ${namespace}::$name

      ### Namespace is "::"
      } else {

         ### ::name
         set answer ::$name
      }

   ### Qualified name
   } else {
      set answer $name
   }

   ### Return the qualified name
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Priv::parse-flags
# -synopsis   : Parses an argument list against a flag list.
# -overview   :
#
#     This procedure parses an argument list against a flag list.  Each flag
#  must carry a parameter which is placed in a variable named after the flag
#  in the calling scope.  If the flag is in the single flags list, it can not
#  appear twice in the argument list:  an error is thrown.  Otherwise, if it
#  is in the multiple flags list, it can appear twice or more times in the
#  argument list:  in that case the variable named after the flag in the
#  calling scope will be set to a list whose elements are the different
#  arguments for its different occurrences.
#
#     The flag arguments (both single and multiple) can be given a form as per
#  K<::Simple::Priv::transform-flag-argument>.
#
#     An error is thrown if there are more than the specified maximum number
#  of arguments after the flags (provided that number is not negative) or a
#  given flag is not in the list.
#
# -arguments  :
#
#  { singleFlagsAndForms      
#                     word-pairs     {Flags/form pairs list}}
#  { multipleFlagsAndForms    
#                     word-pairs     {Multiple flags/form pairs list}}
#  { actualArgs       list           {Actual argument list}}
#  { maxAfterFlagArgs integer        {Maximum number of arguments after the
#                                     flags}}
#  { levels           integer        {Levels up the stack used in error
#                                     reporting; use 0 for not referring to
#                                     the called command}}
#
# -returns    : 
#
#  The list of flags given in the actual argument list.
#
# -effects    :
#
#  * Creates variables named after the flags in the calling scope.
#
#  * The list of flags given in the actual argument list is stored in a
#    variable named V<%GIVEN-FLAGS%> in the calling scope.
#
# -references :
#
#  * The K<::Simple::Proc::Priv::parse-arguments> procedure.
#
#  * The K<::Simple::Priv::transform-flag-argument> procedure.
#
#  * The K<::Simple::Proc::Priv::parse-arguments-checktype> procedure.
#
proc ::Simple::Priv::parse-flags {
   singleFlagsAndForms
   multipleFlagsAndForms
   actualArgs
   maxAfterFlagArgs
   levels
} {
   # All and multiple flags
   set flagsAndForms [concat $singleFlagsAndForms $multipleFlagsAndForms]
   array set flagsArray $flagsAndForms
   array set multipleFlagsArray $multipleFlagsAndForms

   # Foreach flag, create a local variable linked to a variable in the calling
   # scope named after the flag with no leading hyphens.  If the variable does
   # not exist in the calling scope, initialize its value to the empty string
   set formIndex 0
   foreach {flag form} $flagsAndForms {
      set forms($flag) $form
      set variable [string trimleft $flag -]
      upvar $variable ${variable}Local
      if {![info exists ${variable}Local]} {
         set ${variable}Local {}
      }
   }

   ### Loop over arguments
   set argumentIndex 0
   set actualArgsLength [llength $actualArgs]
   while {$argumentIndex < $actualArgsLength} {
      set argument [lindex $actualArgs $argumentIndex]

      ### Argument is flag (starts by an hyphen)
      if {[string match -* $argument]} {
         set variable [string trimleft $argument -]

         ### Bad flag
         if {![info exists flagsArray($argument)]} {

            ### Throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $argument\
               [::Simple::List::enumerate [lsort [array names flagsArray]]]

         ### Good flag
         } else {

            ### It is a multiple flag
            if {[info exists multipleFlagsArray($argument)]} {

               ### Flag previously given
               if {[info exists givenFlags($argument)]} {

                  ### Append the flag parameter to the variable named after the
                  ###    flag, after giving it the appropriate form
                  lappend ${variable}Local\
                     [::Simple::Priv::transform-flag-argument\
                     [lindex $actualArgs [incr argumentIndex]]\
                     $forms($argument)]

               ### Flag not previously given
               } else {

                  ### Set the value of the variable named after the flag to the
                  ###    flag parameter, after giving it the appropriate form
                  set ${variable}Local\
                     [list [::Simple::Priv::transform-flag-argument\
                     [lindex $actualArgs [incr argumentIndex]]\
                     $forms($argument)]]
               }

            ### It is a single flag
            } else {

               ### Assert the flag has not been previously given
               if {[info exists givenFlags($argument)]} {
                  ::Simple::Error::throw\
                     ::Simple::DUPLICATED-FLAG-OR-OPTION flag \"$argument\"
               }

               ### Set the value of the variable named after the flag to the
               ###    flag parameter, after giving it the appropriate form
               incr argumentIndex
               incr argumentIndex -1
               set ${variable}Local [::Simple::Priv::transform-flag-argument\
                  [lindex $actualArgs [incr argumentIndex]]\
                  $forms($argument)]
            }

            ### Flag given
            set givenFlags($argument) {}
         }

      ### Argument is not flag
      } else {

         ### More after flag arguments than expected
         if {$maxAfterFlagArgs >= 0 &&\
            $actualArgsLength - $argumentIndex > $maxAfterFlagArgs} {

            ### Throw error
            if {$levels > 0} {
               ::Simple::Error::throw\
                  ::Simple::TOO-MANY-ARGS [lindex [info level -$levels] 0]
            } else {
               ::Simple::Error::throw\
                  ::Simple::TOO-MANY-ARGS-NO-CALL
            }

         ### No more after flag arguments than expected
         } else {

            ### Exit loop
            break
         }
      }

      # Next argument
      incr argumentIndex
   }

   # Set the list of flags given in the actual argument list in a
   # variable named V<%GIVEN-FLAGS%> in the calling scope
   set givenFlagsNames [array names givenFlags]
   uplevel 1 [list ::set %GIVEN-FLAGS% $givenFlagsNames]

   ### Return the list of flags given in the actual argument list
   return $givenFlagsNames
}

### ===========================================================================
### -command  : ::Simple::Priv::transform-flag-argument 
# -synopsis   : Transforms a flag parameter.
# -overview   :
#
#     This procedure takes a flag parameter and transforms it to a given form. 
#  The allowed forms are:
#
#  * untouched: 
#    the parameter is not modified.
#
#  * trimmed: 
#    the parameter is trimmed.
#
#  * list: 
#    the parameter is "listified".
#
#  * line: 
#    the parameter is converted into a single line.
#
#  * lines: 
#    the parameter is converted into a block of lines by deleting the blank
#    lines at the beginning and at the end of the block.
#
#  * paragraphs:  
#    the parameter is converted into sequence of paragraphs.
#
# -arguments  :
#  { parameter        any            {Flag parameter}}
#  { form             choice {untouched trimmed list line paragraphs}
#                                    {Flag parameter}}
#
# -returns    : The flag parameter with the appropriate form.
#
proc ::Simple::Priv::transform-flag-argument {
   parameter
   form
} {
   ### Switch on the form
   switch -exact $form {

      untouched {
      ### untouched

         # Unmodified parameter
         set answer $parameter
      }

      trimmed {
      ### trimmed

         # Trimmed parameter
         set answer [string trim $parameter]
      }

      list {
      ### list

         # Listified parameter
         set answer [list]
         foreach element $parameter {
            lappend answer $element
         }
      }

      line {
      ### line

         # Parameter converted to a single line
         set answer [::Simple::Priv::convert-to-line $parameter]
      }

      lines {
      ### lines

         # Transform an arbitrary string into a block of lines by deleting the
         # blank lines at the beginning and at the end of the block
         set answer [string trimright [string trim $parameter \n]] 
      }

      paragraphs {
      ### paragraphs

         # Transform an arbitrary string into a sequence of paragraphs. 
         # Paragraphs are separated by blank lines, and each paragraph is
         # converted into a single line as described above.
         regsub -all "\n\[ \t\]*" [string trim $parameter] \n answer
         regsub -all "(\[^\n\])\n(\[^\n\])" $answer {\1 \2} answer
         regsub -all \n\n+ $answer \n\n answer
      }

      default {
      ### default, throw error
         ::Simple::Error::throw ::Simple::BAD-ITEM {flag parameter form} $form
      }
   }

   # Return the flag parameter with the appropriate form
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Priv::convert-to-line
# -synopsis   : Returns whether a flag was given in an actual argument list.
# -overview   :
# 
#     This procedure transforms an arbitrary string into a single line by
#  trimming it and transforming all carriage returns surrounded by any number
#  of blanks into a single space
#
# -arguments  :
#  { string           string         {The string}}
#
# -returns    : The string converted to a single line.
#
proc ::Simple::Priv::convert-to-line {
   string
} {
   regsub -all "\[ \t\n\]*\n\[ \t\n\]*" [string trim $string] { } answer
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Priv::flag-given
# -synopsis   : Returns whether a flag was given in an actual argument list.
# -arguments  :
#  { flag             flag           {Flag}}
#
# -returns    : Whether the flag was given in the actual argument list.
#
# -references :
#
#  * The K<::Simple::Argument::information flaggiven> procedure.
#
#  * The K<::Simple::Option::information given> procedure.
#
proc ::Simple::Priv::flag-given {
   flag
} {
   ### Return whether the flag was given in the actual argument list.
   upvar %GIVEN-FLAGS% flags
   return [expr {[lsearch -exact $flags $flag] != -1}]
}

### ===========================================================================
### -command  : ::Simple::List::enumerate
# -synopsis   : Enumerates the elements of a list.
# -overview   :
#
#     This procedure returns a string built by separating the input list
#  elements with commas except the last one which is preceded by the separator
#  argument which defaults to "or".  If the number of elements in the list is
#  lower than two the list is returned unmodified.
#
# -arguments  :
#  { list             list           {List}}
#  {?separator?       string or      {Modifier for last item}}
#
# -returns    : The enumerated list as a string.
#
proc ::Simple::List::enumerate {
   list
   {separator or}
} {
   set nElements [llength $list]

   # More than one element
   if {$nElements > 1} {
      set answer [concat [join [lrange $list 0 [expr {$nElements - 2}]] {, }]\
         $separator [lindex $list end]]

   # One or zero elements
   } else {
      # Bug 7: K<::Simple::List::enumerate> returns braced string
      #        with one-element list containing spaces
      # Solution: handle this situation within this branch
      set answer [lindex $list 0]
   }

   # Return the enumerated list as a string
   return $answer
}

### ===========================================================================
### -command  : ::Simple::List::globs
# -synopsis   : Performs globbing based on candidates and pattern lists.
# -overview   :
#
#     This procedure returns the elements in a list of candidates matching at
#  least one pattern among those in another list, or those which do not match
#  any pattern if the A<-negate> flag is given.
#
# -arguments  :
#  {-negate           boolflag       {Whether to negate the answer}}
#  { candidates       any-list       {List of candidates}}
#  { patterns         pattern-list   {List of patterns}}
#
# -returns    : The list of matching (or not matching) candidates.
#
proc ::Simple::List::globs {
   arg1
   arg2
   {arg3 %DEFAULT%}
} {
   ### Parse arguments
   if {[string compare $arg1 -negate]} {
      set negate 0
      set candidates $arg1
      set patterns $arg2
   } else {
      set negate 1
      set candidates $arg2
      if {![string compare $arg3 %DEFAULT%]} {
         ::Simple::Error::throw ::Simple::NO-VALUE-FOR-PARAM\
            patterns [lindex [info level 0] 0]
      }
      set patterns $arg3
   }

   ### Negate
   if {$negate} {

      ### Loop over patterns
      set answer [list]
      foreach pattern $patterns {

         ### There is no candidate matching the pattern
         set index [lsearch -glob $candidates $pattern]
         if {$index == -1} {

            ### Short-circuit, return all candidates
            return $candidates

         ### There is at least one candidate matching the pattern
         } else {
            while {$index != -1} {

               ### Delete the matching candidate from the list and
               ###    repeat the process with the same pattern
               set candidates [lreplace $candidates $index $index]
               set index [lsearch -glob $candidates $pattern]
            }
         }
      }
      set answer $candidates

   ### Do not negate
   } else {

      ### Loop over patterns
      set answer [list]
      foreach pattern $patterns {

         ### There is at least one candidate matching the pattern
         set index [lsearch -glob $candidates $pattern]
         while {$index != -1} {

            ### Add the matching candidate to the list of matches
            ###    and repeat the process with the same pattern
            lappend answer [lindex $candidates $index]
            set candidates [lreplace $candidates $index $index]
            set index [lsearch -glob $candidates $pattern]

            # Short-circuit upon no more candidates
            if {[llength $candidates] == 0} {
               return $answer
            }
         }
      }
   }

   ### Return the list of matching (or not matching) candidates
   return $answer
}

### ===========================================================================
### -command  : ::Simple::configure
# -synopsis   : Configures The Simple Development Library options.
#
# -parseflags : false
# -arguments  :
#  {-storemetadata    extbool false  {Whether non-essential details
#                                     (''metadata'') of most items declared
#                                     with The Simple Development Library
#                                     (errors, packages, types, procedures,
#                                     classes, ...) are stored or not}}
#
# -returns    : The empty string.
#
proc ::Simple::configure {
   args
} {
# fixme: an skeleton for the K<configure> and K<cget> procedures should be
#        factored and then used by all packages, at least after P<SimpleType>. 
#        For packages with two or more options, it should be possible to
#        K<configure> them all in one call, but K<cget> shall only accept a
#        single flag.  Further, the creation of the K<configure> and K<cget>
#        procedures should be automaticly done by P<SimpleDeclare>.
   ### No arguments
   if {[llength $args] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::AT-LEAST-ONE-FLAG\
         [::Simple::cget]

   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      set arg2 [lindex $args 1]
      switch -exact -- $arg1 {

         -storemetadata {
         ### -storemetadata given

            ### Invalid argument
            if {[catch {if {$arg2} {}}]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -storemetadata extbool {}
            }

            ### Set the store details package option
            if {$arg2} {
               set ::Simple::%Options%(-storemetadata) 1
            } else {
               set ::Simple::%Options%(-storemetadata) 0
            }
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1 [::Simple::cget]
         }
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::cget
# -synopsis   : Gets The Simple Development Library package options.
# -arguments  :
#  {-storemetadata    boolflag       {Gets whether non-essential details
#                                     (''metadata'') of most items declared
#                                     with The Simple Development Library
#                                     (errors, packages, types, procedures,
#                                     classes, ...) are stored or not}}
#
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::cget {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Return the list of options
      return -storemetadata

   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      switch -exact -- $arg1 {

         -storemetadata {
         ### -storemetadata given
            set ::Simple::%Options%(-storemetadata)
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1 [::Simple::cget]
         }
      }
   }
}

### Install the package
package provide SimpleBase 1.0
