### ===========================================================================
### -package  : SimpleClass
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleClass

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library [incr Tcl] classes declaration.

} -module {

   implementation

} -paradigm {

   procedural

} -overview {

      This package provides a way to declare [incr Tcl] classes and [incr Tk]
   megawidgets in an homogeneous, easily parseable, fully introspectable way. 
   It allows to substitute the comment block which is usually placed on top of
   each declaration by Tcl-handled text. 

      Classes are declared with a single command (K<declare-class>).  Every
   single component needed to either make the class work (such as its
   attributes or methods) or to document it (metadata, such as the class
   synopsis) is provided as arguments to that command.  Components can be
   fully introspected, but metadata is only available if The Simple
   Development Library A<-storemetadata> option is set when the class is
   declared.

      Procedures are provided to declare and delete classes and get
   information about them.

} -usage {

      Use K<declare-class> to declare a class.  Declared classes metadata can
   be obtained via the K<::Simple::Class::information metadata> procedure. 
   Other subcommands of this procedure provide information about declared
   classes:  K<exists>, K<classes>, K<objects>, K<errors>,
   K<exportedcommands>, K<aliases>, K<variables>, K<attributes>,
   K<itkoptions>, K<itkcomponents>, K<commands>, K<methods> and K<api>.  The
   K<::Simple::ItkOption::information> and
   K<::Simple::ItkComponent::information> procedures provide subcommand to
   quet information about the [incr Tk] options and components of an [incr Tk]
   megawidget, respectively.

   Use K<::Simple::Class::delete> to delete the class.

} -keywords {

   {object oriented} class method {[incr Tcl]} {[incr Tk]} widget megawidget

} -examples {

   # Install the package
   package require SimpleClass
   ::Simple::Package::require-and-install SimpleClass

   # Let's store the metadata
   ::Simple::configure -storemetadata true

   # Declare a class
   declare-class {

      Class

   } -synopsis {

      This is the class synopsis.

   } -variable {

      {
         Variable
      } -description {
         This is a class variable
      } -value 50

   } -attribute {

      {
         Attribute
      } -description {
         This is a class attribute
      } -value 50

   } -error {

      {
         Error
      } -message {
         Class error
      } -explanation {
         This is a class error.
      }

   } -command {

      {

         command

      } -synopsis {

         This is the command synopsis.

      } -arguments {
         {-modify     boolflag   {Whether to modify the class variable}}
         { value      integer    {New class variable value}}
      } -effects {

         * Modifies the class variable if -modify is given.

      } -body {

         # Must modify the class variable
         if {[::Simple::Argument::information flaggiven -modify]} {

            # Modify it
            set ::Class::Variable $value
         }

         # Class variable value too big
         if {$::Class::Variable > 99} {

            # Throw the class error
            ::Simple::Error::throw ::Class::Error

         # Correct class variable value
         } else {

            # Display it
            puts "Class variable = $::Class::Variable"
         }
      }

   } -method {

      {

         method

      } -synopsis {

         This is the method synopsis.

      } -arguments {
         {-modify     boolflag   {Whether to modify the object attribute}}
         { value      integer    {New object attribute value}}
      } -effects {

         * Modifies the object attribute if -modify is given.

      } -body {

         # Must modify the object attribute
         if {[::Simple::Argument::information flaggiven -modify]} {

            # Modify it
            set Attribute $value
         }

         # Object attribute value too big
         if {$Attribute > 99} {

            # Throw the class error
            ::Simple::Error::throw ::Class::Error

         # Correct object attribute value
         } else {

            # Display it
            puts "Object $this attribute = $Attribute"
         }
      }
   }

   # Invoke the class command

   # This displays the following:
   #   |Class variable = 50
   ::Class::command 77

   # This displays the following:
   #   |Class variable = 88
   ::Class::command -modify 88

   # This displays the following:
   #   |Class error (::Class::Error)
   catch {::Class::command -modify 120} result
   puts "$result ($::errorCode)"

   # Introspect the class
   # This displays the following:
   #   |Class errors: ::Class::Error
   #   |Class commands: ::Class::command
   #   |Class methods: ::Class::method
   #   |Class variables: ::Class::Variable
   #   |Class attributes: ::Class::Attribute
   puts "Class errors: [::Simple::Class::information errors Class]"
   puts "Class commands: [::Simple::Class::information commands Class]"
   puts "Class methods: [::Simple::Class::information methods Class]"
   puts "Class variables: [::Simple::Class::information variables Class]"
   puts "Class attributes: [::Simple::Class::information attributes Class]"

   # Obtain the class metadata
   # This displays the following:
   #   |-assumptions = 
   #   |  -classtype = 
   #   |  -copyright = 
   #   |    -details = 
   #   |    -effects =
   #   |   -examples = 
   #   |    -extends = 
   #   |    -history = 
   #   |   -keywords = 
   #   |-limitations = 
   #   |   -overview = 
   #   |    -project = 
   #   | -references = 
   #   |    -remarks = 
   #   |  -resources = 
   #   | -stereotype = 
   #   |   -synopsis = This is the command synopsis.
   #   |       -todo = 
   #   |      -usage = 
   #   |    -version = 1.0
   array set metadata [::Simple::Class::information metadata Class]
   foreach element [lsort [array names metadata]] {
      puts [format {%12s = %s} -$element $metadata($element)]
   }

   # Create two objects of the class
   Class object1
   Class object2

   # Invoke the first object methods
   # This displays the following:
   #   |Object ::object1 attribute = 50
   #   |Object ::object1 attribute = 88
   object1 method 77
   object1 method -modify 88

   # Invoke the second object methods
   # This displays the following:
   #   |Object ::object2 attribute = 50
   #   |Object ::object2 attribute = 22
   object2 method 11
   object1 method -modify 22

   # This displays the following:
   #   |Class error (::Class::Error)
   catch {object1 method -modify 120} result
   puts "$result ($::errorCode)"

} -api {

   * K<::Simple::Class::declare-class> class ?-version version?
        ?-project string? ?-synopsis string? ?-overview string? 
        ?-usage string? ?-api string? ?-keywords list? 
        ?-examples string? ?-classtype type? ?-stereotype choice? 
        ?-extends qualifiedname-list? ?-debug extbool? ?-error declaration? 
        ?-variable declaration? ?-attribute declaration? 
        ?-itkoption declaration? ?-itkcomponent declaration? 
        ?-command declaration? ?-method declaration? ?-assumptions string? 
        ?-resources string? ?-limitations string? ?-effects string? 
        ?-details string? ?-references string? ?-remarks string? 
        ?-todo string? ?-history string? ?-copyright string? ?-body script?
     Declares an [incr Tcl] class.

   * K<::Simple::Class::delete> class
     Deletes an [incr Tcl] class.

   * K<::Simple::Class::configure ?-debug extbool?>
     Configures the package options.

   * K<::Simple::Class::cget ?-debug?>
     Gets the package options.

} -requisites {

   {SimpleDeclare 1.0}
   Itcl

} -namespaces {

   ::Simple::Class
   ::Simple::Class::Priv
   ::Simple::ItkOption
   ::Simple::ItkComponent

} -exportedcommands {

   ::Simple::Class::declare-class

} -option {
   {
      -debug

   } -description {

      Whether to create a K<%debug%> debug method for every class.

   } -type extbool -value false

} -history {

   22-mar-2003   Added A<-body> to method declaration, version 0.5.1
   15-jul-2003   Method declaration added to K<declare-class>, version 0.5.2
   20-jan-2004   Comments allowed before embedded declarations in\
                 K<declare-class>, version 0.5.3
   28-jan-2004   Added A<-itkoption> and A<-itkcomponent> to K<declare-class>,\
                 version 0.5.4
   26-oct-2004   Added A<-error> to K<declare-class>, version 0.5.5
   04-apr-2005   Added debug method and A<-debug> option, version 0.5.6
   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -error {

   {
      ::Simple::Class::CONFIGBODY-FOR-NON-PUBLIC-ATTRIBUTE

   } -message {

      -configbody given for non-public attribute "%s"

   } -explanation {

      The declaration of attribute <attribute> failed because "-configbody"
      was provided for a non-public attribute, but that is not allowed

   } -corrective {

      Either remove the "-configbody" or make the attribute public
   }

} -error {

   {
      ::Simple::Class::CONFIGBODY-FOR-ARRAY-ATTRIBUTE

   } -message {

      -configbody given for array attribute "%s"

   } -explanation {

      The declaration of attribute <attribute> failed because "-configbody"
      was provided for an attribute whose type is a derived array type, but
      that is not allowed

   } -corrective {

      Either remove the "-configbody" or change the attribute type
   }

} -error {

   {
      ::Simple::Class::CLASS-COMMAND-CONSTRUCTOR-OR-DESTRUCTOR

   } -message {

      class commands can not be named "%s"

   } -explanation {

      The declaration of a class failed because class command <constructor or
      destructor> is named "constructor" or "destructor", but that is not
      allowed

   } -corrective {

      Use a different name for the class command
   }

} -error {

   {
      ::Simple::Class::CLASS-COMMAND-AND-METHOD-SHARE-BASE-COMMAND

   } -message {

      class command and method share the same base command "%s"

   } -explanation {

      The declaration of a class failed because a class command and a method
      are both subcommands sharing the same base command <base command>, but
      that is not allowed

   } -corrective {

      Use different base commands for class commands and methods
   }

} -error {

   {
      ::Simple::Class::SUBCOMMANDS-WITH-DIFFERENT-ACCESS

   } -message {

      class subcommands "%s" and "%s" have different access modes

   } -explanation {

      The declaration of a class failed because two class subcommands (<class
      subcommand 1> and <class cubcommand 2>) corresponding to the same base
      command have different access modes, but that is not allowed

   } -corrective {

      Declare all subcommands for a given base command with the same access mode
   }

} -error {

   {
      ::Simple::Class::NON-EXISTING-ITKITEM

   } -message {

      class "%s" doesn't have %s "%s"

   } -explanation {

      Class <class> does not have <an option or a component> named
      <bad [incr Tk] option>
   }

} -error {

   {
      ::Simple::Class::SUBMETHODS-WITH-DIFFERENT-ACCESS

   } -message {

      submethods "%s" and "%s" of class "%s" have different access modes

   } -explanation {

      The declaration of class <class> failed because two class submethods
      (<submethod 1> and <submethod 2>) corresponding to the same base method
      have different access modes, but that is not allowed

   } -corrective {

      Declare all submethods for a given base method with the same access mode

   } -translation {
      <class> <submethod 1> <submethod 2>
   }

} -error {

   {
      ::Simple::Class::SHALL-EXTEND-ITK-ARCHETYPE

   } -message {

      class %s with %s shall extend ::itk::Archetype

   } -explanation {

      Class <class> is a megawidget because it has an [incr Tk] <option or
      component>, and so shall extend K<::itk::Archetype>.  Notice that
      K<::itk::Widget> and K<::itk::Toplevel> already extend
      K<::itk::Archetype>
   }

} -error {

   {
      ::Simple::Class::MULTIPLE-OBJECTS-FOR-SINGLETON-CLASS

   } -message {

      multiple objects for singleton class "%s"

   } -explanation {

      More than one object is being created for singleton class <class>,
      but that is not allowed
   }

} -error {

   {
      ::Simple::Class::OBJECT-FOR-ABSTRACT-CLASS

   } -message {

      object for abstract class "%s"

   } -explanation {

      An object for abstrac class <class> is being created, but that is not
      allowed
   }

} -error {

   {
      ::Simple::Class::PUBLIC-METHOD-WITH-BODY-FOR-ABSTRACT-CLASS

   } -message {

      public method "%s" with body for abstract class "%s"

   } -explanation {

      A public method <method> with body is being declared for abstract class
      <class>, but that is not allowed
   }

} -error {

   {
      ::Simple::Class::CONSTRUCTOR-OR-DESTRUCTOR-FOR-ABSTRACT-CLASS

   } -message {

      %s for abstract class "%s"

   } -explanation {

      A <constructor or destructor> has been provided for abstract class
      <class>, but that is not allowed
   }

} -error {

   {
      ::Simple::Class::VIRTUAL-METHOD-CALLED

   } -message {

      virtual method "%s" called

   } -explanation {

      A virtual method <method> (an abstract class method with no body)
      has been called, usually from a derived class

   } -corrective {

      Provide an implementation for the method in the derived class.
   }

} -error {

   {
      ::Simple::Class::UNDECLARED 

   } -message {

      undeclared class "%s"

   } -explanation {

      Class <class> is undeclared, but a declared class is required
   }

} -command {

### ===========================================================================
### -command  : ::Simple::Class::declare-class

{
   ::Simple::Class::declare-class

} -synopsis {

   Declares a class.

} -access {

   exported 

} -overview {
    
      This procedure provides a way to declare [incr Tcl] classes and [incr
   Tk] megawidgets.

   The A<-synopsis> flag is mandatory.

      The A<-classtype> flag specifies a type whose allowed values are objects
   of the class. This type can then be used whenever a type is required, such
   as for the arguments of another class commands or methods.

      The A<-stereotype> flag specifies the class stereotype among one of
   multiobject (the default), abstract or singleton.  Multiobject classes
   support an arbitrary number of objects.  Abstract classes support no
   objects (they are intended to be extended). Finally, singleton classes
   can be instantiated just once.

      The A<-extends> flag specifies the list of classes from which this one
   inherits.

      The A<-variable> flag allows to create class variables (class variables
   in object-oriented parlance; common variables for [incr Tcl]).  Its
   argument is a variable declaration intended to be the arguments of a call
   to K<::Simple::Variable::declare>, except for the added argument
   A<-access>, whose value must be one of public (the default), protected or
   private; this argument specifies the variable access mode.  This flag can
   be given multiple times.

      The A<-attribute> flag allows to create class attributes (class instance
   variables in object-oriented parlance; variables for [incr Tcl]).  Its
   argument is an attribute declaration intended to be the arguments of a call
   to K<::Simple::Variable::declare>, except for the added argument
   A<-access>, equivalent to that for class variables, and A<-configbody>,
   valid for public attributes only, used to specify the attribute
   configuration script (see the K<configbody> [incr Tcl] command).  This flag
   can be given multiple times.

      The A<-error> flag allows to declare class errors.  Its argument is an
   error declaration intended to be the arguments of a call to
   K<::Simple::Error::declare>.  This flag can be given multiple times.
 
      The A<-command> flag allows to declare class commands (class or static
   functions in object-oriented parlance; common procedures for [incr Tcl]). 
   Its argument is a command declaration intended to be the arguments of a
   call to K<::Simple::Declare::declare-proc>, except for the added argument
   A<-access>, whose value must be one of public (the default), private, or
   exported; this argument specifies the command access mode.  This flag can
   be given multiple times.
 
      The A<-method> flag allows to declare class methods (class instance
   functions or methods in object-oriented parlance; methods for [incr Tcl]). 
   Its argument is a method declaration intended to be the arguments of a call
   to K<::Simple::Declare::declare-proc>, except for the fact that A<-alias>
   is not allowed for methods, and the added argument A<-access>, whose value
   must be one of public (the default), protected or private; this argument
   specifies the method access mode.  This flag can be given multiple times.

      If the method name is ''constructor'', it corresponds to the class
   constructor.  In this case, A<-synopsis> is not allowed (it is hardcoded to
   "Class constructor"), and the added argument A<-init>, the script used
   invoke base class constructors, can be specified.

      If the method name is ''destructor'', it corresponds to the class
   destructor.  In this case, A<-synopsis> is not allowed (it is hardcoded to
   "Class destructor"). The destructor default access mode is "private".

      The A<-body> flag specifies the class body, a script evaluated when the
   class is declared.

      The A<-debug> flag allows to override the package option of the same
   name for the class being declared.  This option controls whether to create
   a K<%debug%> debug method.  This method has a single argument:  a script
   which gets evaluated in the object scope.

      The A<-itkoption> and A<-itkcomponent> flags provide support for the
   definition of [incr Tk] megawidgets as follows.

      The A<-itkoption> flag allows to create options for [incr Tk]
   megawidgets.  Its argument is an [incr Tk] option declaration intended to
   be the arguments of a call to K<::Simple::Variable::declare>, except for
   the added argument A<-configbody>, equivalent to that for attributes, and
   A<-resourcename> and A<-resourceclass>:  see the K<itk_option define> [incr
   Tk] command.  This flag can be given multiple times.

      The A<-itkcomponent> flag allows to create components for [incr Tk]
   megawidgets.  Its argument is an [incr Tk] component declaration which
   shall start by the component name and be followed by the following optional
   flags (see the K<itk_component add> [incr Tk] command):

    * A<-description>: the component description.

    * A<-access>: the component access mode (one of public, protected or
       private).

    * A<-ignore>:  list of options to ignore for the megawidget; see the
       K<ignore> command within the K<itk_component add> [incr Tk] command.

    * A<-keep>:  list of options to keep for the megawidget; see the K<keep>
       command within the K<itk_component add> [incr Tk] command.

    * A<-rename>: options to rename for the megawidget. Each option is a
       four-element list; see the K<rename> command within the
       K<itk_component add> [incr Tk] command.

    * A<-resourceclass>: resource class.

    * A<-body>: component body.
   
      The K<usual> command within the K<itk_component add> [incr Tk] command
   is not supported.  The A<-itkcomponents> flag can be given multiple times.

      All other arguments are metadata, and are stored or ignored depending on
   whether The Simple Development Library A<-storemetadata> option is set when
   the class is declared, but for the A<-api> and A<-body> flags which are
   never stored.

} -arguments {

   { class            name           {Class name}}
   {-version          version 1.0    {Class version}}
   {-project          string {}      {Class project}}
   {-synopsis         string {}      {Class synopsis}}
   {-overview         string {}      {Class overview}}
   {-usage            string {}      {Class usage message}}
   {-api              string {}      {Class API}}
   {-keywords         list {}        {Class keywords}}
   {-examples         string {}      {Class usage examples}}
   {-classtype        type %DEFAULT% {Class type name}}
   {-stereotype       choice {multiobject abstract singleton}
                                     {Class stereotype}}
   {-extends          qualifiedname-list {}   
                                     {List of classes from which this one
                                      inherits}}
   {-debug            extbool false  {Whether to create a debug method}}
   {-error            declaration {} {Class error declaration}}
   {-variable         declaration {} {Class variable declaration}}
   {-attribute        declaration {} {Class attribute declaration}}
   {-itkoption        declaration {} {[incr Tk] option declaration}}
   {-itkcomponent     declaration {} {[incr Tk] component declaration}}
   {-command          declaration {} {Class command declaration}}
   {-method           declaration {} {Class method declaration}}
   {-assumptions      string {}      {Class assumptions}}
   {-resources        string {}      {Class required resources}}
   {-limitations      string {}      {Class limitations}}
   {-effects          string {}      {Class effects}}
   {-details          string {}      {Class implementation details}}
   {-references       string {}      {Class references}}
   {-remarks          string {}      {Class remarks}}
   {-todo             string {}      {Class to do list}}
   {-history          string {}      {Class history}}
   {-copyright        string {}      {Class copyright notice}}
   {-body             script {}      {Class body}}

} -returns {

   The empty string.

} -examples {

   # The canonical toaster example
   declare-class {
      Toaster
   } -synopsis {
      A toaster.
   } -overview {
   
         Toaster class with toast and clean methods and a crumb count that
      will burn the toaster if too high
   
   } -error {
      {
         FIRE
      } -message {
         == FIRE! FIRE! ==
      }
   } -attribute {
      {
         crumbs
      } -description {
         Number of crumbs in the toaster
      } -type {
         integer
      } -value 0
   } -method {
      {
         toast
      } -synopsis {
         Toast a given number of slices
      } -arguments {
         { nSlices          integer        {Number of slices to toast}}
      } -returns {
         The number of crumbs after toasting the slices
      } -body {
   
         # Too many crumbs
         if {$crumbs > 50} {
   
            # Catch fire!
            ::Simple::Error::throw FIRE
         }
   
         # Each toasted slice produces 4 crumbs
         incr crumbs [expr {4 * $nSlices}]
   
         # Return the number of crumbs after toasting the slices
         return $crumbs
      }
   } -method {
      {
         clean
      } -synopsis {
         Cleans the toaster
      } -returns {
         The empty string
      } -body {
   
         # Remove all the crumbs
         set crumbs 0
   
         # Return the empty string
         return {}
      }
   }

   declare-class {
      SmartToaster
   } -synopsis {
      A smart toaster
   } -overview {
   
         This toaster has an autoclean feature that automatically cleans the
      toaster when there are lots of crumbs.  How smart!
   
   } -extends {
      ::Toaster
   } -method {
      {
         toast
      } -synopsis {
         Performs autoclean if necessary and toast a given number of slices
      } -arguments {
         { nSlices          integer        {Number of slices to toast}}
      } -returns {
         The number of crumbs after toasting the slices
      } -body {
   
         # Lots of crumbs
         if {$crumbs > 40} {
   
            # Perform autoclean
            clean
         }
   
         # Toast the slices
         set answer [::Toaster::toast $nSlices]
   
         # Return the number of crumbs after toasting the slices
         return $answer
      } 
   }

} -parseflags {

   false

} -effects {
 
   * Creates a class named K<class>.

   * Creates the class commands, which may be exported or have an alias.
 
   * Creates a public K<%debug%> debug method if A<-debug> is given or the
     package option of the same name is set.  

   * Creates a private K<create-components> method if at least one [incr Tk]
     component is given.  This method should be invoked by each megawidget
     object (for example from the megawidget constructor) to create the
     megawidget components.

   * Those of the class body.

} -remarks {
 
   * The order of the flags is irrelevant.
 
} -references {
 
   * The K<::Simple::Class::delete> procedure.

} -details    {

   * Each declared class metadata is stored in a class
     private variable named V<%Metadata%>.

   * Each declared class list of exported commands is stored in a class
     private variable named V<%ExportedCommands%>.

   * Each declared class list of command aliases is stored in a class
     private variable named V<%Aliases%>.

} -body {

   ### Parse and verify the class declaration
   set class [string trim $class]
   ::Simple::Error::Priv::push-location class $class 1
   if {[catch {
      foreach {tmp tmp tmp tmp declarationAndCreationScript\
         installationScript}\
         [::Simple::Class::Priv::parse-and-verify-declaration\
         $class $args 0] break
   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Declare, create and install the class
   eval $declarationAndCreationScript\n$installationScript

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::parse-and-verify-declaration

{
   ::Simple::Class::Priv::parse-and-verify-declaration

} -synopsis {

   Parses and verifies a class declaration.

} -arguments {

   { class            name           {Class name}}
   { actualArgs       list           {Actual class declaration argument list}}
   { declaringPackage boolean        {Whether a package is being declared}}

} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the class type.
    1: the class list of exported commands.
    2: the class list of aliases in the format ||{aliasName aliasedCommand}||.
    3: the class list of errors.
    4: the class declaration and creation script.
    5: the class install script.
#end

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class 4] break

   ### Set defaults
   set version 1.0
   set classtype %DEFAULT%
   set debug [set ::Simple::Class::%Options%(-debug)]

   ### Parse flags
   ::Simple::Priv::parse-flags\
      [list -version trimmed\
      -project trimmed\
      -synopsis line\
      -overview paragraphs\
      -usage paragraphs\
      -api paragraphs\
      -keywords list\
      -examples lines\
      -classtype trimmed\
      -stereotype trimmed\
      -extends list\
      -debug trimmed\
      -assumptions paragraphs\
      -resources paragraphs\
      -limitations paragraphs\
      -effects paragraphs\
      -details paragraphs\
      -references paragraphs\
      -remarks paragraphs\
      -todo paragraphs\
      -history paragraphs\
      -copyright paragraphs\
      -body untouched]\
      [list -error untouched\
      -variable untouched\
      -attribute untouched\
      -itkoption untouched\
      -itkcomponent untouched\
      -command untouched\
      -method untouched] $actualArgs 0 2

   ### Handle A<-synopsis>
   ::Simple::Declare::Priv::assert-synopsis-given

   ### Handle A<-stereotype>
   set constructorInitializationScript {}
   set isAbstract 0
   if {[::Simple::Argument::information flaggiven -stereotype]} {

      # Assert the given stereotype is correct
      ::Simple::Variable::Priv::assert-type-conformation\
         argument 0 -stereotype choice $stereotype 0\
         {multiobject abstract singleton}

      # Switch on stereotype
      switch -exact -- $stereotype {

         singleton {
         # Singleton class

            # Provide a default constructor which throws an
            # error if the class already has one object
            set constructorInitializationScript\
               "::Simple::Class::Priv::assert-no-objects $qualifiedClass"
         }

         abstract {
         # Abstract class

            # Provide a default constructor which throws an error if
            # invoked to create objects of this class directly
            set isAbstract 1
            set constructorInitializationScript\
               "::Simple::Class::Priv::assert-no-this-class \$this\
               $qualifiedClass"
         }
      }
   }

   ### Handle A<-classtype>
   set classDeclarationScript {common %Variables%}
   set afterClassDeclarationScript {}
   set classInstallationScript {}
   if {[::Simple::Argument::information flaggiven -classtype]} {

      ### Declare the class type
      #      The allowed values for these types are objects of the class.
      #      This type needs to be created here so that class variables,
      #   attributes and commands and method arguments can use the class type.
      #      If the class is part of a package declaration, the class type is
      #   created hidden so that package variables, commands and other package
      #   items can use the class type but the type is not publicly visible.
      #   The type is then unhidden upon installing the package.
      set typeDeclarationScript\
         [::Simple::Type::Priv::parse-and-verify-declaration $classtype\
         [list -description "Class $qualifiedClass" -matchingscript\
         [format {return [$value isa %s]} $qualifiedClass]] $declaringPackage]
      eval $typeDeclarationScript

      ### Declare the class type in the class install script
      ::Simple::Script::append\
         classInstallationScript $typeDeclarationScript

      ### Store the class type
      #   So that it can be deleted upon deleting the class via
      #   K<::Simple::Class::delete>
      ::Simple::Script::append afterClassDeclarationScript\
         "set ${qualifiedClass}::%ClassType% $classtype"
   }

   ### Handle A<-extends>
   if {[::Simple::Argument::information flaggiven -extends]} {

      ### Assert the extended classes are qualified
      ::Simple::Variable::Priv::assert-type-conformation\
         argument 0 -extends qualifiedname-list $extends 0

      ### Add the inheritance list to the class declaration script
      ::Simple::Script::append classDeclarationScript "inherit $extends"
   }

   ### Debug method requested
   set debug [{::Simple::Type::convert boolean} $debug]
   if {$debug} {

      ### Add the creation of the K<%debug%> debug
      ###    method to the class declaration script
      ::Simple::Script::append classDeclarationScript {
         public method %debug% {
            script
         } {
            return [eval $script]
         }
      }
   }

   ### Handle A<-variable>

   ### Loop over variables
   foreach variableDeclaration $variable {

      ### Parse the variable declaration
      ::Simple::Declare::Priv::remove-leading-comments variableDeclaration
      foreach {forwardDeclarationScript declarationScript creationScript}\
         [::Simple::Class::Priv::parse-variable-or-similar-declaration\
         variable $qualifiedClass $variableDeclaration $declaringPackage] break

      ### Add the variable forward declaration script
      ##     to the class declaration script
      ::Simple::Script::append classDeclarationScript $forwardDeclarationScript

      ### Add the variable creation script to the after
      ###    class declaration script
      ::Simple::Script::append afterClassDeclarationScript $creationScript

      ### Add the variable declaration script to the
      ###    after class declaration script
      ::Simple::Script::append afterClassDeclarationScript $declarationScript
   }

   ### Handle A<-attribute>

   ### Loop over attributes
   foreach attributeDeclaration $attribute {

      ### Parse the attribute declaration
      ::Simple::Declare::Priv::remove-leading-comments attributeDeclaration
      foreach {forwardDeclarationScript declarationScript creationScript}\
         [::Simple::Class::Priv::parse-variable-or-similar-declaration\
         attribute $qualifiedClass $attributeDeclaration $declaringPackage]\
         break

      ### Add the attribute forward declaration script
      ###    to the class declaration script
      ::Simple::Script::append classDeclarationScript $forwardDeclarationScript

      ### Add the attribute creation script to the
      ###    constructor initialization script
      ::Simple::Script::append constructorInitializationScript $creationScript

      ### Add the attribute declaration script to the
      ###    after class declaration script
      ::Simple::Script::append afterClassDeclarationScript $declarationScript
   }

   ### Handle A<-itkoption>
   if {[llength $itkoption] > 0} {

      ### Add the script which asserts this class extends K<::itk::Archetype>
      ###    to the class declaration script
      ::Simple::Script::append classDeclarationScript\
         "::Simple::Class::Priv::assert-extends-itkarchetype $class\
         [list $extends] option"

      ### Loop over [incr Tk] options
      foreach itkoptionDeclaration $itkoption {

         ### Parse the [incr Tk] option declaration
         ::Simple::Declare::Priv::remove-leading-comments itkoptionDeclaration
         foreach {forwardDeclarationScript declarationScript creationScript}\
            [::Simple::Class::Priv::parse-variable-or-similar-declaration\
            itkoption $qualifiedClass $itkoptionDeclaration $declaringPackage] break

         ### Add the [incr Tk] option forward declaration script
         ###    to the class declaration script
         ::Simple::Script::append\
            classDeclarationScript $forwardDeclarationScript

         # There is no creation script for [incr Tk] options as they can
         # not be arrays and the type monitoring is handled through its
         # A<-configbody>

         ### Add the [incr Tk] option declaration script
         ###    to the after class declaration script
         ::Simple::Script::append\
            afterClassDeclarationScript $declarationScript
      }
   }

   ### Handle A<-itkcomponent>
   set itkComponentCreationScript {}
   set itkComponentsCreationScript {}
   if {[llength $itkcomponent] > 0} {

      ### Add the script which asserts this class extends K<::itk::Archetype>
      ###    to the class declaration script
      ::Simple::Script::append classDeclarationScript\
         "::Simple::Class::Priv::assert-extends-itkarchetype $class\
         [list $extends] component"

      ### Loop over [incr Tk] components
      foreach itkcomponentDeclaration $itkcomponent {

         ### Parse the [incr Tk] component declaration
         ::Simple::Declare::Priv::remove-leading-comments\
            itkcomponentDeclaration
         foreach {itkComponentName itkComponentDeclarationScript\
            itkComponentCreationScript}\
            [::Simple::Class::Priv::parse-itkcomponent-declaration\
            $qualifiedClass $itkcomponentDeclaration] break
         ::Simple::Script::append\
            itkComponentsCreationScript $itkComponentCreationScript

         ### Add the [incr Tk] component declaration script
         ###    to the after class declaration script
         ::Simple::Script::append\
            afterClassDeclarationScript $itkComponentDeclarationScript
      }
   }

   ### At least one [incr Tk] component given
   if {[string compare $itkComponentsCreationScript {}]} {

# fixme:  create a K<create-component> which creates a single component passed
#        as argument; K<create-components> should invoke K<create-component>
#        for each component in turn, in the definition order
      ### Add the creation of the K<create-components> method
      ###    to the class declaration script
      ::Simple::Script::append classDeclarationScript "private method\
         create-components {} [list $itkComponentsCreationScript]"
   }

   ### Handle A<-error>

   ### Loop over errors
   set classErrorCodes [list]
   foreach errorDeclaration $error {

      ### Assert the error code is unqualified
      ::Simple::Declare::Priv::remove-leading-comments errorDeclaration
      set errorCode [string trim [lindex $errorDeclaration 0]]
      ::Simple::Type::assert-is\
         unqualifiedname $errorCode {} {unqualified error code}

      ### Parse and verify the error declaration
      ::Simple::Error::Priv::push-location error $errorCode
      set qualifiedErrorCode ${qualifiedClass}::$errorCode
      set errorDeclarationScript\
         [::Simple::Error::Priv::parse-and-verify-declaration\
         $qualifiedErrorCode [lrange $errorDeclaration 1 end] 0]
      ::Simple::Error::Priv::pop-location

      ### Add the error declaration script to the
      ###    after class declaration script
      ::Simple::Script::append\
         afterClassDeclarationScript $errorDeclarationScript

      ### Add the errors to the class list of errors
      lappend classErrorCodes $qualifiedErrorCode
   }

   ### Handle A<-command>
   set classExportedCommands [list]
   set classAliases [list]
   if {[llength $command] > 0} {

      ### Loop over commands
      foreach commandDeclaration $command {

         ### Parse the command declaration
         ::Simple::Declare::Priv::remove-leading-comments commandDeclaration
         foreach {exportedCommand aliasPair forwardCommandDeclarationScript\
            commandDeclarationAndCreationScript commandInstallationScript tmp}\
            [::Simple::Class::Priv::parse-command-or-similar-declaration\
            $declaringPackage 1 $isAbstract $namespace $unqualifiedClass\
            $project {} $commandDeclaration] break

         ### Add the exported commands to the class list of exported commands
         if {[string compare $exportedCommand {}]} {
            lappend classExportedCommands $exportedCommand
         }

         ### Add the aliases commands to the class list of aliases
         if {[string compare $aliasPair {}]} {
            lappend classAliases $aliasPair
         }

         ### Add the command forward declaration script to the
         ###    class declaration script
         ::Simple::Script::append\
            classDeclarationScript $forwardCommandDeclarationScript

         ### Add the command declaration and creation script
         ###    to the after class declaration script
         ::Simple::Script::append\
            afterClassDeclarationScript $commandDeclarationAndCreationScript

         ### Add the command installation script to the
         ###    class installation script
         ::Simple::Script::append\
            classInstallationScript $commandInstallationScript
      }
   }

   ### Handle A<-method>
   set constructorGiven 0
   if {[llength $method] > 0} {

      ### Loop over methods
      foreach methodDeclaration $method {

         ### Parse the method declaration
         ::Simple::Declare::Priv::remove-leading-comments methodDeclaration
         foreach {tmp tmp forwardMethodDeclarationScript\
            methodDeclarationAndCreationScript methodInstallationScript\
            isConstructor}\
            [::Simple::Class::Priv::parse-command-or-similar-declaration\
            $declaringPackage 0 $isAbstract $namespace $unqualifiedClass\
            $project $constructorInitializationScript $methodDeclaration] break
         if {$isConstructor} {
            set constructorGiven 1
         }

         ### Add the method forward declaration script
         ###    to the class declaration script
         ::Simple::Script::append\
            classDeclarationScript $forwardMethodDeclarationScript

         ### Add the method declaration and creation script to the
         ###    after class declaration script
         ::Simple::Script::append\
            afterClassDeclarationScript $methodDeclarationAndCreationScript

         ### Add the method installation script to the
         ###    class installation script
         ::Simple::Script::append\
            classInstallationScript $methodInstallationScript
      }
   }

   ### No constructor given
   if {!$constructorGiven} {

      ### A<-itkoption> given
      set constructorBody {}
      if {[::Simple::Argument::information flaggiven -itkoption]} {
         ::Simple::Script::append\
            constructorInitializationScript {eval itk_initialize $args}
      }

      ### Constructor initialization script
      if {[string compare $constructorInitializationScript {}]} {

         ### Declare a default constructor containing
         ###    the constructor initialization script
         ::Simple::Script::Header::add constructorBody\
            %PRIV%-declare-class 100 $constructorInitializationScript
         ::Simple::Script::append\
            classDeclarationScript {private constructor args {} {}}
         ::Simple::Script::append afterClassDeclarationScript "::itcl::body\
            ${qualifiedClass}::constructor args [list $constructorBody]"
      }
   }

   ### Compose the class declaration and creation script
   if {[string compare $classtype %DEFAULT%]} {
      set classtypeToStore $classtype
   } else {
      set classtypeToStore {}
   }
   set classDeclarationAndCreationScript "::Simple::Class::Priv::do-declare\
      $qualifiedClass [list $classtype] [list $classDeclarationScript]\
      [list $afterClassDeclarationScript\n$body] [list $classExportedCommands]\
      [list $classAliases] [list [list $version $project $synopsis $overview\
      $usage $keywords $examples $classtypeToStore $stereotype $extends\
      $assumptions $resources $limitations $effects $details $references\
      $remarks $todo $history $copyright]]"

   ### Return the class type, the class list of exported commands, the
   ###    class list of aliases, the class list of errors, the declaration and
   ###    creation script and the class installation script
   return [list $classtype $classExportedCommands $classAliases\
      $classErrorCodes $classDeclarationAndCreationScript\
      $classInstallationScript]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::do-declare

{
   ::Simple::Class::Priv::do-declare

} -synopsis {

   Declares a class.

} -arguments {

   { class            qualifiedname  {Class name}}
   { classtype        name           {Class type name}}
   { creationScript   script         {Class creation script}}
   { afterCreationScript
                      script         {After class creation script}}
   { exportedCommands qualifiedname-list
                                     {Class qualified exported commands}}
   { aliases          list-list      {Class aliases in the format
                                      ||{aliasName aliasedCommand}||}}
   { metadata         any-list       {Class metadata}}

} -effects {
 
   * Creates the class named A<class>.

} -body {

   ### Create the class
   #   Error evaluating the class creation script or body
   if {[catch {::itcl::class $class $creationScript} result]} {

      # Delete the classtype
      if {[string compare $classtype %DEFAULT%]} {
         ::Simple::Type::Priv::delete-with-hidden $classtype 1
      }

      # Rethrow the error
      return -code error -errorcode $::errorCode $result
   }

   ### Evaluate the after class creation script
   #   Error evaluating the after class declaration script
   if {[catch $afterCreationScript result]} {

      # Delete the class
      ::Simple::Class::delete $class

      # Rethrow the error
      return -code error -errorcode $::errorCode $result
   }

   # The class metadata must not be stored
   if {![set ::Simple::%Options%(-storemetadata)]} {
      set metadata [list]
   }

   ### Store the class exported commands
   set ${class}::%ExportedCommands% $exportedCommands

   ### Store the class aliases
   set ${class}::%Aliases% $aliases

   ### Store the class metadata
   set ${class}::%Metadata% $metadata
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::delete

{
   ::Simple::Class::delete

} -synopsis {

   Deletes an [incr Tcl] class.

} -overview {

      This procedure deletes the given class together with its class
   type, if any.

} -arguments {

   { class            name           {Class name}}

} -effects {
 
   * Deletes the given class and its class type, if any.

} -references {
 
   * The K<declare-class> procedure.

} -returns {

   The empty string.

} -body {

   ### Parse, qualify and split the class name
   foreach {namespace unqualifiedClass qualifiedClass}\
      [::Simple::Class::Priv::parse-qualify $class] break

   ### Delete the class type, if any
   if {[info exists ${qualifiedClass}::%ClassType%]} {
      ::Simple::Type::Priv::delete-with-hidden\
         [set ${qualifiedClass}::%ClassType%] 1
   }

   ### Delete the class
   # [uplevel] is needed here due to [incr Tcl] scoping rules
   uplevel 1 [list ::itcl::delete class $qualifiedClass]

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::parse-command-or-similar-declaration

{
   ::Simple::Class::Priv::parse-command-or-similar-declaration

} -synopsis {

   Parses a class command or similar item declaration

} -overview {

      This procedure parses a class command or similar item (constructor,
   destructor or method) declaration.

} -arguments {

   { declaringPackage boolean        {Whether a package is being declared}}
   { isCommand        boolean        {Whether the item is a command}}
   { isAbstract       boolean        {Whether the class is abstract}}
   { namespace        namespace      {Class namespace}}
   { class            unqualifiedname
                                     {Class name}}
   { project          string         {Class project}}
   { constructorInitializationScript
                      script         {Initialization script to be added
                                      to the class constructor}}
   { declaration      list           {Command or similar declaration}}

} -remarks {
 
   * The A<declaration> is a command (or similar) declaration intended to be
     the arguments of a call to K<::Simple::Proc::create> (or similar) as
     follows:

#verbatim
   { item             unqualifiedcommandname
                                     {Item name}}
   {-version          version 1.0    {Item version}}
   {-project          string {}      {Command project (commands only)}}
   {-synopsis         string {}      {Command or method synopsis (not for
                                      constructor nor destructor)}}
   {-overview         string {}      {Item overview}}
   {-access           choice {public protected private exported}
                                     {Item access mode (protected not for
                                      commands and exported for commands
                                      only)}}
   {-alias            name {}        {Command alias (commands only)}}
   {-arguments        extendedarg-list {}
                                     {Item arguments}}
   {-returns          string {}      {Item return value}}
   {-parseflags       extbool true   {Whether to enable flag parsing or
                                      delegate it to the user}}
   {-checktype        extbool false  {Whether to enable argument run-time
                                      type checking for this item}}
   {-keywords         list {}        {Item keywords}}
   {-examples         string {}      {Item usage examples}}
   {-assumptions      string {}      {Item assumptions}}
   {-resources        string {}      {Item required resources}}
   {-limitations      string {}      {Item limitations}}
   {-effects          string {}      {Item effects}}
   {-details          string {}      {Item implementation details}}
   {-references       string {}      {Item references}}
   {-remarks          string {}      {Item remarks}}
   {-todo             string {}      {Item to do list}}
   {-history          string {}      {Item history}}
   {-copyright        string {}      {Item copyright notice}}
   {-init             script {}      {Constructor script used to invoke base
                                      class constructors}}
   {-body             script {}      {Item body}}
#end

   * The A<-synopsis> flag is mandatory, except for the constructor and
     destructor, where this flag can not be given: the synopsis for those
     is hardcoded to "Class constructor" or "Class destructor", respectively.
 
   * The order of the A<-parseflags> and A<-checktype> flags is irrelevant
     but must appear before the item name.
 
   * The order of the other flags is irrelevant also but must appear after
     the item name.
 
} -references {
 
   * The K<::Simple::Proc::create> procedure.
 
   * The K<::Simple::Declare::declare-proc> procedure.

   * The K<::Simple::Class::declare-class> procedure.

} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the qualified command name if exported.
    1: the command alias in the format ||{aliasName aliasedCommand}|| if aliased.
    2: the item forward declaration script.
    3: the item declaration and creation script.
    4: the item install script.
    5: whether the item is the class constructor
#end

} -details {

   * This procedure assumes all flags are contained in a single list argument
     following all the non-flags arguments.

} -body {

   ### Add the location for error reporting
   if {$isCommand} {
      set commandOrMethod command
      set kindLocation {class command}
      set procOrMethod proc
   } else {
      set commandOrMethod method
      set kindLocation method
      set procOrMethod method
   }
   set item [string trim [lindex $declaration 0]]
   ::Simple::Error::Priv::push-location $kindLocation $item

   ### Assert the item name is unqualified
   foreach {tmp emptyNamespace unqualifiedItem}\
      [::Simple::Proc::Priv::parse-qualify-split\
      $item $commandOrMethod 2 0] break
   if {[string compare $emptyNamespace {}]} {
      ::Simple::Error::throw\
         ::Simple::BAD-ITEM "unqualified $commandOrMethod name" $item
   }
   set qualifiedClass ${namespace}${class}
   set qualifiedItem ${qualifiedClass}::$unqualifiedItem

   ### Set defaults
   set version 1.0
   set parseflags 1
   set checktype [set ::Simple::Proc::%Options%(-checktype)]

   ### Parse flags

   # Common flags for all items
   set flags [list -version trimmed\
      -overview paragraphs\
      -access trimmed\
      -arguments list\
      -returns paragraphs\
      -parseflags trimmed\
      -checktype trimmed\
      -keywords list\
      -examples lines\
      -assumptions paragraphs\
      -resources paragraphs\
      -limitations paragraphs\
      -effects paragraphs\
      -details paragraphs\
      -references paragraphs\
      -remarks paragraphs\
      -todo paragraphs\
      -history paragraphs\
      -copyright paragraphs\
      -body untouched]

   ### Identify whether it is command, method, constructor or destructor
   set isConstructor 0
   set isDestructor 0
   set isMethod 0

   # Method, constructor or destructor
   if {!$isCommand} {
      set alias {}
      switch -exact -- $unqualifiedItem {

         constructor {
         # Constructor

            # Add the A<-init> flag
            set isConstructor 1
            lappend flags -init untouched
         }

         destructor {
         # Destructor

            set isDestructor 1
         }

         default {
         # Method

            set isMethod 1
         }
      }

   # Command
   } else {

      # Assert the command is not named "constructor" nor "destrructor"
      if {![string compare $unqualifiedItem constructor] ||\
         ![string compare $unqualifiedItem destructor]} {

         ::Simple::Error::throw\
            ::Simple::Class::CLASS-COMMAND-CONSTRUCTOR-OR-DESTRUCTOR\
            $unqualifiedItem
      }

      # Add the A<-project> and A<-alias> flags
      lappend flags -project trimmed -alias trimmed
   }

   # Assert abstract classes have no constructor nor destructor
   if {$isAbstract && ($isConstructor || $isDestructor)} {
      ::Simple::Error::throw\
         ::Simple::Class::CONSTRUCTOR-OR-DESTRUCTOR-FOR-ABSTRACT-CLASS\
         $unqualifiedItem $qualifiedClass
   }

   # Command or method
   if {$isCommand || $isMethod} {

      # Add the A<-synopsis> flag
      lappend flags -synopsis line
   }

   # Parse flags
   ::Simple::Priv::parse-flags $flags [list] [lrange $declaration 1 end] 0 0

   # Ensure A<-parseflags> and A<-checktype> are booleans
   ::Simple::Proc::Priv::uplevel-ensure-booleans

   ### Handle A<-synopsis>
   if {$isCommand || $isMethod} {
      ::Simple::Declare::Priv::assert-synopsis-given
   } else {

      # Synopsis hardcoded: "Class constructor" or "Class destructor"
      set synopsis "Class $unqualifiedItem"
   }

   ### Handle A<-access>
   # -access given
   if {[::Simple::Argument::information flaggiven -access]} {

      # Assert the given access is correct
      set choices {public private}
      if {$isCommand} {
         lappend choices exported
      } else {
         lappend choices protected
      }
      ::Simple::Variable::Priv::assert-type-conformation\
         argument 0 -access choice $access 0 $choices

   # -access not given for destructor
   } elseif {$isDestructor} {

      # Default access for destructor is protected
      set access protected

   # -access not given for command, method or constructor
   } else {

      # Default access for commands, methods and constructors is public
      set access public
   }

   ### Abstract class method
   if {$isMethod && $isAbstract} {

      ### A<-body> given
      if {[::Simple::Argument::information flaggiven -body]} {

         ### Assert the method is not public
         if {![string compare $access public]} {
            ::Simple::Error::throw\
               ::Simple::Class::PUBLIC-METHOD-WITH-BODY-FOR-ABSTRACT-CLASS\
               $unqualifiedItem $qualifiedClass
         }

      ### A<-body> not given
      } else {

         ### Virtual method: provide a body which throws an error
         set body "::Simple::Error::throw\
               ::Simple::Class::VIRTUAL-METHOD-CALLED $unqualifiedItem"
      }
   }

   ### First section factored out
   set procCommand ::itcl::body
   set fullNamespace $namespace$class
   ::Simple::Declare::Priv::uplevel-declare-proc-1
   if {$isExported} {
      set exportedCommand $qualifiedItem
   } else {
      set exportedCommand {}
   }

   ### Compose the item forward declaration
   #   Notice that the used arguments list in this forward declaration is
   #   simply "args", although the actual arguments list is then supplied
   #   in the item declaration

   #   Constructor
   if {$isConstructor} {

      # Non-empty constructor initialization script
      set constructorBodyModified 0
      if {[string compare $constructorInitializationScript {}]} {

         # Get the constructor body
         set constructorBody [lindex $declarationAndCreationScript end]

         # Add the constructor initialization script to the constructor body
         set constructorBodyModified 1
         ::Simple::Script::Header::add\
            constructorBody %PRIV%-declare-class 100 $constructorInitializationScript
      }

      # A<-init> given
      if {[string compare $init {}]} {

         # Constructor body not modified
         if {!$constructorBodyModified} {

            # Get the constructor body
            set constructorBody [lindex $declarationAndCreationScript end]
         }

         # The constructor has a parse arguments header
         foreach {tag priority header} [lindex [::Simple::Script::Header::get\
            $constructorBody %PRIV%-parse-arguments*] 0] break
         if {[info exists tag]} {

            # Delete the constructor parse arguments header
            set constructorBodyModified 1
            ::Simple::Script::Header::delete constructorBody\
               %PRIV%-parse-arguments*

            # Add the parse arguments header to the A<-init> script
            ::Simple::Script::Header::add init $tag $priority $header
         }
      }

      # Constructor body modified
      if {$constructorBodyModified} {

         # Recreate the constructor declaration and creation
         # script with the modified constructor body
         set declarationAndCreationScript [concat [lrange\
            $declarationAndCreationScript 0 end-1] [list $constructorBody]]
      }

      # Set the actual forward constructor declaration
      set forwardDeclarationScript "$access constructor args [list $init] {}"

   # Command, method or destructor
   } else {

      # If access is exported, change it to public for the forward declaration
      if {![string compare exported $access]} {
         set realAccess public
      } else {
         set realAccess $access
      }
      set forwardDeclarationScript\
         "$realAccess $procOrMethod [list $unqualifiedItem] args {}"

      # Subitem (subcommand or submethod)
      if {[llength $unqualifiedItem] == 2} {

         # Store the base items array in the previous stack frame for
         # persistence across multiple invokations to this procedure
         set baseItem [lindex $unqualifiedItem 0]
         set subcommand [lindex $unqualifiedItem 1]
         upvar baseItems baseItems

         # The base item exists
         if {[info exists baseItems($baseItem)]} {

            # Assert the base item is a command or method as the current item
            foreach {baseProcCommand firstSubcommand baseAccess}\
               $baseItems($baseItem) break
            if {[string compare $baseProcCommand $procOrMethod]} {
               ::Simple::Error::throw\
                  ::Simple::Class::CLASS-COMMAND-AND-METHOD-SHARE-BASE-COMMAND\
                  $baseItem
            }

            # Assert all subcommands have the same access
            if {[string compare $baseAccess $access]} {

               # Class command
               if {[string compare $procOrMethod method]} {
                  ::Simple::Error::throw\
                     ::Simple::Class::SUBCOMMANDS-WITH-DIFFERENT-ACCESS\
                     "${namespace}${class}::$baseItem $firstSubcommand" "$qualifiedItem"

               # Method
               } else {
                  ::Simple::Error::throw\
                     ::Simple::Class::SUBMETHODS-WITH-DIFFERENT-ACCESS\
                     "$baseItem $firstSubcommand" "$baseItem $subcommand"\
                     ${namespace}${class}
               }
            }

         # The base item does not exist
         } else {

            # Compose the base item forward declaration
            set baseItems($baseItem) [list $procOrMethod $subcommand $access]
            ::Simple::Script::append forwardDeclarationScript\
               "$realAccess $procOrMethod $baseItem args {}"
         }
      }
   }

   ### Remove the location
   ::Simple::Error::Priv::pop-location

   ### Return the qualified command name if exported, the command alias
   ###    pair if aliased, the item forward declaration script, the item
   ###    declaration and creation script, the item installation script and
   ###    whether the item is the class constructor
   return [list $exportedCommand $aliasPair $forwardDeclarationScript\
      $declarationAndCreationScript $installationScript $isConstructor]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::parse-variable-or-similar-declaration

{
   ::Simple::Class::Priv::parse-variable-or-similar-declaration

} -synopsis {

   Parses a class variable or similar item declaration

} -overview {

      This procedure parses a class variable or similar item (attribute or
   [incr Tk] option) declaration in a similar way to the
   K<::Simple::Variable::declare> procedure.

} -arguments {

   { kind             choice {variable attribute itkoption}
                                     {Kind of item}}
   { class            qualifiedname  {Class name}}
   { declaration      list           {Variable or similar declaration}}
   { declaringPackage boolean        {Whether a package is being declared}}

} -remarks {
 
   * The A<declaration> is a variable (or similar) declaration intended to be
     the arguments of a call to K<::Simple::Variable::declare> (or similar) as
     follows:

#verbatim
   { item             variablename   {Item name}}
   {-description      string {}      {Item description}}
   {-access           choice {public protected private}
                                     {Item access mode (not for [incr Tk]
                                      options}}
   {-monitortype      extbool false  {Whether to enable run-time type
                                      monitoring}}
   {-type             type any       {Item type}}
   {-choices          any-list {}    {List of choices for choice type
                                      and its derivations}}
   {-value            any %DEFAULT%  {Item value}}
   {-resourcename     word %DEFAULT% {Resource name ([incr Tk] options only}}
   {-resourceclass    word %DEFAULT% {Resource class ([incr Tk] options only}}
   {-configbody       script {}      {Configuration body (not for variables)}}
#end

   * The order of the flags is irrelevant but must appear after the item name
     and their parameters can not start by an hyphen.
 
} -references {
 
   * The K<::Simple::Variable::declare> procedure.

   * The K<::Simple::Class::declare-class> procedure.

} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the item forward declaration script
    1: the item declaration script
    2: the item creation script
#end

} -body {

   ### Add the location for error reporting
   if {[string compare $kind itkoption]} {
      set kindLocation $kind
   } else {
      set kindLocation option
   }
   set item [string trim [lindex $declaration 0]]
   ::Simple::Error::Priv::push-location $kindLocation $item

   ### Assert the item name is unqualified
   ::Simple::Variable::Priv::assert-type-conformation\
       argument 0 -$kind unqualifiedname $item 0

   # Build the qualified item name
   set unqualifiedItem $item
   set qualifiedItem ${class}::$unqualifiedItem

   ### Set defaults
   set access public
   set monitortype [set ::Simple::Variable::%Options%(-monitortype)]
   set type any
   set configbody {}

   # The default value is <undefined> to mimick [incr Tcl] behaviour
   set value <undefined>

   ### Parse flags

   # Common flags for all items
   set flags [list -description paragraphs\
      -monitortype trimmed\
      -type trimmed\
      -choices list\
      -value untouched]

   # Switch on kind
   switch -exact -- $kind {

      variable {
      # variable 

         # Add the A<-access> flag
         set variableOrAttribute 1
         lappend flags -access trimmed
      }

      attribute {
      # attribute 

         # Add the A<-configbody> and A<-access> flags
         set variableOrAttribute 1
         lappend flags -configbody untouched -access trimmed
      }

      itkoption {
      # [incr Tk] option

         # Add the A<-configbody>, A<-resourcename> and A<-resourceclass> flags
         set variableOrAttribute 0
         lappend flags\
            -configbody untouched -resourcename trimmed -resourceclass trimmed
      }
   }

   # Parse flags
   ::Simple::Priv::parse-flags $flags [list] [lrange $declaration 1 end] 0 0

   ### First section factored out
   # If declaring a package, consider hidden types
   set hidden $declaringPackage 
   set types {}
   ::Simple::Variable::Priv::uplevel-declare-1

   ### Assert array types are not given for [incr Tk] options
   if {$isArrayType && !$variableOrAttribute} {
      ::Simple::Error::throw ::Simple::Proc::ARRAY-NOT-ALLOWED\
         $type option $unqualifiedItem
   }

   ### Handle A<-access>
   if {[::Simple::Argument::information flaggiven -access]} {
      ::Simple::Variable::Priv::assert-type-conformation \
          argument 0 -access choice $access 0 {public protected private}
   }

   ### Compose the item forward declaration script
   # Switch on kind
   switch -exact -- $kind {

      variable {
      # variable 

         set forwardDeclarationScript "$access common $unqualifiedItem"
      }

      attribute {
      # attribute 

         set forwardDeclarationScript "$access variable $unqualifiedItem"
      }

      itkoption {
      # [incr Tk] option

         # Resource name not given
         if {![::Simple::Priv::flag-given -resourcename]} {

            # Use the [incr Tk] option name
            set resourcename $unqualifiedItem
         }

         # Resource class not given
         if {![::Simple::Priv::flag-given -resourceclass]} {

            # Use the capitalised [incr Tk] option name
            set resourceclass [string toupper $resourcename 0 0]
         }

         set forwardDeclarationScript\
            "itk_option define -$unqualifiedItem $resourcename $resourceclass"
      }
   }

   # Add the value to the forward item declaration for no scalar items
   if {!$isArrayType} {
      append forwardDeclarationScript " [list $value]"
   }

   ### Monitor type
   set monitorTypeViaTrace 0
   if {$monitortype} {

      ### No [incr Tk] option
      if {$variableOrAttribute} {

         ### Perform type monitoring via variable trace
         set monitorTypeViaTrace 1

      ### [incr Tk] option
      } else {

         ### Perform type monitoring at the [incr Tk]
         ###    option configuration body
         ::Simple::Script::Header::add configbody %PRIV%-declare-class 100\
            "::Simple::Variable::Priv::assert-type-conformation\
            option 0 -$unqualifiedItem $type\
            \$itk_option(-$unqualifiedItem) 0 [list $choices]"
      }
   }

   ### Handle A<-configbody>
   if {[::Simple::Priv::flag-given -configbody]} {

      ### Assert A<-configbody> is given for [incr Tk] options
      ###    or public attributes only
      if {![string compare $kind attribute] &&\
         [string compare $access public]} {

         ::Simple::Error::throw\
            ::Simple::Class::CONFIGBODY-FOR-NON-PUBLIC-ATTRIBUTE\
            $unqualifiedItem
      }

      ### Assert A<-configbody> is not given for array attributes
      if {[string match *-array $type]} {
         ::Simple::Error::throw\
            ::Simple::Class::CONFIGBODY-FOR-ARRAY-ATTRIBUTE $unqualifiedItem
      }
   }

   ### Non-empty configuration body
   if {[string compare $configbody {}]} {

      ### Add the attribute or [incr Tk] option configuration body
      ###    to the forward item declaration script
      append forwardDeclarationScript " [list $configbody]"
   }

   ### Compose the item declaration script
   set internalRepWithoutDescription [list $type $choices]
   if {$variableOrAttribute} {
      lappend internalRepWithoutDescription $access $valueGiven $value
   } else {
      lappend internalRepWithoutDescription $resourcename $resourceclass\
         $valueGiven $value $monitortype
   }
   set declarationScript "::Simple::Variable::Priv::do-declare\
      $kind ${class}:: $unqualifiedItem [list $description]\
      [list $internalRepWithoutDescription]"

   ### Compose the item creation script
   set creationScript "::Simple::Variable::Priv::do-create\
      $qualifiedItem $monitorTypeViaTrace $valueGiven $type [list $value]"

   ### Remove the location
   ::Simple::Error::Priv::pop-location

   ### Return the item forward declaration, the item initialization
   ###    and the declaration scripts
   return [list $forwardDeclarationScript $declarationScript $creationScript]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::parse-itkcomponent-declaration

{
   ::Simple::Class::Priv::parse-itkcomponent-declaration

} -synopsis {

   Parses an [incr Tk] component declaration.

} -overview {

      This procedure parses an [incr Tk] component declaration.  It takes into
   account ignored, kept and renamed options. 
   
} -limitations {

   * The K<usual> clause is not supported.

} -arguments {

   { class            qualifiedname  {Class name}}
   { declaration      list           {[incr Tk] component declaration}}

} -remarks {
 
   * The A<declaration> is a [incr Tk] component declaration as follows:

#verbatim
   { component        unqualifiedname
                                     {Component name}}
   {-description      string {}      {Component description}}
   {-access           choice {public protected private}
                                     {Component access mode}}
   {-ignore           flag-list %DEFAULT%
                                     {Options to ignore for the megawidget}}
   {-keep             flag-list %DEFAULT%
                                     {Options to keep for the megawidget}}
   {-rename           list-list %DEFAULT%
                                     {Options to rename for the megawidget}}
   {-resourceclass    word Button    {Resource class}}
   {-body             script {}      {Component body}}
#end

   * The order of the flags is irrelevant but must appear after the item name
     and their parameters can not start by an hyphen.

} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the [incr Tk] component name
    1: the [incr Tk] component declaration script
    2: the [incr Tk] component creation script
#end

} -details {

   * This procedure assumes all flags are contained in a single list argument
     following all the non-flags arguments.

} -body {

   ### Assert the error code is unqualified
   set component [string trim [lindex $declaration 0]]
   ::Simple::Type::assert-is\
      unqualifiedname $component {} {unqualified component name}

   ### Add the location for error reporting
   ::Simple::Error::Priv::push-location component $component

   ### Set defaults
   set access public
   set body {}
   set resourceclass Button

   ### Parse flags
   # Parse flags
   ::Simple::Priv::parse-flags\
      [list -description paragraphs\
      -access trimmed\
      -ignore list\
      -keep list\
      -rename list\
      -resourceclass trimmed\
      -body untouched]\
      [list] [lrange $declaration 1 end] 0 0

   ### Handle A<-access>

   # Private or protected
   if {[string compare $access public]} {

      # -private or -protected flag
      set accessWithHyphen -$access

   # Public
   } else {

      # No flag needed for public access
      set accessWithHyphen {}
   }

   ### Handle A<-ignore>, A<-keep> and A<-rename>
   set optionsScript {}
   foreach option {ignore keep rename} {
      if {[::Simple::Priv::flag-given -$option]} {
         foreach item [set $option] {
            ::Simple::Script::append optionsScript "$option $item"
         }
      }
   }
   set renamedOptions $rename
   set rename [list]
   foreach renamedOption $renamedOptions {
      lappend rename [lrange $renamedOption 0 end]
   }

   ### Compose the [incr Tk] component declaration script
   set declarationScript "::Simple::Variable::Priv::do-declare\
      itkcomponent ${class}:: $component [list $description]\
      [list [list $access $ignore $keep $rename $resourceclass]]"

   ### Compose the [incr Tk] component creation script
   set creationScript "itk_component add $accessWithHyphen -- $component\
      [list $body] [list $optionsScript]"

   ### Remove the location
   ::Simple::Error::Priv::pop-location

   ### Return the [incr Tk] component name, declaration and creation scripts
   return [list $component $declarationScript $creationScript]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::assert-extends-itkarchetype

{
   ::Simple::Class::Priv::assert-extends-itkarchetype

} -synopsis {

   Asserts a class extends K<::itk::Archetype>.

} -details {

      This procedure asserts that a class extends K<::itk::Archetype>.  This
   is donde by assessign whether its A<-extends> list directly includes
   K<::itk::Archetype>, K<::itk::Widget> or K<::itk::Toplevel> (these last two
   classes inherit from K<::itk::Archetype>).  Otherwise, the inheritance list
   is transversed looking for a K<::itk::Archetype> inheritance.  This two
   step algorithm is required because K<::itk::Archetype>, K<::itk::Widget>
   and K<::itk::Toplevel> are not full-fledged [incr Tcl] classes (at least
   they do not support the K<info heritage> command) unless at least one
   object of those classes has already been created.

} -arguments {

   { class            name           {Class name}}
   { extends          qualifiedname-list      
                                     {List of classes from which this one
                                      inherits}}
   { optionOrComponent
                      choice {option component}
                                     {Option or component}}
} -returns {

   The empty string.

} -body {

   ### The list of extended classes does not include K<::itk::Archetype>
   if {[lsearch -exact $extends ::itk::Archetype] == -1 &&\
      [lsearch -exact $extends ::itk::Widget] == -1 &&\
      [lsearch -exact $extends ::itk::Toplevel] == -1} {

      ### Loop over extended classes
      set found 0
      foreach extendedClass $extends {

         ### Get the full list of classes this extended class extends
         set extendedClasses [namespace eval $extendedClass\
            {[uplevel 1 set extendedClass]::info heritage}]

         ### The full list of extended classes includes K<::itk::Archetype>
         if {[lsearch -exact $extendedClasses ::itk::Archetype] != -1} {

            # Short-circuit
            set found 1
            break
         }
      }

      ### Class K<::itk::Archetype> not found in the inheritance list
      if {!$found} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Class::SHALL-EXTEND-ITK-ARCHETYPE\
            $class $optionOrComponent
      }
   }
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::assert-no-objects

{
   ::Simple::Class::Priv::assert-no-objects

} -synopsis {

   Throws an error if a class has more than one object.

} -overview {

      This procedure is invoked by singleton constructors to assert that
   a single object is created.

} -arguments {

   { class            qualifiedname  {Class name}}

} -returns {

   The empty string.

} -body {

   ### Assert the class has a single object
   if {[llength [::itcl::find objects -class $class]] > 1} {
      ::Simple::Error::throw\
         ::Simple::Class::MULTIPLE-OBJECTS-FOR-SINGLETON-CLASS $class
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::assert-no-this-class

{
   ::Simple::Class::Priv::assert-no-this-class

} -synopsis {

   Throws an error if an object of the given abstract class is being created.

} -overview {

      This procedure is invoked by abstract class constructors.  It checks
   whether an object for that abstract class is being created (as opossed to
   an object for a class derived from it) and, is so, throws an error.
      
} -arguments {

   { this             name           {Object name}}
   { abstractClass    qualifiedname  {Abstract class name}}

} -returns {

   The empty string.

} -body {

   ### Assert the object is not being created for the given (abstract) class
   #   The first class in the object inheritance list gives the topmost one
   if {![string compare $abstractClass [lindex [$this info heritage] 0]]} {
      ::Simple::Error::throw\
         ::Simple::Class::OBJECT-FOR-ABSTRACT-CLASS $abstractClass
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::delete-package-objects

{
   ::Simple::Class::Priv::delete-package-objects

} -synopsis {

   Deletes all objects of classes belonging to a package.

} -overview {

      This procedure deletes all objects of classes belonging to a package. 
   It is used by P<SimplePackage> for deleting all package objects upon
   uninstalling a package.

} -arguments {

   { packagePair      nameandversion {Package name and version}}

} -returns {

   The empty string.

} -body {

   ### Loop over package classes
   foreach class $::Simple::Package::PackageClasses($packagePair) {

      ### Delete all class objects
      eval ::itcl::delete object [::itcl::find objects -class $class]
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::Priv::parse-qualify

{
   ::Simple::Class::Priv::parse-qualify

} -synopsis {

   Parses and qualifies a class name.

} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the class namespace
    1: the unqualified class name
    2: the qualified class name
#end

} -arguments {

   { class            name           {Class name}}
   {?levels?          integer 3      {Levels up the stack in which to operate}}

} -body {

   ### Parse, qualify and split the class name
   foreach {qualifiedClass namespace unqualifiedClass}\
      [::Simple::Priv::qualify-split $class $levels] break

   ### Return the class namespace, unqualifed and qualified names
   return [list $namespace $unqualifiedClass $qualifiedClass]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::configure

{
   ::Simple::Class::configure

} -synopsis {

   Configures the package options.

} -arguments {

   {-debug            extbool false  {Whether to create a debug method
                                      for every class}}

} -returns {

   The empty string.

} -parseflags {

   false

} -body {

   ### No arguments
   if {[llength $args] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::AT-LEAST-ONE-FLAG\
         [::Simple::List::enumerate [::Simple::Class::cget]]

   ### Arguments
   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      set arg2 [lindex $args 1]
      switch -exact -- $arg1 {

         -debug {
         ### -debug given

            ### Invalid argument
            if {![::Simple::Type::Priv::is-basic extbool $arg2]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -debug extbool {}
            }

            ### Set the create debug method option
            set ::Simple::Class::%Options%(-debug)\
               [{::Simple::Type::convert boolean} $arg2]
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1\
               [::Simple::List::enumerate [::Simple::Class::cget]]
         }
      }
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Class::cget

{
   ::Simple::Class::cget

} -synopsis {

   Gets the package options.

} -arguments {

   {-debug            boolflag       {Gets whether to create a debug method
                                      for every class}}
} -returns {

   The requested option value or the whole list of options if none specified.

} -parseflags {

   false

} -body {

   ### No arguments
   if {[llength $args] == 0} {

      ### Return the list of options
      return -debug

   ### Arguments
   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      switch -exact -- $arg1 {

         -debug {
         ### -debug given
            set ::Simple::Class::%Options%(-debug)
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1\
               [::Simple::List::enumerate [::Simple::Class::cget]]
         }
      }
   }
}
}
