### ===========================================================================
### -package  : SimpleDeclare-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library packages and procedures declaration.
# -module     : extra
# -api        :
#
#  * K<::Simple::Package::information packageclass> class
#    Returns the declared package a class belongs to.
#
#  * K<::Simple::Package::information metadata> package ?version?
#    Returns a declared package metadata.
#
#  * K<::Simple::Package::information optiondescription> package version flag
#    Returns a declared package option description.
#
#  * K<::Simple::Package::information optiontype> package version flag
#    Returns a declared package option type.
#
#  * K<::Simple::Package::information optionchoices> package version flag
#    Returns a declared package option choices.
#
#  * K<::Simple::Package::information optionvalue> package version flag
#    Returns a declared package option value.
#
#  * K<::Simple::Package::information api> package ?version?
#    Returns a package public API signatures.
#
#  * K<::Simple::Proc::information metadata> procedure
#    Returns a declared procedure metadata.
#
#  * K<::Simple::Proc::information access> procedure
#    Returns a declared procedure access mode.
#
#  * K<::Simple::Proc::information alias> procedure
#    Returns a declared procedure alias.
#
# -history    :
#
#  26-oct-2004   Added A<-error> to K<declare-class>, version 0.5.6
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleDeclare-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Package::information packageclass
# -synopsis   : Returns the declared package a class belongs to.
# -overview   :
#
#     This procedure returns the declared package the given class belongs to,
#  or the empty string if either no declared package owns the given class or
#  the class does not exist.
#
# -returns    : The declared package a class belongs to.
#
proc-ext {
   ::Simple::Package::information packageclass
} {
   { class            name           {Class name}}
} {
   ### Parse and qualify the class name
   set class [string trim $class]
   foreach {qualifiedClass namespace unqualifiedClass}\
      [::Simple::Proc::Priv::parse-qualify-split $class class 2] break

   ### Loop over declared packages
   set answer {}
   foreach declaredPackage [array names ::Simple::Package::DeclaredPackages] {

      ### This package owns the class
      if {[lsearch -exact\
         $::Simple::Package::PackageClasses($declaredPackage)\
         $qualifiedClass] != -1} {

         set answer $declaredPackage
         break
      }
   }

   ### Return the declared package the class belongs to
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information metadata
# -synopsis   : Returns a declared package metadata.
# -overview   :
#
#     This procedure returns a list of pairs suitable to be given as argument
#  to the K<array set> command.  Each element is one of the flags given to
#  K<declare-package> upon package declaration (with no hyphen), but for
#  A<-api> and A<-firsttimeinstall> (which are not stored), the item
#  declarations (A<-error>, A<-type>, A<-class>, A<-variable>, A<-command> and
#  A<-option>; these are obtained using K<::Simple::Package::information
#  errors> or equivalent) as well as the items stored by
#  K<::Simple::Package::declare> (A<requisites>, A<namespaces>,
#  A<exportedcommands> and A<aliases>; these are obtained using specific
#  K<::Simple::Package::information> subcommands).
#
#     Thus, the actual list of flags is the following:  A<-project>,
#  A<-synopsis>, A<-module>, A<-overview>, A<-usage>, A<-keywords>,
#  A<-examples>, A<-paradigm>, A<-assumptions>, A<-resources>,
#  A<-limitations>, A<-effects>, A<-details>, A<-references>, A<-remarks>,
#  A<-todo>, A<-cmid>, A<-sprs>, A<-smrs>, A<-history> and A<-copyright>.
#
#     Notice that no data is returned unless The Simple Development Library
#  A<-storemetadata> option is set when the package was declared.  
#
#     If a particular flag was not given in the package declaration, the empty
#  string is returned as that flag contents.
#
# -returns    : The package metadata as an item/contents pairs list.
#
proc-ext {
   ::Simple::Package::information metadata
} {
   { package          package        {Package name}}
   {?version?         version %DEFAULT%
                                     {Package version}}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### The package metadata was stored
   set answer [list]
   if {[llength $::Simple::Package::PackageMetadata($packagePair)] > 0} {
      foreach flag [list project synopsis module overview usage keywords\
         examples paradigm assumptions resources limitations effects details\
         references remarks todo cmid sprs smrs history copyright] contents\
         $::Simple::Package::PackageMetadata($packagePair) {

         ### Add the metadata to the list
         lappend answer $flag $contents
      }
   }

   ### Return the package metadata
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information optiondescription
# -synopsis   : Returns a declared package option description.
# -overview   :
#
#  This procedure returns the description of a declared package option.
#
# -limitations:
#
#  * The package must be installed.
#
# -returns    : The package option description.
#
proc-ext {
   ::Simple::Package::information optiondescription
} {
   { package          package        {Package name}}
   { version          version        {Package version}}
   { flag             flag           {Option flag}}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Assert the option is a valid option for the package
   set mainNamespace\
      [lindex [{::Simple::Package::information namespaces} $package $version] 0]
   set optionVariable ${mainNamespace}::%Options%($flag)
   if {![::Simple::Variable::information exists $optionVariable]} {
      ::Simple::Error::throw ::Simple::Declare::NON-EXISTING-OPTION\
         $packagePair $flag
   }

   ### Return the package option description
   return [{::Simple::Variable::information description} $optionVariable]
}

### ===========================================================================
### -command  : ::Simple::Package::information optiontype
# -synopsis   : Returns a declared package option type.
# -overview   :
#
#     This procedure returns the type of a declared package option.
#
# -limitations:
#
#  * The package must be installed.
#
# -returns    : The package option type.
#
proc-ext {
   ::Simple::Package::information optiontype
} {
   { package          package        {Package name}}
   { version          version        {Package version}}
   { flag             flag           {Option flag}}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Assert the option is a valid option for the package
   set mainNamespace\
      [lindex [{::Simple::Package::information namespaces} $package $version] 0]
   set optionVariable ${mainNamespace}::%Options%($flag)
   if {![::Simple::Variable::information exists $optionVariable]} {
      ::Simple::Error::throw ::Simple::Declare::NON-EXISTING-OPTION\
         $packagePair $flag
   }

   ### Return the package option type
   return [{::Simple::Variable::information type} $optionVariable]
}

### ===========================================================================
### -command  : ::Simple::Package::information optionchoices
# -synopsis   : Returns a declared package option choices.
# -overview   :
#
#     This procedure returns the choices of a declared package option.
#
# -limitations:
#
#  * The package must be installed.
#
# -returns    : The package option choices.
#
proc-ext {
   ::Simple::Package::information optionchoices
} {
   { package          package        {Package name}}
   { version          version        {Package version}}
   { flag             flag           {Option flag}}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Assert the option is a valid option for the package
   set mainNamespace\
      [lindex [{::Simple::Package::information namespaces} $package $version] 0]
   set optionVariable ${mainNamespace}::%Options%($flag)
   if {![::Simple::Variable::information exists $optionVariable]} {
      ::Simple::Error::throw ::Simple::Declare::NON-EXISTING-OPTION\
         $packagePair $flag
   }

   ### Return the package option choices
   return [{::Simple::Variable::information choices} $optionVariable]
}

### ===========================================================================
### -command  : ::Simple::Package::information optionvalue
# -synopsis   : Returns a declared package option value.
# -overview   :
#
#     This procedure returns the value of a declared package option.
#
# -limitations:
#
#  * The package must be installed.
#
# -returns    : The package option value.
#
proc-ext {
   ::Simple::Package::information optionvalue
} {
   { package          package        {Package name}}
   { version          version        {Package version}}
   { flag             flag           {Option flag}}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Assert the option is a valid option for the package
   set mainNamespace\
      [lindex [{::Simple::Package::information namespaces} $package $version] 0]
   set optionVariable ${mainNamespace}::%Options%($flag)
   if {![::Simple::Variable::information exists $optionVariable]} {
      ::Simple::Error::throw ::Simple::Declare::NON-EXISTING-OPTION\
         $packagePair $flag
   }

   ### Return the package option value
   return [set $optionVariable]
}

### ===========================================================================
### -command  : ::Simple::Package::information api
# -synopsis   : Returns a package public API signatures.
# -overview   :
#
#     This procedure returns a string containing a package public API
#  signatures.  It is mainly intended for assessing a package public API
#  in its test module.
#
#     Package public and exported commands are included, as well as package
#  classes public and exported commands, public constructor, public destructor
#  and public methods.  If existing, aliases are used instead of the original
#  command or method name.
#
# -references :
#
#  * The K<::Simple::Package::information commands> command.
#
#  * The K<::Simple::Class::information commands> command.
#
#  * The K<::Simple::Class::information methods> command.
#
#  * The K<::Simple::Class::information api> command.
#
# -remarks    :
#
#  * If existing, aliases are used instead of the original
#    command or method name.
#
#  * The returned list does not contain base commands.
#
#  * The returned list does not contain commands with a percent sign in its
#    name are those are assumed to be internal to The Simple Development
#    Library.
#
#  * Includes the items in the given package extra package, if any.
#
# -effects    :
#
#  * Requires and installs the given package extra package, if any.
#
# -limitations:
#
#  * The package must be installed.
#
# -returns    : The given package public API signatures as follows:
#
##verbatim
#     Commands:
#        <command 1 API>
#        <command 2 API>
#        ...
#     Class <class 1>
#        (extends <class list>)
#        Class constructor:
#           <class 1 constructor API>
#        Class destructor:
#           <class 1 destructor API>
#        Class commands:
#           <class 1 command 1 API>
#           <class 1 command 2 API>
#           ...
#        Class methods:
#           <class 1 method 1 API>
#           <class 1 method 2 API>
#           ...
#     Class <class 2>
#        ...
##end
#
proc-ext {
   ::Simple::Package::information api
} {
   { package         package         {Package name}}
   {?version?        version %DEFAULT%
                                     {Package version}}
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package $version]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the package is installed
   ::Simple::Package::Priv::assert-package-installed $packagePair

   ### Require and install the extra package, if any
   set packageVersion [lindex $packagePair 1]
   set extraPackagePair [list $package-extra $packageVersion]
   set noExtraPackage [catch {::Simple::Package::require-and-install\
      -exact $package-extra $packageVersion}]

   ### Get the package aliases and exported commands
   set packageExportedCommands\
      $::Simple::Package::PackageExportedCommands($packagePair)
   set packageAliases $::Simple::Package::PackageAliases($packagePair)

   # Add the extra package aliases, if any
   if {!$noExtraPackage} {
      set packageAliases [concat $packageAliases\
         $::Simple::Package::PackageAliases($extraPackagePair)]
   }

   ### Get the package public and exported command APIs
   set commands [{Simple::Package::information commands} -noclasses $package]

   ### There are public or exported commands
   set answer {}
   if {[llength $commands] > 0} {

      ### Loop over the public or exported commands
      set tmp [list]
      foreach command [lsort $commands] {

         ### Add the command API
         lappend tmp [::Simple::Declare::Priv::command-api $command\
            $packageExportedCommands $packageAliases]
      }

      ### Add the command APIs
      append answer Commands:\n
      foreach commandAPI [lsort $tmp] {
         append answer "   $commandAPI\n"
      }
   }

   ### Get the package classes
   set packageClasses $::Simple::Package::PackageClasses($packagePair)

   # Add the extra package classes, if any
   if {!$noExtraPackage} {
      set packageClasses [concat $packageClasses\
         $::Simple::Package::PackageClasses($extraPackagePair)]
   }

   ### Loop over clases
   foreach class $packageClasses {

      ### Add the class name
      append answer "Class $class\n"

      ### Add the class API
      foreach line [split [{::Simple::Class::information api} $class] \n] {
         append answer "   $line\n"
      }
   }

   ### Return the given package public API signatures
   return [string trimright $answer \n]
}

### ===========================================================================
### -command  : ::Simple::Proc::information metadata
# -synopsis   : Returns a declared procedure metadata.
# -overview   :
#
#     This procedure returns a list of pairs suitable to be given as argument
#  to the K<array set> command.  Each element is one of the flags given to
#  K<declare-proc> upon procedure declaration (with no hypen), but for
#  A<-arguments>, A<-access>, A<-alias> and A<-body> (these are obtained using
#  specific K<::Simple::Proc::information> subcommands), and the items stored
#  by K<proc-ext> (A<parseflags> and A<checktype>; these are also obtained
#  using specific K<::Simple::Proc::information> subcommands).
#
#     Thus, the actual list of flags is:  A<-version>, A<-project>,
#  A<-synopsis>, A<-overview>, A<-returns>, A<-keywords>, A<-examples>,
#  A<-assumptions>, A<-resources>, A<-limitations>, A<-effects>, A<-details>,
#  A<-references>, A<-remarks>, A<-todo>, A<-history> and A<-copyright>.
#
#     Notice that no data is returned unless The Simple Development Library
#  A<-storemetadata> option is set when the procedure was declared.  
#
#     If a particular flag was not given in the procedure declaration, the
#  empty string is returned as that flag contents.
#
# -returns    : The declared procedure metadata as an item/contents pairs list.
#
proc-ext {
   ::Simple::Proc::information metadata
} {
   { procedure        commandname    {Procedure name}}
} {
   ### Parse and qualify the procedure name
   set procedure [string trim $procedure]
   foreach {qualifiedProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure procedure 2]\
      break

   ### Assert the procedure was declared
   if {![info exists\
      ${namespace}%ProcData%(%Metadata%,$unqualifiedProcedure)]} {

      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         $procedure {a declared procedure}
   }

   ### The procedure metadata was stored
   set answer [list]
   if {[llength\
      [set ${namespace}%ProcData%(%Metadata%,$unqualifiedProcedure)]] > 0} {

      foreach flag [list version project synopsis overview returns keywords\
         examples assumptions resources limitations effects details\
         references remarks todo history copyright] contents\
         [set ${namespace}%ProcData%(%Metadata%,$unqualifiedProcedure)] {

         ### Add the metadata to the list
         lappend answer $flag $contents
      }
   }

   ### Return the declared procedure metadata
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information access
# -synopsis   : Returns a declared procedure access mode.
# -returns    : The declared procedure access mode.
#
proc-ext {
   ::Simple::Proc::information access
} {
   { procedure        commandname    {Procedure name}}
} {
   ### Parse and qualify the procedure name
   set procedure [string trim $procedure]
   foreach {qualifiedProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure procedure 2]\
      break

   ### Assert the procedure was declared
   if {![info exists\
      ${namespace}%ProcData%(%Metadata%,$unqualifiedProcedure)]} {

      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         $procedure {a declared procedure}
   }

   ### Return the declared procedure access mode
   return [set ${namespace}%ProcData%(%Access%,$unqualifiedProcedure)]
}

### ===========================================================================
### -command  : ::Simple::Proc::information alias
# -synopsis   : Returns a declared procedure alias.
# -returns    : The declared procedure alias.
#
proc-ext {
   ::Simple::Proc::information alias
} {
   { procedure        commandname    {Procedure name}}
} {
   ### Parse and qualify the procedure name
   set procedure [string trim $procedure]
   foreach {qualifiedProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure procedure 2]\
      break

   ### Assert the procedure was declared
   if {![info exists\
      ${namespace}%ProcData%(%Metadata%,$unqualifiedProcedure)]} {

      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         $procedure {a declared procedure}
   }

   ### Return the declared procedure alias
   return [set ${namespace}%ProcData%(%Alias%,$unqualifiedProcedure)]
}

### ===========================================================================
### -command  : ::Simple::Declare::Priv::command-api
# -synopsis   : Returns a command or method API signatures.
# -returns    : The given command or method API signature
# -remarks    :
#
#  * If existing, aliases are used instead of the original
#    command or method name.
#
proc-ext {
   ::Simple::Declare::Priv::command-api
} {
   { item             commandname    {Item name}}
   { exportedcommands qualifiedname-list
                                     {Package qualified exported commands}}
   { aliases          list-list      {Aliases in the format
                                      ||{aliasName aliasedCommand}||}}
   {?PADMarkup?       boolean 0      {Whether to generate PAD markup}}
} {
   # Exported item
   if {[lsearch -exact $exportedcommands $item] != -1} {

      # Use the unqualified name
      regexp -- {^.*::(.*)$} $item -> itemOrAlias

   # Aliased item
   } elseif {[set index [lsearch -glob $aliases "* $item"]] > -1} {

      # Use the alias
      set itemOrAlias [lindex [lindex $aliases $index] 0]

   # Non-exported nor aliased item
   } else {

      # Use the name
      set itemOrAlias $item
   }

   ### Get the item API signature
   set usage [{::Simple::Proc::information usage} $item]
   if {$PADMarkup} {
      set answer ||$itemOrAlias||
      if {[string compare $usage {}]} {
         append answer " ''$usage''"
      }
   } else {
      set answer $itemOrAlias
      if {[string compare $usage {}]} {
         append answer " $usage"
      }
   }

   ### Return the given item API signature
   return $answer
}
}
