### ===========================================================================
### -package  : SimpleDeclare
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library packages and procedures declaration.
# -module     : implementation
# -overview   :
#
#     This package provides a way to declare packages and procedures in an
#  homogeneous, easily parseable, fully introspectable way.  It allows to
#  substitute the comment block which is usually placed on top of each
#  declaration (as this very same overview!) by Tcl-handled text.  
#
#     Both items (packages and procedures) can be declared with a single
#  command (K<declare-package> or K<declare-proc>, respectively).  Every
#  single component needed to either make the item work (such as a procedure
#  argument default value) or to document it (metadata, such as a procedure
#  argument textual description) is provided as arguments to that command. 
#  Both items and components can be fully introspected, but metadata is only
#  available if The Simple Development Library A<-storemetadata> option is set
#  when the item is declared.
#
# -usage      :
#
#     Use K<declare-package> to declare a package.  Declared packages metadata
#  can be obtained via the K<::Simple::Package::information metadata>
#  procedure.  Use K<::Simple::Package::undeclare> to undeclare the package.
#
#     The K<declare-package> command encompasses the funcionality provided by
#  the K<::Simple::Package::declare> command of the P<SimpleDeclare> package,
#  so it is recommended not to directly use the later.
#
#     Use K<declare-proc> to declare a procedure.  Declared procedures
#  metadata given to K<declare-proc> can be obtained via the
#  K<::Simple::Proc::information metadata> procedure.  Use
#  K<::Simple::Proc::delete> to delete the procedure.
#
#     The K<declare-proc> command encompasses the funcionality provided by the
#  K<proc-ext> command of the P<SimpleProc> package, so it is recommended not
#  to use the later.  Further, the A<-command> flag of K<declare-package>
#  allows to declare package commands from within the package declaration, so
#  that K<declare-proc> should only be used for declaration of commands
#  outside packages, should that be needed.
#
# -keywords   : package procedure declaration
# -api        :
#
#  * K<declare-package> package ?-version version? ?-project string?
#       ?-synopsis string? ?-module choice? ?-overview string? ?-usage string?
#       ?-api string? ?-exportedcommands qualifiedname-list? ?-aliases list? 
#       ?-keywords list? ?-examples string? ?-paradigm choice?
#       ?-namespaces namespace-list? ?-error declaration? ?-type declaration?
#       ?-class declaration? ?-variable declaration? ?-command declaration? 
#       ?-option declaration? ?-requisites packagerequisite-list? 
#       ?-assumptions string? ?-resources string? ?-limitations string? 
#       ?-effects string? ?-details string? ?-references string? 
#       ?-remarks string? ?-todo string? ?-cmid string? ?-sprs string? 
#       ?-smrs string? ?-history string? ?-copyright string? ?-install script?
#       ?-uninstall script? ?-firsttimeinstall script?
#    Declares a package.
#
#  * K<declare-proc> procedure ?-version version? ?-project string?
#       ?-synopsis string? ?-overview string? ?-access choice? ?-alias name? 
#       ?-arguments extendedarg-list? ?-returns string? ?-parseflags extbool? 
#       ?-checktype extbool? ?-keywords list? ?-examples string? 
#       ?-assumptions string? ?-resources string? ?-limitations string? 
#       ?-effects string? ?-details string? ?-references string?
#       ?-remarks string? ?-todo string? ?-history string? ?-copyright string?
#       ?-body script?
#    Declares a procedure.
#
#  * K<::Simple::Declare::configure>
#    Configures the package options.
#
#  * K<::Simple::Declare::cget>
#    Gets the package options.
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#  ::Simple::Package::require-and-install SimpleDeclare
#  
#  # Let's store the metadata
#  ::Simple::configure -storemetadata true
#  
#  # Declare a package which declares a procedure
#  declare-package {
#  
#     Package
#  
#  } -synopsis {
#  
#     This is the package synopsis.
#  
#  } -namespaces {
#  
#     ::Namespace
#  
#  } -variable {
#  
#     {
#        ::Namespace::Variable
#     } -description {
#        This is a package variable
#     } -value 50
#  
#  } -error {
#  
#     {
#        ::Namespace::Error
#     } -message {
#        Package error
#     } -explanation {
#        This is a package error.
#     }
#  
#  } -command {
#  
#     {
#  
#        ::Namespace::procedure
#  
#     } -synopsis {
#  
#        This is the procedure synopsis.
#  
#     } -arguments {
#        {-modify     boolflag   {Whether to modify the package variable}}
#        { variable   integer    {New package variable value}}
#     } -effects {
#  
#        * Modifies the package variable if -modify is given.
#  
#     } -body {
#
#        # Must modify the package variable
#        if {[::Simple::Argument::information flaggiven -modify]} {
#  
#           # Modify it
#           set ::Namespace::Variable $variable
#        }
#  
#        # Package variable value too big
#        if {$::Namespace::Variable > 99} {
#  
#           # Throw the package error
#           ::Simple::Error::throw ::Namespace::Error
#  
#        # Correct package variable value
#        } else {
#  
#           # Display it
#           puts "Package variable = $::Namespace::Variable"
#        }
#     }
#  }
#  
#  # Install the package
#  ::Simple::Package::install Package
#  
#  # Invoke the package procedure
#  
#  # This displays the following:
#  #   |Package variable = 50
#  ::Namespace::procedure 77
#  
#  # This displays the following:
#  #   |Package variable = 88
#  ::Namespace::procedure -modify 88
#  
#  # This displays the following:
#  #   |Package error (::Namespace::Error)
#  catch {::Namespace::procedure -modify 120} result
#  puts "$result ($::errorCode)"
#  
#  # Obtain the package metadata
#  # This displays the following:
#  #   |-assumptions = 
#  #   |  -copyright = 
#  #   |    -details = 
#  #   |    -effects = * Modifies the package variable if -modify is given.
#  #   |   -examples = 
#  #   |    -history = 
#  #   |   -keywords = 
#  #   |-limitations = 
#  #   |   -overview = 
#  #   |    -project = 
#  #   | -references = 
#  #   |    -remarks = 
#  #   |  -resources = 
#  #   |    -returns = 
#  #   |   -synopsis = This is the procedure synopsis.
#  #   |       -todo = 
#  #   |    -version = 1.0
#  array set metadata\
#     [::Simple::Proc::information metadata ::Namespace::procedure]
#  foreach element [lsort [array names metadata]] {
#     puts [format {%12s = %s} -$element $metadata($element)]
#  }
#
# -history    :
#
#  10-sep-2001   First public release, version 0.4
#  20-oct-2001   Fixed bugs related to types and typed variables, version 0.4.1
#  26-apr-2002   Fixed bug related to procedure body trimmming, version 0.4.4
#  17-feb-2003   Second public release, version 0.5
#  22-feb-2003   Added A<-command> to K<declare-package>, version 0.5.1
#  01-mar-2003   Added K<information option*> subcommands, version 0.5.2
#  22-mar-2003   Added A<-body> to K<declare-proc>, version 0.5.3
#  30-mar-2003   Extra package, version 0.5.4
#  09-jan-2004   P<SimpleClass> requirement for object packages now\
#                optional, version 0.5.5
#  20-jan-2004   Comments allowed before embedded declarations in\
#                K<declare-package>, version 0.5.6
#  18-apr-2004   Alias can be qualified, version 0.5.7
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleDeclare 1.0 -requisites {

   SimpleProc

} -namespaces {

   ::Simple::Declare
   ::Simple::Declare::Priv

} -exportedcommands {

   ::Simple::Declare::declare-package
   ::Simple::Declare::declare-proc

} -install {

### Create the package variables

### Create the package options

} -uninstall {

### Delete the package variables

### Delete the package options

} -firsttimeinstall {

### Declare the package types

### Declare the package errors

::Simple::Error::declare {
   ::Simple::Declare::REQUIRES-NAMESPACES 
} -message {
   %s requires -namespaces
} -explanation {
   Giving the flag <flag> requires giving the A<-namespaces> flag also
} -corrective {
   Provide the A<-namespaces> flag in the package declaration
}

::Simple::Error::declare {
   ::Simple::Declare::OPTION-ARRAY 
} -message {
   derived array type "%s" for option "%s"
} -explanation {
   Option <option> has a derived array type <type>, but that is not allowed
} -translation {<type> <option>}

::Simple::Error::declare {
   ::Simple::Declare::PRIV-PROC-IN-NON-PRIV-NAMESPACE
} -message {
   private procedure "%s" can't live in non-private namespace "%s"
} -explanation {
   The procedure <procedure> is private because its access mode is "private",
   but it is living in a non-private namespace <namespace> not-matching
   the pattern ||*::Priv*||
} -corrective {
   Either declare the procedure as non-private or place it in a private
   namespace
}

::Simple::Error::declare {
   ::Simple::Declare::NON-PRIV-PROC-IN-PRIV-NAMESPACE
} -message {
   non-private procedure "%s" can't live in private namespace "%s"
} -explanation {
   The procedure <procedure> is non-private because its access mode is not
   "private", but it is living in a private namespace <namespace> matching
   the pattern ||*::Priv*||
} -corrective {
   Either declare the procedure as private or place it in a non-private
   namespace
}

::Simple::Error::declare {
   ::Simple::Declare::EXPORTED-OR-ALIASED-GLOBAL-PROC
} -message {
   %s global namespace procedure "%s"
} -explanation {
   A procedure is trying to be <exported or aliased> from the global
   namespace, but that is not allowed
} -translation {<exported or aliased> <procedure>}

::Simple::Error::declare {
   ::Simple::Declare::NOT-DECLARED-AS-EXPORTED 
} -message {
   package "%s" declaration does not include "%s" as an exported command
} -explanation {
   Exported procedure <procedure> belongs to package <package> (because the
   procedure namespace belongs to the package), but it was not specified as an
   exported command when the package was declared (flag A<-exportedcommands>
   of K<::Simple::Package::declare> or K<declare-package>)
} -corrective {
   Either declare the procedure non-exported, add the command to the
   A<-exportedcommands> list in the package declaration or remove the
   A<-exportedcommands> flag
} -translation {<package> <procedure>}

::Simple::Error::declare {
   ::Simple::Declare::DECLARED-AS-EXPORTED 
} -message {
   package "%s" declaration includes "%s" as an exported command
} -explanation {
   Non-exported procedure <procedure> belongs to package <package> (because
   the procedure namespace belongs to the package), but it was specified
   as an exported command when the package was declared (flag
   A<-exportedcommands> of K<::Simple::Package::declare> or
   K<declare-package>)
} -corrective {
   Either declare the procedure exported, delete the command from the
   A<-exportedcommands> list in the package declaration or remove the
   A<-exportedcommands> flag
} -translation {<package> <procedure>}

::Simple::Error::declare {
   ::Simple::Declare::NOT-DECLARED-AS-ALIAS 
} -message {
   package "%s" declaration does not include "%s" as an alias for command "%s"
} -explanation {
   Procedure <procedure> belongs to package <package> (because the procedure
   namespace belongs to the package) and it is declared with an alias for
   <alias>, but this alias was not specified when the package was declared
   (flag A<-aliases> of K<::Simple::Package::declare> or K<declare-package>)
} -corrective {
   Either declare the procedure non-aliased, add the alias to the
   A<-aliases> list in the package declaration or remove the A<-aliases> flag
} -translation {<package> <alias> <procedure>}

::Simple::Error::declare {
   ::Simple::Declare::DECLARED-AS-ALIAS 
} -message {
   package "%s" declaration includes "%s" as an alias for command "%s"
} -explanation {
   Procedure <procedure> belongs to package <package> (because the procedure
   namespace belongs to the package) and it is declared with no alias,
   but alias <alias> was specified when the package was declared
   (flag A<-aliases> of K<::Simple::Package::declare> or K<declare-package>)
} -corrective {
   Either declare the procedure as aliased, delete the alias from the
   A<-aliases> list in the package declaration or remove the A<-aliases> flag
} -translation {<package> <alias> <procedure>}

::Simple::Error::declare {
   ::Simple::Declare::ALIAS-NAMESPACE-NOT-THAT-OF-THE-COMMAND
} -message {
   alias namespace "%s" is not that of the command "%s"
} -explanation {
   A <command alias> is qualified with a namespace which is not that of the
   <command>, but that is not allowed.
} -corrective {
   Either make the alias unqualified or use the command namespace.
}

::Simple::Error::declare {
   ::Simple::Declare::NON-EXISTING-OPTION 
} -message {
   package "%s" doesn't have an option "%s"
} -explanation {
   Package <package> does not have an option named <bad option>
} -corrective {
   Provide an option from the package list of options
}

::Simple::Error::declare {
   ::Simple::Declare::CLASS-IN-NON-OBJECT-PACKAGE
} -message {
   class declaration in non-object package
} -explanation {
   A class has tried to be declared in a package whose paradigm is not object
} -corrective {
   Either delete the A<-class> flags in the package declaration
   or ensure its A<-paradigm> is "object"
}

::Simple::Error::declare {
   ::Simple::Declare::NO-NON-EXTRA-PACKAGE
} -message {
   non-extra package "%s" is not declared
} -explanation {
   An extra package is being declared, but the non-extra one, <package>, has
   not been declared before
} -corrective {
   Ensure the non-extra package is declared before the extra one
}

::Simple::Error::declare {
   ::Simple::Declare::FLAG-NOT-ALLOWED-EXTRA
} -message {
   flag "%s" not allowed in extra package
} -explanation {
   The flag <flag> is not allowed in extra packages.
}

::Simple::Error::declare {
   ::Simple::Declare::NAMESPACE-NOT-OWNED
} -message {
   class namespace "%s" not owned by package "%s" nor any of its requisites
} -explanation {
   Class <class> in package <package> has tried to be declared in a namespace
   not owned by the package nor any of its direct requisites
} -corrective {
   Either place the class in a different namespace, add the namespace to the
   package namespaces or add the package owning the namespace to the package
   list of requisites.   
}

### ===========================================================================
### -command  : ::Simple::Declare::declare-package
# -synopsis   : Declares a package.
# -access     : exported
# -overview   :
#
#     This procedure provides a way to declare packages.  It acts as an
#  interface to K<::Simple::Package::declare> to which the A<package>,
#  A<-version>, A<-requisites>, A<-namespaces>, A<-exportedcommands>,
#  A<-aliases>, A<-install>, A<-uninstall> and A<-firsttimeinstall> arguments
#  are passed.
#
#  The A<-synopsis> flag is mandatory.  
#
#     The A<-module> flag can be set to either "implementation", "test",
#  "extra" or "demonstration", specifying whether the Tcl package is the
#  actual package implementation, the test package (see X<simpletest>), the
#  extra package or a demonstration package.  For extra packages, several
#  flags are forbidden.  In addition, extra packages classes can live in and
#  aliases can point to the namespaces of the non-extra one plus those of its
#  directly required packages.
#
#     The A<-paradigm> flag can be set to either "procedural" or "object". 
#  Object packages imply the loading of P<SimpleClass>, and allow the
#  declaration of package classes.
#
#     The A<-variable> flag allows to create package variables.  Its argument
#  is a variable declaration intended to be the arguments of a call to
#  K<::Simple::Variable::declare>.  This flag can be given multiple times.
#
#     The A<-option> flag allows to create package options.  Its argument is
#  an option declaration.  The first element is the option flag and the rest
#  is intended to be the arguments of a call to K<::Simple::Variable::declare>
#  but for the variable name.  A value is required for each option, so that
#  option declarations must have a A<-value> flag.  Derived array types
#  are not allowed for options.  Options are internally stored in the package
#  main namespace (that is the first namespace in the list of the package
#  namespaces).  For this reason, giving the A<-option> flag requires the
#  A<-namespaces> flag also.  This flag can be given multiple times.  The
#  package options values are configured and got via the K<configure> and
#  K<cget> package commands, respectively.  These two procedures must
#  currently be created by the user.
#
#     The A<-type> flag allows to declare package types.  Its argument is a
#  type declaration intended to be the arguments of a call to
#  K<::Simple::Type::declare>.  This flag can be given multiple times.
#
#     The A<-error> flag allows to declare package errors.  Its argument is an
#  error declaration intended to be the arguments of a call to
#  K<::Simple::Error::declare>.  This flag can be given multiple times.
#
#     The A<-command> flag allows to declare package commands.  Its
#  argument is a command declaration intended to be the arguments of a call
#  to K<::Simple::Declare::declare-proc>.  This flag can be given multiple
#  times.
#
#     The A<-class> flag allows to declare package classes.  Its argument is a
#  class declaration intended to be the arguments of a call to
#  K<::Simple::Class::declare-class>.  This flag can be given multiple times. 
#  The classes must live in either one of the package namespaces or to in a
#  namespace belonging to one of the package direct requisites.
#
#     The variables, options, types and errors are created in the package
#  install script and deleted in the package uninstall script, so that
#  uninstalling and then reinstalling a package resets them all.  Uninstalling
#  a package also deletes all the objects of the package classes.
#
#     The commands and classes are created in the package first time install
#  script, so that they are deleted upon deleting the corresponding
#  namespaces, which is only done when the package is undeclared.
#  
#     All other arguments are metadata, and are stored or ignored depending on
#  whether The Simple Development Library A<-storemetadata> option is set when
#  the package is declared, but for the A<-api> and A<-firsttimeinstall> flags
#  which are never stored.
#
# -keywords   : package declaration
# -returns    : The empty string.
# -effects    :
#
#  * Creates the package namespaces.
#
#  * Requires the required packages.
#
#  * Exports exported commands from their namespaces.
#
#  * Provides the package.
#
# -remarks    :
#
#  * The A<-namespaces>, A<-exportedcommands>, A<-aliases>, A<-requisites> and
#    A<-option> flags are forbidden for "extra" packages.
#
#  * The A<-aliases> and A<-exportedcommands> are optional but, if given,
#    their contents must exactly match the actual aliases and exported
#    commands within the package, respectively.
#
#  * The order of the flags is irrelevant but their parameters can not start
#    by an hyphen.
#
# -references :
#
#  * The K<::Simple::Package::declare> procedure.
#
# -details    :
#
#  * Package options are stored in a variable named ''%Options%(flag)''
#    (where ''flag'' is the option flag) in the package main namespace.
#
proc-ext -parseflags false {
   ::Simple::Declare::declare-package
} {
   { package          unqualifiedname
                                     {Package name}}
   {-version          version 1.0    {Package version}}
   {-project          string {}      {Package project}}
   {-synopsis         string {}      {Package synopsis}}
   {-module           choice {implementation test extra demonstration}
                                     {Package module}}
   {-overview         string {}      {Package overview}}
   {-usage            string {}      {Package usage message}}
   {-api              string {}      {Package API}}
   {-exportedcommands qualifiedname-list {}
                                     {Package qualified exported commands}}
   {-aliases          list-list {}   {Package aliases in the format
                                      ||{aliasName aliasedCommand}||}}
   {-keywords         list {}        {Package keywords}}
   {-examples         string {}      {Package usage examples}}
   {-paradigm         choice {procedural object}
                                     {Package paradigm}}
   {-namespaces       namespace-list {}
                                     {Package namespaces}}
   {-error            declaration {} {Package error declaration}}
   {-type             declaration {} {Package type declaration}}
   {-class            declaration {} {Package class declaration}}
   {-variable         declaration {} {Package variable declaration}}
   {-command          declaration {} {Package command declaration}}
   {-option           declaration {} {Package option declaration}}
   {-requisites       packagerequisite-list {}       
                                     {Package requisites}}
   {-assumptions      string {}      {Package assumptions}}
   {-resources        string {}      {Package required resources}}
   {-limitations      string {}      {Package limitations}}
   {-effects          string {}      {Package effects}}
   {-details          string {}      {Package implementation details}}
   {-references       string {}      {Package references}}
   {-remarks          string {}      {Package remarks}}
   {-todo             string {}      {Package to do list}}
   {-cmid             string {}      {Package configuration management
                                      identification}}
   {-sprs             string {}      {Package software problem reports}}
   {-smrs             string {}      {Package software modification reports}}
   {-history          string {}      {Package history}}
   {-copyright        string {}      {Package copyright notice}}
   {-install          script {}      {Package install script}}
   {-uninstall        script {}      {Package uninstall script}}
   {-firsttimeinstall script {}      {Package first time install script}}
} {
   ### Parse and verify the package declaration
   set package [string trim $package]
   ::Simple::Error::Priv::push-location package $package 1
   if {[catch {

      ### Set defaults
      set version 1.0
      set paradigm procedural

      ### Parse flags
      set flags [::Simple::Priv::parse-flags\
         [list -version trimmed\
         -project trimmed\
         -synopsis line\
         -module trimmed\
         -overview paragraphs\
         -usage paragraphs\
         -api untouched\
         -exportedcommands list\
         -aliases list\
         -keywords list\
         -examples lines\
         -paradigm trimmed\
         -namespaces list\
         -requisites list\
         -assumptions paragraphs\
         -resources paragraphs\
         -limitations paragraphs\
         -effects paragraphs\
         -details paragraphs\
         -references paragraphs\
         -remarks paragraphs\
         -todo paragraphs\
         -cmid paragraphs\
         -sprs paragraphs\
         -smrs paragraphs\
         -history paragraphs\
         -copyright paragraphs\
         -install untouched\
         -uninstall untouched\
         -firsttimeinstall untouched]\
         [list -error untouched\
         -type untouched\
         -class untouched\
         -variable untouched\
         -command untouched\
         -option untouched] $args 0 1]

      ### Extra package
      set isNotExtra [string compare $module extra]
      if {!$isNotExtra} {

         ### Assert the non-extra package has been declared for extra packages 
         regsub {^(.*)-extra$} $package \\1 nonExtraPackage
         if {![info exists\
            ::Simple::Package::DeclaredPackages($nonExtraPackage\ $version)]} {

            ::Simple::Error::throw\
               ::Simple::Declare::NO-NON-EXTRA-PACKAGE "$nonExtraPackage $version"
         }

         ### Assert no forbidden flags for extra packages have been given
         foreach flag\
            {-namespaces -exportedcommands -aliases -requisites -option} {

            if {[::Simple::Argument::information flaggiven $flag]} {
               ::Simple::Error::throw\
                  ::Simple::Declare::FLAG-NOT-ALLOWED-EXTRA $flag
            }
         }

      } else {
         set nonExtraPackage $package
      }

      ### Handle A<-paradigm>
      if {[::Simple::Argument::information flaggiven -paradigm]} {
         ::Simple::Argument::check-type -paradigm
      }

      ### First section factored out
      ::Simple::Package::Priv::uplevel-declare-1

      # For extra packages, allow to place classes and aliases in
      # the namespaces of the required packages of the non-extra one
      if {!$isNotExtra} {
         set requisitePackages [list]
         foreach requiredPackage [{::Simple::Package::information\
            required} $nonExtraPackage] {

            lappend requisitePackages [lindex $requiredPackage 0]
         }
      }

      ### Handle A<-synopsis>
      ::Simple::Declare::Priv::assert-synopsis-given

      ### Handle A<-module>
      if {[::Simple::Argument::information flaggiven -module]} {
         ::Simple::Argument::check-type -module
      }

      ### Build the first time install, install and uninstall headers
      set installHeader {}
      set uninstallTrailer {}
      set firstTimeInstallHeader {}

      ### Procedural package
      if {[string compare $paradigm object]} {
      
         ### Assert A<-class> is not given
         if {[::Simple::Argument::information flaggiven -class]} {
            ::Simple::Error::throw ::Simple::Declare::CLASS-IN-NON-OBJECT-PACKAGE
         }

      ### Object package
      } else {

         ### Install P<SimpleClass>
         #   Save and restore the location stack
         set locationStack $::Simple::Error::Priv::LocationStack
         ::Simple::Package::require-and-install SimpleClass
         set ::Simple::Error::Priv::LocationStack $locationStack
      }

      ### Handle A<-type>
      # Types need to be created before other items in case those items
      # use the new types. However, on uninstall the reverse is true

      ### Loop over types
      set types [list]
      foreach typeDeclaration $type {

         ### Declare the type
         #      The type is created hidden so that package variables,
         #   commands and other package items can use the class type but
         #   the type is not publicly visible.  The type is then unhidden
         #   upon installing the package.
         ::Simple::Declare::Priv::remove-leading-comments typeDeclaration
         set typeName [string trim [lindex $typeDeclaration 0]]
         ::Simple::Error::Priv::push-location type $typeName
         set typeDeclarationScript\
            [::Simple::Type::Priv::parse-and-verify-declaration\
            $typeName [lrange $typeDeclaration 1 end] 1]
         eval $typeDeclarationScript
         ::Simple::Error::Priv::pop-location

         ### Add the type declaration script to the package install header
         ::Simple::Script::append installHeader $typeDeclarationScript

         ### Add the type to the package types
         lappend types $typeName
      }

      ### Get the list of namespaces the package can use
      #   For extra packages, this list includes those of the non-extra one
      set allowedNamespaces $namespaces
      if {!$isNotExtra} {
         set allowedNamespaces [concat $allowedNamespaces\
            $::Simple::Package::PackageNamespaces($nonExtraPackage\ $version)]
      }

      ### Handle A<-class>
      # Classes need to be created before other items in case those items
      # use the class types. However, on uninstall the reverse is true
      set classes [list]
      set actualExportedCommands [list]
      set actualAliases [list]

      ### Loop over classes
      foreach classDeclaration $class {

         ### Assert the class name is qualified
         ::Simple::Declare::Priv::remove-leading-comments classDeclaration
         set className [string trim [lindex $classDeclaration 0]]
         ::Simple::Type::assert-is\
            qualifiedname $className {} {qualified class name}

         ### Add the location for error reporting
         ::Simple::Error::Priv::push-location class $className

         ### Assert the class namespace either belongs to the package or
         ###    is a namespace of one of its package requisites

         # The class namespace does not belong to this package
         set classNamespace [namespace qualifiers $className]
         if {[lsearch -exact $allowedNamespaces $classNamespace] == -1} {

            # Assert the class name does not match a
            # declared package namespace or class
            catch {unset packageOwns}
            foreach {packageOwns packageOwnsVersion}\
               [{::Simple::Package::information packagenamespace}\
               $className] break
            if {![info exists packageOwns]} {
               foreach {packageOwns packageOwnsVersion}\
                  [{::Simple::Package::information packageclass}\
                  $className] break
            }
            if {[info exists packageOwns]} {
               ::Simple::Error::throw\
                  ::Simple::Package::NAMESPACE-OR-CLASS-OWNED\
                  $className $packageOwns
            }

            # The class namespace does not belong to a required package
            foreach {packageOwns packageOwnsVersion}\
               [{::Simple::Package::information packagenamespace}\
               $classNamespace] break
            if {![info exists packageOwns] ||\
               [lsearch -exact $requisitePackages $packageOwns] == -1} {

               # Throw error
               ::Simple::Error::throw ::Simple::Declare::NAMESPACE-NOT-OWNED\
                  $classNamespace $nonExtraPackage
            }
         }

         ### Parse and verify the class declaration
         foreach {classType classExportedCommands classAliases\
            classErrorCodes classDeclarationAndCreationScript\
            classInstallationScript}\
            [::Simple::Class::Priv::parse-and-verify-declaration\
            $className [lrange $classDeclaration 1 end] 1] break
         if {[llength $classExportedCommands] > 0} {
            lappend actualExportedCommands $classExportedCommands
         }
         if {[llength $classAliases] > 0} {
            set actualAliases [concat $actualAliases $classAliases]
         }

         ### Remove the location
         ::Simple::Error::Priv::pop-location

         ### Add the class declaration and creation script to the
         ###    package first time install header
         ::Simple::Script::append\
            firstTimeInstallHeader $classDeclarationAndCreationScript

         ### Add the class installation script in the
         ###    package install header
         ::Simple::Script::append installHeader $classInstallationScript

         ### Add the class error deletion script to the
         ###    package uninstall trailer
         foreach classErrorCode $classErrorCodes {
            ::Simple::Script::append\
               uninstallTrailer "::Simple::Error::delete $classErrorCode"
         }

         ### Add the class to the package classes
         lappend classes $className

         ### Class type given
         if {[string compare $classType %DEFAULT%]} {

            ### Add the class type to the package types
            lappend types $classType
         }
      }

      ### Handle A<-variable>
      set variables [list]
      if {[::Simple::Argument::information flaggiven -variable]} {

         ### Assert A<-namespaces> has been given or this in an extra package
         if {[llength $namespaces] == 0 && $isNotExtra} {
            ::Simple::Error::throw\
               ::Simple::Declare::REQUIRES-NAMESPACES -variable
         }

         ### Loop over variables
         foreach variableDeclaration $variable {

            ### Assert the variable name is qualified
            ::Simple::Declare::Priv::remove-leading-comments\
               variableDeclaration
            set variableName [string trim [lindex $variableDeclaration 0]]
            ::Simple::Type::assert-is\
               qualifiedname $variableName {} {qualified variable name}

            ### Add the location for error reporting
            ::Simple::Error::Priv::push-location variable $variableName

            ### Assert the variable namespace belongs to the package
            ::Simple::Package::Priv::assert-package-owns-namespace-or-class\
               $packagePair $variableName $allowedNamespaces

            ### Parse and verify the variable declaration
            foreach {variableDeclarationScript variableCreationScript}\
               [::Simple::Variable::Priv::parse-and-verify-declaration\
               $variableName [lrange $variableDeclaration 1 end] 1] break

            ### Remove the location
            ::Simple::Error::Priv::pop-location

            ### Add the variable declaration and creation scripts to the
            ###    package install header
            ::Simple::Script::append installHeader $variableDeclarationScript
            ::Simple::Script::append installHeader $variableCreationScript

            ### Add the variable deletion script to the
            ###    package uninstall trailer
            ::Simple::Script::append\
               uninstallTrailer "::Simple::Variable::delete $variableName"

            ### Add the variable to the package variables
            lappend variables $variableName
         }
      }

      ### Handle A<-option>
      set options [list]
      if {[::Simple::Argument::information flaggiven -option]} {

         ### Get the package main namespace
         if {[llength $namespaces] == 0} {
            ::Simple::Error::throw\
               ::Simple::Declare::REQUIRES-NAMESPACES -option
         }
         set mainNamespace [lindex $namespaces 0]

         ### Loop over options
         foreach optionDeclaration $option {

            ### Assert the flag name is valid
            ::Simple::Declare::Priv::remove-leading-comments optionDeclaration
            set optionFlag [string trim [lindex $optionDeclaration 0]]
            ::Simple::Type::assert-is flag $optionFlag {} {flag name}

            ### Add the location for error reporting
            ::Simple::Error::Priv::push-location option $optionFlag

            ### Assert a value is given for the option
            set optionDeclaration [lrange $optionDeclaration 1 end]
            if {[lsearch -exact $optionDeclaration -value] == -1} {
               ::Simple::Error::throw ::Simple::FLAG-REQUIRED -value
            }

            ### Assert the option type is not a derived array type
            set typeIndex [lsearch -exact $optionDeclaration -type]
            set type [lindex $optionDeclaration [incr typeIndex]]
            if {$typeIndex > -1 && [string match *-array $type]} {
               ::Simple::Error::throw\
                  ::Simple::Declare::OPTION-ARRAY $type $optionFlag
            }

            ### Parse and verify the option declaration
            # The option variable is the element whose index is the option flag
            # of the "%Options%" array in the package main namespace
            set optionVariableName ${mainNamespace}::%Options%($optionFlag)
            foreach {optionDeclarationScript optionCreationScript}\
               [::Simple::Variable::Priv::parse-and-verify-declaration\
               $optionVariableName $optionDeclaration 1] break

            ### Remove the location
            ::Simple::Error::Priv::pop-location

            ### Add the option declaration and creation scripts to the
            ###    package install header
            ::Simple::Script::append installHeader $optionDeclarationScript
            ::Simple::Script::append installHeader $optionCreationScript

            ### Add the option deletion script to the package uninstall trailer
            ::Simple::Script::append\
               uninstallTrailer "::Simple::Variable::delete $optionVariableName"

            ### Add the option to the package options
            lappend options $optionFlag
         }
      }

      ### Handle A<-error>
      set errorCodes [list]
      if {[::Simple::Argument::information flaggiven -error]} {

         ### Assert A<-namespaces> has been given or this in an extra package
         if {[llength $namespaces] == 0 && $isNotExtra} {
            ::Simple::Error::throw ::Simple::Declare::REQUIRES-NAMESPACES -error
         }

         ### Loop over errors
         foreach errorDeclaration $error {

            ### Assert the error code is qualified
            ::Simple::Declare::Priv::remove-leading-comments errorDeclaration
            set errorCode [string trim [lindex $errorDeclaration 0]]
            ::Simple::Type::assert-is\
               qualifiedname $errorCode {} {qualified error code}

            ### Add the location for error reporting
            ::Simple::Error::Priv::push-location error $errorCode

            ### Assert the error code namespace belongs to the package
            ::Simple::Package::Priv::assert-package-owns-namespace-or-class\
               $packagePair $errorCode $allowedNamespaces

            ### Parse and verify the error declaration
            set errorDeclarationScript\
               [::Simple::Error::Priv::parse-and-verify-declaration\
               $errorCode [lrange $errorDeclaration 1 end] 0]

            ### Remove the location
            ::Simple::Error::Priv::pop-location

            ### Add the error declaration/deletion scripts to the
            ###    package install/uninstall header/trailer
            ::Simple::Script::append installHeader $errorDeclarationScript
            ::Simple::Script::append\
               uninstallTrailer "::Simple::Error::delete $errorCode"

            ### Add the error to the package errors
            lappend errorCodes $errorCode
         }
      }

      ### Handle A<-command>
      if {[::Simple::Argument::information flaggiven -command]} {

         ### Assert A<-namespaces> has been given or this in an extra package
         if {[llength $namespaces] == 0 && $isNotExtra} {
            ::Simple::Error::throw\
               ::Simple::Declare::REQUIRES-NAMESPACES -command
         }

         ### Loop over commands
         foreach commandDeclaration $command {

            ### Assert the command name is qualified
            ::Simple::Declare::Priv::remove-leading-comments commandDeclaration
            set commandName [string trim [lindex $commandDeclaration 0]]
            ::Simple::Type::assert-is\
               qualifiedname [lindex $commandName 0] {} {qualified command name}

            ### Add the location for error reporting
            ::Simple::Error::Priv::push-location command $commandName

            ### Assert the command namespace belongs to the package
            ::Simple::Package::Priv::assert-package-owns-namespace-or-class\
               $packagePair $commandName $allowedNamespaces

            ### Parse and verify the command declaration
            foreach {isExported aliasPair\
               commandDeclarationAndCreationScript commandInstallationScript}\
               [::Simple::Declare::Priv::parse-and-verify-procedure-declaration\
               $commandName [lrange $commandDeclaration 1 end] 1]\
               break
            if {$isExported} {
               lappend actualExportedCommands $commandName
            }
            if {[string compare $aliasPair {}]} {
               lappend actualAliases $aliasPair
            }

            ### Remove the location
            ::Simple::Error::Priv::pop-location

            ### Add the command declaration and creation script to the
            ###    package first time install header
            ::Simple::Script::append\
               firstTimeInstallHeader $commandDeclarationAndCreationScript

            ### Add the command installation script to the package install header
            ::Simple::Script::append installHeader $commandInstallationScript
         }
      }

      # A<-exportedcommands> given
      if {[::Simple::Argument::information flaggiven -exportedcommands]} {
         set exportedCommandsGiven 1
      } else {

         # Use the actual list of exported commands
         set exportedCommandsGiven 0
         set exportedcommands $actualExportedCommands
      }

      # A<-aliases> given
      if {[::Simple::Argument::information flaggiven -aliases]} {
         set aliasesGiven 1

      } else {

         # Use the actual list of aliases
         set aliasesGiven 0
         set aliases $actualAliases
      }

      ### Second section factored out
      ::Simple::Package::Priv::uplevel-declare-2

      ### Handle A<-exportedcommands>
      # A<-exportedcommands> given
      if {$exportedCommandsGiven} {

         # Assert the declared and actual exported commands match
         foreach exportedCommand $exportedcommands {
            set index [lsearch -exact $actualExportedCommands $exportedCommand]
            if {$index == -1 } {
               ::Simple::Error::throw ::Simple::Declare::DECLARED-AS-EXPORTED\
                  $packagePair $exportedCommand
            }
            set actualExportedCommands\
               [lreplace $actualExportedCommands $index $index]
         }

         # All commands declared as exported are actually exported
         # Is there any command exported not declared as such?
         if {[llength $actualExportedCommands] > 0} {
            ::Simple::Error::throw ::Simple::Declare::NOT-DECLARED-AS-EXPORTED\
               $packagePair [lindex $actualExportedCommands 0]
         }
      }

      ### Handle A<-aliases>
      if {$aliasesGiven} {

         # Assert the declared and actual aliases match
         foreach aliasPair $aliases {
            set index [lsearch -exact $actualAliases $aliasPair]
            if {$index == -1 } {
               foreach {alias aliased} $aliasPair break
               ::Simple::Error::throw ::Simple::Declare::DECLARED-AS-ALIAS\
                  $packagePair $alias $aliased
            }
            set actualAliases [lreplace $actualAliases $index $index]
         }

         # All declared aliases in actual aliases
         # Is there any alias not declared as such?
         if {[llength $actualAliases] > 0} {
               foreach {alias aliased} [lindex $actualAliases 0] break
            ::Simple::Error::throw ::Simple::Declare::NOT-DECLARED-AS-ALIAS\
               $packagePair $alias $aliased
         }
      }

      ### Add the package first time install header to the package
      ###    first time install script
      if {[string compare $firstTimeInstallHeader {}]} {
         ::Simple::Script::Header::add\
            firsttimeinstall %PRIV%-declare-package 100 $firstTimeInstallHeader
      }

      ### Add the package install header to the package install script
      if {[string compare $installHeader {}]} {
         ::Simple::Script::Header::add\
            install %PRIV%-declare-package 100 $installHeader
      }

      ### Add the package uninstall trailer to the package uninstall script
      if {[string compare $uninstallTrailer {}]} {
         ::Simple::Script::Trailer::add uninstall %PRIV%-declare-package 100\
            $uninstallTrailer
      }

   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Compose the package metadata
   #   The A<-api> flag is never stored.  Data items not explicitly
   #   included here are stored by K<::Simple::Package::declare>, except for
   #   A<-firsttimeinstall> which is never stored.
   set metadata [list $project $synopsis $module $overview $usage $keywords\
      $examples $paradigm $assumptions $resources $limitations $effects\
      $details $references $remarks $todo $cmid $sprs $smrs $history\
      $copyright]

   ### Declare the package
   ::Simple::Package::Priv::do-declare\
      $packagePair $requisites $namespaces $exportedcommands\
      $aliases $install $uninstall $firsttimeinstall\
      $variables $options $errorCodes $types $classes $metadata

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Declare::declare-proc
# -synopsis   : Declares a procedure.
# -access     : exported
# -overview   :
#
#     This procedure provides a way to declare other procedures.  It acts as
#  an interface to K<::Simple::Proc::create> to which the A<-parseflags>,
#  A<-checktype>, A<procedure>, A<-arguments> and A<-body> arguments are
#  passed.
#
#  The A<-synopsis> flag is mandatory.
#
#     If A<-access> is not given, it is set to either "private" or "public"
#  depending on whether the namespace is private or not, respectively; private
#  namespaces are those matching the pattern ||*::Priv||.  If given, an error
#  is thrown if a public procedure is placed in a private namespace or
#  viceversa.
#
#     If the A<-access> mode is "exported" and the procedure namespace is not
#  the global one, the procedure must be exported from its namespace and
#  imported into the global one.  If the procedure does not belong to a
#  declared package (because the command namespace does not belong to any
#  declared package), the export and import into the global namespace is
#  performed here; otherwise it is the package declaration responsibility to
#  perform the import and the export, and the procedure must have been
#  specified as an exported command when the package was declared (flag
#  A<-exportedcommands> of K<::Simple::Package::declare> or
#  K<declare-package>; the flag is optional in the later case), otherwise an
#  error is thrown.  If a command was included in the A<-exportedcommands>
#  list of its package declaration but it is not declared as exported here, an
#  error is thrown as well.
#
#     If the A<-alias> flag is given, an alias for the procedure must be
#  created.  If the procedure does not belong to a declared package (because
#  the command namespace does not belong to any declared package), the alias
#  is created here; otherwise it is the package declaration responsibility to
#  create the alias, and the alias must have been specified when the package
#  was declared (flag A<-aliases> of K<::Simple::Package::declare> or
#  K<declare-package>; the flag is optional in the later case), otherwise an
#  error is thrown.  In any case, the alias must be unqualified or qualified,
#  but in the later case the namespace must be that of the command.
#
#     All other arguments are metadata, and are ignored or stored depending on
#  whether The Simple Development Library A<-storemetadata> option is set when
#  the package is declared.
#
# -keywords   : procedure declaration
# -returns    :
#
#  A list with the following elements:
#
##table;delimiter=":";ignore="true"
#   0: the procedure declaration and creation script.
#   1: the procedure install script.
##end
#
# -examples   :
#
#  # A procedure exported into the global namespace
#  declare-proc {
#     ::foo::bar
#  } -synopsis {
#     An example of exported procedure.
#  } -access {
#     exported
#  } -arguments {
#     { int1             integer        {First argument}}
#     {-flag             float 1.0      {A flag}}
#     { int2             integer        {Second argument}}
#     { int3             integer        {Third argument}}
#  } -checktype {
#     true
#  } -body {
#     exported
#     puts "int1      = <$int1>"
#     puts "flag      = <$flag> "
#     puts "int2      = <$int2>"
#     puts "int3      = <$int3>"
#  }
#
# -effects    :
#
#  * Creates the procedure named A<procedure>.
#
#  * If the procedure A<-access> mode is "exported", the procedure namespace
#    is not the global one and the procedure does not belong to a declared
#    package, the procedure is exported from its namespace and imported into
#    the global one.
#
#  * If the A<-alias> flag is given and the procedure does not belong to a
#    declared package, an alias for the procedure is created.
#
# -remarks    :
#
#  * The order of the A<-parseflags> and A<-checktype> flags is irrelevant but
#    must appear before the procedure name.
#
#  * The order of the other flags is irrelevant also but must appear after the
#    procedure name and their parameters can not start by an hyphen.
#
# -references :
#
#  * The K<::Simple::Proc::create> procedure.
#
proc-ext -parseflags false {
   ::Simple::Declare::declare-proc
} {
   { procedure        commandname    {Command name}}
   {-version          version 1.0    {Command version}}
   {-project          string {}      {Command project}}
   {-synopsis         string {}      {Command synopsis}}
   {-overview         string {}      {Command overview}}
   {-access           choice {public private exported}
                                     {Command access mode}}
   {-alias            name %DEFAULT% {Command alias}}
   {-arguments        extendedarg-list {}
                                     {Command arguments}}
   {-returns          string {}      {Command return value}}
   {-parseflags       extbool true   {Whether to enable flag parsing or
                                      delegate it to the user}}
   {-checktype        extbool false  {Whether to enable argument run-time
                                      type checking for this procedure}}
   {-keywords         list {}        {Command keywords}}
   {-examples         string {}      {Command usage examples}}
   {-assumptions      string {}      {Command assumptions}}
   {-resources        string {}      {Command required resources}}
   {-limitations      string {}      {Command limitations}}
   {-effects          string {}      {Command effects}}
   {-details          string {}      {Command implementation details}}
   {-references       string {}      {Command references}}
   {-remarks          string {}      {Command remarks}}
   {-todo             string {}      {Command to do list}}
   {-history          string {}      {Command history}}
   {-copyright        string {}      {Command copyright notice}}
   {-body             script {}      {Command body}}
} {
# fixme: A<-project> only apply to standalone items, not those within
#        programs, packages or classes.
#        mismo con las clases
   ### Parse and verify the procedure declaration
   set procedure [string trim $procedure]
   ::Simple::Error::Priv::push-location command $procedure 1
   if {[catch {
      foreach {tmp tmp declarationAndCreationScript installationScript}\
         [::Simple::Declare::Priv::parse-and-verify-procedure-declaration\
         $procedure $args 0] break
   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Declare, create and install the procedure
   eval $declarationAndCreationScript\n$installationScript
}

### ===========================================================================
### -command  : ::Simple::Declare::Priv::parse-and-verify-procedure-declaration
# -synopsis   : Parses and verifies a procedure declaration.
# -returns    :
#
#  A list with the following elements:
#
##table;delimiter=":";ignore="true"
#   0: whether the procedure is exported.
#   1: procedure alias pair if aliased.
#   2: the procedure declaration and creation script.
#   3: the procedure installation script.
##end
#
proc-ext {
   ::Simple::Declare::Priv::parse-and-verify-procedure-declaration
} {
   { procedure        commandname    {Command name}}
   { actualArgs       list           {Actual command declaration argument
                                      list}}
   { declaringPackage boolean        {Whether a package is being declared}}
} {
   ### Parse and qualify the procedure name
   foreach {qualifiedProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure procedure 3]\
      break

   ### Set defaults
   set version 1.0
   set parseflags 1
   set checktype [set ::Simple::Proc::%Options%(-checktype)]

   ### Parse flags
   ::Simple::Priv::parse-flags\
      [list -version trimmed\
      -project trimmed\
      -synopsis line\
      -overview paragraphs\
      -access trimmed\
      -alias trimmed\
      -arguments list\
      -returns paragraphs\
      -parseflags trimmed\
      -checktype trimmed\
      -keywords list\
      -examples lines\
      -assumptions paragraphs\
      -resources paragraphs\
      -limitations paragraphs\
      -effects paragraphs\
      -details paragraphs\
      -references paragraphs\
      -remarks paragraphs\
      -todo paragraphs\
      -history paragraphs\
      -copyright paragraphs\
      -body untouched]\
      [list] $actualArgs 0 2

   # Ensure A<-parseflags> and A<-checktype> are booleans
   ::Simple::Proc::Priv::uplevel-ensure-booleans

   ### Handle A<-synopsis>
   ::Simple::Declare::Priv::assert-synopsis-given

   ### Handle A<-access>
   # -access given
   if {[::Simple::Argument::information flaggiven -access]} {

      # Assert the type conformation
      ::Simple::Variable::Priv::assert-type-conformation \
          argument 0 -access choice $access 0 {public private exported}

      # The procedure access mode is private
      if {![string compare private $access]} {

         # Assert the procedure lives in a private namespace
         if {![string match *::Priv* $namespace]} {
            ::Simple::Error::throw\
               ::Simple::Declare::PRIV-PROC-IN-NON-PRIV-NAMESPACE\
               $unqualifiedProcedure $namespace
         }

      # The procedure access mode is not private
      } else {

         # Assert the command does not live in a private namespace
         if {[string match *::Priv* $namespace]} {
            ::Simple::Error::throw\
               ::Simple::Declare::NON-PRIV-PROC-IN-PRIV-NAMESPACE\
               $unqualifiedProcedure $namespace
         }
      }

   # -access not given
   } else {

      # The procedure access mode is private or public depending
      # on whether it lives in a private namespace
      if {[string match *::Priv* $namespace]} {
         set access private
      } else {
         set access public
      }
   }

   ### First section factored out
   set procCommand ::proc
   set fullNamespace $namespace
   set isMethod 0
   set item $procedure
   set qualifiedItem $qualifiedProcedure
   set unqualifiedItem $unqualifiedProcedure
   ::Simple::Declare::Priv::uplevel-declare-proc-1

   ### Return whether the procedure is exported, the procedure alias pair if
   ###    aliased, the procedure declaration and creation script and the
   ###    procedure installation script
   return [list $isExported $aliasPair $declarationAndCreationScript\
      $installationScript]
}

### ===========================================================================
### -command  : ::Simple::Declare::Priv::uplevel-declare-proc-1
# -synopsis   : Declares a procedure (first section factored out).
# -remarks    :
#
#  * This is a full K<uplevel> procedure used to factor code out.
#
# -effects    :
#
#  * Modifies V<isExported>, V<aliasPair>, V<declarationAndCreationScript>
#    and V<installationScript>.
#
proc-ext {
   ::Simple::Declare::Priv::uplevel-declare-proc-1
} {
} {
   ### Up one level
   uplevel 1 {

      ### The procedure access mode is exported
      set isExported 0
      if {![string compare exported $access]} {

         ### Assert the command is not exported from the global namespace
         if {![string compare $namespace ::]} {
            ::Simple::Error::throw\
               ::Simple::Declare::EXPORTED-OR-ALIASED-GLOBAL-PROC\
               exported $unqualifiedItem
         }

         ### Flag it as exported
         set isExported 1
      }

      ### There is an alias for the procedure
      set aliasPair {}
      if {[{::Simple::Argument::information flaggiven} -alias]} {

         ### Assert the command does not live in the global namespace
         if {![string compare $namespace ::]} {
            ::Simple::Error::throw\
               ::Simple::Declare::EXPORTED-OR-ALIASED-GLOBAL-PROC\
               aliased $unqualifiedItem
         }

         ### Flag it as aliased
         set aliasPair "$alias $qualifiedItem"

         ### The command does not belong to any package being declared
         if {!$declaringPackage} {

            ### Assert the alias namespace, if qualified,
            ###    matches the command namespace
            if {[string match ::* $alias]} {
               set aliasNamespace [namespace qualifiers $alias]
               if {[string compare $namespace ${aliasNamespace}::]} {
                  ::Simple::Error::throw\
                     ::Simple::Declare::ALIAS-NAMESPACE-NOT-THAT-OF-THE-COMMAND\
                     $aliasNamespace $qualifiedItem
               }
            }
         }
      }

      ### Compose the item declaration and creation script

      ### Compose the item metadata
      #   Data items not explicitly included here are stored by K<proc-ext>
      set metadata [list $version $project $synopsis $overview\
         $returns $keywords $examples $assumptions $resources $limitations\
         $effects $details $references $remarks $todo $history $copyright]

      ### Parse and verify the procedure declaration
      set declarationAndCreationScript\
         [::Simple::Proc::Priv::parse-and-verify-declaration $parseflags\
         $checktype $checktypeGiven $qualifiedItem $arguments $body $access\
         $alias $metadata $declaringPackage $procCommand $isMethod]

      ### Compose the item install script

      ### The command does not belong to any package being declared
      set installationScript {}
      if {!$declaringPackage} {

         ### Item to be exported
         if {$isExported} {

            ### Add the export and import scripts to the
            ###    item installation script
            ::Simple::Script::append installationScript\
               "namespace eval $fullNamespace\
               {namespace export [list $unqualifiedItem]}"
            ::Simple::Script::append installationScript\
               "namespace eval :: {namespace import [list $qualifiedItem]}"
         }

         ### Item to be aliased
         if {[string compare $aliasPair {}]} {

            #### Add the alias creation script to the item installation script
            ::Simple::Script::append installationScript\
               "interp alias {} $alias {} [list $qualifiedItem]"
         }
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Declare::Priv::assert-synopsis-given
# -synopsis   : Asserts V<-synopsis> has been given.
# -returns    : The empty string.
#
proc-ext {
   ::Simple::Declare::Priv::assert-synopsis-given
} {
} {
   if {![uplevel 1 ::Simple::Argument::information flaggiven -synopsis]} {
      ::Simple::Error::throw ::Simple::FLAG-REQUIRED -synopsis
   }
   return {}
}

### ===========================================================================
### -command  : ::Simple::Declare::Priv::remove-leading-comments
# -synopsis   : Removes leading comments from an item declaration.
# -overview   :
#
#     This procedure removes leading comments from an item (command, class,
#  method, ...) declaration.  This is done so that comments can be place at
#  the beginning of these item declarations.
#
# -returns    : The empty string.
# -effects    :
#
#  * Removes leading comments from the item declaration named A<itemName>
#    in the calling scope.
#
proc-ext {
   ::Simple::Declare::Priv::remove-leading-comments
} {
   { itemName         name           {Item declaration name}}
} {
   ### Trim and remove leading comments from the item declaration
   upvar $itemName item
   regexp -expanded {\
      ^(\s*\#[^\n]*)*               # Zero or more comment lines
      \s*(.*)\s*$                   # Trim white space
   } $item -> tmp item

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Declare::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Declare::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Declare::cget
# -synopsis   : Gets the package options.
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Declare::cget {
} {
   return [list]
}

# Arrange for the commands in the extra package living in
# namespaces not owned by this package to load it
set tmp [list ::Simple::Package::require-and-install\
   -exact SimpleDeclare-extra 1.0]
set ::auto_index(::Simple::Package::information\ packageclass) $tmp
set ::auto_index(::Simple::Package::information\ metadata) $tmp
set ::auto_index(::Simple::Package::information\ optiondescription) $tmp
set ::auto_index(::Simple::Package::information\ optiontype) $tmp
set ::auto_index(::Simple::Package::information\ optionchoices) $tmp
set ::auto_index(::Simple::Package::information\ optionvalue) $tmp
set ::auto_index(::Simple::Package::information\ api) $tmp
set ::auto_index(::Simple::Proc::information\ metadata) $tmp
set ::auto_index(::Simple::Proc::information\ access) $tmp
set ::auto_index(::Simple::Proc::information\ alias) $tmp
unset tmp

# Augment the list of subcommands for the K<::Simple::Package::information>
# base command
lappend ::Simple::Package::InformationSubcommands\
   api metadata optionchoices optiondescription optiontype optionvalue\
   packageclass
}
