### ===========================================================================
### -package  : SimpleError-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library error handling.
# -module     : extra
# -api        :
#
#  * K<::Simple::Error::information exists> errorCode
#    Returns whether an error has been declared.
#
#  * K<::Simple::Error::information declared> ?pattern?
#    Returns the list of declared errors.
#
#  * K<::Simple::Error::information message> errorCode
#    Returns an error message format string.
#
#  * K<::Simple::Error::information explanation> errorCode
#    Returns an error explanation.
#
#  * K<::Simple::Error::information corrective> errorCode
#    Returns an error corrective action.
#
#  * K<::Simple::Error::information explained> errorCode
#    Returns an error explained error message.
#
#  * K<::Simple::Error::delete> errorCode
#    Deletes an error.
#
#  * K<::Simple::Error::explain> errorCode
#    Explains an error.
#
#  * K<::Simple::Error::catch-explain> script ?variableName?
#    Catches an error and explains it.
#
# -history    :
#
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  28-oct-2004   Error codes can be unqualified, version 0.5.4
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleError-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Error::information
# -synopsis   : Base command for package K<information> subcommands.
# -access     : private
# -overview   :
#
#     This is the base command for the K<::Simple::Error::information>
#  subcommand procedures.
#
# -arguments  :
#  { args             list           {Subcommand and subcommand arguments}}
#
# -returns    : The subcommand procedure return value.
#
proc ::Simple::Error::information {
   args
} {
   #========= %BASE-COMMAND%

   ### Get the subcommand and full command names
   set subcommand [lindex $args 0]
   set fullCommand ::Simple::Error::information\ $subcommand

   ### Evaluate the subcommand procedure
   ### Error from subcommand procedure
   #   The K<uplevel> here ensures that the call to the base command
   #   is hidden to the command procedure body
   if {[catch {uplevel 1 [list $fullCommand] [lrange $args 1 end]} result]} {

      ### Compose the usage string
      switch -exact -- $subcommand {
         declared {
            set usage ?pattern?
         }
         exists -
         message -
         explained -
         explanation -
         corrective {
            set usage errorCode
         }
         default {
            set usage {}
         }
      }

      ### Process the subcommand procedure error
      ::Simple::Error::Priv::subcommand-error\
         ::Simple::Error::information $result $usage\
         [list declared exists message explained explanation corrective]

   ### No error from subcommand procedure
   } else {

      ### Return the subcommand procedure return value
      return $result
   }
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::subcommand-error
# -synopsis   : Processes a subcommand procedure call error.
# -overview   :
#
#     This procedure is called when a call to a subcommand procedure created
#  via the K<::Simple::Subcommand::create> procedure returns an error.  The
#  error is parsed to intercept the different "wrong number of arguments" Tcl
#  errors as well as the error generated in case the subcommand does not
#  exist; in that cases an error is thrown informing either about the argument
#  list for that subcommand or about the allowed subcommands, respectively. 
#  Otherwise the error is an user error from the subcommand procedure, thus
#  being passed up to the calling scope.
#
# -arguments  :
#  { command          qualifiedname  {Qualified command name}}
#  { result           string         {Error from the subcommand procedure call}}
#  {?usage?           string %DEFAULT%
#                                    {Subcommand argument usage string}}
#  {?subcommands?     word-list %DEFAULT%
#                                    {Subcommand list}}
#
# -returns    : The subcommand user error.
#
proc ::Simple::Error::Priv::subcommand-error {
   command
   result
   {usage %DEFAULT%}
   {subcommands %DEFAULT%}
} {
   ### Get args from the subcommand procedure
   #   K<upvar> is used instead of adding an args argument to this
   #   procedure because it is not possible to discriminate
   #   between an empty args and no args at all
   upvar args args

   ### Get the subcommand
   set subcommand [lindex $args 0]

   ### No subcommand given
   if {[llength $args] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::NO-SUBCOMMAND [info level -1]

   ### Error while compiling
   # Bug 9: bad handling of errors while compiling subcommands
   # Solution: correct the identification of an error while compiling
   } elseif {[string first "\n    while compiling\n" $::errorInfo] > -1} {

      ### Rethrow the subcommand procedure error with original V<errorInfo>
      return -code error\
         -errorinfo $::errorInfo -errorcode $::errorCode $result

   ### Non-existing subcommand
   # Bug 8: bad handling of "invalid command name" errors within subcommands
   # Solution: correct the identification of an invalid call to the base command
   } elseif {[string match "invalid command name \"* $subcommand\"" $result]} {

      ### Throw error
      if {![string compare $subcommands %DEFAULT%]} {
         set subcommands\
            [{::Simple::Subcommand::information subcommands} $command]
      }
      ::Simple::Error::throw ::Simple::BAD-SUBCOMMAND $subcommand\
         [::Simple::List::enumerate [lsort $subcommands]]

# fixme: tcl 8.4a2 gives a single error when not suplying enough arguments to
#        a command or when suplying more arguments that expected 
#           wrong # args: should be "foo arg1 arg2 ..." 
#        instead of the various different errors than previous tcl versions gave
   ### Invalid arguments
   } elseif {[string match {wrong # args: should be "*"} $result] ||
      [string match {called "*" with too many arguments} $result] ||
      [string match {no value given for parameter*} $result] ||
      [string match {no parameter given for flag*} $result]} {

      ### Throw error
      #   If A<usage> has not been given the procedure argument list is got
      #   using K<::Simple::Proc::information> if available or
      #   K<::Simple::Priv::proc-information-usage> otherwise
      if {![string compare $usage %DEFAULT%]} {
         if {[::Simple::Priv::proc-information-exists\
            ::Simple::Proc information]} {

            set usage [{::Simple::Proc::information usage}\
               $command\ $subcommand]
         } else {
            set usage [::Simple::Priv::proc-information-usage\
               $command\ $subcommand]
         }
      }
      if {[string length $usage] == 0} {
         ::Simple::Error::throw ::Simple::BAD-SUBCOMMAND-ARGS-EMPTY\
            [concat [lindex [info level -1] 0] $subcommand]
      } else {
         ::Simple::Error::throw ::Simple::BAD-SUBCOMMAND-ARGS\
            [concat [lindex [info level -1] 0] $subcommand] $usage
      }

   ### User error
   } else {

      ### Rethrow the subcommand procedure error
      return -code error -errorcode $::errorCode $result
   }
}

### ===========================================================================
### -command  : ::Simple::Error::information exists
# -synopsis   : Returns whether an error has been declared.
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : Whether the error has been declared.
#
proc {::Simple::Error::information exists} {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Return whether the error has been declared
   return [info exists ${namespace}%Errors%($errorCodeTail)]
}

### ===========================================================================
### -command  : ::Simple::Error::information declared
# -synopsis   : Returns the list of declared errors.
# -overview   :
#
#     This procedure returns the list of declared errors matching the given
#  pattern.
#
# -arguments  :
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns    : The list of declared errors matching the given pattern.
# -remarks    :
#
#  * Use no pattern to obtain the complete list of declared errors.
#
proc {::Simple::Error::information declared} {
   {pattern {}}
} {
   ### Loop over namespaces or classes with declared errors
   set index 0
   foreach namespaceOrClass\
      [array names ::Simple::Error::NamespacesAndClasses] {

      ### Assume there are no declared errors in the namespace or class
      set noErrors 1

      ### Loop over declared errors
      foreach errorCode [array names ${namespaceOrClass}%Errors%] {

         ### Add the error to the list of declared errors
         set declaredErrors(${namespaceOrClass}$errorCode) {}

         ### There is at least one declared error in the namespace or class
         set noErrors 0
      }

      ### Deleted namespace or no declared errors in it
      if {$noErrors} {

         ### Delete the namespace or class from the list of
         ###    namespaces or classes with declared errors
         unset ::Simple::Error::NamespacesAndClasses($namespaceOrClass)
      }

      ### Next namespace or class
      incr index
   }

   ### Return the list of declared errors
   if {[string compare $pattern {}]} {
      return [array names declaredErrors $pattern]
   } else {
      return [array names declaredErrors]
   }
}

### ===========================================================================
### -command  : ::Simple::Error::information message
# -synopsis   : Returns an error message format string.
# -overview   :
#
#  This procedure returns a declared error message format string.
#
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : The error message format string.
#
proc {::Simple::Error::information message} {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Undeclared error
   if {![info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode
   }

   ### Return the error message format string
   return [lindex [set ${namespace}%Errors%($errorCodeTail)] 0]
}

### ===========================================================================
### -command  : ::Simple::Error::information explanation
# -synopsis   : Returns an error explanation.
# -overview   :
#
#  This procedure returns a declared error explanation.
#
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : The error explanation.
#
proc {::Simple::Error::information explanation} {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Undeclared error
   if {![info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode
   }

   ### Return the error explanation
   return [lindex [set ${namespace}%Errors%($errorCodeTail)] 1]
}

### ===========================================================================
### -command  : ::Simple::Error::information corrective
# -synopsis   : Returns an error corrective action.
# -overview   :
#
#  This procedure returns a declared error corrective action.
#
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : The error corrective action.
#
proc {::Simple::Error::information corrective} {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Undeclared error
   if {![info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode
   }

   ### Return the error corrective action
   return [lindex [set ${namespace}%Errors%($errorCodeTail)] 2]
}

### ===========================================================================
### -command  : ::Simple::Error::information explained
# -synopsis   : Returns an error explained error message.
# -overview   :
#
#  This procedure returns a declared error explained error message.
#
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : The error explained error message.
#
proc {::Simple::Error::information explained} {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Undeclared error
   if {![info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode
   }

   ### Return the error explained error message
   return [lindex [set ${namespace}%Errors%($errorCodeTail)] 3]
}

### ===========================================================================
### -command  : ::Simple::Error::delete
# -synopsis   : Deletes an error.
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : The empty string.
#
proc ::Simple::Error::delete {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Undeclared error
   if {![info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode

   ### Declared error
   } else {

      ### Delete the error
      unset ${namespace}%Errors%($errorCodeTail)
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Error::explain
# -synopsis   : Explains an error.
# -overview   :
#
#     This procedure explains an error previously declared via
#  K<::Simple::Error::declare>.
#
#     If The Simple Development Library A<-storemetadata> option was not set
#  or no error explanation was provided when the error was declared, the empty
#  string is returned; otherwise the full error explanation is returned as a
#  three elements list.
#
#     If the explained error message was provided when the error was declared,
#  it is used.  Otherwise it is composed from the unexplained error message
#  with the format specifiers substituted by the name of the error run-time
#  arguments they represent.  The list of the run-time arguments are extracted
#  from the error explanation, in which they appear as strings between angle
#  brackets.  For example, if the error message is
#
#  >  "no file named %s in directory %s"
#
#  an appropriate error explanation would be
#
#  >  "the file named <file> in directory <directory> could not be found",
#
#  so that the error message included in this procedure return value would be
#
#  >  "no file named <file> in directory <directory>".
#
#     The translation string argument of K<::Simple::Error::declare> is used
#  here when the order in which the run-time arguments appear in the error
#  explanation is different from their order in the error message.  In this
#  case, the translation string is a sequence of items between angle
#  brackets which correspond to the names of the error message run-time
#  arguments.
#
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#
# -returns    : A list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0: the explained error message, or unexplained if the error\
#        explanation is not available.
#     1: the error explanation, or the empty string if not available.
#     2: the corrective action, or the empty string if not available.
##end
#
# -examples   :
#
#  # Let's store the error details
#  ::Simple::configure -storemetadata true
#
#  # Declare an error
#  ::Simple::Error::declare {
#     FILE-NOT-FOUND-1 
#  } -message {
#     no file named %s in directory %s
#  } -explanation {
#     There is no file named <file> in directory <dir>
#  } -corrective {
#     Make sure both the file and directory are correct
#  }
#
#  # Explain the error
#  # This displays the following:
#  #   |FILE-NOT-FOUND-1
#  #   |no file named <file> in directory <dir>
#  #   |There is no file named <file> in directory <dir>.
#  #   |Make sure both the file and directory are correct.
#  foreach line [::Simple::Error::explain FILE-NOT-FOUND-1] {puts $line}
#
#  # Declare another error
#  ::Simple::Error::declare {
#     FILE-NOT-FOUND-2 
#  } -message {
#     directory %s has no file named %s
#  } -explanation {
#     There is no file named <file> in directory <dir>
#  } -corrective {
#     Make sure both the file and directory are correct
#  } {<dir> <file>}
#
#  # Explain the error
#  # This displays the following:
#  #   |FILE-NOT-FOUND-2
#  #   |directory <dir> has no file named <file>
#  #   |There is no file named <file> in directory <dir>.
#  #   |Make sure both the file and directory are correct.
#  foreach line [::Simple::Error::explain FILE-NOT-FOUND-2] {puts $line}
#
#  # Declare another error
#  ::Simple::Error::declare {
#     FILE-NOT-FOUND-3 
#  } -message {
#     directory %s has no file named %s
#  } -explanation {
#     There is no file named <file> nor <file>.BACKUP in directory <dir>
#  } -corrective {
#     Make sure both the file and directory are correct
#  } -explained {
#     directory <dir> has no file named <file> or <file>.BACKUP
#  }
#
#  # Explain the error
#  # This displays the following:
#  #   |FILE-NOT-FOUND-3
#  #   |directory <dir> has no file named <file> or <file>.BACKUP
#  #   |There is no file named <file> in directory <dir>.
#  #   |Make sure both the file and directory are correct.
#  foreach line [::Simple::Error::explain FILE-NOT-FOUND-3] {puts $line}
#
proc ::Simple::Error::explain {
   errorCode
} {
   ### Qualify and split the error code
   foreach {tmp namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Undeclared error
   if {![info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode

   ### Declared error
   } else {

      ### Get the error details
      foreach {message explanation corrective explained translation}\
         [set ${namespace}%Errors%($errorCodeTail)] break

      ### Error explanation available
      if {[string compare $explanation {}]} {

         ### There is no explained error message
         if {![string compare $explained {}]} {

            ### There is no translation string
            if {![string compare $translation {}]} {

               ### Use the error explanation
               set translation $explanation
            }

            ### Compose the translation list from the
            ###    angle-bracketed items in the translation string
            #     The "@" prepended to A<translation> is to allow for
            #  a starting item between brackets
            set translationList [list]
            while {[regexp -- {^[^<]*(.)(<[^>]*>)(.*)$} @$translation\
               -> char item translation]} {
            
               # Skip angle-bracketed items preceed by an upppercase letter
               # as those correspond to The Simple Development Library markup
               if {![regexp -- {^[A-Z]$} $char]} {
                  lappend translationList $item
               }
            }
            ### Format the error message
            set explained [::Simple::Error::Priv::format-error-message\
               $namespace $errorCodeTail $translationList 1]
         }

         ### Set the error explanation
         set answer [list $explained $explanation $corrective]

      ### Error explanation not available
      } else {

         ### Set the error explanation to the unexplained message
         set answer [list $message {} {}]
      }
   }

   ### Return the error explanation
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Error::catch-explain
# -synopsis   : Catches an error and explains it.
# -overview   :
#
#     This procedure is similar to the K<catch> command.  It evaluates the
#  given script catching any error it may throw.  If the script throws a
#  declared error, the V<::errorInfo> variable is modified to include an
#  explanation of the error.  Otherwise, the procedure acts exactly as
#  K<catch> except that the call to this procedure is deleted from
#  V<::errorInfo>.
#
# -arguments  :
#  { script           script         {Script to evaluate}}
#  {?variableName?    variablename %DEFAULT% 
#                                    {Name of the variable in the calling
#                                     scope in which the script evaluation
#                                     result is to be placed}}
#
# -returns    : The result of the evaluated script.
# -effects    :
#
#  * Those of the evaluated script.
#
proc ::Simple::Error::catch-explain {
   script
   {variableName %DEFAULT%}
} {
   ### Evaluate the script
   set return [catch {uplevel 1 $script} result]

   ### Error evaluating the script
   if {$return == 1} {

      ### Declared error
      set errorInfoLines [split $::errorInfo \n]
      if {[::Simple::Error::information exists $::errorCode]} {

         ### Initialize V<::errorInfo> with the error message
         set ::errorInfo [lindex $errorInfoLines 0]\n

         ### Insert the error explanation into V<::errorInfo>
         #   The error explanation is indented with four spaces
         set errorExplanation\
            "===================== Error explanation ======================"
         append errorExplanation "\nError code       : $::errorCode"
         foreach {message explanation corrective}\
            [::Simple::Error::explain $::errorCode] break
         append errorExplanation "\nError message    : $message"
         if {[string compare $explanation {}]} {
            append errorExplanation "\nExplanation      : $explanation"
         } 
         if {[string compare $corrective {}]} {
            append errorExplanation "\nCorrective action: $corrective"
         }
         append errorExplanation\
            "\n================== End of error explanation =================="
         regsub -all -- \n $errorExplanation "\n    " errorExplanation
         append ::errorInfo "    $errorExplanation"

         ### Skip three lines from the beginning of the original V<::errorInfo>
         #   This corresponds to the call to K<::Simple::Error::throw>
         set skipFromBeginning 3

      ### Undeclared error
      } else {

         ### Initialize V<::errorInfo>
         set ::errorInfo {}

         ### Skip no lines from the beginning of the original V<::errorInfo>
         set skipFromBeginning 0
      }

      ### Skip three lines from the end of the original V<::errorInfo>
      #   This corresponds to the K<uplevel> in the evaluation of the
      #   script above
      set skipFromEnd 4

      ### Finish the composition of V<::errorInfo>
      set errorInfoTail [join [lrange $errorInfoLines $skipFromBeginning\
         [expr {[llength $errorInfoLines] - $skipFromEnd}]] \n]
      if {[string compare $errorInfoTail {}]} {
         if {[string compare $::errorInfo {}]} {
            append ::errorInfo \n $errorInfoTail
         } else {
            set ::errorInfo $errorInfoTail
         }
      }
   }

   ### Variable name given
   if {[string compare $variableName %DEFAULT%]} {

      ### Set the evaluation script result in the calling scope variable
      uplevel 1 [list ::set $variableName $result]
   }

   ### Return the result of the evaluated script
   return $return
}
}
