### ===========================================================================
### -package  : SimpleError
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library error handling.
# -module     : implementation
# -overview   :
#
#  This package allows to handle errors.  
#
#     In its simplest form, an error is an ''error code'' and an ''error
#  message''.  The ''error code'' has the form of a Tcl qualified name and is
#  used for indexing the error.  The ''error message'' is a format string in
#  the sense of the K<format> Tcl command used for formatting the actual error
#  message with parameters when thrown.
#
#     Procedures are provided to declare, delete, throw and explain errors,
#  get information about an error, obtain the list of errors and catch errors
#  explaining them.
#
# -usage      :
#
#     Errors are declared via K<::Simple::Error::declare> or embedded within
#  programs (K<declare-program>), packages (K<declare-package>) or classes
#  (K<declare-class>).  The error code and error message are compulsory while
#  the error explanation, corrective action, translation string and explained
#  error message, used by K<::Simple::Error::explain> to explain an error, are
#  optional.
#
#     Declared errors can be thrown via K<::Simple::Error::throw> and
#  explained via K<::Simple::Error::explain>.  The error code of thrown errors
#  (as reported by the V<::errorCode> variable) is the error code (with the
#  form of a qualified name) given to the error upon declaration. 
#  K<::Simple::Error::throw> accepts an A<-extra> argument that, if given,
#  makes the V<::errorCode> of the thrown error a two element list where the
#  first element is the error code of the declared error and the second is the
#  parameter of the A<-extra> flag.
#
#     The K<::Simple::Error::information> procedure provides several
#  subcommands to query information about declared errors:  K<exists>,
#  K<declared>, K<message>, K<explanation>, K<corrective> and K<explained>.
#
#     K<::Simple::Error::catch-explain> is similar to the K<catch> command but
#  explains any declared error the evaluated script may throw.
#
#  Finally, use K<::Simple::Error::delete> to delete a declared error.
#
# -keywords   : error catch explain throw
# -api        :
#
#  * K<::Simple::Error::declare> errorCode ?-message formatstring?
#       ?-explanation string? ?-corrective string? ?-explained string?
#       ?-translation string?
#    Declares an error.
#
#  * K<::Simple::Error::throw> ?-extra string? errorCode args
#    Throws an error.
#
#  * K<::Simple::Error::configure>
#    Configures the package options.
#
#  * K<::Simple::Error::cget>
#    Gets the package options.
#
# -variable   :
#
#  {
#     ::Simple::Error::NamespacesAndClasses
#
#  } -description {
#
#     Array containing for each type its description
#
#  } -type any-array
#
# -examples   :
#
#  # Install the package
#  package require SimpleError
#  package require SimpleError-extra
#
#  # Let's store the error details
#  ::Simple::configure -storemetadata true
#
#  # Declare an error
#  ::Simple::Error::declare {
#     FILE-NOT-FOUND 
#  } -message {
#     no file named %s in directory %s
#  } -explanation {
#     There is no file named <file> in directory <dir>
#  } -corrective {
#     Make sure both the file and directory are correct
#  }
#
#  # The error is now declared
#  # This displays the following:
#  #   |1
#  puts [::Simple::Error::information exists FILE-NOT-FOUND]
#
#  # Get the error correction action
#  # This displays the following:
#  #   |Make sure both the file and directory are correct
#  puts [::Simple::Error::information corrective FILE-NOT-FOUND]
#
#  # Throw the error, then catch and explain it
#  # This displays the following:
#  #   |no file named foo.tcl in directory /usr/home/
#  #   |    ===================== Error explanation ======================
#  #   |    Error code       : ::FILE-NOT-FOUND
#  #   |    Error message    : no file named <file> in directory <dir>
#  #   |    Explanation      : There is no file named <file> in directory <dir>
#  #   |    Corrective action: Make sure both the file and directory are correct
#  #   |    ================== End of error explanation ==================
#  ::Simple::Error::catch-explain {
#     ::Simple::Error::throw FILE-NOT-FOUND foo.tcl /usr/home/
#     } result
#  puts $::errorInfo
#
#  # Delete the error
#  ::Simple::Error::delete FILE-NOT-FOUND
#
#  # The error is not declared anymore
#  # This displays the following:
#  #   |0
#  puts [::Simple::Error::information exists FILE-NOT-FOUND]
#
# -details    :
#
#  * Each declared error internal representation is stored in one element of
#    an array named V<%Errors%> in the namespace pointed to by the error code
#    namespace qualifiers.  The name of the array element is the error code
#    unqualified name.  The array element is a list with the following
#    elements (only the first is compulsory):
#
##table;delimiter=":";ignore="true"
#     0: error message format string
#     1: error explanation
#     2: error corrective action
#     3: error explained error message
#     4: error translation string
##end
#
# -todo       :
#
#  * Provide a K<::Simple::Error::modify> procedure.
#
# -history    :
#
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  09-mar-2003   K<::Simple::Error::explain> returns a list, version 0.5.1
#  22-mar-2003   Added A<-message> to K<::Simple::Error::declare>, version 0.5.2
#  07-sep-2004   Bug 7 correction, version 0.5.3
#  28-oct-2004   Error codes can be unqualified, version 0.5.4
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Create the package namespaces
namespace eval ::Simple::Error {}
namespace eval ::Simple::Error::Priv {}

### Declare the package
#   Utility procedure to declare the package
proc ::Simple::Error::Priv::declare-package {} {
   ::Simple::Package::declare SimpleError 1.0 -requisites {
   
      {SimpleBase 1.0}

   } -namespaces {

      ::Simple::Error
      ::Simple::Error::Priv
   }

   # Delete this procedure
   rename [info level 0] {}
}

### Create the package variables

### Create the package options

### Declare the package types
#   Utility procedure to declare the package types
proc ::Simple::Error::Priv::declare-types {
} {
   # Error code
   ::Simple::Type::declare {
      errorcode
   } -description {
      Error code: ''::Simple::Error::FORMAT-ERROR'', ...
   } -matchingscript {
      return [::Simple::Type::Priv::is-basic name $value]
   }

   # Delete this procedure
   rename [info level 0] {}
}

### Declare the package errors
#   Utility procedure to declare the package errors
proc ::Simple::Error::Priv::declare-errors {} {

   ::Simple::Error::declare {
      ::Simple::Error::FORMAT-ERROR 
   } -message {
      error formatting error "%s" (error message is {%s}, arguments are {%s})
   } -explanation {
      An error ocurred when formatting error <error code>, whose error message
      is <error message>, with run-time arguments <arguments>
   } -corrective {
      Check the error message format and the number of run-time arguments.
      Check in particular that the A<-explanation> contains the same number
      of items between brackets than format specifiers ("%s", "%d", ...) in
      the error message
   }

   ::Simple::Error::declare {
      ::Simple::Error::UNDECLARED-ERROR 
   } -message {
      undeclared error "%s"
   } -explanation {
      Error <error code> has not been declared
   } -corrective {
      Use K<::Simple::Error::declare> to declare the error
   }

   # Delete this procedure
   rename [info level 0] {}
}

### ===========================================================================
### -command  : ::Simple::Error::declare
# -synopsis   : Declares an error.
# -overview   :
#
#  This procedure is used to declare a new error.
#
#  An error is made of six elements:  
#
#     1. the error code, 
#
#     1. the error messsage format string,
#
#     1. the error explanation, 
#
#     1. the corrective action, 
#
#     1. the explained error message and
#  
#     1. the error translation string.  
#
#  The first two are compulsory while the last four are optional.
#
#     The error code must be a Tcl name.  The qualifiers point to the
#  namespace in which the error information will be stored; the namespace must
#  exist; if the error code is unqualified, it is qualified in the calling
#  scope.  The unqualified name is the actual code of the error to be thrown
#  via K<::Simple::Error::throw>.  The error message is a format string as
#  accepted by the K<format> command.  This format string is used by
#  K<::Simple::Error::throw> to format the error message.
#
#     The error explanation, corrective action, explained error message and
#  translation string are used by K<::Simple::Error::explain>.
#
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#  {-message          formatstring {}
#                                    {Error message format string}}
#  {-explanation      string {}      {Error explanation}}
#  {-corrective       string {}      {Corrective action}}
#  {-explained        string {}      {Explained error message}}
#  {-translation      string {}      {Translation string}}
#
# -returns    : The empty string.
# -remarks    :
#
#  * The order of the flags is irrelevant but their parameters can not start
#    by an hyphen.
#
proc ::Simple::Error::declare {
   errorCode
   args
} {
   ### Parse and verify the error declaration
   set errorCode [string trim $errorCode]
   ::Simple::Error::Priv::push-location error $errorCode 1
   if {[catch {
      set errorDeclaration\
         [::Simple::Error::Priv::parse-and-verify-declaration $errorCode $args]
   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Assert the error namespace exists
   foreach {tmp errorCodeNamespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break
   if {[catch {namespace children $errorCodeNamespace}]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING\
         namespace [namespace qualifiers $errorCodeNamespace]
   }

   ### Declare the error
   eval $errorDeclaration

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::parse-and-verify-declaration
# -synopsis   : Parses and verifies an error declaration.
# -arguments  :
#  { errorCode        errorcode      {Error code}}
#  { actualArgs       list           {Actual error declaration argument list}}
#  {?levels?          integer 2      {Levels up the stack used in error
#                                     reporting; use 0 for not referring to
#                                     the called command}}
#
# -returns    : The error declaration script.
#
proc ::Simple::Error::Priv::parse-and-verify-declaration {
   errorCode
   actualArgs
   {levels 2}
} {
   ### Parse flags
   ::Simple::Priv::parse-flags\
      [list -message line\
      -explanation paragraphs\
      -corrective paragraphs\
      -explained line\
      -translation untouched]\
      [list] $actualArgs 0 $levels

   ### Assert A<-message> has been given
   if {![::Simple::Priv::flag-given -message]} {
      ::Simple::Error::throw ::Simple::FLAG-REQUIRED -message
   }

   ### Compose the variable declaration script
   foreach {tmp errorCodeNamespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode 3] break
   set declarationScript [list ::Simple::Error::Priv::do-declare\
      $errorCodeNamespace $errorCodeTail $message\
      $explanation $corrective $explained $translation]

   ### Return the error declaration script
   return $declarationScript
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::do-declare
# -synopsis   : Declares an error.
# -arguments  :
#
#  { errorCodeNamespace
#                     namespace      {Error code namespace}}
#  { errorCodeTail    unqualifiedname
#                                    {Error code tail}}
#  { message          formatstring   {Error message format string}}
#  { explanation      string         {Error explanation}}
#  { corrective       string         {Corrective action}}
#  { explained        string         {Explained error message}}
#  { translation      string         {Translation string}}
#
# -returns    : The empty string.
#
proc ::Simple::Error::Priv::do-declare {
   errorCodeNamespace
   errorCodeTail
   message
   explanation
   corrective
   explained
   translation
} {
   ### Add the namespace to the list of namespaces with declared errors
   set ::Simple::Error::NamespacesAndClasses($errorCodeNamespace) [list]

   ### Compose the error internal representation

   # The error details must be stored
   set errorInternalRepresentation [list $message]
   if {[set ::Simple::%Options%(-storemetadata)]} {

      # Store the error details
      lappend errorInternalRepresentation\
         $explanation $corrective $explained $translation

   # Error details must not be stored
   } else {

      # Do not store the error details
      lappend errorInternalRepresentation {} {} {} {}
   }

   ### Store the error internal representation
   set ${errorCodeNamespace}%Errors%($errorCodeTail)\
      $errorInternalRepresentation

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::transform-format-string
# -synopsis   : Transforms a format string specifiers to the string specifier.
# -overview   :
#
#     This procedure transforms all format specifiers within a format string
#  to be string specifiers (that is, "%s") or to an arbitrary string if given.
#
# -arguments  :
#  { formatString     formatstring   {Format string}}
#  {?substitution?    string %s      {Substitution string}}
#
# -returns    : The transformed format string
#
proc ::Simple::Error::Priv::transform-format-string {
   formatString
   {substitution %s}
} {
   ### Protect the "%%" format specifiers
   regsub -all -- %% $formatString @%%@ formatString

   ### Transform all specifiers to the string specifier
   regsub -all --\
      {%([0-9]+\$)?[-+ 0#]?([0-9]+|\*)?(\.[0-9]+)?[hl]?[duioxXcsfeEgG]}\
      $formatString $substitution formatString

   ### Revert all "%%" format specifiers to its original form
   regsub -all -- @%%@ $formatString %% answer

   ### Return the transformed format string
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Error::throw
# -synopsis   : Throws an error.
# -overview   :
#
#  This procedure throws a declared error.
#
# -arguments  :
#  {-extra            string {}      {Extra data}}
#  { errorCode        errorcode      {Error code}}
#  { args             list           {Error run-time arguments}}
#
# -returns    : None; the error is thrown.
# -effects    :
#
#  * The error is thrown.
#
proc ::Simple::Error::throw {
   args
} {
   ### Parse flags
   set extra %DEFAULT%
   set flags [::Simple::Priv::parse-flags\
      [list -extra untouched]\
      [list] $args -1 1]

   ### Parse remaining arguments
   set remainingArgs [lrange $args [expr {2 * [llength $flags]}] end]
   set nRemainingArgs [llength $remainingArgs]
   if {$nRemainingArgs < 1} {
      ::Simple::Error::throw ::Simple::NO-VALUE-FOR-PARAM errorCode\
         [lindex [info level 0] 0]
   }
   set errorCode [lindex $remainingArgs 0]
   set args [lrange $remainingArgs 1 end]

   ### Qualify and split the error code
   foreach {qualifiedErrorCode namespace errorCodeTail}\
      [::Simple::Priv::qualify-split $errorCode] break

   ### Declared error
   if {[info exists ${namespace}%Errors%($errorCodeTail)]} {

      ### Format the error message
      set formattedErrorMessage [::Simple::Error::Priv::format-error-message\
         $namespace $errorCodeTail $args]

      ### A<-extra> given
      if {[string compare $extra %DEFAULT%]} {

         ### Throw the error (with the extra data)
         return -code error\
            -errorcode [list $qualifiedErrorCode $extra] $formattedErrorMessage

      ### A<-extra> not given
      } else {

         ### Throw the error
         return -code error\
            -errorcode $qualifiedErrorCode $formattedErrorMessage
      }

   ### Undeclared error
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::UNDECLARED-ERROR $errorCode
   }
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::format-error-message
# -synopsis   : Formats an error message.
# -overview   :
#
#  This procedure formats a declared error message.
#
# -arguments  :
#  { namespace        namespace      {Error namespace}}
#  { errorCodeTail    unqualifiedname
#                                    {Error code unqualified name}}
#  { arguments        list           {Error run-time arguments}}
#  {?transform?       boolean 0      {Whether to transform the error message
#                                     format string specifiers to the string
#                                     specifier}}
#
# -returns    : The formatted error message.
#
proc ::Simple::Error::Priv::format-error-message {
   namespace
   errorCodeTail
   arguments
   {transform 0}
} {
   ### Get the error message format string
   set message [lindex [set ${namespace}%Errors%($errorCodeTail)] 0]

   ### Transform the error message format string specifiers
   ###    to the string specifier if so requested
   if {$transform} {
      set message [::Simple::Error::Priv::transform-format-string $message]
   }

   ### Error formatting the error message
   if {[catch {set answer [eval format [list $message] $arguments]}]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Error::FORMAT-ERROR\
         $namespace$errorCodeTail $message $arguments
   }

   ### Return the formatted error message
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::push-location
# -synopsis   : Pushes a location or initializes the location stack.
# -overview   :
#
#     This procedure pushes a location or initializes the location stack
#  (depending on the A<init> argument) used by the
#  K<::Simple::Error::Priv::rationalize-call-stack> procedure.
#
# -arguments  :
#  { kind             word           {Kind of item (procedure, error, ...)}}
#  { name             name           {Item name}}
#  {?init?            boolean 0      {Whether to initialize the location
#                                     stack}}
# -returns    : The empty string.
#
proc ::Simple::Error::Priv::push-location {
   kind
   name
   {init 0}
} {
   ### The location stack shall be initialised
   if {$init && [info exists ::Simple::Error::Priv::LocationStack]} {

      ### Initialise it
      unset ::Simple::Error::Priv::LocationStack
   }

   ### Name given
   if {[string compare $name {}]} {

      ### Add the kind/name pair to the location stack
      lappend ::Simple::Error::Priv::LocationStack "$kind \"$name\""

   ### Name not given
   } else {

      ### Add the kind to the location stack
      lappend ::Simple::Error::Priv::LocationStack $kind
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::pop-location
# -synopsis   : Pops a location from the location stack.
# -overview   :
#
#     This procedure pops a location from the location stack used by the
#  K<::Simple::Error::Priv::rationalize-call-stack> procedure.
#
# -returns    : The empty string.
#
proc ::Simple::Error::Priv::pop-location {
} {
   ### Pop a location from the location stack
   set ::Simple::Error::Priv::LocationStack\
      [lrange $::Simple::Error::Priv::LocationStack 0 end-1]

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::rationalize-call-stack
# -synopsis   : Returns a script which throws a rationalized error.
# -overview   :
#
#     This procedure returns a script which throws an error with the call
#  stack rationalized.  This means that the error V<errorInfo> is set to a
#  message which indicates the stack of items which were being declared when
#  the errors was thrown.  For example, if the error happens in a variable
#  declaration the message is ''(declaring variable "<variable name>")'' and
#  if it happens in an error within a class within a package, then it would
#  set to ''(declaring package "<package name>", class "<class name>", error
#  "<error code>")''.
#
# -returns    : The script which throws a rationalized error.
#
proc ::Simple::Error::Priv::rationalize-call-stack {
} {
   set errorMessage [lindex [::Simple::Error::Priv::split-errorinfo] 0]
   set errorInfo "$errorMessage\n    while declaring\
      [join $::Simple::Error::Priv::LocationStack {, }]"
   return [list return -code error -errorcode $::errorCode\
      -errorinfo $errorInfo $errorMessage]
}

### ===========================================================================
### -command  : ::Simple::Error::Priv::split-errorinfo
# -synopsis   : Splits V<::errorInfo into error message and call stack
# -returns    : A list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0: the error message
#     1: the error call stack
##end
#
proc ::Simple::Error::Priv::split-errorinfo {
} {
   # Take as error message everything up to a line starting
   # with four spaces followed by "while "
   regexp {^(.*)\n(    while .*)$} $::errorInfo -> errorMessage callStack
   return [list $errorMessage $callStack]
}

### ===========================================================================
### -command  : ::Simple::Error::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Error::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Error::cget
# -synopsis   : Gets the package options.
# -returns    :
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Error::cget {
} {
   return [list]
}

### ===========================================================================

# Declare this and previous package errors
::Simple::Priv::declare-errors
::Simple::Error::Priv::declare-errors

### Install the package
package provide SimpleError 1.0
