### -package  : SimpleNamespace
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : 
#
#  The Simple Development Library namespaces and qualified names handling.
#
# -module     : implementation
# -overview   :
#
#     This package provides utilities for namespaces and qualified names
#  handling.  Procedures are provided for qualifying names, obtaining
#  qualified names qualifiers and assessing whether a namespace exists.
#
# -usage      :
#
#      Use K<::Simple::Namespace::qualifiers> to obtain a namespace leading
#   qualifiers, K<::Simple::Namespace::qualify> to qualify a name and
#   K<::Simple::Namespace::information exists> to assess whether a namespace
#   exists.
#
# -keywords   : namespace qualify qualifiers
# -api        :
#
#  * K<::Simple::Namespace::qualifiers> name
#    Returns leading namespace qualifiers.
#
#  * K<::Simple::Namespace::qualify> name ?levels?
#    Qualifies a name up the calling stack.
#
#  * K<::Simple::Namespace::information exists> namespace
#    Returns whether a namespace exists.
#
#  * K<::Simple::Namespace::configure>
#    Configures the package options.
#
#  * K<::Simple::Namespace::cget>
#    Gets the package options.
#
# -history    :
#
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleNamespace 1.0 -requisites {

   {SimplePackage 1.0}

} -namespaces {

   ::Simple::Namespace
   ::Simple::Namespace::Priv

} -install {

### Create the package variables

### Create the package options

} -uninstall {

### Delete the package variables

### Delete the package options

} -firsttimeinstall {

### Declare the package types

### Declare the package errors

### ===========================================================================
### -command  : ::Simple::Namespace::qualifiers
# -synopsis   : Returns leading namespace qualifiers.
# -overview   :
#
#     This procedure returns any leading namespace qualifers for a variable or
#  procedure name.  It is similar to the K<namespace qualifiers> command but
#  works in such a way that for any name the concatenation of the value
#  returned by this procedure and the result of K<namespace tail>
#  equals the original name as follows:
#
##table;delimiter="'";headings="true";ignore="true"
#  Name         ' K<namespace qualifiers> ' This procedure ' K<namespace tail>
#  {}           ' {}                      ' {}             ' {}
#  foo          ' {}                      ' {}             ' foo
#  foo::bar     ' foo                     ' foo::          ' bar
#  foo::::bar   ' foo                     ' foo::::        ' bar
#  ::           ' {}                      ' ::             ' {}
#  ::foo        ' {}                      ' ::             ' foo
#  ::foo::      ' ::foo                   ' ::foo::        ' {}
#  ::foo::bar   ' ::foo                   ' ::foo::        ' bar
#  ::foo::bar:: ' ::foo::bar              ' ::foo::bar::   ' {}
##end
#
# -keywords   : namespace qualify qualifier
# -arguments  :
#  { name             variableorcommandname 
#                                    {Variable or procedure name}}
#
# -returns    : The leading namespace qualifers for the given name.
#
proc ::Simple::Namespace::qualifiers {
   name
} {
   if {[regexp -- {^(.*::).*$} $name -> namespace]} {
      return $namespace
   }
}

### ===========================================================================
### -command  : ::Simple::Namespace::qualify
# -synopsis   : Qualifies a name up the calling stack.
# -overview   :
#
#     This procedure qualifies a variable or procedure name.  It is somewhat
#  similar to the K<auto_qualify> procedure but the namespace is taken from a
#  scope up the calling stack.  By default, one level up the stack is used.
#
# -keywords   : namespace qualify qualifier
# -arguments  :
#  { name             variableorcommandname 
#                                    {Variable or procedure name}}
#  {?levels?          integer 1      {Levels up the stack in which to operate}}
#
# -returns    : The qualified name.
#
# -limitations:
#
#  * This procedure must be called from another procedure.  If the A<levels>
#    optional argument is specified, the corresponding number of levels up the
#    stack are necessary.
#
# -references :
#
#  * The K<auto_qualify> procedure.
#
#  * The K<::Simple::Priv::qualify> procedure.
#
#  * The K<::Simple::Priv::qualify-split> procedure.
#
proc ::Simple::Namespace::qualify {
   name
   {levels 1}
} {
   ### Return the qualified name
   return [::Simple::Priv::qualify $name [incr levels]]
}

### ===========================================================================
### -command  : ::Simple::Namespace::information
# -synopsis   : Base command for package K<information> subcommands.
# -access     : private
# -overview   :
#
#     This is the base command for the K<::Simple::Namespace::information>
#  subcommand procedures.
#
# -arguments  :
#  { args             list           {Subcommand and subcommand arguments}}
#
# -returns    : The subcommand procedure return value.
#
proc ::Simple::Namespace::information {
   args
} {
   #========= %BASE-COMMAND%

   ### Get the subcommand and full command names
   set subcommand [lindex $args 0]
   set fullCommand ::Simple::Namespace::information\ $subcommand

   ### Evaluate the subcommand procedure
   ### Error from subcommand procedure
   #   The K<uplevel> here ensures that the call to the base command
   #   is hidden to the command procedure body
   if {[catch {uplevel 1 [list $fullCommand] [lrange $args 1 end]} result]} {

      ### Compose the correct arguments
      switch -exact -- $subcommand {
         exists {
            set usage namespace
         }
         default {
            set usage {}
         }
      }

      ### Process the subcommand procedure error
      ::Simple::Error::Priv::subcommand-error\
         ::Simple::Namespace::information $result $usage exists

   ### No error from subcommand procedure
   } else {

      ### Return the subcommand procedure return value
      return $result
   }
}

### ===========================================================================
### -command  : ::Simple::Namespace::information exists
# -synopsis   : Returns whether a namespace exists.
# -arguments  :
#  { namespace        namespace      {Namespace}}
#
# -returns    : Whether the namespace exists.
#
proc {::Simple::Namespace::information exists} {
   namespace
} {
   if {[string compare $namespace {}]} {
      set answer [expr {![catch {namespace parent $namespace}]}]
   } else {
      set answer 1
   }

   ### Returns whether the namespace exists.
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Namespace::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Namespace::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Namespace::cget
# -synopsis   : Gets the package options.
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Namespace::cget {
} {
   return [list]
}

}
