### ===========================================================================
### -package  : SimpleOption
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleOption

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library command line options parser.

} -module {

   implementation

} -paradigm {

   procedural

} -overview {

      This package provides a powerful command line options parser.  Features
   include:

   * Each option has a primary flag and an optional secondary flag, so that
     any of them can be used in the command line.

   * The special option "--" is used to collect the arguments at the end of
     the command line not belonging to other options.  In the command line,
     "--" can be used to inhibit the processing of the arguments following it.

   * Each option can optionally have a name.  The name is used in the options
     usage string and other references to the option.

   * Each option has a type to which the option values must conform. Types
     are implemented via the P<SimpleType> package, so a wealth of predefined
     types are available and further types can be defined.

   * Options with numeric types can optionally have a numeric range to which
     the option values must conform.

   * Options with type "choice" or derived from it have a list of allowed
     choices to which the option values must conform.

   * Each option can optionally have a default value.

   * Each option can optionally have a range of values to which the option
     number of values must conform.  By default, each option admits a single
     value.

   * Options can be optional or compulsory (not really ''options'' then, as
     Harald Kirsh, X<clig>'s author, first noted).

   * The presence or absence of certain options can be made optional,
     compulsory or disallowed depending on the presence or absence of other
     options.  For two options, the possibilites are:

     > * ||-a -b||: both -a and -b are compulsory

     > * ||-a [-b]||: -a is compulsory and -b is optional

     > * ||[-a -b]||: both -a and -b are optional, but if one is given the
         other must be also given.

     > * ||[-a [-b\]]||: both -a and -b are optional, but -b can not be given
         unless -a is also present.

     > * ||(-a | -b)||: one of -a or -b must be given.

     > * ||[(-a | -b)]||: both -a and -b are optional, but both can not be
         simultaneously given.

     > Other relations are possible (and are accepted) but are equivalente to
        one of the ones above.  For example, the last relation ||[(-a | -b)]||
        can also be expressed as either ||(-a | [-b])|| or ||([-a] | -b)||.

     > These basic usage relations between two options can be combined to
        yield almost any imaginable usage relations between three or more
        options and beyond.

   * The package automatically builds the options usage string.

      Procedures are provided to declare options, declare their usage
   relations, reset the declared options and their usage relations, parse a
   command line with respect to the declared options and usage relations and
   get the options usage string.  After a command line has been parsed,
   procedures are provided to assess whether a certain option was given
   in the command line, to assess whether it is available (that is, whether it
   has a default value or has been given in the command line), to get the
   value of an available option and to assess whether the value of an
   available option is that option default value.

} -usage {

      Options are declared via the K<::Simple::Option::declare> procedure. 
   The declaration of each option includes its flag or flags, its name, its
   type, its description, its number of values and its values numeric
   range, choices and default value.

      The declared options usage relations is specified via the
   K<::Simple::Option::declare-usage> procedure.  Use
   K<::Simple::Option::reset> to reset the declared options and usage
   information.

      The K<::Simple::Option::information> procedure provides several
   subcommands to query information about the options and the actual command
   line:  K<available>, K<choices>, K<declared>, K<default>, K<defaultgiven>,
   K<description>, K<given>, K<help>, K<isdefault>, K<name>, K<range>,
   K<secondary>, K<type>, K<usage>, K<value> and K<values>.
   
} -keywords {

   option flag parameter {command line}

} -examples {

   # Install the package
   package require SimpleDevLibPackage 1.0
   ::Simple::Package::require-and-install SimpleOption

   # Declare the program options
   ::Simple::Option::declare {-p --processing} -type choice\ 
      -choices {A B C} -description {processing option}
   ::Simple::Option::declare {-v --verbosity} -name verbosity -type integer\ 
      -range -1:4 -description {set the verbosity level}
   ::Simple::Option::declare {-w --width} -name width -type integer\ 
      -default 80 -description {set the width of the output}
   ::Simple::Option::declare {-a --all} -type boolflag\ 
      -description {process all files}

   # The special option "--" is used to collect the arguments at the
   # end of the command line not belonging to other options
   ::Simple::Option::declare -- -name files -type string\ 
      -values 3: -description {files to process}

   # Declare the options usage:  -p is compulsory, -v and -w are
   # optional, one of -a or a list of files must be given.
   ::Simple::Option::declare-usage {-p {O -v} {O -w} {X -a --}}

   # Display the options usage string
   # This displays the following:
   #   |-p choice [-v verbosity] [-w width] (-a | files(3:))
   puts [::Simple::Option::information usage]

   # Display the help message
   # This displays the following:
   #   |-p, --processing           processing option (A, B or C)
   #   |-v, --verbosity VERBOSITY  set the verbosity level
   #   |-w, --width WIDTH          set the width of the output (default 80)
   #   |-a, --all                  process all files
   #   |files                      files to process
   foreach optionPair [::Simple::Option::information help] {
      foreach {reference description} $optionPair break
      puts [format {%-26s %s} $reference $description]
   }

   # Now parse some possible command lines

   # This displays the following:
   #   |"-p" (or "--processing") required
   catch {::Simple::Option::parse {-w 132 -a}} result
   puts $result

   # This displays the following:
   #   |invalid value "D" for option "-p" of type "choice": must be A, B or C
   catch {::Simple::Option::parse {-p D -a}} result
   puts $result

   # This displays the following:
   #   |invalid value "foo" for option "-w" of type "integer"
   catch {::Simple::Option::parse {-p A -w foo -a}} result
   puts $result

   # This displays the following:
   #   |one of "-a" (or "--all") or "files" required
   catch {::Simple::Option::parse {-p A}} result
   puts $result

   # This displays the following:
   #   |too few values, 3: files required, got 2
   catch {::Simple::Option::parse {-p A file1 file2}} result
   puts $result

   # This displays the following:
   #   |invalid value "7" for option "-v": value out of range "-1:4"
   catch {::Simple::Option::parse {-p A -v 7 -a}} result
   puts $result

   # A valid command line
   ::Simple::Option::parse {-p A -v 3 -a}

   # Assess the available options values
   # This displays the following:
   #   |The value of option -p is A
   #   |The value of option -v is 3
   #   |The value of option -w is 80
   #   |The value of option -a is 1
   #   |Option -- not available
   foreach option [::Simple::Option::information declared] {
      if {[::Simple::Option::information available $option]} {
         puts "The value of option $option is\
            [::Simple::Option::information value $option]"
      } else {
         puts "Option $option not available"
      }
   }

} -requisites {

   {SimpleDeclare 1.0}

} -namespaces {

   ::Simple::Option
   ::Simple::Option::Priv

} -api {

   * K<::Simple::Option::declare-usage> usageGroups
     Specifies the options usage.

   * K<::Simple::Option::information usage>
     Returns the options usage string.

   * K<::Simple::Option::information isdefault> option
     Returns whether the value of an option is the default one.

   * K<::Simple::Option::information given> option
     Returns whether an option was given in the command line.

   * K<::Simple::Option::information help>
     Returns the option flag(s) and description pairs

   * K<::Simple::Option::declare> optionPair ?-name word? ?-type type?
        ?-range number-range? ?-choices any-list? ?-default any? 
        ?-values decimal-range? ?-description string?
     Declares an option.

   * K<::Simple::Option::parse> arguments
     Parses a command line.

   * K<::Simple::Option::information> value option
     Returns an option value.

   * K<::Simple::Option::information available> option
     Returns whether an option is available.
 
   * K<::Simple::Option::configure>
     Configures the package options.

   * K<::Simple::Option::cget>
     Gets the package options.

} -details {

      The following is a side-by-side comparison between the functionality
   provided by this package for command line options and that for procedure
   arguments provided by the P<SimpleProc> package:

   **Run-time type checking**

   P<SimpleProc>:  no implicit arguments run-time type checking by default
   (this can be changed by setting the A<-checktype> P<SimpleProc> or using
   the A<-checktype> flag for K<proc-ext> or K<declare-proc>).

   P<SimpleOption>: implicit options run-time type checking.

   **Flag or option given**
  
   P<SimpleProc>: K<::Simple::Argument::information flaggiven> returns
   whether a flag argument has been given in the procedure call.

   P<SimpleOption>: K<::Simple::Option::information given> returns
   whether an option was given in the command line.

   **Values availability**

   P<SimpleProc>: all arguments are available as both flags and optional
   arguments require a default value upon declaration.

   P<SimpleOption>:  not all options are available, only those for which a
   default value has been defined, boolean flags and options in the actual
   command line.  Use K<::Simple::Option::information available> to assess
   whether an option is available.

   **Values retrieval**

   P<SimpleProc>: all argument values go to variables.

   P<SimpleOption>: no option value goes to any variable, it must be
   fetched with K<::Simple::Option::information value>.

   **Usage string**

   P<SimpleProc>: the procedure argument usage string is got via
   K<::Simple::Proc::information usage>.

   P<SimpleOption>: the options usage is got via
   K<::Simple::Option::information usage>.

   **List of values**

   P<SimpleProc>: K<::Simple::Proc::information args> returns a
   procedure argument names list.

   P<SimpleOption>: K<::Simple::Option::information declared> returns the
   list of declared options.

   **Value equal to the default**

   P<SimpleProc>: K<::Simple::Argument::information isdefault>
   returns whether an argument value is equal to its default.

   P<SimpleOption>: K<::Simple::Option::information isdefault> returns
   whether an option value is equal to its default.

} -references {

   * The command line interpreter generator (X<clig>) at
     {{http://wsd.iitb.fhg.de/~kir/clighome}}.

} -todo {

   * Add K<::Simple::Option::information requires> and
     K<::Simple::Option::information excludes> which returns the list of
     options a given option requires or excludes, respectively.

   * Allow more than one set of options, identifying them by name.

} -history {

   10-sep-2001   First public release, version 0.4
   17-feb-2003   Second public release, version 0.5
   30-mar-2003   Extra package, version 0.5.1
   11-jul-2003   Package renamed to P<SimpleOption>, version 0.5.2
   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -variable {

   {
      ::Simple::Option::PrimaryOptions

   } -description {

      List of primary declared options

   } -type flag-list -value {}

} -variable {

   {
      ::Simple::Option::Usage

   } -description {

      Options usage string

   } -type string -value {}

} -variable {

   {
      ::Simple::Option::ActualValues

   } -description {

      Array containing for each primary option flag its actual value

   } -type any-array 

} -variable {

   {
      ::Simple::Option::DefaultValues

   } -description {

      Array containing for each primary option flag its default value

   } -type any-array 

} -variable {

   # fixme: the type of this variable is flag-list-list
   {
      ::Simple::Option::CompulsoryOptionsSets

   } -description {

      List of compulsory options sets

   } -type any-list -value {}

} -variable {

   {
      ::Simple::Option::RequiredOptions

   } -description {

      Array containing for each primary option flag its required options

   } -type flag-list-array 

} -variable {

   {
      ::Simple::Option::ExcludedOptions

   } -description {

      Array containing for each primary option flag its excluded options

   } -type flag-list-array 

} -variable {

   {
      ::Simple::Option::Options

   } -description {

      Array containing for each primary option flag a list with the following
      elements:

#table;delimiter=".";ignore="true"
       0. primary option flag
       1. secondary option flag
       2. option name
       3. option type
       4. option choices (empty if type is not "choice" nor\
          derived from it)
       5. option range (":" by default)
       6. option number of values
       7. option description
#end

   } -type any-list-array 

} -variable {

   {
      ::Simple::Option::PrimaryGivenOptions

   } -description {

      Array containing for each primary option flag given in the last call to
      K<::Simple::Option::parse> its values

   } -type flag-list-array 

} -error {

   {
      ::Simple::Option::DOUBLE-HYPHEN-NOT-AS-SECONDARY

   } -message {

      "--" can not be a secondary option flag

   } -explanation {

      An option declaration failed because "--" was supplied
      as a secondary option flag, but that is not allowed
   }

} -error {

   {
      ::Simple::Option::DOUBLE-HYPHEN-NOT-SECONDARY

   } -message {

      "--" can not have a secondary option flag

   } -explanation {

      The declaration of option "--" failed because it was accompanied with a
      secondary option flag, but that is not allowed
   }

} -error {

   {
      ::Simple::Option::DOUBLE-HYPHEN-REQUIRES-NAME

   } -message {

      option "--" requires a name

   } -explanation {

      The declaration of option "--" failed because it lacks the customary
      option name
   }

} -error {

   {
      ::Simple::Option::DOUBLE-HYPHEN-NOT-ALLOWED

   } -message {

      "--" not allowed

   } -explanation {

      "--" in the command line is not allowed because it was not declared,
      that is, only arguments associated to options are allowed.
   }

} -error {

   {
      ::Simple::Option::ALREADY-DECLARED

   } -message {

      option "%s" already declared

   } -explanation {

      Option <flag> has already been declared
   }

} -error {

   {
      ::Simple::Option::ALREADY-DECLARED-AS-SECONDARY

   } -message {

      option "%s" already declared as secondary of "%s"

   } -explanation {

      Option <secondary flag> has already been declared as the secondary flag
      of <primary flag>
   }

} -error {

   {
      ::Simple::Option::NON-EXISTING-TYPE

   } -message {

      non-existing type "%s" for option "%s"

   } -explanation {

      The declaration of option <flag> failed because the type <type> does not
      exists


   } -translation {
      <type> <flag>
   }

} -error {

   {
      ::Simple::Option::NO-RANGE-ALLOWED

   } -message {

      no range allowed for option "%s" of type "%s"

   } -explanation {

      The declaration of option <flag> failed because a range was supplied,
      but the type <type> does not allow ranges
   }

} -error {

   {
      ::Simple::Option::BAD-RANGE

   } -message {

      invalid range "%s" for option "%s" of type "%s"

   } -explanation {

      The declaration of option <flag> failed because <range> is not a valid
      range for type <type>

   } -translation {
      <range> <flag> <type>
   }

} -error {

   {
      ::Simple::Option::VALUE-OUT-OF-RANGE

   } -message {

      invalid%s value "%s" for %s: value out of range "%s"

   } -explanation {

      The declaration of option <flag> failed, or its command line value is
      incorrect, because the value <value> is out of the range <range>

   } -explained {

      invalid[ default] value "<value>" for (option <flag>|<option name>): 
      value out of range "<range>"

   } -corrective {

      Either supply a value within the range or modify the range
   }

} -error {

   {
      ::Simple::Option::NO-VALUES-ALLOWED

   } -message {

      no values allowed for option "%s" of type "boolflag"

   } -explanation {

      The declaration of option <flag> failed because the type boolflag
      does not allow specifying a value
   }

} -error {

   {
      ::Simple::Option::USAGE-NOT-SPECIFIED

   } -message {

      usage for option "%s" not specified

   } -explanation {

      Option <flag> was not specified in the declaration of the options usage
      via K<::Simple::Option::declare-usage>

   } -corrective {

      Give all declared options when calling
      K<::Simple::Option::declare-usage>
   }

} -error {

   {
      ::Simple::Option::NO-NON-OPTIONAL-OPTION-IN-OPTIONAL-GROUP

   } -message {

      no non-optional option within optional group "%s"

   } -explanation {

      A call to K<::Simple::Option::declare-usage> includes a <bad optional
      group>.  Optional groups must include at least one non-optional option,
      that is, not within another nested optional group
   }

} -error {

   {
      ::Simple::Option::NOT-ENOUGH-ALTERNATIVES-IN-EXCLUSIVE-GROUP

   } -message {

      not enough alternatives in exclusive group "%s"

   } -explanation {

      A call to K<::Simple::Option::declare-usage> includes a <bad exclusive
      group>.  An exclusive group is made of an "X" specifier followed by two
      or more major option groups
   }

} -error {

   {
      ::Simple::Option::BAD-MAJOR-GROUP

   } -message {

      bad major group "%s"

   } -explanation {

      A call to K<::Simple::Option::declare-usage> includes a <bad major
      group>.  A major group is either a single option, a group, an exclusive
      group or an optional group
   }

} -error {

   {
      ::Simple::Option::EXCLUSIVE-GROUP-AS-ALTERNATIVE

   } -message {

      exclusive group "%s" is an alternative for previous exclusive group

   } -explanation {

      A call to K<::Simple::Option::declare-usage> includes an <exclusive
      group> which is by itself an alternative for a previous exclusive group. 
      For example, the call K<::Simple::Option::declare-usage> ||\{{X -a {X -b
      -c}\}}|| would give rise to this error as the exclusive group ||{X -b
      -c}|| is an alternative for the exclusive group ||{X -a ...}||

   } -translation {
      <exclusive group>
   }

} -error {

   {
      ::Simple::Option::UNDECLARED-OPTION

   } -message {

      unrecognized option "%s"

   } -explanation {

      Option <flag> has not been declared

   } -corrective {

      Either use K<::Simple::Option::declare> to declare the option <flag> or
      get rid of the offending option
   }

} -error {

   {
      ::Simple::Option::NOT-A-PRIMARY-OPTION

   } -message {

      not a primary option flag "%s"

   } -explanation {

      Option flag <flag> is not a primary option flag, but the invoked
      procedure requires the use of primary flags
   }

} -error {

   {
      ::Simple::Option::OPTIONS-EXCLUDE-EACH-OTHER

   } -message {

      options "%s" and "%s" exclude each other

   } -explanation {

      Options <flag 1> and <flag 2> exclude each other, but both have been
      given simultaneously

   } -corrective {

      Provide one of the two flags only
   }

} -error {

   {
      ::Simple::Option::OPTION-REQUIRED

   } -message {

      %s required

   } -explanation {

      Option <required flag> is required, but it has not been given
   }

} -error {

   {
      ::Simple::Option::OPTION-GIVEN-REQUIRED

   } -message {

      %s given, %s required

   } -explanation {

      Option <given flag> was given, option that requires option <required
      flag>, but the later has not been given
   }

} -error {

   {
      ::Simple::Option::ONE-OF-OPTIONS-REQUIRED

   } -message {

      one of %s required

   } -explanation {

      One of the options in the <flags list> is required
   }

} -error {

   {
      ::Simple::Option::OPTION-GIVEN-ONE-OF-REQUIRED

   } -message {

      %s given, one of %s required

   } -explanation {

      Option <given flag> was given, but one of the options in the <required
      flags list> is required by the given one
   }

} -error {

   {
      ::Simple::Option::EXTRA-VALUES

   } -message {

      extra value(s) "%s"

   } -explanation {

      The values <values> in the command line can not be used
   }

} -error {

   {
      ::Simple::Option::TOO-MANY-OR-FEW-VALUES-1

   } -message {

      too %s values for option "%s", %s required, got %s

   } -explanation {

      Too <may or few> values have been given for option <option>.  The
      required range is <range>, but the number of given values is <actual
      number of values>
   }

} -error {

   {
      ::Simple::Option::TOO-MANY-OR-FEW-VALUES-2

   } -message {

      too %s values, %s %s required, got %s

   } -explanation {

      Too <may or few> <option name>, the required range is <range>, but the
      number of given values is <actual number of values>

   } -translation {
      <many or few> <range> <option name> <actual number of values>
   }

} -error {

   {
      ::Simple::Option::VALUE-NOT-AVAILABLE

   } -message {

      value not available for option "%s"

   } -explanation {

      The option <flag> value has been requested, but it is not available

   } -corrective {

      Use K<::Simple::Option::information available> to assess whether an
      option value is available or not
   }

} -command {

### ===========================================================================
### -command  : ::Simple::Option::declare

{
   ::Simple::Option::declare

} -synopsis {

   Declares an option.

} -overview {

   This procedure is used to declare a new option.

      Each option has a primary flag and an optional secondary flag, so that
   any of them can be used in the command line.  The special option "--" is
   used to collect the arguments at the end of the command line not belonging
   to other options.  In the command line, "--" can be used to inhibit the
   processing of the arguments following it.

      Use A<-name> to set the option name. The option name is used in the
   options usage string and other references to the option.

      If A<-type> is given, the option type is set to the given type,
   otherwise the option type is set to A<any>. 

   If A<-range> is given, the option values must be in the given range.

      When the option type is A<choice> or derived from it, A<-choices> is
   required.  This specifies the list of allowed choices to which the option
   values must conform.

   Use A<-default> to specify a default value for the option.

      If A<-values> is given, the number of allowed values for the
   option is limited by the given range.

} -arguments {
   { optionPair       flag-list      {Primary option flag and optional
                                      secondary flag}}
   {-name             word %DEFAULT% {Option name}}
   {-type             type any       {Option values type}}
   {-range            number-range : {Option values numeric range}}
   {-choices          any-list {}    {List of choices for choice type
                                      and its derivations}}
   {-default          any %DEFAULT%  {Option default value}}
   {-values           decimal-range 1
                                     {Number of values range}}
   {-description      string {}      {Option description}}
} -returns {

   The empty string.

} -body {

   ### Validate the option flags
   set nFlags [llength $optionPair]
   if {$nFlags > 0} {
      foreach {primary secondary} $optionPair break
      set isNotDoubleHyphen [string compare $primary --]
   }
   if {$nFlags > 0 && ![string compare $secondary --]} {
      ::Simple::Error::throw ::Simple::Option::DOUBLE-HYPHEN-NOT-AS-SECONDARY
   }
   if {$nFlags == 0 || $nFlags > 2 || ($isNotDoubleHyphen &&\
      (![::Simple::Type::Priv::is-basic flag $primary] ||\
      ($nFlags == 2 && ![::Simple::Type::Priv::is-basic flag $secondary])))} {

      ::Simple::Error::throw ::Simple::BAD-VALUE $optionPair argument\
         optionPair flag-list {}
   }

   #### Assert "--" does not have a secondary flag
   if {!$isNotDoubleHyphen && [string compare $secondary {}]} {
      ::Simple::Error::throw ::Simple::Option::DOUBLE-HYPHEN-NOT-SECONDARY
   }

   ### Assert the option has not been declared
   if {[info exists ::Simple::Option::Options($primary)]} {
      ::Simple::Error::throw ::Simple::Option::ALREADY-DECLARED $primary
   }
   if {$nFlags == 2 && [info exists ::Simple::Option::Options($secondary)]} {
      set primary [lindex $::Simple::Option::Options($secondary) 0]
      ::Simple::Error::throw ::Simple::Option::ALREADY-DECLARED-AS-SECONDARY\
         $secondary $primary
   }

   #### Assert an option name is given for "--"
   if {!$isNotDoubleHyphen &&\
      ![{::Simple::Argument::information flaggiven} -name]} {

      ::Simple::Error::throw ::Simple::Option::DOUBLE-HYPHEN-REQUIRES-NAME
   }

   ### Assert the type exists
   if {![::Simple::Type::information exists $type]} {
      ::Simple::Error::throw ::Simple::Option::NON-EXISTING-TYPE $type $primary
   }

   ### Assert it is not a derived array type
   if {[string match *-array $type]} {
      ::Simple::Error::throw ::Simple::Proc::ARRAY-NOT-ALLOWED\
         $type option $primary
   }

   # Remember whether it is a boolean option
   set booleanOption [expr {![string compare $type boolflag]}]

   ### Range given
   set rangeGiven [{::Simple::Argument::information flaggiven} -range]
   if {$rangeGiven} {

      ### Assert the range is valid
      if {[catch {set tmp\
         [::Simple::Type::Priv::is-with-hidden $type-range $range 0 {} 2]}]} {

         ::Simple::Error::throw ::Simple::Option::NO-RANGE-ALLOWED\
            $primary $type
      } elseif {!$tmp} {
         ::Simple::Error::throw ::Simple::Option::BAD-RANGE\
            $range $primary $type
      }

   ### Range not given
   } else {

      ### Set the range to the empty string if the type is not numeric
      if {[lsearch -exact [list integer float decimal hexadecimal\
         binary octal anybase number] $type] == -1} {

         set range {}
      }
   }

   ### Handle A<-choices>
   set choicesGiven [::Simple::Priv::flag-given -choices]
   if {[regexp choice(-list)? $type]} {
      if {!$choicesGiven || [llength $choices] < 2} {
         ::Simple::Error::throw\
            ::Simple::Variable::AT-LEAST-TWO-CHOICES option $primary
      }
   } else {
      if {$choicesGiven} {
         ::Simple::Error::throw\
            ::Simple::Variable::NO-CHOICES-ALLOWED option $primary $type
      } else {
         set choices [list]
      }
   }

   ### Default value given
   if {[{::Simple::Argument::information flaggiven} -default]} {

      ### Assert it is not a boolean option
      if {$booleanOption} {
         ::Simple::Error::throw ::Simple::Proc::DEFAULT-NOT-ALLOWED\
            option $primary $type
      }

      ### Assert the default value conforms to the type
      #   The option choices only apply if the
      #   type is choice or choice-list
      ::Simple::Variable::Priv::assert-type-conformation\
         option 0 $primary $type $default 0 $choices

      ### Assert the default value conforms to the range, if given
      if {$rangeGiven} {
         if {[::Simple::Type::outside-range $range $default]} {
            ::Simple::Error::throw ::Simple::Option::VALUE-OUT-OF-RANGE\
               { default} $default [format {option "%s"} $primary] $range
         }
      }

   ### No default value given, boolean option
   } elseif {$booleanOption} {

      ### Boolean options have a default value of 0
      set default 0
   }

   ### Values given
   if {[{::Simple::Argument::information flaggiven} -values]} {

      ### Assert it is not a boolean option
      if {$booleanOption} {
         ::Simple::Error::throw ::Simple::Option::NO-VALUES-ALLOWED $primary
      }

      ### Assert the values range is valid
      ::Simple::Argument::check-type -values

   ### Values not given
   } else {

      ### Boolean option
      if {$booleanOption} {

         ### Boolean options can not have values
         set values 0
      }
   }

   ### Store the option

   # Store the default usage information
   append ::Simple::Option::Usage "\[$primary\] "
   lappend ::Simple::Option::PrimaryOptions $primary

   # Convert the description to a single line
   set description [::Simple::Priv::convert-to-line $description]

   # Store the primary flag internal representation
   set ::Simple::Option::Options($primary) [list $primary $secondary $name\
      $type $choices $range $values $description]
      
   # Store the secondary flag internal representation
   if {[string compare $secondary {}]} {
      set ::Simple::Option::Options($secondary)\
         $::Simple::Option::Options($primary)
   }

   # Store the option default value, if available
   if {![{::Simple::Argument::information isdefault} -default]} {
      set ::Simple::Option::DefaultValues($primary) $default
      set ::Simple::Option::ActualValues($primary) $default
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::declare-usage

{
   ::Simple::Option::declare-usage

} -synopsis {

   Specifies the options usage.

} -overview {

      By default, all options are optional and independent.  This procedure
   allows to alter that behaviour by specifying relationships between options
   via usage option groups.

      Each usage option group contains options whose presence or absence in
   the command line is optional, compulsory or disallowed depending on the
   presence or absence of other options from the same usage option group in
   the command line.  Options in different usage option groups are, thus,
   fully independent.  These relationships are specified via an
   special-purpose syntax which is described below.

      Each usage option group is either a single option primary flag, an
   optional group or an exclusive group: 
   
   * An optional group is made of an "O" specifier followed by one or more
     elements among single options, optional groups or exclusive groups. At
     least one element within an optional group (either directly or within
     other contained groups) must be a required option.
    
     Options in an optional group are optional in the command line.
   
   * An exclusive group is made of an "X" specifier followed by two or more
     exclusive major groups.  Exclusive major groups are made of one or more
     elements among single options, optional groups or exclusive groups. At
     least one exclusive major group must not be an exclusive group.

     Options in different exclusive major groups of an exclusive group are
     mutually exclusive in the command line.
   
   These rules are summarised in the following pseudo-BNF syntax:

      UsageGroup     :=
         (SingleOption | OptionalGroup | ExclusiveGroup)
      OptionalGroup  :=
         O NoOptionalUsageGroup [UsageGroup ...]
      NoOptionalUsageGroup :=
         (SingleOption | ExclusiveGroup)
      NoExclusiveUsageGroup :=
         (SingleOption | OptionalGroup)
      ExclusiveGroup :=
         X ExclusiveMajorGroup ExclusiveMajorGroup [ExclusiveMajorGroup ...]
      ExclusiveMajorGroup :=
         NoExclusiveUsageGroup [UsageGroup ...]

      If this procedure is invoked, the usage of all declared options
   must be specified.

} -arguments {

   { usageGroups      list           {Usage option groups}}

} -returns {

   The empty string.

} -body {

   ### Reset a possible previous call to this procedure
   ::Simple::Option::Priv::reset-declare-usage

   ### Loop over usage option groups
   set specifiedOptions [list]
   foreach usageGroup $usageGroups {

      ### Declare the usage option group usage
      foreach {usageGroupRequiredOptions usageGroupSpecifiedOptions}\
         [::Simple::Option::Priv::declare-usage-group-1 $usageGroup [list] 0]\
         break
      set ::Simple::Option::CompulsoryOptionsSets [concat\
         $::Simple::Option::CompulsoryOptionsSets $usageGroupRequiredOptions]
      set specifiedOptions\
         [concat $specifiedOptions $usageGroupSpecifiedOptions]
   }
   set ::Simple::Option::Usage [string trimright $::Simple::Option::Usage]

   ### Assert the usage of all primary declared options has been specified
   foreach primary $::Simple::Option::PrimaryOptions {
      if {[lsearch -exact $specifiedOptions $primary] == -1} {
         ::Simple::Option::Priv::reset-declare-usage
         ::Simple::Error::throw ::Simple::Option::USAGE-NOT-SPECIFIED $primary
      }
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::declare-usage-group-1

{
   ::Simple::Option::Priv::declare-usage-group-1

} -synopsis {

   Specifies a usage option group usage.

} -arguments {

   { usageGroup       list           {Usage option group}}
   { requiredOptions  list           {Required options list}}
   { exclusiveMajorGroup
                      boolean        {Whether the group is an exclusive
                                      major group}}

} -returns {

   A list with the following elements:

#table;delimiter=":";ignore="true"
    0: the group required options.
    1: the group specified options.
#end

} -body {

   ### Switch on specifier
   set nGroupElements [llength $usageGroup]
   set groupFirst [lindex $usageGroup 0]
   set groupRest [lrange $usageGroup 1 end]
   set groupRequiredOptions [list]
   set groupSpecifiedOptions [list]
   switch -exact -- $groupFirst {

      O {
      ### Optional group

         ### Assert the group has at least two elements
         if {$nGroupElements < 2} {
            ::Simple::Error::throw ::Simple::BAD-ITEM\
               {option flag} $usageGroup
         }

         ### Add the opening optional symbol to the usage string
         append ::Simple::Option::Usage {[}

         ### Specify the major group or groups usage
         foreach {optionalGroupRequiredOptions groupSpecifiedOptions}\
            [::Simple::Option::Priv::declare-usage-group-2\
            $groupRest $requiredOptions $groupSpecifiedOptions 0] break

         ### Assert the group has at least one non-optional option
         if {[llength $optionalGroupRequiredOptions] == 0} {
            ::Simple::Error::throw\
               ::Simple::Option::NO-NON-OPTIONAL-OPTION-IN-OPTIONAL-GROUP\
               $usageGroup
         }

         ### Add the closing optional symbol to the usage string
         set ::Simple::Option::Usage\
            "[string trimright $::Simple::Option::Usage]\] "
      }

      X {
      ### Exclusive group

         ### Assert the group has at least three elements
         if {$nGroupElements < 3} {
            ::Simple::Error::throw\
               ::Simple::Option::NOT-ENOUGH-ALTERNATIVES-IN-EXCLUSIVE-GROUP\
               $usageGroup
         }

         ### Add the opening exclusive group symbol to the usage string
         append ::Simple::Option::Usage (

         ### Specify the major groups usage
         foreach {groupRequiredOptions groupSpecifiedOptions}\
            [::Simple::Option::Priv::declare-usage-group-2\
            $groupRest $requiredOptions $groupSpecifiedOptions 1] break

         ### Add the closing exclusive group symbol to the usage string
         set ::Simple::Option::Usage\
            "[string trimright $::Simple::Option::Usage {| }]) "
      }

      default {
      ### Exclusive major group or single option flag

         ### Exclusive major group and more than one element
         if {$exclusiveMajorGroup && $nGroupElements > 1} {

            ### Specify the exclusive major group usage
            foreach {groupRequiredOptions groupSpecifiedOptions}\
               [::Simple::Option::Priv::declare-usage-group-2\
               $usageGroup $requiredOptions $groupSpecifiedOptions 0] break

         ### Non-exclusive major group or just one element
         } else {

            ### Assert the group has one element
            if {$nGroupElements != 1} {
               ::Simple::Error::throw ::Simple::BAD-ITEM\
                  {option flag} $usageGroup
            }

            ### Assert the only element is a single option flag or "--"
            if {[string compare $groupFirst --] &&\
               ![::Simple::Type::Priv::is-basic flag $groupFirst]} {

               ::Simple::Error::throw ::Simple::BAD-ITEM\
                  {option flag or "--"} $groupFirst
            }

            ### Specify the single option flag usage
            ::Simple::Option::Priv::declare-usage-option\
               $groupFirst $requiredOptions
            set groupRequiredOptions $groupFirst
            set groupSpecifiedOptions $groupFirst

            ### Add the option flag to the usage string
            append ::Simple::Option::Usage "$groupFirst "
         }
      }
   }

   ### Return the group required and specified options
   return [list $groupRequiredOptions $groupSpecifiedOptions]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::declare-usage-option

{
   ::Simple::Option::Priv::declare-usage-option

} -synopsis {

   Specifies a single option usage.

} -arguments {

   { option           flag           {Option flag}}
   { requiredOptions  list           {Option required options list}}

} -returns {

   The empty string.

} -body {

   ### Assert the option has been declared
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0]\
      $option]} {

      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Assert the option usage has not been specified twice
   if {[info exists ::Simple::Option::RequiredOptions($option)]} {
      ::Simple::Error::throw ::Simple::DUPLICATED-FLAG-OR-OPTION\
         option \"$option\"
   }

   ### Store the option required options
   set ::Simple::Option::RequiredOptions($option) $requiredOptions

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::reset-declare-usage

{
   ::Simple::Option::Priv::reset-declare-usage

} -synopsis {

   Resets a previous call to K<::Simple::Option::declare-usage>.

} -returns {

   The empty string.

} -body {

   set   ::Simple::Option::Usage {}
   set   ::Simple::Option::CompulsoryOptionsSets [list]
   catch {unset ::Simple::Option::RequiredOptions}
   catch {unset ::Simple::Option::ExcludedOptions}

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::reset-parse

{
   ::Simple::Option::Priv::reset-parse

} -synopsis {

   Reset a previous call to K<::Simple::Option::parse>.

} -returns {

   The empty string.

} -body {

   catch {unset ::Simple::Option::ActualValues}
   if {[info exists ::Simple::Option::DefaultValues]} {
      array set ::Simple::Option::ActualValues\
         [array get ::Simple::Option::DefaultValues]
   }
   catch {unset ::Simple::Option::PrimaryGivenOptions}

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::parse

{
   ::Simple::Option::parse

} -synopsis {

   Parses a command line.

} -overview {

      This is the procedure which parses a command line and assesses its
   compliance against the declared command line options and their usage.

} -arguments {

   { commandLine      list           {Command line}}

} -references {

   * The K<::Simple::Program::parse-command-line> procedure.

} -returns {

   The empty string.

} -body {

   ### Reset a possible previous call to this procedure
   ::Simple::Option::Priv::reset-parse

   ### Phase 1: check the options independently

   ### Explicit "--"
   set doubleHyphenIndex [lsearch -exact $commandLine --]
   if {$doubleHyphenIndex != -1} {

      ### Extract the "--" values
      ::Simple::Option::Priv::extract-double-hyphen-values\
         $doubleHyphenIndex

      ### Store the "--" option and its values
      set ::Simple::Option::PrimaryGivenOptions(--) $values
      set givenOptions(--) {}
   }

   ### Loop over remaining arguments
   while {[llength $commandLine] > 0}  {

      ### The first remaining argument is an option flag
      set argument [lindex $commandLine 0]
      if {[::Simple::Type::Priv::is-basic flag $argument]} {

         ### Assert the option flag is a primary or secondary
         ###    declared option flag
         if {![info exists ::Simple::Option::Options($argument)]} {
            ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $argument
         }

         ### If secondary, grab the primary
         foreach {primary secondary name type choices range valuesRange\
            description} $::Simple::Option::Options($argument) break

         ### Assert the option is not duplicated
         if {[info exists ::Simple::Option::PrimaryGivenOptions($primary)]} {
            ::Simple::Error::throw ::Simple::DUPLICATED-FLAG-OR-OPTION\
               option [::Simple::Option::Priv::option-reference $primary]
         }

         ### Extract the option values
         set lastFlag 1
         set nArgs [llength $commandLine]
         for {set index 1} {$index < $nArgs} {incr index} {
            if {[::Simple::Type::Priv::is-basic flag\
               [lindex $commandLine $index]]} {

               set lastFlag 0
               break
            }
         }

         ### Possible implicit "--"
         #   Last flag and "--" is a primary declared option flag
         if {$lastFlag && $doubleHyphenIndex == -1 &&\
            [info exists ::Simple::Option::Options(--)]} {

            ### Take the maximum allowed number of values
            # fixme: we should provide some sort of heuristic here to split
            #        the remaining values between the last flag and "--"
            #        instead of of just trying to get the maximum number of
            #        values for the last flag.  Ideally, we should take
            #        values one by one, checking its conformance to the
            #        flag type, and throw up on the first non-conformance or
            #        when the maximum allowed number of values have been
            #        taken:  the remaining values are the implicit "--"
            #        values
            set maxNValues [lindex [split $valuesRange :] end]
            if {[string compare $maxNValues {}]} {
               set values [lrange $commandLine 1 $maxNValues]
               set commandLine [lrange $commandLine [incr maxNValues] end]
            } else  {
               set values [lrange $commandLine 1 end]
               set commandLine [list]
            }

         ### No implicit "--"
         } else {
            set values [lrange $commandLine 1 [incr index -1]]
            set commandLine [lrange $commandLine [incr index] end]
         }

         ### Check the option values
         ::Simple::Option::Priv::check-values $argument $values

         ### Boolean flag
         if {![string compare $type boolflag]} {

            ### The value of given boolean flags is 1
            set values 1
         }

         ### Store the given argument and its values
         set ::Simple::Option::PrimaryGivenOptions($primary) $values
         set givenOptions($argument) {}

      ### The first remaining argument is not an option flag
      } else  {

         ### Implicit "--"

         ### Extract the "--" values
         ::Simple::Option::Priv::extract-double-hyphen-values -1

         ### Store the "--" option and its values
         set ::Simple::Option::PrimaryGivenOptions(--) $values
         set givenOptions(--) {}
      }
   }

   ### Phase 2: check the option groups

   ### Assert all compulsory options have been given
   #   Loop over compulsory options sets
   foreach compulsoryOptionsSet $::Simple::Option::CompulsoryOptionsSets {
      ::Simple::Option::Priv::assert-options-given $compulsoryOptionsSet 1
   }

   ### Loop over given options
   foreach givenOption [array names givenOptions] {
      foreach {primary secondary name type choices range valuesRange\
         description} $::Simple::Option::Options($givenOption) break

      ### Assert this option required options have been given
      #   Loop over required options
      if {[info exists ::Simple::Option::RequiredOptions($primary)]} {
         foreach requiredOptions $::Simple::Option::RequiredOptions($primary) {
            ::Simple::Option::Priv::assert-options-given\
               $requiredOptions 0 $givenOption
         }
      }

      ### Assert none of this option excluded options have been given
      if {[info exists ::Simple::Option::ExcludedOptions($primary)]} {
         foreach excludedOption $::Simple::Option::ExcludedOptions($primary) {
            if {[info exists ::Simple::Option::PrimaryGivenOptions($excludedOption)]} {
               if {![info exists givenOptions($excludedOption)]} {
                  foreach {primary excludedOption name type choices range\
                     valuesRange description}\
                     $::Simple::Option::Options($excludedOption) break
               }
               set givenOption\
                  [::Simple::Option::Priv::option-or-double-hyphen-name\
                  $givenOption]
               set excludedOption\
                  [::Simple::Option::Priv::option-or-double-hyphen-name\
                  $excludedOption]
               ::Simple::Error::throw\
                  ::Simple::Option::OPTIONS-EXCLUDE-EACH-OTHER $givenOption\
                  $excludedOption
            }
         }
      }
   }

   ### Parsing successful
   #   This updates the values of the given options while 
   #   leaving the non-given ones to their default values
   array set ::Simple::Option::ActualValues\
      [array get ::Simple::Option::PrimaryGivenOptions]

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::option-or-double-hyphen-name

{
   ::Simple::Option::Priv::option-or-double-hyphen-name

} -synopsis {

   Returns the given option or, if it is the double hyphen, its name

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The given option or, if it is the double hyphen, its name

} -body {

   if {[string compare $option --]} {
      set answer $option
   } else {
      set answer [lindex $::Simple::Option::Options(--) 2]
   }
   return $answer

}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::assert-options-given

{
   ::Simple::Option::Priv::assert-options-given

} -synopsis {

   Asserts that the required options have been given

} -overview {

      This procedure asserts that either all compulsory options have been
   given or that a given option required options have been given.

} -arguments {

   { options          list           {Compulsory or required options list}}
   { compulsory       boolean        {Whether the options are required or
                                      compulsory}}
   {?givenOption?     flag %DEFAULT% {Given option}}

} -returns {

   The empty string.

} -body {

   # No option in the set has been given
   set found 0
   foreach option $options {
      if {[info exists ::Simple::Option::PrimaryGivenOptions($option)]} {
         set found 1
         break
      }
   }
   if {!$found} {

      # One option in the set
      if {[llength $options] == 1} {
         if {$compulsory} {
            ::Simple::Error::throw ::Simple::Option::OPTION-REQUIRED\
               [::Simple::Option::Priv::option-reference $options]
         } else {
            ::Simple::Error::throw ::Simple::Option::OPTION-GIVEN-REQUIRED\
               [::Simple::Option::Priv::option-reference $givenOption 1]\
               [::Simple::Option::Priv::option-reference $options]
         }

      # More than one option in the set
      } else {
         set optionsReference [list]
         foreach option $options {
            lappend optionsReference\
               [::Simple::Option::Priv::option-reference $option]
         }
         if {$compulsory} {
            ::Simple::Error::throw ::Simple::Option::ONE-OF-OPTIONS-REQUIRED\
               [::Simple::List::enumerate $optionsReference]
         } else {
            ::Simple::Error::throw\
               ::Simple::Option::OPTION-GIVEN-ONE-OF-REQUIRED\
               [::Simple::Option::Priv::option-reference $givenOption 1]\
               [::Simple::List::enumerate $optionsReference]
         }
      }
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::check-values

{
   ::Simple::Option::Priv::check-values

} -synopsis {

   Checks an option values.

} -arguments {

   { option           flag           {Option flag}}
   { values           any-list       {Option values}}

} -returns {

   The empty string.

} -body {

   ### Check number of values
   foreach {primary secondary name type choices range valuesRange\
      description} $::Simple::Option::Options($option) break
   set nValues [llength $values]
   set outsideRange [::Simple::Type::outside-range $valuesRange $nValues]
   if {$outsideRange != 0} {
      if {$outsideRange == -1} {
         set manyOrFew few
      } else  {
         set manyOrFew many
      }
      if {[string compare $option --]} {
         ::Simple::Error::throw ::Simple::Option::TOO-MANY-OR-FEW-VALUES-1\
            $manyOrFew $option $valuesRange $nValues
      } else {
         ::Simple::Error::throw ::Simple::Option::TOO-MANY-OR-FEW-VALUES-2\
            $manyOrFew $valuesRange $name $nValues
      }
   }

   ### Check the values type
   foreach argument $values {
      ::Simple::Variable::Priv::assert-type-conformation\
         option 0 $option $type $argument 0 $choices
   }

   ### Assert the values conform to the range, if given
   if {[string compare $range :] && [string compare $range {}]} {
      foreach argument $values {
         if {[::Simple::Type::outside-range $range $argument]} {
            if {[string compare $option --]} {
               set optionReference [format {option "%s"} $option]
            } else {
               set optionReference $name
            }
            ::Simple::Error::throw ::Simple::Option::VALUE-OUT-OF-RANGE\
               {} $argument $optionReference $range
         }
      }
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information value

{
   ::Simple::Option::information value

} -synopsis {

   Returns an option value.

} -overview {

      This procedure returns an option value, if available.  Use
   K<::Simple::Option::information available> to assess whether an option is
   available or not. The option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   The option value.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Assert the option value is available
   if {![info exists ::Simple::Option::ActualValues($option)]} {
      ::Simple::Error::throw ::Simple::Option::VALUE-NOT-AVAILABLE $option
   }

   ### Return the option value
   return $::Simple::Option::ActualValues($option)
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information available

{
   ::Simple::Option::information available

} -synopsis {

   Returns whether an option is available.

} -overview {

      This procedure returns whether an option is available or not.  The
   option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   Whether the option is available.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return whether the option is available
   return [info exists ::Simple::Option::ActualValues($option)]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information isdefault

{
   ::Simple::Option::information isdefault

} -synopsis {

   Returns whether the value of an option is the default one.

} -overview {

      This procedure returns whether an option is available and its actual
   value matches its default value.  For unavailable options, this procedure
   returns always false. The option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

      Whether the option is available and its actual value matches its
   default value.

} -references {

   * The K<::Simple::Argument::information isdefault> procedure.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Available option with a default value
   if {[info exists ::Simple::Option::ActualValues($option)] &&\
      [info exists ::Simple::Option::DefaultValues($option)]} {

      ### Get whether the option default and actual values match
      set answer [expr {![string compare\
         $::Simple::Option::ActualValues($option)\
         $::Simple::Option::DefaultValues($option)]}]

   ### Unavailable option
   } else {

      ### Return false
      set answer 0
   }

   ### Return wheter the option is available and its
   ###    value value matches its default value
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information given

{
   ::Simple::Option::information given

} -synopsis {

   Returns whether an option was given in the command line.

} -overview {

      This procedure returns whether an option was given in the last
   call to K<::Simple::Option::parse>. The option flag must be primary.

} -arguments {

   { option           flag           {Option flag}}

} -returns {

   Whether the option was given in the command line.

} -references {

   * The K<::Simple::Priv::flag-given> procedure.

   * The K<::Simple::Argument::information flaggiven> procedure.

} -body {

   ### Assert the option flag is a primary or secondary
   ###    declared option flag
   if {![info exists ::Simple::Option::Options($option)]} {
      ::Simple::Error::throw ::Simple::Option::UNDECLARED-OPTION $option
   }

   ### Assert the option is primary
   if {[string compare [lindex $::Simple::Option::Options($option) 0] $option]} {
      ::Simple::Error::throw ::Simple::Option::NOT-A-PRIMARY-OPTION $option
   }

   ### Return whether the option was given in the command line
   return [info exists ::Simple::Option::PrimaryGivenOptions($option)]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information help

{
   ::Simple::Option::information help

} -synopsis {

   Returns the option flag(s) and description pairs

} -overview {

      This procedure returns a list of pairs.  Each element in the list
   corresponds to an option.
   
      The first element of each pair is the option reference, including the
   primary and secondary option flags (if any) plus the option name or type,
   when applicable.

      The second element of each pair is the option description, including the
   string given as argument to the A<-description> flag of
   K<::Simple::Option::declare> plus the list of options and default value,
   when applicable.

      The return value of this procedure can be used to provide a help
   message; this is used by the K<declare-program> procedure of the
   P<SimpleProgram> package.

} -returns {

   The option flag(s) and description pairs

} -body {

   ### Loop over declared options
   set answer [list]
   set maxReferenceLength 0
   foreach option $::Simple::Option::PrimaryOptions {

      ### Option is not "--"
      foreach {primary secondary name type choices range valuesRange\
         description} $::Simple::Option::Options($option) break
      if {[string compare $primary --]} {

         ### Use as reference the primary option flag plus
         ###    the secondary, if any
         if {[string compare $secondary {}] &&\
            [string compare $primary $secondary]} {

            set reference "$primary, $secondary"
         } elseif {[string match --* $primary]} {

            # Indent by four spaces primary option flags with two dashes so
            # that those get aligned, assuming the primary option flag is
            # always of the form "-1" and the secondaries of the form
            # "--123..."
            set reference "    $primary"

         } else {
            set reference $primary
         }

         ### Add to the reference the option name, if
         ###    available, or the option type otherwise
         if {[string compare %DEFAULT% $name]} {
            append reference " [string toupper $name]"
         } else {
            if {[string compare $type boolflag] &&\
               ![regexp choice(-list)? $type]} {

               append reference " [string toupper $type]"
            }
         }

      ### Option is "--"
      } else {

         ### Use as reference the "--" name
         set reference $name
      }

      ### Get the option description
      # The option type is choice or choice-list
      if {[regexp choice(-list)? $type]} {

         # Add "(<choices>" to the description
         append description " ([::Simple::List::enumerate $choices]"

         # There is a default value, add "default <default>" to the description
         if {[info exists ::Simple::Option::DefaultValues($option)]} {
            append description\
               ", default $::Simple::Option::DefaultValues($option)"
         }

         # Add ")" to the description
         append description )

      } else {

         # There is a default value  andit is not a boolean flag,
         # add "(default <default>)" to the description
         if {[info exists ::Simple::Option::DefaultValues($option)] &&
            [string compare $type boolflag]} {

            append description\
               " (default $::Simple::Option::DefaultValues($option))"
         }
      }

      ### Add this option reference and description to the list
      lappend answer [list $reference $description]
   }

   ### Return the option flag(s) and description pairs
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::information usage

{
   ::Simple::Option::information usage

} -synopsis {

   Returns the options usage string.

} -overview {

      This procedure returns the options usage string as declared by the
   K<::Simple::Option::declare-usage> procedure, or the default usage string
   (all options are optional) if that procedure has not been used.
   
      Each declared option appears in the options usage string with its
   description, which is either the option name (if available) or the option
   type.  If the option range is non-trivial (that is, different from "1")
   that is also added, between parenthesis.

      Optional option groups appear between square brackets, such as in ||[-a
   -b]||.  Exclusive option groups appear between parenthesis whith the
   different alternatives separated by a vertical bar, such as in ||(-a | -b
   -c)||.

      If no options have been declared, this procedure returns the empty
   string.

} -returns {

   The options usage string

} -references {

   * The K<::Simple::Proc::information usage> procedure.

} -examples {

   # Declare some options
   ::Simple::Option::declare  -a
   ::Simple::Option::declare --b -type boolflag
   ::Simple::Option::declare  -c -type integer -values 2:
   ::Simple::Option::declare --d -type float -range 3:4
   ::Simple::Option::declare  -- -name ARGS

   # Declare their usage
   ::Simple::Option::declare-usage {{O -a --b {X -c --d} {O --}}}

   # Get the usage string
   # This displays the following:
   #   |[-a any --b (-c integer(2:) | --d float) [ARGS]]
   puts [::Simple::Option::information usage]

} -body {

   ### Loop over primary declared options
   set answer $::Simple::Option::Usage
   foreach primary $::Simple::Option::PrimaryOptions {
      foreach {primary secondary name type choices range values description}\
         $::Simple::Option::Options($primary) break

      ### Option is "--"
      if {![string compare $primary --]} {

         ### Set the argument description to the "--" name
         set valuesDescription $name

         ### Non-trivial values range
         if {[string compare $values 1]} {

            ### Add the values range to the values description
            append valuesDescription ($values)
         }

         ### Substitute "--" by the values description
         #   ensuring that it is not part of another longer option flag (that
         #   is, ensure it is not followed by any allowed character for an
         #   option flag or is at the end of the string)
         regsub -- --\(\[^_a-zA-Z0-9%\]|$\)\
            $answer $valuesDescription\\1 answer

      ### Not a boolean option
      } elseif {[string compare $type boolflag]} {

         ### Option name available
         if {[string compare $name %DEFAULT%]} {

            ### Set the argument description to the option name
            set valuesDescription " $name"

         ### Option name not available
         } else {

            ### Set the argument description to the option type
            set valuesDescription " $type"
         }

         ### Non-trivial values range
         if {[string compare $values 1]} {

            ### Add the values range to the values description
            append valuesDescription ($values)
         }

         ### Add the values description to the option
         #   Substitute the option flag by that same option flag followed
         #   by the values description, ensuring that the option flag is
         #   not part of another longer option flag (that is, ensure it is
         #   not followed by any allowed character for an option flag or is
         #   at the end of the string)
         regsub -- $primary\(\[^_a-zA-Z0-9%\]|$\)\
            $answer $primary$valuesDescription\\1 answer
      }
   }

   ### Return the options usage string
   return [string trimright $answer]
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::Priv::option-reference

{
   ::Simple::Option::Priv::option-reference

} -synopsis {

   Returns a reference to an option.

} -overview {

      This procedure returns a reference to an option.  If it is "--", it
   returns its name.  If it is not but the A<useAsGiven> argument is true,
   then it returns the option flag.  Otherwise it returns the primary option
   flag including its secondary flag, if any.

} -arguments {

   { option           flag           {Option flag}}
   {?useAsGiven?      boolean 0      {Whether to use the option as given}}

} -returns {

   A reference to the option.

} -body {

   ### Option is not "--"
   foreach {primary secondary name type choices range valuesRange\
      description} $::Simple::Option::Options($option) break
   if {[string compare $primary --]} {

      ### Use as given
      if {$useAsGiven} {

         ### Set the reference to the given option flag
         set answer \"$option\"

      ### Not use as given
      } else {

         ### Set the reference to the primary option flag and
         ###    include the secondary, if any
         set answer \"$primary\"
         if {[string compare $secondary {}] &&\
            [string compare $primary $secondary]} {

            append answer [format { (or "%s")} $secondary]
         }
      }

   ### Option is "--"
   } else {

      ### Set the reference to the "--" name
      set answer \"$name\"
   }

   ### Return a reference to the option
   return $answer
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::configure

{
   ::Simple::Option::configure

} -synopsis {

   Configures the package options.

} -returns {

   The "package with no options" error is thrown.

} -body {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}
} -command {

### ===========================================================================
### -command  : ::Simple::Option::cget

{
   ::Simple::Option::cget

} -synopsis {

   Gets the package options.

} -returns {

   The requested option value or the whole list of options if none specified.

} -body {

   return [list]
}
}
