### ===========================================================================
### -package  : SimplePackage
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library packages handling.
# -module     : implementation
# -overview   :
#
#     This package improves the Tcl facilities for package loading and version
#  control.  It works by each package declaring itself.  This declaration
#  includes such information as the namespaces used by the package, its
#  exported commands and aliases or its package requisites, so that this
#  information can not only be queried later, but also used for other
#  purposes.  As an example, it is possible to obtain the list of exported
#  commands from a package plus those exported from all packages it may
#  require, recursively, before actually modifying any namespace not owned
#  by it.
#
#     This package also augments K<unknown> in order to autoload ''extra''
#  packages.  An extra package works as a companion to a regular package.  It
#  is intended to contain the package functionality less frequently used. 
#  That way, when an application requires a package, only the regular package
#  is actually loaded, leaving the functionality in the extra package to be
#  autoloaded in case it becomes necessary.  Nevertheless, extra packages live
#  independently of the package they extend.  For example, extra packages are
#  not uninstalled when the non-extra package is uninstalled.  The
#  K<::Simple::Package::information *> commands do not, in general, return
#  information about items in extra packages.  Exceptions to this rule are the
#  information commands which return a current snapshot of the package
#  namespaces such as K<::Simple::Package::information commands>, as the extra
#  package may add commands to those namespaces.  Another exception is the
#  K<::Simple::Package::information api> which provides a package full API,
#  including its public items in its extra package, if any.
#
# -usage      :
#
#     Packages may be either known or unknown to Tcl, depending on whether or
#  not K<package ifneeded> has been invoked for the package.  In turn, known
#  packages may be either available or provided, depending on whether or not
#  K<package provide> has been invoked for them.  In turn, provided packages
#  may be either declared or undeclared, depending on whether
#  K<::Simple::Package::declare> (or, alternatively, K<declare-package> if
#  P<SimpleDeclare> is used) have been invoked for them.  Finally, all
#  undeclared packages are considered to be installed (that is, ready to be
#  used) but declared packages may be either installed or uninstalled,
#  depending on whether or not K<::Simple::Package::install> has been invoked
#  for them.  The following drawing sketches the situation:
#
##verbatim
#             .===========.     .-----------.
#             |  unknown  |     |   known   | *
#             `==========='     `-----------'
#                                 |       |
#                      .===========.     .-----------.
#                      | available |     | provided  |
#                      `==========='     `-----------'
#                                          |       |
#                               .===========.     .-----------.
#                               | declared  | *   |undeclared |
#                               `==========='     `-----------'
#                                 |       |         |
#                      .-----------.     .===========.
#                      |uninstalled| *   | installed | *
#                      `-----------'     `==========='
##end
#
#     It is possible to get the list of packages in those states which have an
#  asterisk by its box in the drawing above (''known'', ''declared'',
#  ''uninstalled'' and ''installed'') via the K<::Simple::Package::information
#  ''state''> commands where ''state'' is the corresponding state.
#
#     Each package can be considered to be in one of the four main states
#  depicted with double-line boxes in the drawing above (''unknown'',
#  ''available'', ''declared'' or ''installed'').  The
#  K<::Simple::Package::information state> command returns the state of a
#  package among these main four.
#
#     At the beginning, all packages are ''unknown''.  Where Tcl evaluates all
#  of the F<pkgIndex.tcl> files in the V<auto_path> directories, K<package
#  ifneeded> is typically executed for some packages which become
#  ''available''.  These packages can then be required with
#  K<::Simple::Package::require> or K<package require>, which should invoke
#  K<package provide> so that the packages become ''provided''.
#
#     During this process, K<::Simple::Package::declare> can be either invoked
#  or not.  In the later case, the package is considered ''installed'',
#  although a full install might require manually importing the package
#  namespaces via K<namespace import>.  In the former case, the package
#  becomes ''declared'', state in which many package details can be queried
#  (see below) before actually installing the package using
#  K<::Simple::Package::install>.  With this, the declared package finally
#  reaches the ''installed'' state.
#
#     Another possibility is to use K<::Simple::Package::require-and-install>
#  which requires the package and, if declared, automatically installs it.
#
#     It is also possible to go backwards in the above described route by
#  using K<::Simple::Package::uninstall>, K<::Simple::Package::undeclare> and
#  K<::Simple::Package::forget> or K<package forget> (or substituting the
#  first two calls by K<::Simple::Package::uninstall-and-undeclare>).  The
#  following drawings sketches these transitions.  In the drawings the prefix
#  ||::Simple::Package::|| has been abbreviated to ||::S::P::||.  First, for
#  undeclared (regular Tcl) packages:
#
##verbatim
#                                 .===========.
#                                 |  unknown  |
#                ::S::P::forget .>`==========='-.
#             OR package forget |               | package ifneeded
#                               |               |
#                               `-.===========.<'
#                                 | available |
#                               .>`==========='-. ::S::P::require OR
#                      ::S::P:: |               | package require OR
#       uninstall-and-undeclare |               | ::S::P::install-and-require
#                               `-.===========.<'
#                                 | installed |
#                                 `==========='
##end
#
# Second, for packages declared using The Simple Development Library:
#
##verbatim
#                                 .===========.
#                                 |  unknown  |
#                ::S::P::forget .>`==========='-.
#             OR package forget |               | package ifneeded
#                               |               |
#                               `-.===========.<'
#                                 | available |
#     ,----------->-------------.>`==========='-.----------->-------------.
#     |                         |               | (::S::P::require OR     |
#     |       ::S::P::undeclare |               | package require) AND    |
#     |                         `-.===========.<' ::S::P::declare         |
# ::S::P::                        | declared  |                       ::S::P::
# uninstall-and-undeclare       .>`==========='-.          require-and-install
#     |                         |               |                         |
#     |       ::S::P::uninstall |               | ::S::P::install         |
#     |                         `-.===========.<'                         |
#     |                           | installed |                           |
#     `-----------<---------------`==========='<------------<-------------'
##end
#
#     Notice that in the case of undeclared (regular Tcl) packages, a full
#  installation might require manually importing the package namespaces via
#  K<namespace import> and, conversely, a full uninstallation might require
#  manually forgetting the package namespaces via K<namespace forget>.  This
#  is not necessary for declared packages as the installation and
#  uninstallation procedures take care of that.
#
#     The purpose of having ''declared'' and ''installed'' states is twofold. 
#  First, it is possible to require a package and get information about it
#  without any effect in namespaces not owned by it until it is installed. 
#  Second, it is possible to ''reset'' the status of a package to its default
#  configuration by uninstalling and reinstalling it.
#
#     The K<::Simple::Package::information> command provides a myriad of
#  subcommands to query information about declared packages.
#
#     Notice that it is very easy to convert regular Tcl packages into
#  declared ones, thus taking benefit of all the functionality provided by
#  P<SimplePackage>, by simply adding a proper call to
#  K<::Simple::Package::declare>, for example in the corresponding entry of
#  the F<pkgIndex.tcl> file.
#
#     Upon requiring a package which gets declared in the process, all its
#  required packages are also required, recursively.  Something similar
#  happens when installing a declared package, which results in all its
#  required packages being installed as well, recursively.  Notice, though,
#  that the contrary is not true; that is, uninstalling a package does not
#  uninstall any other package.
#
#     After a package has been installed, it is possible to import all its
#  exported commands into another namespace via K<::Simple::Package::import>
#  and unimport them via K<::Simple::Package::unimport>.
#
#     Finall, use K<::Simple::Package::resync> to resynchronize the
#  information about available packages.
#
# -keywords   : package alias export import install load require unknown
# -api        :
#
#  * K<::Simple::Package::information state> package ?version?
#    Returns the state of a package.
#
#  * K<::Simple::Package::require> ?-exact? package ?version?
#    Requires a package.
#
#  * K<::Simple::Package::declare> package version
#       ?-requisites packagerequisite-list? ?-namespaces namespace-list?
#       ?-exportedcommands qualifiedname-list? ?-aliases list-list?
#       ?-install script? ?-uninstall script? ?-firsttimeinstall script?
#    Declares a package.
#
#  * K<::Simple::Package::information packagenamespace> namespace
#    Returns the declared package a namespace belongs to.
#
#  * K<::Simple::Package::information collisions> package ?namespace?
#    Returns the list of collisions upon importing a declared package.
#
#  * K<::Simple::Package::install> package
#    Installs a package.
#
#  * K<::Simple::Package::require-and-install> ?-exact? package ?version?
#    Requires and installs a package.
#
#  * K<::unknown> args
#    Augments the unknown mechanism to autoload extra packages.
#
#  * K<::Simple::Package::configure>
#    Configures the package options.
#
#  * K<::Simple::Package::cget>
#    Gets the package options.
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageRequisites
#
#  } -description {
#
#     Array containing for each package its requisites in the
#     format ||{?-exact?  package ?version?}||
#
#  } -type packagerequisite-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageRequiredPackages
#
#  } -description {
#
#     Array containing for each package its required packages in
#     the format ||{package version}||
#
#  } -type nameandversion-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageExportedCommands
#
#  } -description {
#
#     Array containing for each package its exported commands
#
#  } -type qualifiedname-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageAliases
#
#  } -description {
#
#     Array containing for each package its aliases
#
#  } -type list-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageInstallScript
#
#  } -description {
#
#     Array containing for each package its install script
#
#  } -type script-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageFirstInstallScript
#
#  } -description {
#
#     Array containing for each package its first time install
#     script
#
#  } -type script-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageUninstallScript
#
#  } -description {
#
#     Array containing for each package its uninstall script
#
#  } -type script-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageNamespaces
#
#  } -description {
#
#     Array containing for each package its namespaces
#
#  } -type namespace-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageVariables
#
#  } -description {
#
#     Array containing for each package its variables
#
#  } -type qualifiedname-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageOptions
#
#  } -description {
#
#     Array containing for each package its options
#
#  } -type flag-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageErrors
#
#  } -description {
#
#     Array containing for each package its errors
#
#  } -type qualifiedname-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageTypes
#
#  } -description {
#
#     Array containing for each package its types
#
#  } -type type-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageClasses
#
#  } -description {
#
#     Array containing for each package its classes
#
#  } -type namespace-list-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageMetadata
#
#  } -description {
#
#     Array containing for each package its metadata
#
#  } -type any-pairs-array
#
# -variable   :
#
#  {
#     ::Simple::Package::InstalledPackages
#
#  } -description {
#
#     Array whos elements are the installed declared packages
#
#  } -type any-array
#
# -variable   :
#
#  {
#     ::Simple::Package::DeclaredPackages
#
#  } -description {
#
#     Array whos elements are the declared packages
#
#  } -type any-array
#
# -variable   :
#
#  {
#     ::Simple::Package::PackageImportedInto
#
#  } -description {
#
#     Array containing for each package the list of packages it
#     has been imported into
#
#  } -type namespace-array
#
# -variable   :
#
#  {
#     ::Simple::Package::InformationSubcommands
#
#  } -description {
#
#     List of subcommands for the K<::Simple::Package::information>
#     base command
#
#  } -type word-list
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#
#  # Create an ifneeded script for a package which does not declare it
#  package ifneeded UndeclaredPackage 1.0 {
#     package provide UndeclaredPackage 1.0
#  }
#
#  # Create an ifneeded script for a package declaring it
#  package ifneeded DeclaredPackage 1.0 {
#     ::Simple::Package::declare DeclaredPackage 1.0 -requisites {
#        {UndeclaredPackage 1.0}
#     } -namespaces {
#        ::Declared
#     } -exportedcommands {
#        ::Declared::export
#     } -aliases {
#        {alias ::Declared::public}
#     } -install {
#        puts {Installing DeclaredPackage 1.0}
#        set ::Declared::PackageVariable 1
#     } -uninstall {
#        puts {Uninstalling DeclaredPackage 1.0}
#     } -firsttimeinstall {
#        puts {   for the first time!}
#        proc ::Declared::export {} {puts export}
#        proc ::Declared::public {} {puts public}
#     }
#  }
#
#  # No packages initially installed
#  # This displays the following:
#  #   |installed: {}
#  puts "installed: {[::Simple::Package::information installed *eclared*]}"
#
#  # Now require the declared package
#  ::Simple::Package::require DeclaredPackage
#
#  # The undeclared package has been installed
#  # This displays the following:
#  #   |installed: {{UndeclaredPackage 1.0}}
#  puts "installed: {[::Simple::Package::information installed *eclared*]}"
#
#  # Create a couple of commands which would collide with the package exported
#  # commands and aliases upon installing it
#  proc export {} {}
#  proc alias {} {}
#
#  # Which commands are exported by the declared package ?
#  # This displays the following:
#  #   |export: {export}
#  puts "export: {[::Simple::Package::information\ 
#     exportedcommands DeclaredPackage]}"
#
#  # Which commands are exported by the declared package ?
#  # This displays the following:
#  #   |aliases: {{alias ::Declared::public}}
#  puts "aliases: {[::Simple::Package::information aliases DeclaredPackage]}"
#
#  # Would those command or aliases collide with other commands or aliases
#  # in the global namespace ?
#  # This displays the following:
#  #   |collisions: {::Declared::export alias}
#  puts "collisions:\ 
#     {[::Simple::Package::information collisions DeclaredPackage]}"
#
#  # Let's remove the colliding commands and assess again whether the
#  # collisions remain
#  rename export {}
#  rename alias {}
#  # This displays the following:
#  #   |collisions: {}
#  puts "collisions:\ 
#     {[::Simple::Package::information collisions DeclaredPackage]}"
#
#  # No collisions, we may safely install the declared package
#  # This displays the following:
#  #   |Installing DeclaredPackage 1.0
#  #   |   for the first time!
#  ::Simple::Package::install DeclaredPackage
#
#  # Increase and display the package variable
#  # This displays the following:
#  #   |2
#  puts [incr ::Declared::PackageVariable]
#
#  # Into which namespaces has the declared package been imported ?
#  # This displays the following:
#  #   |importedinto: {::}
#  puts "importedinto:\ 
#     {[::Simple::Package::information importedinto DeclaredPackage]}"
#
#  # Which commands are provided by the declared package ?
#  # This displays the following:
#  #   |commands: {::Declared::export ::Declared::public}
#  puts "commands:\ 
#     {[::Simple::Package::information commands DeclaredPackage]}"
#
#  # Ok, use them
#  # This displays the following:
#  #   |public
#  ::Declared::public
#  # This displays the following:
#  #   |export
#  export
#
#  # We are done, let's get rid of the declared package
#  # This displays the following:
#  #   |Uninstalling DeclaredPackage 1.0
#  ::Simple::Package::uninstall DeclaredPackage
#
#  # The package commands are no longer available
#  # This displays the following:
#  #   |invalid command name "export"
#  catch {export} result
#  puts $result
#
#  # The undeclared package remains installed
#  # This displays the following:
#  #   |installed: {{UndeclaredPackage 1.0}}
#  puts "installed: {[::Simple::Package::information installed *eclared*]}"
#
#  # We need the declared package again!
#  # This displays the following:
#  #   |Installing DeclaredPackage 1.0
#  ::Simple::Package::require-and-install DeclaredPackage
#
#  # Verify that the package variable has its default value
#  # This displays the following:
#  #   |1
#  puts $::Declared::PackageVariable
#
# -todo       :
#
#  * Currently all packages required by a given package are installed when
#    that package is installed.  It would be nice to have a mechanism so that
#    packages not explicitly installed and no longer required by any
#    other could be automatically uninstalled.
#
#  * Provide a K<::Simple::Package::reset> command to reset the state of an
#    installed package by uninstalling and reinstalling it. If effect, this
#    may reduce to evaluate the uninstall and install scripts in turn.
#
#  * Provide something similar to K<::Simple::Package::information collisions>
#    but informing about the collisions when importing a namespace (instead of
#    a package) into another namespace.
#
#  * Allow a ||{package version}|| pair where a package name is expected in
#    the user procedures, so that it is possible to call the procedure either
#    with a single argument ||{package version}|| or ||package|| or with two
#    arguments ||package|| and ||version||
#
#  * Detect circular dependencies.
#
#  * Provide a K<::Simple::Package::modify> command.
#
#  * The install and uninstall package scripts should be stored as procedures
#    instead as of variables in order to benefit from its byte compilation.
#
# -history    :
#
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  23-apr-2002   "Extra" packages functionality, version 0.4.1
#  17-feb-2003   Third public release, version 0.5
#  29-mar-2003   Extra package, version 0.5.1
#  18-apr-2004   Alias can be qualified, version 0.5.2
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
#   Utility procedure to declare the package
namespace eval ::Simple::Package::Priv {}
proc ::Simple::Package::Priv::declare-package {} {

# Require P<SimpleBase> and P<SimpleError>
package require SimpleBase
package require SimpleError

::Simple::Package::declare SimplePackage 1.0 -requisites {

   {SimpleError 1.0}

} -namespaces {

   ::Simple::Package
   ::Simple::Package::Priv

} -install {

### Create the package variables
array set ::Simple::Package::PackageRequisites [list]
array set ::Simple::Package::PackageRequiredPackages [list]
array set ::Simple::Package::PackageExportedCommands [list]
array set ::Simple::Package::PackageAliases [list]
array set ::Simple::Package::PackageInstallScript [list]
array set ::Simple::Package::PackageFirstInstallScript [list]
array set ::Simple::Package::PackageUninstallScript [list]
array set ::Simple::Package::PackageNamespaces [list]
array set ::Simple::Package::PackageVariables [list]
array set ::Simple::Package::PackageOptions [list]
array set ::Simple::Package::PackageErrors [list]
array set ::Simple::Package::PackageTypes [list]
array set ::Simple::Package::PackageClasses [list]
array set ::Simple::Package::PackageMetadata [list]
array set ::Simple::Package::InstalledPackages [list]
array set ::Simple::Package::DeclaredPackages [list]
array set ::Simple::Package::PackageImportedInto [list]
set ::Simple::Package::InformationSubcommands [list\
   aliases aliasesall classes collisions commands declared\
   declaredexportedcommands errors exportedcommands exportedcommandsall\
   importedinto install installed known namespaces options packagealias\
   packagecommand packagenamespace privatecommands publiccommands required\
   requiredall requiredby requisites state types uninstall uninstalled\
   variables]

### Create the package options

} -uninstall {

### Delete the package variables
unset ::Simple::Package::PackageRequisites
unset ::Simple::Package::PackageRequiredPackages
unset ::Simple::Package::PackageExportedCommands
unset ::Simple::Package::PackageAliases
unset ::Simple::Package::PackageInstallScript
unset ::Simple::Package::PackageFirstInstallScript
unset ::Simple::Package::PackageUninstallScript
unset ::Simple::Package::PackageNamespaces
unset ::Simple::Package::PackageVariables
unset ::Simple::Package::PackageOptions
unset ::Simple::Package::PackageErrors
unset ::Simple::Package::PackageTypes
unset ::Simple::Package::PackageClasses
unset ::Simple::Package::PackageMetadata
unset ::Simple::Package::InstalledPackages
unset ::Simple::Package::DeclaredPackages
unset ::Simple::Package::PackageImportedInto
unset ::Simple::Package::InformationSucommands

### Delete the package options

} -firsttimeinstall {

### Declare the package types
#   Utility procedure to declare the package types
proc ::Simple::Package::Priv::declare-types {
} {
   # Package name
   ::Simple::Type::declare {
      package
   } -description {
      Package name: ''Tcl'', ...
   } -matchingscript {
      ::Simple::Package::Priv::assert-valid-package $value
      return 1
   }

   # Package version
   ::Simple::Type::declare {
      version
   } -description {
      Package version: ''1.1'', ...
   } -matchingscript {
      ::Simple::Package::Priv::assert-valid-version $value
      return 1
   }

   # Package name and version pair
   ::Simple::Type::declare {
      nameandversion
   } -description {
      Package name and version as ||{package version}||: ''{Tcl 8.4}''
   } -matchingscript {
      ::Simple::Package::Priv::assert-valid-package [lindex $value 0]
      ::Simple::Package::Priv::assert-valid-version [lindex $value 1]
      return [expr {[llength $value] == 2}]
   }

   # Package requisite
   ::Simple::Type::declare {
      packagerequisite
   } -description {
      Package requisite as ||{?-exact? package ?requisiteVersion?}||: ''Tcl 8'', ''-exact Tk 8.3'', ...
   } -matchingscript {
      ::Simple::Package::Priv::assert-valid-requisite $value
      return 1
   }

   # Delete this procedure
   rename [info level 0] {}
}

### Declare the package errors
::Simple::Error::declare {
   ::Simple::Package::NAMESPACE-OR-CLASS-OWNED 
} -message {
   namespace or class "%s" owned by package "%s"
} -explanation {
   Namespace or class <namespace or class> is already owned by package
   <package> and a namespace or class can not be shared by two packages
}

::Simple::Error::declare {
   ::Simple::Package::NAMESPACE-OR-CLASS-NOT-OWNED 
} -message {
   namespace or class "%s" not owned by package "%s"
} -explanation {
   Package <package> is using the namespace or class <namespace or class> as
   if it were its owner, but that is not the case
} -translation {<package> <namespace or class>}

::Simple::Error::declare {
   ::Simple::Package::NAMESPACE-OR-CLASS-NOT-OWNED-NOR-REQUISITES
} -message {
   namespace or class "%s" not owned by package "%s" nor any of its requisites
} -explanation {
   Package <package> has tried to be declared with an alias in a namespace or
   class <namespace or class> not owned by the package nor any of its direct
   requisites
} -translation {<package> <namespace or class>}

::Simple::Error::declare {
   ::Simple::Package::UNDECLARED 
} -message {
   undeclared package "%s"
} -explanation {
   Package <package> is undeclared, but a declared package is required
} -corrective {
   Use K<::Simple::Package::declare> or K<declare-package> to declare the
   package
}

::Simple::Error::declare {
   ::Simple::Package::UNINSTALLED 
} -message {
   uninstalled package "%s"
} -explanation {
   Package <package> is uninstalled, but an installed package is required
} -corrective {
   Use K<::Simple::Package::install> to install the package
}

::Simple::Error::declare {
   ::Simple::Package::INSTALLED 
} -message {
   installed package "%s"
} -explanation {
   Package <package> is installed, but an uninstalled package is required
} -corrective {
   Use K<::Simple::Package::uninstall> to uninstall the package
}

::Simple::Error::declare {
   ::Simple::Package::ERROR-EVAL-SCRIPT 
} -message {
   error evaluating package "%s" %s script: %s
} -explanation {
   Error <error> was thrown when evaluating the <install, first time install
   or uninstall> script for package <package>
} -corrective {
   Correct the corresponding script
} -translation {<package> <install, first time install or uninstall> <error>}

::Simple::Error::declare {
   ::Simple::Package::ALREADY-DECLARED 
} -message {
   package "%s %s" already declared
} -explanation {
   Version <version> of package <package> has already been declared via
   K<::Simple::Package::declare> or K<declare-package>
} -corrective {
   Use K<::Simple::Package::undeclare> to undeclare the package
} -translation {<package> <version>}

::Simple::Error::declare {
   ::Simple::Package::COLLISION 
} -message {
   couldn't %s package "%s": command (or alias) "%s" in %s collides with
   %s "%s" (package "%s")
} -explanation {
   It was impossible to <import or install> package <package> because
   the existing command or alias <command or alias> in namespace
   <namespace> collides (that is, their names are identical) with
   <command or alias> <exported command or alias> from package <colliding
   package>.  This last package either is the one being imported or
   installed or is required by it
} -corrective {
   Either delete or rename the command or alias <command or alias>.  Use
   K<::Simple::Package::information collisions> to ensure there are
   no collisions when importing or installing a package
}

::Simple::Error::declare {
   ::Simple::Package::COLLISION-BETWEEN-PACKAGES 
} -message {
   couldn't %s package "%s": command (or alias) "%s" is exported (or
   created) by both package "%s"%s and "%s"%s
} -explanation {
   It was impossible to <import or install> package <package> because it
   exports (or creates) command (or alias) <command or alias> in two
   packages <package 1> and <package 2>.  This means that
   package <package> is badly formed
} -corrective {
   Correct package <package>
} -explained {
   couldn't <import or install> package "<package>": command (or alias)
   "<command or alias>" is exported or created by both package "<package
   1>" [(namespace "<namespace 1>")] and "<package 2>" [(namespace
   "<namespace 2>")]
}

::Simple::Error::declare {
   ::Simple::Package::ALREADY-IMPORTED 
} -message {
   package "%s" already imported into namespace "%s"
} -explanation {
   It was impossible to import package <package> into namespace
   <namespace> because it was already imported into it
} -corrective {
   Use K<::Simple::Package::unimport> to unimport the package from the
   namespace
}

::Simple::Error::declare {
   ::Simple::Package::NOT-IMPORTED 
} -message {
   package "%s" not imported into namespace "%s"
} -explanation {
   It was impossible to unimport package <package> from namespace
   <namespace> because it was not imported into it
} -corrective {
   Use K<::Simple::Package::import> to import the package into the namespace
}

::Simple::Error::declare {
   ::Simple::Package::EMPTY-NAMESPACES-LIST 
} -message {
   empty namespaces list
} -explanation {
   The list of namespaces given in a package declaration is empty, but this is
   not allowed
}

}

# Declare previous packages
::Simple::Priv::declare-package
::Simple::Error::Priv::declare-package

# Delete this procedure
rename [info level 0] {}

}

### ===========================================================================
### -command  : ::Simple::Package::information state
# -synopsis   : Returns the state of a package.
# -overview   :
#
#     This procedure returns the state of the given package, one of
#  ''unknown'', ''available'', ''declared'' or ''installed''.  If no version
#  is given, the returned state is that of the known package with higher
#  version.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    :  
#
#     The state of the given package, one of ''unknown'', ''available'',
#  ''declared'' or ''installed''.
#
proc {::Simple::Package::information state} {
   package
   {version %DEFAULT%}
} {
   ### Version given
   set versionGiven [string compare $version %DEFAULT%]
   if {$versionGiven} {

      ### Get the version equivalent to that given
      set version\
         [::Simple::Package::Priv::equivalent-version $package $version]
   }

   ### Provided package
   set providedVersion [package provide $package]
   if {[string compare $providedVersion {}] && (!$versionGiven\
      || ![package vcompare $providedVersion $version])} {

      ### Build the package pair
      set packagePair [list $package $providedVersion]

      ### Declared but not installed package
      if {[info exists ::Simple::Package::DeclaredPackages($packagePair)] &&\
         ![info exists ::Simple::Package::InstalledPackages($packagePair)]} {

         ### Declared package
         set answer declared

      ### Undeclared ot declared and installed package
      } else {

         ### Installed package
         set answer installed
      }

   ### Non-provided package
   } else {

      ### Version given
      if {$versionGiven} {

         ### Check whether there is a package ifneeded script for that version
         set available [llength [package ifneeded $package $version]]

      ### Version not given
      } else {

         ### Check whether there is any version available
         set available [llength [package versions $package]]
      }

      ### Available package
      if {$available} {

         ### Available package
         set answer available

      ### Not available package
      } else  {

         ### Unknown package
         set answer unknown
      }
   }

   ### Return the state of the package
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::require
# -synopsis   : Requires a package.
# -overview   :
#
#     This procedure requires a package, that is, changes its state from
#  ''available'' to either ''required'' or ''installed'' depending on whether
#  the package is declared upon requiring it.
#
#  No error is thrown if the package is already provided.
#
#  An error is thrown if the package is not available.
#
# -arguments  :
#  {-exact            boolflag       {Only the specified version is acceptable}}
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    : The required package version.
#
proc ::Simple::Package::require {
   arg1
   {arg2 {}}
   {arg3 {}}
} {
   ### Require the package
   return [namespace eval :: package require $arg1 $arg2 $arg3]
}

### ===========================================================================
### -command  : ::Simple::Package::declare
# -synopsis   : Declares a package.
# -overview   :
#
#     This procedure is used to declare a package including its required
#  packages (which are required), namespaces (which are created), exported
#  commands (which are exported), aliases as well as the install, first time
#  install and uninstall scripts.
#
#     The aliases are created when the package is installed via the
#  K<::Simple::Package::install> command.  The install, first time install and
#  uninstall scripts are evaluated in the global namespace when the package is
#  installed via the K<::Simple::Package::install> command, installed for the
#  first time via the K<::Simple::Package::install> command, or uninstalled
#  via the K<::Simple::Package::uninstall> command, respectively.
#
#     An error is thrown if the package is already declared; use
#  K<::Simple::Package::undeclare> to undeclare a package.
#
#     Aliases can be qualified or not.  Qualified aliases must point to either
#  one of the package namespaces or to a namespace belonging to one of the
#  package direct requisites.
#
#     The most prominent difference between an alias and an exported command
#  is that the alias effect is interpreter-wide while that of the exported
#  command is namespace-wide.  As all package exported commands are imported
#  into the global namespace upon installing a package, place in which are
#  accessible even from other namespaces, that difference is in practice
#  almost irrelevant.  Aliases are thus provided by this package so that a
#  command can be accessed through two names, as the
#  K<::Simple::Subcommand::create> command which can also be accessed as
#  K<proc-sub>.  Notice that the alias is valid to invoke the command only,
#  not being a full-fledged substitute at all times when a command name is
#  expected.  Notice also that a command call through an alias is slightly
#  slower than calling the command directly.
#
#     The first time install script should contain code intended to be
#  evaluated once, when a package is installed for the first time, as it gets
#  evaluated at that time and deleted afterwards.  On the other hand, the
#  install script is evaluated whenever a package is installed, including
#  reinstallations.  The rationale is that a package can be ''reset'' by
#  uninstalling and uninstalling it back.  The uninstall script is the
#  opposite to the install script. 
#
# -arguments  :
#  { package          package        {Package name}}
#  { version          version        {Package version}}
#  {-requisites       packagerequisite-list {}
#                                    {Package requisites}}
#  {-namespaces       namespace-list {}
#                                    {Package namespaces}}
#  {-exportedcommands qualifiedname-list {}
#                                    {Package qualified exported commands}}
#  {-aliases          list-list {}   {Package aliases in the format
#                                     ||{aliasName aliasedCommand}||}}
#  {-install          script {}      {Package install script}}
#  {-uninstall        script {}      {Package uninstall script}}
#  {-firsttimeinstall script {}      {Package first time install script}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Creates the package namespaces.
#
#  * Requires the required packages.
#
#  * Exports exported commands from their namespaces.
#
#  * Provides the package.
#
# -remarks    :
#
#  * The order of the flags is irrelevant but their parameters can not start
#    by an hyphen.
#
proc ::Simple::Package::declare {
   package
   version
   args
} {
   ### Parse and verify the package declaration
   if {[catch {

      # Parse flags
      set flags [::Simple::Priv::parse-flags\
         [list -requisites list\
         -namespaces list\
         -exportedcommands list\
         -aliases list\
         -install untouched\
         -uninstall untouched\
         -firsttimeinstall untouched]\
         [list] $args 0 1]

      # First section factored out
      ::Simple::Package::Priv::uplevel-declare-1

      # Second section factored out
      set classes [list]
      ::Simple::Package::Priv::uplevel-declare-2

   }]} {

      # Rethrow the error after rationalizing the stack call
      ::Simple::Error::Priv::push-location package $package 1
      eval [::Simple::Error::Priv::rationalize-call-stack]
   }

   ### Declare the package
   ::Simple::Package::Priv::do-declare\
      $packagePair $requisites $namespaces $exportedcommands\
      $aliases $install $uninstall $firsttimeinstall\
      [list] [list] [list] [list] [list] [list]

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::uplevel-declare-1
# -synopsis   : Declares a package (first section factored out).
# -remarks    :
#
#  * This is a full K<uplevel> procedure used to factor code out.
#
# -effects    :
#
#  * Modifies V<version>, V<requisitePackages> and V<packagePair>.
#
proc ::Simple::Package::Priv::uplevel-declare-1 {
} {
   ### Up one level
   uplevel 1 {

      ### Build the package pair
      set originalVersion $version
      set version\
         [::Simple::Package::Priv::equivalent-version $package $version]
      set packagePair [list $package $version]

      ### Assert the package has not been declared
      if {[info exists ::Simple::Package::DeclaredPackages($packagePair)]} {
         ::Simple::Error::throw\
            ::Simple::Package::ALREADY-DECLARED $package $originalVersion
      }

      ### Handle A<-requisites>
      #   Assert the requisites are valid
      set requisitePackages [list]
      foreach requisite $requisites {
         lappend requisitePackages\
            [::Simple::Package::Priv::assert-valid-requisite $requisite]
      }

      ### Handle A<-namespaces>
      if {[lsearch -exact $flags -namespaces] != -1} {

         # Assert no namespace belongs to other package
         ::Simple::Package::Priv::assert-namespaces-not-owned\
            $packagePair $namespaces
      }

      ### Require the required packages
      ::Simple::Package::Priv::require-packages $requisites
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::uplevel-declare-2
# -synopsis   : Declares a package (second section factored out).
# -remarks    :
#
#  * This is a full K<uplevel> procedure used to factor code out.
#
# -effects    :
#
#  * Modifies V<aliases>.
#
proc ::Simple::Package::Priv::uplevel-declare-2 {
} {
   ### Up one level
   uplevel 1 {

      # Loop over package aliases
      set namespacesAndClasses [concat $namespaces $classes]
      set newAliases [list]
      foreach aliasPair $aliases {
         foreach {alias aliased} $aliasPair break
         lappend newAliases [list $alias $aliased]

         # Assert the alias namespace, if qualified, is owned by the package
         #    or one of its requisites
         if {[string match ::* $alias]} {
            ::Simple::Package::Priv::assert-package-owns-namespace-or-class\
               $packagePair $alias $namespacesAndClasses $requisitePackages
         }

         # Assert the aliased command is qualified
         # The alias can be both qualified or not
         if {![string match ::* $aliased]} {
            ::Simple::Error::throw ::Simple::BAD-ITEM\
               {qualified aliased command name} $aliased
         }

         # Assert the aliased command namespace or class
         #    is owned by the package
         ::Simple::Package::Priv::assert-package-owns-namespace-or-class\
            $packagePair $aliased $namespacesAndClasses
      }
      set aliases $newAliases

      ### Loop over exported commands
      foreach exportedCommand $exportedcommands {

         ### Assert the exported command is qualified
         if {![string match ::* $exportedCommand]} {
            ::Simple::Error::throw ::Simple::BAD-ITEM\
               {qualified exported command name} $exportedCommand
         }

         ###    Assert the exported command namespace or
         ### class is owned by the package
         ::Simple::Package::Priv::assert-package-owns-namespace-or-class\
            $packagePair $exportedCommand $namespacesAndClasses
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::require-packages
# -synopsis   : Requires a list of packages.
# -arguments  :
#  { requisites       packagerequisite-list
#                                    {Package requisites}}
# -returns    : 
#
#  The list of required packages in the format ||{package version}||.
#
# -effects    :
#
#  * Requires the given packages.
#
proc ::Simple::Package::Priv::require-packages {
   requisites
} {
   ### Require the required packages
   set answer [list]
   foreach requisite $requisites {

      # -exact given
      if {[llength $requisite] == 3} {
         set requisitePackage [lindex $requisite 1]

      # -exact not given
      } else {
         set requisitePackage [lindex $requisite 0]
      }

      # Require the package
      lappend answer\
         [list $requisitePackage [::Simple::Package::require $requisite]]
   }

   ### Return the list of required packages in the format ||{package version}||
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::do-declare
# -synopsis   : Declares a package.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { requisites       packagerequisite-list
#                                    {Package requisites}}
#  { namespaces       namespace-list {Package namespaces}}
#  { exportedCommands qualifiedname-list
#                                    {Package qualified exported commands}}
#  { aliases          list-list      {Package aliases in the format
#                                     ||{aliasName aliasedCommand}||}}
#  { install          script         {Package install script}}
#  { uninstall        script         {Package uninstall script}}
#  { firstTimeInstall script         {Package first time install script}}
#  { variables        qualifiedname-list
#                                    {Package variables}}
#  { options          flag-list      {Package options}}
#  { errors           qualifiedname-list
#                                    {Package errors}}
#  { types            type-list      {Package types}}
#  { classes          qualifiedname-list
#                                    {Package classes}}
#  { metadata         any-pairs      {Package metadata}}
#
# -effects    :
#
#  * Creates the package namespaces (but does not create the classes nor its
#    associated namespaces).
#
#  * Exports commands from their namespaces.
#
#  * Provides the package.
#
proc ::Simple::Package::Priv::do-declare {
   packagePair
   requisites
   namespaces
   exportedCommands
   aliases
   install 
   uninstall
   firstTimeInstall
   variables
   options
   errors
   types
   classes
   metadata
} {
   ### Require the required packages
   set requiredPackages [::Simple::Package::Priv::require-packages $requisites]

   ### Store the package namespaces
   set ::Simple::Package::PackageNamespaces($packagePair) $namespaces

   ### Store the package variables
   set ::Simple::Package::PackageVariables($packagePair) $variables

   ### Store the package options
   set ::Simple::Package::PackageOptions($packagePair) $options

   ### Store the package errors
   set ::Simple::Package::PackageErrors($packagePair) $errors

   ### Store the package types
   set ::Simple::Package::PackageTypes($packagePair) $types

   ### Store the package classes
   set ::Simple::Package::PackageClasses($packagePair) $classes

   ### Store the package requisites
   set ::Simple::Package::PackageRequisites($packagePair) $requisites

   ### Store the package required packages
   set ::Simple::Package::PackageRequiredPackages($packagePair)\
      $requiredPackages

   ### Create the namespaces
   ::Simple::Package::Priv::create-namespaces $namespaces

   ### Store the package exported commands
   set ::Simple::Package::PackageExportedCommands($packagePair)\
      $exportedCommands

   ### Export the exported commands
   ::Simple::Package::Priv::export $packagePair $exportedCommands

   ### Store the package aliases
   set ::Simple::Package::PackageAliases($packagePair) $aliases

   ### Store the package install script
   set ::Simple::Package::PackageInstallScript($packagePair) $install

   ### Store the package first time install script
   set ::Simple::Package::PackageFirstInstallScript($packagePair)\
      $firstTimeInstall

   ### Store the package uninstall script
   set ::Simple::Package::PackageUninstallScript($packagePair) $uninstall

   # The package metadata must not be stored
   if {![set ::Simple::%Options%(-storemetadata)]} {
      set metadata [list]
   }

   ### Store the package metadata
   set ::Simple::Package::PackageMetadata($packagePair) $metadata

   ### Add this package to the list of declared packages
   ::Simple::Package::Priv::add-declared $packagePair

   ### Provide the package
   package provide [lindex $packagePair 0] [lindex $packagePair 1]
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-valid-requisite
# -synopsis   : Throws an error if a requisite format is not valid.
# -arguments  :
#  { requisite        packagerequisite
#                                    {Requisite}}
#
# -returns    : The requisite package name.
#
proc ::Simple::Package::Priv::assert-valid-requisite {
   requisite
} {
   # One, two or three elements list
   set requisiteLength [llength $requisite]
   if {$requisiteLength < 1 || $requisiteLength > 3} {
      ::Simple::Error::throw ::Simple::BAD-ITEM requisite $requisite
   }

   # Switch on number of elements
   switch -exact -- $requisiteLength {
      1 {
         set package $requisite
         set requisiteVersion 0
      }

      2 {
         foreach {package requisiteVersion} $requisite break
      }

      3 {
         if {[string compare [lindex $requisite 0] -exact]} {
            ::Simple::Error::throw ::Simple::BAD-ITEM requisite $requisite
         }
         foreach {tmp package requisiteVersion} $requisite break
      }
   }

   # Assert the package name is valid
   ::Simple::Package::Priv::assert-valid-package $package

   # Assert the package version is valid
   ::Simple::Package::Priv::assert-valid-version $requisiteVersion

   # Return the requisite package name
   return $package
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-valid-package
# -synopsis   : Throws an error if a package name is not valid.
# -arguments  :
#  { package          package        {Package name}}
# -details    :
#
#  * The package name can have the form "Img::jpg" because Jan Nijtmans'
#    Img package uses that.
#
proc ::Simple::Package::Priv::assert-valid-package {
   package
} {
   if {![regexp\
      {^([_a-zA-Z0-9%][-_a-zA-Z0-9%]*((::[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)*)|)$}\
      $package]} {

      ::Simple::Error::throw ::Simple::BAD-ITEM {package name} $package
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-valid-version
# -synopsis   : Throws an error if a version is not valid.
# -arguments  :
#  { version          version        {Version}}
#
proc ::Simple::Package::Priv::assert-valid-version {
   version
} {
   package vcompare 0 $version
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::add-declared
# -synopsis   : Adds a package to the list of declared packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::add-declared {
   packagePair
} {
   set ::Simple::Package::DeclaredPackages($packagePair) {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::add-installed
# -synopsis   : Adds a package to the list of installed packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::add-installed {
   packagePair
} {
   set ::Simple::Package::InstalledPackages($packagePair) {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-namespaces-not-owned
# -synopsis   : Asserts no namespaces from a list are owned by any package.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { namespaces       namespace-list {Namespaces}}
#
# -returns    : The empty string.
#
proc ::Simple::Package::Priv::assert-namespaces-not-owned {
   packagePair
   namespaces
} {
   ### Assert the namespaces list is not empty
   if {[llength $namespaces] == 0} {
      ::Simple::Error::throw ::Simple::Package::EMPTY-NAMESPACES-LIST
   }

   ### Loop over namespaces
   foreach namespace $namespaces {
      foreach declaredPackage\
         [array names ::Simple::Package::DeclaredPackages] {

         # Bug 6: bad handling of "A loads B which requires A" situation
         # Solution: skip the very same package
         if {[string compare $packagePair $declaredPackage] == 0} {
            continue
         }

         if {[lsearch -exact [::Simple::Package::Priv::namespaces-and-classes\
            $declaredPackage] $namespace] != -1} {

            ::Simple::Error::throw ::Simple::Package::NAMESPACE-OR-CLASS-OWNED\
               $namespace $declaredPackage
         }
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::namespaces-and-classes
# -synopsis   : Returns a declared package list of namespaces and classes.
# -returns    : The package list of namespaces and classes.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::namespaces-and-classes {
   packagePair
} {
   return [concat $::Simple::Package::PackageNamespaces($packagePair)\
      $::Simple::Package::PackageClasses($packagePair)]
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::create-namespaces
# -synopsis   : Creates a list of namespaces.
# -arguments  :
#  { namespaces       namespace-list {Namespaces}}
#
# -remarks    :
#
#  * The namespaces may already exists and no error is thrown.
#
# -effects    :
#
#  * Creates the package namespaces.
#
proc ::Simple::Package::Priv::create-namespaces {
   namespaces
} {
   foreach namespace $namespaces {
      namespace eval $namespace {}
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::export
# -synopsis   : Exports a list of commands.
# -overview   :
#
#     This procedure exports a list of commands.  The corresponding namespace,
#  which must belong to the given package, is extracted from each
#  qualified command name.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { exportedCommands qualifiedname-list
#                                    {Qualified exported commands}}
# -effects    :
#
#  * Exports commands from their namespaces.
#
proc ::Simple::Package::Priv::export {
   packagePair
   exportedCommands
} {
   ### Loop over exported commands
   foreach exportedCommand $exportedCommands {

      ### Export the command
      set namespace [namespace qualifiers $exportedCommand]
      if {![string compare $namespace {}]} {
         set namespace ::
      }
      namespace eval $namespace\
         [format {namespace export %s} [list [namespace tail $exportedCommand]]]
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-package-owns-namespace-or-class
#
# -synopsis   : Throws an error if a namespace or class is not owned by a package.
#
# -overview   :
#
#     This procedure throws an error if the given command namespace
#  or class is not owned by the given package. The package namespaces and
#  classes are assumed to be given in the V<namespacesAndClasses> argument.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { item             qualifiedname  {Qualified item name}}
#  { namespacesAndClasses
#                     namespace-list {Namespaces and classes list}}
#  {?requisitePackages?
#                     package-list %DEFAULT%
#                                    {Package requisites}}
#
# -returns    : The empty string.
#
proc ::Simple::Package::Priv::assert-package-owns-namespace-or-class {
   packagePair
   item
   namespacesAndClasses
   {requisitePackages %DEFAULT%}
} {

   ### Get the item namespace or class
   set namespaceOrClass [namespace qualifiers $item]
   if {![string compare $namespaceOrClass {}]} {
      set namespaceOrClass ::
   }

   ### The namespace or class does not belong to the package
   if {[lsearch -exact $namespacesAndClasses $namespaceOrClass] == -1} {

      # The namespace does not belong to a required package
      foreach {packageOwns packageOwnsVersion}\
         [{::Simple::Package::information packagenamespace}\
         $namespaceOrClass] break
      if {![info exists packageOwns] ||\
         [lsearch -exact $requisitePackages $packageOwns] == -1} {

         # Throw error
         if {[llength $requisitePackages] != 1 ||\
            [string compare $requisitePackages %DEFAULT%]} {

            ::Simple::Error::throw\
               ::Simple::Package::NAMESPACE-OR-CLASS-NOT-OWNED-NOR-REQUISITES\
               $namespaceOrClass $packagePair
         } else {
            ::Simple::Error::throw\
               ::Simple::Package::NAMESPACE-OR-CLASS-NOT-OWNED\
               $namespaceOrClass $packagePair
         }
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::equivalent-version
# -synopsis   : Returns a package equivalent version.
# -overview   :
#
#     This procedure returns the given package version possibly modified to
#  exactly match an already known equivalent version.  For example, if
#  K<package ifneeded> has been evaluated for "foo 1.0" and this procedure is
#  called for "foo 1", the returned version is 1.0.
#
# -arguments  :
#  { package          package        {Package name}}
#  { version          version        {Package version}}
#
# -returns    : The package equivalent version.
#
proc ::Simple::Package::Priv::equivalent-version {
   package
   version
} {
   ### Assert the given version is valid
   ::Simple::Package::Priv::assert-valid-version $version

   ### Return the given version by default
   set answer $version

   ### Loop over known versions
   #   The provided version has precedence over equivalent ifneeded version
   foreach knownVersion\
      [concat [package provide $package] [package versions $package]] {

      ### Given version equivalent to known version
      if {![package vcompare $version $knownVersion]} {

         ### Return the known version
         set answer $knownVersion
         break
      }
   }

   ### Return the package equivalent version
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::package-pair
# -synopsis   : Validates a package name and version and returns the pair.
# -overview   :
#
#     This procedure validates a package version, if given, or gets that of
#  any declared package with the given name otherwise; in that case, if there
#  is no declared package with that name, an error is thrown, unless the
#  optional argument A<allowInstalled> is set, case in which if the package is
#  not installed, the error is finally thrown.
#
#  The package name and version pair is returned.
#
# -arguments  :
#  { package          package        {Package name}}
#  { version          version        {Package version}}
#  {?allowInstalled?  boolean 0      {Allow installed packages}}
#
# -returns    : The validated package name and version pair.
# -limitations:
#
#  * The version is assumed to have not been given to the caller of this
#    procedure if its value is equal to ''%DEFAULT%''.
#
proc ::Simple::Package::Priv::package-pair {
   package
   version
   {allowInstalled 0}
} {
   ### Version given
   if {[string compare $version %DEFAULT%]} {

      ### Get the version equivalent to that given
      set version\
         [::Simple::Package::Priv::equivalent-version $package $version]

   ### No version given
   } else {

      ### There is a declared package with that name
      set packages\
         [array names ::Simple::Package::DeclaredPackages "$package *"]
      if {[llength $packages] > 0} {

         ### Get its version
         set version [lindex [lindex $packages 0] 1]

      ### There is no declared package with that name
      } else {

         ### Installed packages allowed
         if {$allowInstalled} {

            ### Package not installed
            set version [package provide $package]
            if {![string compare $version {}]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::Package::UNINSTALLED $package
            }
            ### Installed package

         ### Installed packages not allowed
         } else {

            ### Throw error
            ::Simple::Error::throw ::Simple::Package::UNDECLARED $package
         }
      }
   }

   ### Return the package name and version pair
   return [list $package $version]
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-namespace-exists
# -synopsis   : Throws an error if a namespace does not exist.
# -arguments  :
#  { namespace        namespace      {Namespace}}
#
proc ::Simple::Package::Priv::assert-namespace-exists {
   namespace
} {
   if {[catch {namespace eval :: "namespace children $namespace"}]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING\
         namespace $namespace
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-package-declared
# -synopsis   : Throws an error if a package has not been declared.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::assert-package-declared {
   packagePair
} {
   if {![info exists ::Simple::Package::DeclaredPackages($packagePair)]} {
      ::Simple::Error::throw ::Simple::Package::UNDECLARED $packagePair
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-package-not-installed
# -synopsis   : Throws an error if a package is installed.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::assert-package-not-installed {
   packagePair
} {
   if {[info exists ::Simple::Package::InstalledPackages($packagePair)]} {
      ::Simple::Error::throw ::Simple::Package::INSTALLED $packagePair
   }
}

### ===========================================================================
### -command  : ::Simple::Package::information packagenamespace
# -synopsis   : Returns the declared package a namespace belongs to.
# -overview   :
#
#     This procedure returns the declared package the given namespace belongs
#  to, or the empty string if either no declared package owns the given
#  namespace or the namespace does not exist.
#
# -arguments  :
#  { namespace        namespace      {Namespace}}
#
# -returns    : The declared package a namespace belongs to.
#
proc {::Simple::Package::information packagenamespace} {
   namespace
} {
   ### Loop over declared packages
   set answer {}
   foreach declaredPackage [array names ::Simple::Package::DeclaredPackages] {

      ### This package owns the namespace
      if {[lsearch -exact\
         $::Simple::Package::PackageNamespaces($declaredPackage)\
         $namespace] != -1} {

         set answer $declaredPackage
         break
      }
   }

   ### Return the declared package the namespace belongs to
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::required-not-installed-recursive
# -synopsis   : Returns a package full list of uninstalled required packages.
# -overview   :
#
#     This procedure returns the given package list of required packages plus
#  all not installed packages it may require.  It only takes into account the
#  required packages of declared packages.
#
# -returns    : The package full list of uninstalled required packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -references :
#
#  * The K<::Simple::Package::information requiredall> command.
#
#  * The K<::Simple::Package::Priv::required-not-instaled> command.
#
# -details    :
#
#  This command is recursive.
#
# -remarks    :
#
#  * No provision is made for removing duplicated packages, use
#    K<::Simple::Package::Priv::required-not-installed> instead.
#
proc ::Simple::Package::Priv::required-not-installed-recursive {
   packagePair
} {
   ### Loop over required packages
   set answer [list]
   foreach requiredPackagePair\
      $::Simple::Package::PackageRequiredPackages($packagePair) {

      ### The package has been declared
      if {[info exists\
         ::Simple::Package::DeclaredPackages($requiredPackagePair)]} {

         ### The package has not been installed
         if {![info exists\
            ::Simple::Package::InstalledPackages($requiredPackagePair)]} {

            ### Add this package
            lappend answer $requiredPackagePair

            ### Add this package uninstalled required packages
            set answer [concat $answer\
               [::Simple::Package::Priv::required-not-installed-recursive\
               $requiredPackagePair]]
         }
      }
   }

   ### Return the package full list of required packages
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::required-not-installed
# -synopsis   : Returns a package full list of uninstalled required packages.
# -overview   :
#
#     This procedure returns the given package list of required packages plus
#  all not installed packages it may require.  It only takes into account the
#  required packages of declared packages.  In effect, it removes the
#  duplicates from the list returned by
#  K<::Simple::Package::Priv::required-not-installed-recursive>.
#
# -returns    : The package full list of uninstalled required packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -references :
#
#  * The K<::Simple::Package::information requiredall> command.
#
#  * The K<::Simple::Package::Priv::required-not-instaled> command.
#
proc ::Simple::Package::Priv::required-not-installed {
   packagePair
} {
   foreach requiredPackagePair\
      [::Simple::Package::Priv::required-not-installed-recursive $packagePair] {

      set tmp($requiredPackagePair) {}
   }
   return [array names tmp]
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::exported-commands
# -synopsis   : Returns a declared package current list of exported commands.
# -overview   :
#
#     This procedure returns a package current list of exported commands as
#  given by the K<namespace export> command in each of its namespaces and
#  classes, that is, a snapshot of the currently exported commands is taken. 
#  Thus, the returned list of exported commands might be different from the
#  "exported commands list" argument to K<::Simple::Package::declare> if any
#  command was exported after the package declaration.
#
# -returns    : The package list of qualified exported commands.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -references :
#
#  * The K<::Simple::Package::information declaredexportedcommands> command.
#
#  * The K<::Simple::Package::information exportedcommandsall> command.
#
#  * The K<::Simple::Package::information exportedcommands> command.
#
# -remarks    :
#
#  * Both package namespaces and classes are taken into account.
#
#  * The returned list may contain base commands.
#
#  * The returned list may contain commands with a percent sign in its name
#    are those are assumed to be internal to The Simple Development Library.
#
proc ::Simple::Package::Priv::exported-commands {
   packagePair
} {
   ### Loop over package namespaces and classes
   set answer [list]
   foreach namespaceOrClass\
      [::Simple::Package::Priv::namespaces-and-classes $packagePair] {

      ### Get the exported commands list from this namespace or class
      namespace eval $namespaceOrClass {
         set ::Simple::Package::%TMP% [namespace export]
      }

      ### Loop over exported commands
      foreach command ${::Simple::Package::%TMP%} {

         ### Add the command to the package list of exported commands
         lappend answer ${namespaceOrClass}::$command
      }
   }

   ### Return the package list of qualified exported commands
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::aliases
# -synopsis   : Returns a declared package list of aliases.
# -returns    : The package list of aliases.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::aliases {
   packagePair
} {
   ### Loop over package aliases
   set answer [list]
   foreach aliasPair $::Simple::Package::PackageAliases($packagePair) {

      ### Add the alias to the package list of alias
      lappend answer [lindex $aliasPair 0]
   }

   ### Return the package list of aliases
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::information collisions
# -synopsis   : Returns the list of collisions upon importing a declared package.
# -overview   :
#
#     This procedure returns the list of qualified exported commands and
#  aliases (qualified or not) which would collide if the given package were
#  imported into the given namespace.  Collisions occur when one of the
#  package exported commands is named as an already existing command in the
#  target namespace or as a alias, or when one of the package aliases is named
#  as an already existing global command or alias.
#
#     If no namespace is given, it defaults to the global one.  If the target
#  namespace is the global one, the list of collisions takes into account not
#  only the exported command and aliases from the given package but also those
#  from any not previously installed package it may require.
#
# -arguments  :
#  { package          package        {Package name}}
#  {?namespace?       namespace ::   {Target namespace}}
#
# -returns    : 
#
#     The list of colliding qualified exported commands and aliases (qualified
#  or not) upon importing the package.
#
proc {::Simple::Package::information collisions} {
   package
   {namespace ::}
} {
   ::Simple::Package::Priv::collisions $package $namespace %DEFAULT%
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::collisions
# -synopsis   : Returns the list of collisions upon importing a declared package.
#
# -arguments  :
#  { package          package        {Package name}}
#  { namespace        namespace      {Target namespace}}
#  { packagesToImport nameandversion-list
#                                    {Lists of packages to import}}
#
# -returns    : 
#
#     The list of colliding qualified exported commands and aliases (qualified
#  or not) upon importing a package.
#
# -details    :
#
#     This procedure is the implementation for
#  K<::Simple::Package::information collisions> with an extra argument which
#  is internally used to to avoid a call to
#  K<::Simple::Packages::Priv::packages-to-import> if the result of this
#  procedure is available.
#
proc {::Simple::Package::Priv::collisions} {
   package
   namespace
   packagesToImport
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package %DEFAULT%]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Assert the namespace exists
   ::Simple::Package::Priv::assert-namespace-exists $namespace

   ### Get all packages to import
   if {![string compare $packagesToImport %DEFAULT%]} {
      set packagesToImport\
         [::Simple::Package::Priv::packages-to-import $packagePair $namespace]
   }

   ### Get all packages to import exported commands and aliases
   #   Loop over packages to import
   set collisionCandidates [list]
   foreach packageToImport $packagesToImport {

      # Add this package exported commands and aliases
      set collisionCandidates [concat $collisionCandidates\
         [::Simple::Package::Priv::exported-commands $packageToImport]\
         [::Simple::Package::Priv::aliases $packageToImport]]
   }

   ### Get the namespace commands
   set namespaceCommandNames [list]
   foreach namespaceCommand [info commands ${namespace}::*] {
      lappend namespaceCommandNames [namespace tail $namespaceCommand]
   }

   ### Loop over collision candidates
   set answer [list]
   foreach collisionCandidate $collisionCandidates {

      ### There is a namespace command named as the collision cndidate
      if {[lsearch -exact $namespaceCommandNames\
         [namespace tail $collisionCandidate]] != -1} {

         ### Add this command to the list of collisions
         lappend answer $collisionCandidate
      }
   }

   ### Return the list of collisions
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::packages-to-import
# -synopsis   : Returns the full list of packages to import.
# -overview   :
#
#     This procedure returns the full list of packages to import into a given
#  namespace upon importing the given package.  If the target namespace is not
#  the global one, that list is made of the given package only; this
#  corresponds to the use by K<::Simple::Package::import>.  If the target
#  namespace is the global one, the list includes not only the given package
#  but also any not previously installed package it may require.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { namespace        namespace      {Namespace}}
#
# -returns    : The full list of packages to import.
#
proc ::Simple::Package::Priv::packages-to-import {
   packagePair
   namespace
} {
   ### Non-global namespace
   if {[string compare $namespace ::]} {

      ### The package is the only one to import
      set answer [list $packagePair]

   ### Global namespace
   } else {

      ### Add the package full list of uninstalled required
      ###    packages to the list of packages to import
      set answer\
         [::Simple::Package::Priv::required-not-installed $packagePair]

      ### Add the package to the list of packages to import
      lappend answer $packagePair
   }

   ### Return the full list of packages to import
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Package::install
# -synopsis   : Installs a package.
# -overview   :
#
#  This procedure installs a declared package.
#
#     The following tasks are performed in this same order:
#
#  * All package required packages not previously installed are installed.
#
#  * All package aliases are created.
#
#  * The package install script is evaluated in the global namespace.
#
#  * If this is the first time the package is installed after its declaration,
#    the package first time install script is evaluated in the global
#    namespace.
#
#  * All package exported commands (either from the package itself or from any
#    not previously installed package it may require) are imported into the
#    global namespace.
#
#     The install script is evaluated every time the package is installed,
#  while the first time install script is only evaluated the first time the
#  package is installed after its declaration; upon evaluation, the first time
#  install script is deleted.  The first time install script is intended for
#  the initialization of the package static items, notably its commands.  On
#  the other hand, the install script is intended for the initialization of
#  the package non-static items such as its variables.  Notice that all the
#  packages required by the package being installed are guaranteed to be
#  installed when evaluating the install and first time install scripts.
#
#     An error is thrown if the package is already installed; use
#  K<::Simple::Package::uninstall> to uninstall a package.
#
#     An error is thrown if importing the package into the global namespace
#  would result in a collision; use K<::Simple::Package::information
#  collisions> to ensure there are no collisions.
#
# -arguments  :
#  { package          package        {Package name}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Creates the package aliases.
#
#  * Those of the evaluated scripts.
#
#  * Imports the package exported commands into the global namespace.
#
proc ::Simple::Package::install {
   package
} {
   ### Build the package pair
   set packagePair [::Simple::Package::Priv::package-pair $package %DEFAULT%]

   ### Assert the package has been declared
   ::Simple::Package::Priv::assert-package-declared $packagePair

   ### Continue with the second part of the installation
   ::Simple::Package::Priv::install-2 $packagePair

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::install-2
# -synopsis   : Installs a package (second part).
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::install-2 {
   packagePair
} {
   ### Assert the package is not installed
   ::Simple::Package::Priv::assert-package-not-installed $packagePair

   ### Assert no collision will occur when importing
   ###    the package into the global namespace
   ::Simple::Package::Priv::assert-no-collisions $packagePair ::

   ### Install the package
   ::Simple::Package::Priv::do-install $packagePair
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::do-install
# -synopsis   : Installs a package.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
# -effects    :
#
#  * Creates the package aliases.
#
#  * Those of the evaluated scripts.
#
#  * Imports the package exported commands into the global namespace.
#
proc ::Simple::Package::Priv::do-install {
   packagePair
} {
   ### Install the package required packages
   ::Simple::Package::Priv::install-required $packagePair

   ### Error evaluating the package install script
   if {[catch {namespace eval ::\
      $::Simple::Package::PackageInstallScript($packagePair)} result]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Package::ERROR-EVAL-SCRIPT\
         $packagePair install $result

   ### No error evaluating the package install script
   } else {

      ### Error evaluating the package first time install script
      if {[catch {namespace eval ::\
         $::Simple::Package::PackageFirstInstallScript($packagePair)} result]} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Package::ERROR-EVAL-SCRIPT\
            $packagePair {first time install} $result

      ### No error evaluating the package first time install script
      } else {

         ### Create the package aliases
         foreach aliasPair $::Simple::Package::PackageAliases($packagePair) {
            foreach {alias aliased} $aliasPair break
            interp alias {} $alias {} $aliased
         }

         ### Unhide the package types
         if {[llength $::Simple::Package::PackageTypes($packagePair)] > 0} {
            ::Simple::Type::Priv::unhide\
               $::Simple::Package::PackageTypes($packagePair)
         }

         ### Clear the package first time install script
         set ::Simple::Package::PackageFirstInstallScript($packagePair) {}

         ### Import the package into the global namespace
         ::Simple::Package::Priv::do-import $packagePair ::

         ### Add this package to the list of installed packages
         ::Simple::Package::Priv::add-installed $packagePair
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::assert-no-collisions
# -synopsis   : Throws an error if importing a package would give a collision.
# -overview   :
#
#     This procedure throws an error if importing the given package into the
#  given namespace would result in a collision.  Collisions occur when one of
#  the package exported commands is named as an already existing command in
#  the target namespace or as a global alias, or when one of the package
#  aliases is named as an already existing global command or alias.
#
#     If the target namespace is the global one, the list of collisions takes
#  into account not only the exported command and aliases from the given
#  package but also those from any not previously installed package it may
#  require.
#
#  There are two possibilities:
#
#  * A namespace command or global alias collides with an exported command
#    or package alias.
#
#  * An exported command or package alias from a namespace collides with
#    another from a second namespace.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { namespace        namespace      {Namespace}}
#
proc ::Simple::Package::Priv::assert-no-collisions {
   packagePair
   namespace
} {
   ### Get the list of collisions
   foreach {package version} $packagePair break
   set packagesToImport\
      [::Simple::Package::Priv::packages-to-import $packagePair $namespace]
   set collisions [{::Simple::Package::Priv::collisions}\
      $package $namespace $packagesToImport]

   ### Non-global namespace
   if {[string compare $namespace ::]} {

      ### Importing to a non-global namespace
      set actionMessage import
      set namespaceMessage "namespace \"$namespace\""

   ### Global namespace
   } else  {

      ### Installing to the global namespace
      set actionMessage install
      set namespaceMessage {the global namespace}
   }

   ### A namespace command collides with an exported command or alias
   if {[llength $collisions] != 0} {

      ### The collision is an exported command
      set collision [lindex $collisions 0]
      if {[string match ::* $collision]} {

         ### Throw error
         set collidingPackage [{::Simple::Package::information\
            packagenamespace} [namespace qualifiers $collision]]
         ::Simple::Error::throw ::Simple::Package::COLLISION\
            $actionMessage $packagePair [namespace tail $collision]\
            $namespaceMessage {exported command} $collision $collidingPackage

      ### The collision is an alias
      } else {

         ### Throw error
         set collidingPackage\
            [{::Simple::Package::information packagealias} $collision]
         ::Simple::Error::throw ::Simple::Package::COLLISION\
            $actionMessage $packagePair [namespace tail $collision]\
            $namespaceMessage alias $collision $collidingPackage
      }
   }

   ### Loop over packages to import
   foreach packageToImport $packagesToImport {

      ### Loop over exported commands and aliases
      foreach commandOrAlias [concat\
         [::Simple::Package::Priv::exported-commands $packageToImport]\
         [::Simple::Package::Priv::aliases $packageToImport]] {

         ### The command or alias would have been
         ###    previously imported or created
         set commandOrAliasName [namespace tail $commandOrAlias]
         if {[info exists\
            importedCommandsAndCreatedAlias($commandOrAliasName)]} {

            ### Throw error
            foreach {package namespace}\
               $importedCommandsAndCreatedAlias($commandOrAliasName) break
            if {[string compare $namespace {}]} {
               set namespaceMessage1 " (namespace \"$namespace\")"
            } else {
               set namespaceMessage1 {}
            }
            set namespace [namespace qualifiers $commandOrAlias]
            if {[string compare $namespace {}]} {
               set namespaceMessage2 " (namespace \"$namespace\")"
            } else {
               set namespaceMessage2 {}
            }
            ::Simple::Error::throw\
               ::Simple::Package::COLLISION-BETWEEN-PACKAGES $actionMessage\
               $packagePair $commandOrAliasName $package\
               $namespaceMessage1 $packageToImport $namespaceMessage2

         ### The command or alias would not have
         ###    been previously imported or created
         } else {

            ### Register the command or alias as previously imported or created
            set importedCommandsAndCreatedAlias($commandOrAliasName)\
               [list $packageToImport [namespace qualifiers $commandOrAlias]]
         }
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::install-required
# -synopsis   : Installs a package required packages.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#
proc ::Simple::Package::Priv::install-required {
   packagePair
} {
   ### Loop over required packages
   foreach requiredPackagePair\
      $::Simple::Package::PackageRequiredPackages($packagePair) {

      ### Declared but not installed package
      if {[info exists\
         ::Simple::Package::DeclaredPackages($requiredPackagePair)] && ![info\
         exists ::Simple::Package::InstalledPackages($requiredPackagePair)]} {

         ### Install the package
         ::Simple::Package::Priv::do-install $requiredPackagePair
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::do-import
# -synopsis   : Imports a package exported commands into a namespace.
# -overview   :
#
#     This procedure loops over a package namespaces and classes and imports
#  their exported commands into the given namespace.
#
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { namespace        namespace      {Namespace}}
#
# -effects    :
#
#  * Imports a package exported commands into the namespace.
#
proc ::Simple::Package::Priv::do-import {
   packagePair
   namespace
} {
   ### Import the package exported commands into the namespace
   set ::Simple::Package::%TMP%\
      [::Simple::Package::Priv::namespaces-and-classes $packagePair]
   namespace eval $namespace {
      foreach packageNamespace ${::Simple::Package::%TMP%} {

         # Skip the global namespace
         if {[string compare $packageNamespace ::]} {
            namespace import ${packageNamespace}::*
         }
      }
   }

   ### Add the namespace to the package list of namespaces imported into
   ::Simple::Package::Priv::add-importedinto $packagePair $namespace
}

### ===========================================================================
### -command  : ::Simple::Package::Priv::add-importedinto
# -synopsis   : Adds a namespace to a package list of namespaces imported into.
# -arguments  :
#  { packagePair      nameandversion {Package name and version}}
#  { namespace        namespace      {Namespace}}
#
proc ::Simple::Package::Priv::add-importedinto {
   packagePair
   namespace
} {
   set ::Simple::Package::PackageImportedInto($packagePair%$namespace)\
      $namespace
}

### ===========================================================================
### -command  : ::Simple::Package::require-and-install
# -synopsis   : Requires and installs a package.
# -overview   :
#
#     This procedure changes a package state from ''available'' to
#  ''installed''.  If the package is declared upon requiring it, the package
#  is then installed.  Notice that otherwise a full installation might require
#  manually importing the package namespaces via K<namespace import>.
#
#     No error is thrown if the package is already provided or installed.
#
#     An error is thrown if the package is not available.
#
# -arguments  :
#  {-exact            boolflag       {Only the specified version is acceptable}}
#  { package          package        {Package name}}
#  {?version?         version %DEFAULT%
#                                    {Package version}}
#
# -returns    : The required package version.
#
proc ::Simple::Package::require-and-install {
   args
} {
   ### Require the package
   set answer [::Simple::Package::require $args]

   # -exact given
   if {[llength $args] == 3} {
      set package [lindex $args 1]

   # -exact not given
   } else {
      set package [lindex $args 0]
   }

   ### Declared but not installed package
   set packagePair [list $package $answer]
   if {[info exists ::Simple::Package::DeclaredPackages($packagePair)] &&\
      ![info exists ::Simple::Package::InstalledPackages($packagePair)]} {

      ### Continue with the second part of the installation
      ::Simple::Package::Priv::install-2 $packagePair
   }

   ### Return the required package version
   return $answer
}

### ===========================================================================
### -command  : ::unknown
# -synopsis   : Augments the unknown mechanism to autoload extra packages.
# -overview   :
#
#     This command augments K<unknown> in order to autoload extra packages.
#   Whenever a non-existent command is invoked, and the command namespace
#   belongs to a declared package, and that package is installed, and there
#   is an extra package for that package, and it is not installed yet, then
#   it is arranged for that extra package to be autoloaded.
#
# -arguments  :
#  { args             list           {Non-existent command and arguments}}
#
proc ::unknown {
   args
} [format {

   ### There is no autoloading script for the command
   set command [lindex $args 0]
   if {![info exists ::auto_index($command)]} {

      # P<SimpleNamespace> is required
      ::Simple::Package::require-and-install SimpleNamespace
      set qualifiedCommand [::Simple::Namespace::qualify $command]
      set namespace [namespace qualifiers $qualifiedCommand]

      ### The command namespace exists
      if {[string compare $namespace {}] &&\
         [{::Simple::Namespace::information exists} $namespace]} {

         ### The command namespace belongs to a declared and installed package
         set packagePair\
            [{::Simple::Package::information packagenamespace} $namespace]
         foreach {package version} $packagePair break
         if {[string length $packagePair] > 0 &&
            ![string compare [{::Simple::Package::information state}\
            $package $version] installed]} {

            ### There is an extra package for the one owning the command
            ###    namespace and it is not installed yet
            append package -extra
            set extraState\
	            [{::Simple::Package::information state} $package $version]
	         if {[lsearch [list available declared] $extraState] != -1} {

	            ### Arrange for the extra package to be autoloaded
               set ::auto_index($qualifiedCommand) [list\
                  Simple::Package::require-and-install $package $version]
	         }
         }
      }
   }

   ### What follows is the original unknown procedure body
   %s

} [info body unknown]]

### ===========================================================================
### -command  : ::Simple::Package::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Package::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Package::cget
# -synopsis   : Gets the package options.
# -returns    : 
#
#     The requested option value or the whole list of options if none
#  specified.
#
proc ::Simple::Package::cget {
} {
   return [list]
}

### ===========================================================================

# Declare the package
::Simple::Package::Priv::declare-package

### Install the package
::Simple::Package::install SimplePackage
