### ===========================================================================
### -package  : SimpleProc-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library procedures with extended argument syntax.
# -module     : extra
# -api        :
#
#  * K<::Simple::Proc::convert extended> canonical
#    Converts an argument from canonical to extended format.
#
#  * K<::Simple::Proc::delete> procedure
#    Deletes a procedure.
#
#  * K<::Simple::Proc::move> sourceProcedure targetProcedure
#    Renames a procedure.
#
#  * K<::Simple::Proc::information issubcommand> procedure
#    Returns whether a procedure is a base command procedure.
#
#  * K<::Simple::Proc::information subcommands> command
#    Returns a base command procedure subcommand list.
#
#  * K<::Simple::Proc::information runtimeparsing> procedure
#    Returns whether a procedure run-time parses its arguments.
#
#  * K<::Simple::Proc::information parseflags> procedure
#    Returns whether a procedure has flag parsing enabled.
#
#  * K<::Simple::Proc::information body> procedure
#    Returns a procedure body.
#
#  * K<::Simple::Proc::information args> procedure
#    Returns a procedure argument names list.
#
#  * K<::Simple::Proc::information usage> procedure
#    Returns a procedure argument usage string.
#
#  * K<::Simple::Proc::information arguments> procedure
#    Returns a procedure argument list.
#
#  * K<::Simple::Argument::information description> procedure argument
#    Returns an extended procedure argument description.
#
#  * K<::Simple::Argument::information givenflags>
#    Returns the flags and their parameters in an extended procedure call.
#
#  * K<::Simple::Script::Header::delete> scriptName ?pattern?
#    Deletes headers from a script.
#
#  * K<::Simple::Script::Trailer::get> script ?pattern?
#    Gets trailers from a script.
#
#  * K<::Simple::Script::Trailer::delete> scriptName ?pattern?
#    Deletes trailers from a script.
#
#  * K<::Simple::Proc::resolve-alias> procedure
#    Resolves command aliases.
#
# -history    :
#
#  24-oct-2004   Copy functionality deleted, version 0.5.2
#  03-dec-2004   P<SimpleRegProc> functionality moved here and changes to 
#                K<::Simple::Proc::information> subcommands, version 0.5.3
#  11-may-2005   Bug 11 correction, version 0.5.4
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleProc-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Proc::delete
# -synopsis   : Deletes a procedure.
# -overview   :
#
#     This command is similar to the K<rename> command when its second
#  argument is the empty string, but works for extended procedures as well as
#  for regular procedures.
#
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Deletes the procedure named A<procedure>.
#
# -remarks    :
#
#  * No aliases for the procedure are deleted.
#
proc ::Simple::Proc::delete {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {qualifiedProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Subcommand procedure
   if {[llength $procedure] == 2} {

      ### Save the base command name
      ::Simple::Priv::save-name [lindex $procedure 0]

      ### Delete the subcommand procedure
      ::Simple::Subcommand::delete $qualifiedProcedure

      ### Delete the base command auxiliary data
      ::Simple::Proc::Priv::reset $namespace[lindex $unqualifiedProcedure 0]

   ### Non-subcommand procedure
   } else {

      ### Delete the regular procedure
      uplevel 1 [list ::rename $procedure {}]
   }

   ### Delete the procedure auxiliary data
   ::Simple::Proc::Priv::reset $qualifiedProcedure

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Proc::move
# -synopsis   : Renames a procedure.
# -overview   :
#
#     This command is similar to the K<rename> command when its second
#  argument is non-empty, but works for extended procedures as well as for
#  regular procedures.
#
# -arguments  :
#  { sourceProcedure  name           {Source procedure name}}
#  { targetProcedure  name           {Target procedure name}}
#
# -returns    : The empty string.
# -remarks    :
#
#  * If the target procedure namespace does not exists, it is created.
#    This mimics the behaviour of the K<rename> command.
#
# -todo       :
#
#  * Allow to rename from subcommand to regular procedures and viceversa.
#
# -effects    :
#
#  * Renames the procedure named K<sourceProcedure> to K<targetProcedure>.
#
proc ::Simple::Proc::move {
   sourceProcedure
   targetProcedure
} {
   ### Parse, qualify and split the procedure names
   foreach {qualifiedSourceProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $sourceProcedure] break
   foreach {qualifiedTargetProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $targetProcedure] break

   ### Assert either both or none of the procedures are subcommands
   set sourceIsSubcommand [expr {[llength $sourceProcedure] == 2}]
   set targetIsSubcommand [expr {[llength $targetProcedure] == 2}]
   if {$sourceIsSubcommand && !$targetIsSubcommand} {
      ::Simple::Error::throw\
         ::Simple::Proc::CANT-BETWEEN-SUBCO-AND-NO-SUBCO\
         rename {} $sourceProcedure non- $targetProcedure
   } elseif {!$sourceIsSubcommand && $targetIsSubcommand} {
      ::Simple::Error::throw\
         ::Simple::Proc::CANT-BETWEEN-SUBCO-AND-NO-SUBCO\
         rename non- $sourceProcedure {} $targetProcedure
   }

   ### Extended procedure
   if {[{::Simple::Proc::information isextended} $qualifiedSourceProcedure]} {

      ### Get the procedure necessary information
      set checkType\
         [{::Simple::Proc::information checktype} $qualifiedSourceProcedure]
      set interpFlags\
         [{::Simple::Proc::information parseflags} $qualifiedSourceProcedure]
      set arguments\
         [{::Simple::Proc::information arguments} $qualifiedSourceProcedure]
      set body [info body $qualifiedSourceProcedure]

      ### Save the procedure name
      ::Simple::Priv::save-name $sourceProcedure

      ### Delete the procedure
      ::Simple::Proc::delete $qualifiedSourceProcedure

      ### Delete the run-time argument parsing header, if any
      ::Simple::Script::Header::delete body %PRIV%-parse-arguments*

      ### Create the target namespace
      namespace eval [namespace qualifiers $qualifiedTargetProcedure] {}

      ### Create the target procedure
      ::Simple::Proc::create -checktype $checkType\
         -parseflags $interpFlags $qualifiedTargetProcedure $arguments $body

   ### Non existing or non-extended procedure
   } else {

      ### Subcommand procedure
      if {$sourceIsSubcommand} {

         ### Save the base command name
         ::Simple::Priv::save-name [lindex $sourceProcedure 0]

         ### Rename the subcommand procedure
         ::Simple::Subcommand::move\
            $qualifiedSourceProcedure $qualifiedTargetProcedure

      ### Non-subcommand procedure
      } else {

         ### Rename the regular procedure
         uplevel 1 [list ::rename $sourceProcedure $qualifiedTargetProcedure]
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::parse-arguments-checktype
# -synopsis   : Parses an argument list and checks their types in run-time.
# -arguments  :
#  { namespace        name           {Procedure namespace}}
#  { procedure        unqualifiedname
#                                    {Unqualified procedure name}}
#  { actualArgs       list           {Actual argument list}}
#
# -returns    : The list of flags given in the actual argument list.
# -effects    :
#
#  * Creates variables named after the command named K<procedure>
#    arguments in the calling scope.
#
# -references :
#
#  * The K<::Simple::Proc::Priv::parse-arguments> procedure.
#
#  * The K<::Simple::Error::Priv::parse-options> procedure.
#
proc ::Simple::Proc::Priv::parse-arguments-checktype {
   namespace
   procedure
   actualArgs
} {
   # Argument list
   set arguments [set ${namespace}%ProcData%(%ArgInternal%,$procedure)]

   # First argument
   set argument [lindex $arguments [set argumentIndex 0]]

   # First actual argument
   set actualArg [lindex $actualArgs [set actualArgIndex 0]]

   ### Loop over actual arguments
   set givenFlags [list]
   set actualArgsLlength [llength $actualArgs]
   while {$actualArgIndex < $actualArgsLlength} {

      ### Loop over arguments
      while 1 {

         # If the argument name is empty we are beyond the last one
         set name [lindex $argument 2]
         set isFlag [lindex $argument 0]
         if {![string compare $name {}]} {
            ::Simple::Error::throw ::Simple::TOO-MANY-ARGS\
               [{::Simple::Proc::information caller}]
         }

         ### Parse flags
         if {$isFlag} {

            ### Handle match
            set requiresParameter [lindex $argument 1]
            if {![string compare $actualArg -$name]} {

               # Flag given
               lappend givenFlags -$name

               ### Parameter required
               if {$requiresParameter} {

                  # Take parameter from next actual argument
                  incr actualArgIndex
                  if {$actualArgIndex >= $actualArgsLlength} {
                     ::Simple::Error::throw\
                        ::Simple::Proc::NO-PARAM-FOR-FLAG -$name\
                        [{::Simple::Proc::information caller}]
                  }
                  set parameter [lindex $actualArgs $actualArgIndex]

                  ### Check the parameter type
                  ::Simple::Variable::Priv::assert-type-conformation argument\
                     0 -$name [lindex $argument 4] $parameter 0 [lindex $argument 5]

                  # Set the flag in the calling scope
                  uplevel 1 [list ::set $name $parameter]

               ### Parameter not required
               } else {

                  # Set the boolflag in the calling scope: given
                  uplevel 1 [list ::set $name 1]
               }

               ### Eat argument
               set argument [lindex $arguments [incr argumentIndex]]

               ### Next actual argument
               break

            ### Handle non-match
            } else {

               ### Parameter required
               if {$requiresParameter} {

                  # Set the argument in the calling scope,
                  # taking the parameter from the default value
                  set defaultValue [lindex $argument 3]
                  uplevel 1 [list ::set $name $defaultValue]

               ### Parameter not required
               } else {

                  # Set the boolflag in the calling scope: not given
                  uplevel 1 [list ::set $name 0]
               }

               ### Eat argument
               set argument [lindex $arguments [incr argumentIndex]]

               ### Next argument
               continue
            }

         ### Handle non-flags
         } else {

            ### Check the argument type
            ::Simple::Variable::Priv::assert-type-conformation argument\
               0 $name [lindex $argument 4] $actualArg 0 [lindex $argument 5]

            ### Handle the special argument "args"
            if {![string compare $name args]} {

               # Get all remaining actual arguments
               uplevel 1 [list ::set args [lrange $actualArgs $actualArgIndex end]]

               # Return the list of flags given in the actual argument list
               return $givenFlags
            }

            # Set the argument in the calling scope
            uplevel 1 [list ::set $name $actualArg]

            ### Eat argument
            set argument [lindex $arguments [incr argumentIndex]]

            ### Next actual argument
            break
         }
      }

      ### Eat actual argument
      set actualArg [lindex $actualArgs [incr actualArgIndex]]
   }

   ### Loop over not used arguments
   set argumentsLlength [llength $arguments]
   while {$argumentIndex < $argumentsLlength} {

      set name [lindex $argument  2]
      if {![string compare $name args]} {

         # No remaining actual arguments
         uplevel 1 [list ::set args [list]]

         # Return the list of flags given in the actual argument list
         return $givenFlags
      }

      set defaultGiven [lindex $argument 1]
      if {$defaultGiven} {

         # Set the argument in the calling scope,
         # taking the parameter from the default value
         set defaultValue [lindex $argument 3]
         uplevel 1 [list ::set $name $defaultValue]

      } else {

         # boolflag
         if {[lindex $argument 0]} {

            # Set the boolflag in the calling scope: not given
            uplevel 1 [list ::set $name 0]
         } else {
            ::Simple::Error::throw ::Simple::NO-VALUE-FOR-PARAM $name\
               [{::Simple::Proc::information caller}]
         }
      }

      ### Eat non-used argument
      set argument [lindex $arguments [incr argumentIndex]]
   }

   # If this point is reached, this procedure does not use the
   # special argument "args", so unset it in the calling scope
   uplevel 1 [list ::unset args]

   # Return the list of flags given in the actual argument list
   set givenFlags
}

### ===========================================================================
### -command  : ::Simple::Proc::information issubcommand
# -synopsis   : Returns whether a procedure is a base command procedure.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : Whether the procedure is a base command procedure.
# -references :
#
#  * The K<::Simple::Subcommand::information exists> procedure.
#
proc-sub {::Simple::Proc::information issubcommand} {
   procedure
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Subcommand procedure
   if {[llength $procedure] > 1} {

      ### Throw error
      ::Simple::Error::throw ::Simple::BAD-ITEM\
         {non-subcommand procedure name} [::Simple::Priv::restore-name]
   }

   ### Qualify the procedure name
   set procedure [::Simple::Priv::qualify $procedure]

   ### Return whether the procedure is a base command procedure
   return [{::Simple::Subcommand::information exists} $procedure]
}

### ===========================================================================
### -command  : ::Simple::Proc::information subcommands
# -synopsis   : Returns a base command procedure subcommand list.
# -arguments  :
#  { command          name           {Base command name}}
#
# -returns    : The base command procedure subcommand list.
# -references :
#
#  * The K<::Simple::Subcommand::information subcommands> procedure.
#
# -remarks    :
#
#  * Procedures containing a percent sign in its name are assumed to be
#    internal to The Simple Development Library and are excluded from the
#    output.
#
proc-sub {::Simple::Proc::information subcommands} {
   command
} {
   ### Save the base command name
   ::Simple::Priv::save-name $command

   ### Subcommand procedure
   if {[llength $command] > 1} {

      ### Throw error
      ::Simple::Error::throw ::Simple::BAD-ITEM\
         {non-subcommand procedure name} [::Simple::Priv::restore-name]
   }

   ### Qualify the base command name
   set command [::Simple::Priv::qualify $command]

   ### Returns the base command procedure subcommand list
   return [{::Simple::Subcommand::information subcommands} $command]
}

### ===========================================================================
### -command  : ::Simple::Proc::information runtimeparsing
# -synopsis   : Returns whether a procedure run-time parses its arguments.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : Whether the procedure run-time parses its arguments.
# -remarks    :
#
#  * This procedure returns false for non-existing or non-extended procedures.
#
proc-sub {::Simple::Proc::information runtimeparsing} {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Get whether the procedure run-time parses its arguments
      set answer [expr {[llength [::Simple::Script::Header::get\
         [::Simple::Proc::Priv::info-body $namespace $unqualifiedProcedure]\
         %PRIV%-parse-arguments*]] > 0}]

   ### Non existing or non-extended procedure
   } else {

      ### Return false
      set answer 0
   }

   ### Return whether the procedure run-time parses its arguments
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information parseflags
# -synopsis   : Returns whether a procedure has flag parsing enabled.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : Whether the procedure has flag parsing enabled.
# -remarks    :
#
#  * This procedure returns false for non-existing or non-extended procedures.
#
proc-sub {::Simple::Proc::information parseflags} {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Get whether the procedure has flag parsing enabled
      set answer\
         [set ${namespace}%ProcData%(%ParseFlags%,$unqualifiedProcedure)]

   ### Non existing or non-extended procedure
   } else {

      ### Return false
      set answer 0
   }

   ### Return whether the procedure has flag parsing enabled
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information body
# -synopsis   : Returns a procedure body.
# -overview   :
#
#     This command is similar to the K<info body>, but works for extended
#  procedures as well as for regular procedures.
#
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : The procedure body.
#
proc-sub {::Simple::Proc::information body} {
   procedure
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Procedure does not exist
   if {[catch {set body\
      [::Simple::Proc::Priv::info-body $namespace $unqualifiedProcedure]}]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         [::Simple::Priv::restore-name] {a procedure}

   ### Procedure exists
   } else {

      ### Delete the run-time argument parsing header, if any
      ::Simple::Script::Header::delete body %PRIV%-parse-arguments*

      ### Return the procedure body
      return $body
   }
}

### ===========================================================================
### -command  : ::Simple::Proc::information args
# -synopsis   : Returns a procedure argument names list.
# -overview   :
#
#     This command is similar to the K<info args> command, but works for
#  extended procedures as well as for regular procedures.
#
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : The procedure argument names list.
#
proc-sub {::Simple::Proc::information args} {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Return the extended procedure argument names list
      set answer [set ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)]

   ### Non existing or non-extended procedure
   } else {

      ### Try with a regular procedure
      set answer [uplevel 1 [list ::info args $procedure]]
   }

   ### Return the procedure argument names list
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information usage
# -synopsis   : Returns a procedure argument usage string.
# -returns    : The procedure argument usage string.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -examples   :
#
#  # A procedure with several arguments, including flags
#  proc-ext name {
#     { arg1             integer        {First regular argument}}
#     {-boolflag         boolflag       {Boolean flag}}
#     {-flag1            string  FLAG1  {First flag}}
#     {-flag2            string  FLAG2  {Second flag}}
#     { arg2             integer        {Second regular argument}}
#  } {
#  }
#  
#  # This displays the following:
#  #   |arg1 ?-boolflag? ?-flag1 string? ?-flag2 string? arg2
#  puts [::Simple::Proc::information usage name]
#
# -remarks    :
#
#  * This command works for extended procedures as well as for regular
#    procedures.
#
# -references :
#
#  * The K<::Simple::Option::information usage> procedure.
#
proc-sub {::Simple::Proc::information usage} {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Compose the extended procedure usage string
      set answer [list]
      foreach argument\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         name [set ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] {

         # Flag
         if {[lindex $argument 0]} {
            set type [lindex $argument 4]
            if {[string compare $type boolflag]} {
               lappend answer ?$name $type?
            } else {
               lappend answer ?$name?
            }

         # Optional
         } elseif {[lindex $argument 1]} {

            lappend answer ?$name?

         # Non-flag nor optional
         } else {
            lappend answer $name
         }
      }

      ### Return the extended procedure usage string

   ### Non existing or non-extended procedure
   } else {

      ### Try with a regular procedure
      set answer\
         [uplevel 1 [list ::Simple::Priv::proc-information-usage $procedure]]
   }

   ### Return the procedure argument usage string
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information arguments
# -synopsis   : Returns a procedure argument list.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : The procedure argument list.
#
proc-sub {::Simple::Proc::information arguments} {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Compose the extended procedure argument list
      set answer [list]
      foreach argument [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         name [set ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] {

         # Flag or optional
         if {![lindex $argument 0] && [lindex $argument 1]} {
            set name ?$name?
         }

         # choice or choice-list
         set type [lindex $argument 4]
         if {[string match choice* $type]} {

            lappend answer [list $name $type [lindex $argument 5]\
               [lindex $argument 6]]

         # Optional
         } elseif {[lindex $argument 1]} {
            lappend answer [list $name $type [lindex $argument 3]\
               [lindex $argument 6]]

         # Otherwise
         } else {
            lappend answer [list $name $type [lindex $argument 6]]
         }
      }

      ### Return the extended procedure argument list

   ### Non existing or non-extended procedure
   } else {

      ### Try with a regular procedure
      set answer\
         [uplevel 1 [list ::Simple::Priv::proc-information-arguments $procedure]]
   }

   ### Return the procedure argument list
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Argument::information description
# -synopsis   : Returns an extended procedure argument description.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#  { argument         unqualifiedname
#                                    {Argument name}}
#
# -returns    : The extended procedure argument desription.
#
proc-sub {::Simple::Argument::information description} {
   procedure
   argument
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Locate the argument
      set position [lsearch -exact [set\
         ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] $argument]

      ### Non-existing argument
      if {$position == -1} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-ARGUMENT\
            [::Simple::Priv::restore-name] $argument
      }

      ### Return the extended procedure argument description
      return [lindex [lindex\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         $position] 6]

   ### Non-extended procedure
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         [::Simple::Priv::restore-name] {an extended procedure}
   }
}

### ===========================================================================
### -command  : ::Simple::Argument::information givenflags
# -synopsis   : Returns the flags and their parameters in an extended procedure call.
# -overview   :
#
#     This procedure returns the list of flags and their parameters given in
#  an extended procedure call.  The format of the returned list is suitable to
#  be given as argument to the K<array get> command: the array indices would be
#  the flags and the array values their parameters.
#
#  Notice that for boolean flags (type "boolflag") the value is always ''1''.
#
# -keywords   : procedure argument flag
# -returns    : 
#
#  The list of flags and their parameters given in the actual argument list.
#
# -references :
#
#  * The K<::Simple::Argument::information flaggiven> procedure.
#
#  * The K<::Simple::Priv::flag-given> procedure.
#
#  * The K<::Simple::Option::information usage> procedure.
#
# -limitations:
#
#  * This procedure must be called from another procedure.
#
proc-sub {::Simple::Argument::information givenflags} {
} {
   ### Return the list of flags given in the extended procedure call
   upvar %GIVEN-FLAGS% flags

   set answer [list]
   foreach flag $flags {

      # Convert the flag to an unqualified name
      set variable [string trimleft $flag -]
      lappend answer $flag [uplevel 1 ::set $variable]
   }

   ### Return the list of flags and their parameters given in the actual argument list
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Script::Header::delete
# -synopsis   : Deletes headers from a script.
# -overview   :
#
#     This procedure deletes script headers added by
#  K<::Simple::Script::Header::add> matching the given pattern.
#
# -arguments  :
#  { scriptName       name           {Script name}}
#  {?pattern?         pattern {}     {Pattern}}
#
# -effects    :
#
#  * Deletes the headers from the script called A<scriptName> in the calling
#    scope.
#
# -returns    : The number of deleted headers.
# -remarks    :
#
#  * Use no pattern to delete all headers.
#
proc ::Simple::Script::Header::delete {
   scriptName
   {pattern {}}
} {
   ### The script does not exist
   upvar $scriptName script
   if {![info exists script]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING script $scriptName
   }

   ### Headers present
   set answer 0
   if {[regexp -- "^($::Simple::Script::MarkerStart\
      %HEADER-BEGIN%.*\n$::Simple::Script::MarkerStart\
      %HEADER-END%\[^\n\]*)(.*)$" $script -> headers afterHeaders]} {

      set headers [::Simple::Script::Priv::split-headers-or-trailers\
         $headers]
      # A<afterHeaders> contains the script after the headers
      # A<headers> contains the list of headers

      ### Tag pattern given
      set finalHeaders [list]
      if {[string compare $pattern {}]} {

         ### Loop over headers
         set RE "^$::Simple::Script::MarkerStart %HEADER-BEGIN%\
            +(\[0-9\]+) (\[^\n\]+)\n.*$"
         foreach header $headers {

            regexp -- $RE $header -> priority tag
            # A<priority> contains the priority of this header
            # A<tag> contains the tag of this header

            ### Tag does not match pattern
            if {![string match $pattern $tag]} {

               ### Do not delete this header
               lappend finalHeaders $header

            ### Tag matches pattern
            } else {

               ### Delete this header
               incr answer
            }
         }
      ### Tag pattern not given
      } else {

         ### Delete all headers
         set finalHeaders [list]

         ### Get the number of deleted headers
         set answer [regexp --\
            "$::Simple::Script::MarkerStart %HEADER-BEGIN" $headers]
      }

      ### Rebuild the script
      if {[llength $finalHeaders] == 0} {

         # Remove the leading \n
         set script [string range $afterHeaders 1 end]

      } else {
         set script [join $finalHeaders \n]$afterHeaders
      }
   }

   ### Return the number of deleted headers
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Script::Trailer::get
# -synopsis   : Gets trailers from a script.
# -overview   :
#
#     This procedure returns the list of trailers from a script added by
#  K<::Simple::Script::Trailer::add> matching the given pattern.
#
# -arguments  :
#  { script           script         {Script}}
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns    : 
#
#     The list of script trailers matching the given pattern in the format
#  ||{tag priority header}||.
#
# -remarks    :
#
#  * Use no pattern to get all trailers.
#
proc ::Simple::Script::Trailer::get {
   script
   {pattern {}}
} {
   ### Trailers present
   set index\
      [string first "$::Simple::Script::MarkerStart %TRAILER-BEGIN" $script]
   if {$index != -1} {

      set fullTrailers [::Simple::Script::Priv::split-headers-or-trailers\
         [string range $script $index end]]
      set beforeTrailers [string range $script 0 [incr index -1]]
      # A<fullTrailers> contains the list of fullTrailers
      # A<beforeTrailers> contains the script before the trailers

      ### Loop over full trailers
      set RE "^$::Simple::Script::MarkerStart %TRAILER-BEGIN%\
         +(\[0-9\]+) (\[^\n\]+)\n(.*)\n$::Simple::Script::MarkerStart\
         %TRAILER-END%"
      foreach fullTrailer $fullTrailers {

         regexp -- $RE $fullTrailer -> priority tag trailer
         # A<priority> contains the priority of this trailer
         # A<tag> contains the tag of this trailer

         ### Pattern given and the tag matches it
         if {![string compare $pattern {}] || [string match $pattern $tag]} {

            ### Append tag/priority pair
            lappend answer [list $tag $priority $trailer]
         }
      }
   }

   ### Return the list of matching trailers
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Script::Trailer::delete
# -synopsis   : Deletes trailers from a script.
# -overview   :
#
#     This procedure deletes script trailers added by
#  K<::Simple::Script::Trailer::add> matching the given pattern.
#
# -arguments  :
#  { scriptName       name           {Script name}}
#  {?pattern?         pattern {}     {Pattern}}
#
# -effects    :
#
#  * Deletes the trailers from the script called A<scriptName> in the calling
#    scope.
#
# -returns    : The number of deleted trailers.
# -remarks    :
#
#  * Use no pattern to delete all trailers.
#
proc ::Simple::Script::Trailer::delete {
   scriptName
   {pattern {}}
} {
   ### The script does not exist
   upvar $scriptName script
   if {![info exists script]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING script $scriptName
   }

   ### Trailers present
   set answer 0
   set index\
      [string first "$::Simple::Script::MarkerStart %TRAILER-BEGIN" $script]
   if {$index != -1} {

      set trailers [::Simple::Script::Priv::split-headers-or-trailers\
         [string range $script $index end]]
      set beforeTrailers [string range $script 0 [incr index -2]]
      # A<trailers> contains the list of trailers
      # A<beforeTrailers> contains the script before the trailers

      ### Tag pattern given
      set finalTrailers [list]
      if {[string compare $pattern {}]} {

         ### Loop over trailers
         set RE "^$::Simple::Script::MarkerStart %TRAILER-BEGIN%\
            +(\[0-9\]+) (\[^\n\]+)\n.*$"
         foreach trailer $trailers {

            regexp -- $RE $trailer -> priority tag
            # A<priority> contains the priority of this trailer
            # A<tag> contains the tag of this trailer

            ### Tag does not match pattern
            if {![string match $pattern $tag]} {

               ### Do not delete this trailer
               lappend finalTrailers $trailer

            ### Tag matches pattern
            } else {

               ### Delete this trailer
               incr answer
            }
         }

      ### Tag pattern not given
      } else {

         ### Delete all trailers
         set finalTrailers [list]

         ### Get the number of deleted trailers
         set answer [regexp --\
            "$::Simple::Script::MarkerStart %TRAILER-BEGIN" $trailers]
      }

      ### Rebuild the script
      if {[llength $finalTrailers] == 0} {
         set script $beforeTrailers
      } else {
         set script $beforeTrailers\n[join $finalTrailers \n]
      }
   }

   ### Return the number of deleted trailers
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Script::Priv::split-headers-or-trailers
# -synopsis   : Splits a block of headers or trailers.
#
# -arguments  :
#  { block            string         {The headers or trailers block}}
#
# -returns    : The list of headers or trailers.
#
proc ::Simple::Script::Priv::split-headers-or-trailers {
   block
} {
   set answer [list]
   set RE "^(.*)($::Simple::Script::MarkerStart\
      .*${::Simple::Script::MarkerStart}(\[^\n\]*))"
   while {[regexp -- $RE $block -> block headerOrTrailer]} {
      set answer [linsert $answer 0 $headerOrTrailer]
   }
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::resolve-alias
# -synopsis   : Resolves command aliases.
# -overview   :
#
#  This procedure resolves one aliasing level for a command, if existing.
#
# -arguments  :
#  { procedure        name           {Procedure name}}
#
# -returns    : The alias resolved command name
#
# -limitations:
#
#  * Just one aliasing level is resolved.
#
proc ::Simple::Proc::resolve-alias {
   procedure
} {
   ### It is an alias
   set alias [interp alias {} $procedure]
   if {[string compare $alias {}]} {

      ### Use the aliased command
      set answer $alias

   ### It is not an alias
   } else {

      ### Use the original command
      set answer $procedure
   }

   ### Return the alias solved command name
   return $answer
}
}
