### ===========================================================================
### -package  : SimpleProc
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library procedures with extended argument syntax.
# -module     : implementation
# -overview   :
#
#     This package provides a way to create procedures with extended argument
#  syntax.  This syntax includes argument types (with optional run-time type
#  checking) and flag arguments with or without parameters which may be
#  automatically parsed or left for the user to parse.
#
#     Procedures are provided to create, delete, rename and get information
#  about procedures with extended argument syntax, as well as others to handle
#  arguments in the body of those procedures, such as assess whether certain
#  flag has been given in the command line or assess whether the value of a
#  flag is that flag value.
#
#     This package also allows to handle headers and trailers for code
#  scripts.  Each header or trailer is identified by a tag and, within a
#  script, headers and trailers are ordered by increasing priority. 
#  Procedures are provided to add, get and delete script headers and trailers.
#
# -usage      :
#
#     Extended syntax procedures are created via the K<::Simple::Proc::create>
#  procedure.  The declaration specifies whether to perform run-time checking
#  of the arguments types as well as whether to delegate the flag parsing to
#  the user.
#
#     Extended argument syntax procedures can be deleted via
#  K<::Simple::Proc::delete> or renamed via K<::Simple::Proc::move>.  The
#  K<::Simple::Proc::information> procedure provides several subcommands to
#  query information about extended argument syntax procedures: 
#  K<issubcommand>, K<subcommands>, K<runtimeparsing>, K<checktype>,
#  K<parseflags>, K<exists>, K<body>, K<args>, K<usage>, K<arguments>,
#  K<type>, K<default>, K<choices>, K<metadata> and K<description>.
#
#     The syntax of the extended arguments is similar to that of the arguments
#  used by the P<opt> package distributed with Tcl, so the upgrade for users
#  using that should be easy; in fact this package was implemented as an
#  evolution of the P<opt> package for two reasons:  1) P<opt> is A<slow> and
#  2) it was deprecated from Tcl 8.1 (made private, actually).  
#
#     This argument syntax is of the form ||{name type ?value? 
#  description}||.  There, "name" is the argument name, "type" its type,
#  "value" its default value or choice list and "description" a string
#  describing the argument.  Types used by P<opt> are of the form "-float",
#  while The Simple Development Library removes the hyphen to yield "float";
#  further, some P<opt> types are abbreviated (such as "int") while The Simple
#  Development Library uses mostly unabbreviated names such as "integer".  See
#  the P<SimpleType> package for further information about types.  Extended
#  arguments are further described in the K<proc-ext> procedure.
#
#     The presence or absence of a regular flag in an extended procedure call
#  argument list can be obtained via the K<::Simple::Argument::information
#  flaggiven> procedure, while the complete list of all given flags is
#  returned by K<::Simple::Argument::information givenflags>.  For flags or
#  optional arguments, K<::Simple::Argument::information isdefault> returns
#  whether an extended procedure argument actual value matches its
#  corresponding default value.  Other subcommands of the
#  K<::Simple::Argument::information> command are <description>,
#  K<defaultgiven>, K<default>, K<type> and K<choices>.
#
#     This package encompasses the funcionality provided by the
#  P<SimpleSubcommand> package, so it is recommended not to directly use the
#  later if using the former.  In fact, all procedures provided by this
#  package work for regular procedures also, so this is the only package
#  needed to handle regular or extended argument syntax procedures as well as
#  regular or subcommand procedures.  Nevertheless, see the P<SimpleDeclare>
#  package K<declare-proc> command, which encompasses K<proc-ext> in turn.
#
#     The K<::Simple::Proc::convert> pair of procedures convert between
#  extended and Tcl canonical procedure argument formats.
#
#     K<::Simple::Script::append> allows to append scripts.  The
#  K<::Simple::Script::Header::add> and K<::Simple::Script::Trailer::add>
#  procedures prepend a header or append a trailer to a script, respectively. 
#  Headers and trailers can be got or deleted with the
#  K<::Simple::Script::Header::get>, K<::Simple::Script::Trailer::get>,
#  K<::Simple::Script::Header::delete> and
#  K<::Simple::Script::Trailer::delete> procedures.
#
# -keywords   : procedure extended argument type flag
# -variable   :
#
#  {
#     ::Simple::Script::MarkerStart
#
#  } -description {
#
#     Begin and an end of script header or trailer marker
#
#  } -type string -value #=========
#
# -variable   :
#
#  {
#     ::Simple::Script::MarkerFormat
#
#  } -description {
#
#     Begin and an end of script header or trailer marker format
#
#  } -type formatstring -value "$::Simple::Script::MarkerStart %-15s %4d %s"
#
# -option     :
#
#  {
#     -checktype
#
#  } -description {
#
#        Whether argument run-time type checking is performed for procedures
#     created via the K<::Simple::Proc::create> procedure in order to check
#     whether the actual arguments match their type.  This has a performance
#     penalty which should be considered.  Notice that only those procedures
#     created via the K<::Simple::Proc::create> procedure A<after> modifying
#     this option are affected by the change.
#
#  } -type boolean -value 0
#
# -option     :
#
#  {
#     -afterflags
#
#  } -description {
#
#        If the number of arguments after the last flag in procedures created
#     via the K<::Simple::Proc::create> procedure exceedes the value of this
#     package option, those arguments are handled via an auxiliary "after
#     flags" procedure.  The optimum value of this parameter depends on the
#     Tcl version used.  The Simple Development Library default value
#     corresponds to the latest Tcl version.  Notice that only those
#     procedures created via the K<::Simple::Proc::create> procedure A<after>
#     modifying this option are affected by the change.
#
#  } -type integer -value 5
#
# -api        :
#
#  * K<proc-ext> ?-parseflags extbool? ?-checktype extbool? procedure
#       arguments body
#    Creates a procedure with extended argument syntax.
#
#  * K<::Simple::Proc::information exists> procedure
#    Returns whether a procedure exists.
#
#  * K<::Simple::Proc::information isextended> procedure
#    Returns whether a procedure is an extended procedure.
#
#  * K<::Simple::Argument::information defaultgiven> procedure argument
#    Returns whether a procedure argument has a default value.
#
#  * K<::Simple::Argument::information default> procedure argument
#    Returns a procedure argument default value.
#
#  * K<::Simple::Argument::information flaggiven> flag
#    Returns whether a flag was given in an extended procedure call.
#
#  * K<::Simple::Argument::information isdefault> argument
#    Returns whether an argument value is the default one.
#
#  * K<::Simple::Proc::information caller> ?levels?
#    Returns the caller procedure name.
#
#  * K<::Simple::Script::Header::add> scriptName tag priority header
#    Prepends a header to a script.
#
#  * K<::Simple::Script::Trailer::add> scriptName tag priority trailer
#    Appends a trailer to a script.
#
#  * K<::Simple::Script::append> scriptName scriptToAppend
#    Appends a non-empty script to another one.
#
#  * K<::Simple::Script::Header::get> script ?pattern?
#    Gets headers from a script.
#
#  * K<::Simple::Argument::information type> procedure argument
#    Returns an extended procedure argument type.
#
#  * K<::Simple::Proc::information checktype> procedure
#    Returns whether a procedure performs argument type checking.
#
#  * K<::Simple::Argument::check-type> args
#    Perform runtime type checking on named arguments.
#
#  * K<::Simple::Argument::information choices> procedure argument
#    Returns an extended procedure argument choices list.
#
#  * K<::Simple::Proc::configure> ?-checktype extbool? ?-afterflags integer?
#    Configures the package options.
#
#  * K<::Simple::Proc::cget> ?-checktype? ?-afterflags?
#    Gets the package options.
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#  ::Simple::Package::require-and-install SimpleProc
#
#  # Configure The Simple Development Library to store the metadata
#  ::Simple::configure -storemetadata true
#
#  # Enable arguments run-time type checking
#  ::Simple::Proc::configure -checktype true
#
#  # Create a procedure which takes an
#  # optional flag and a required integer
#  proc-ext {
#     zero
#  } {
#     {-zero             boolflag       {Zero flag}}
#     { integer          integer        {Integer}}
#  } {
#     if {$zero} {
#        puts 0
#     } else {
#        puts $integer
#     }
#  }
#
#  # Get some info about the procedure
#  # This displays the following:
#  #   |0
#  #   |?-zero? integer
#  #   |Zero flag
#  puts [::Simple::Proc::information issubcommand zero]
#  puts [::Simple::Proc::information usage zero]
#  puts [::Simple::Argument::information description zero -zero]
#
#  # Valid calls
#  # This displays the following:
#  #   |23
#  #   |0
#  #   |0
#  zero 23
#  zero 0
#  zero -zero 23
#
#  # Invalid call
#  # This displays the following:
#  #   |invalid value "foo" for argument "integer" of type "integer"
#  catch {zero foo} result
#  puts $result
#
#  # Create a couple of subcommand procedures
#  proc-ext {
#     command foo
#  } {
#     { integer          integer        {Integer}}
#     { choice           choice {1 2}   {Choices}}
#  } {
#     if {$integer > $choice} {
#        puts {The integer is greater than the choice}
#     } else {
#        puts {The integer is lower or equal than the choice}
#     }
#  }
#
#  proc-ext {
#     command bar
#  } {
#     {?integer?         integer 1      {Integer}}
#  } {
#     if {[::Simple::Argument::information isdefault integer]} {
#        puts {Default value}
#     } else {
#        puts {Non-default value}
#     }
#  }
#
#  # This displays the following:
#  #   |The integer is lower or equal than the choice
#  #   |The integer is greater than the choice
#  #   |Default value
#  #   |Non-default value
#  command foo 1 2
#  command foo 8 1
#  command bar
#  command bar 8
#
#  # Get some info about the subcommand
#  # This displays the following:
#  #   |1
#  #   |foo bar
#  puts [::Simple::Proc::information issubcommand command]
#  puts [::Simple::Proc::information subcommands command]
#
#  # Delete one of the subcommands
#  ::Simple::Proc::delete {command bar}
#
#  # Get some info about the subcommand
#  # This displays the following:
#  #   |foo
#  puts [::Simple::Proc::information subcommands command]
#
# -details    :
#
#  * For the sake of performance, this package does not implement some of the
#    more esoteric features of the P<opt> package such as argument names
#    abbreviation, type guessing, completion of incomplete arguments or out of
#    order flags (this last feature might be added in a future revision of the
#    package).
#
#  > For procedures with flags and a small number of arguments P<SimpleProc>
#    is about A<20> times faster than the P<opt> package, or A<4> times if
#    run-time type checking is enabled (notice that P<opt> always checks
#    the type of its arguments).
#
#  > Also, several tricks are applied to improve performance (see the details
#    section of the K<::Simple::Proc::create> procedure) which result in
#    typical procedures (those which do not require argument run-time type
#    checking and have no flags in the argument list) A<behaving identically>
#    to regular Tcl procedures (no run-time parsing is required).
#
#  > For a ten non-flag arguments procedure, P<opt 0.4.4.1> requires 2810
#    microseconds per call (Tcl 8.4.9, 1.0 GHz Pentium III Mobile, Linux Red
#    Hat 7.3) while P<SimpleProc> uses a mere 3 or 740 microseconds when
#    run-time type checking is disabled or enabled, respectively; that means
#    performance improvement factors of A<936> or A<3.8>, respectively.  For a
#    ten flag arguments procedure the figures are 2924 and 188 microseconds,
#    a factor of A<16.6>.  For a procedure with no arguments the figures are
#    90 and 2, a factor of A<45>.  Even for procedures with flags, their
#    parsing can be delegated to the user so, again, no overhead is introduced
#    if desired.
#
#  * The argument list internal representation of each extended procedure is
#    stored in an element of the V<%ProcData%> array in the procedure
#    namespace whose index is ''%ArgInternal%,<unqualified procedure name>''. 
#    This array element is a list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0:  whether the argument is a flag
#     1:  whether the argument has a default value
#     2:  argument name
#     3:  argument default value
#     4:  argument type
#     5:  argument choices (empty if type is not "choice" nor derived from it)
#     6:  argument description
##end
#
#  * The argument names list of each extended procedure argument list is
#    stored in an element of the V<%ProcData%> array in the procedure
#    namespace whose index is ''%ArgNames%,unqualified procedure name''.
#
#  * The indices of the first and last flags in the argument list of each
#    extended procedure argument is stored in an element of the V<%ProcData%>
#    array in the procedure namespace whose index is
#    ''%FlagsIndices%,unqualified procedure name''.
#
#  * Whether the flag parsing is enabled or has been delegated to the user for
#    each extended procedure is stored in an element of the V<%ProcData%>
#    array in the procedure namespace whose index is
#    ''%ParseFlags%,unqualified procedure name''.
#
#  * The procedure access mode is stored in an element of the V<%ProcData%>
#    array in the procedure namespace whose index is ''%Access%,unqualified
#    procedure name''.
#
#  * The procedure alias is stored in an element of the V<%ProcData%> array in
#    the procedure namespace whose index is ''%Alias%,unqualified procedure
#    name''.
#
#  * The procedure metadata is stored in an element of the V<%ProcData%> array
#    in the procedure namespace whose index is ''%Metadata%,unqualified
#    procedure name''.
#
# -remarks    :
#
#  * This package is optimized for speed and not for maintenability.  In
#    particular the run-time parsing of the argument list is done by one of
#    two different private procedures
#    (K<::Simple::Proc::Priv::parse-arguments> and
#    K<::Simple::Proc::Priv::parse-arguments-checktype>) depending on
#    whether run-time type checking is enabled or not, even though they share
#    most of the code.
#
# -todo       :
#
#  * Support out of order flags.
#
#  * Allow derived arrary types as extended arguments.
#
#  * It might be useful to have a way to enable and disable A<-checktype> for
#    existing procedures.
#
#  * Provide a package option or flag to K<proc-ext> to complain on bad flags
#    throwing a ||::Simple::BAD-FLAG|| error.  Currently a bad flag is assumed
#    to be a non-flag parameter.
#
#  * Implement a mechanism to pass arguments by reference.  This could sit on
#    top of a new procedure named K<::Simple::Argument::upvar> which performs
#    the K<upvar>.  It should have an A<-array> flag which specifies whether
#    the entity being linked via K<upvar> is an array or a regular variable. 
#    It should also have an A<-exists> (or A<-force>) flag which, if given,
#    asserts the array/value does exists.  Independently on whether A<-exists>
#    is given or not, if the linked entity does exists we should assert that
#    it is an array or a regular variable, as directed by A<-array> being
#    given or not.
#
# -history    :
#
#  19-feb-1999   Unreleased first version 0.1
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  30-mar-2003   Extra package, version 0.5.1
#  24-oct-2004   Copy functionality deleted, version 0.5.2
#  03-dec-2004   ||SimpleRegProc|| functionality moved here and changes to\
#                K<::Simple::Proc::information> subcommands, version 0.5.3
#  11-may-2005   Bug 11 correction, version 0.5.4
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleProc 1.0 -requisites {

   {SimpleVariable 1.0}

} -namespaces {

   ::Simple::Proc
   ::Simple::Proc::Priv
   ::Simple::Argument
   ::Simple::Script
   ::Simple::Script::Header
   ::Simple::Script::Trailer
   ::Simple::Script::Priv

} -aliases {

   {proc-ext ::Simple::Proc::create}

} -install {

### Create the package variables

::Simple::Variable::declare {
   ::Simple::Script::MarkerStart
} -description {
   Start of the begin an end of script header or trailer  markers format
} -type string -value #=========

::Simple::Variable::declare {
   ::Simple::Script::MarkerFormat
} -description {
   Begin and an end of script header or trailer markers format
} -type formatstring -value "#========= %-15s %4d %s"

### Create the package options

# A<-checktype>
::Simple::Variable::declare {
   ::Simple::Proc::%Options%(-checktype)
} -description {
   Whether to perform run-time type checking of the procedures arguments
} -type boolean -value 0

# A<-afterflags>
::Simple::Variable::declare {
   ::Simple::Proc::%Options%(-afterflags)
} -description {
   Minimum number of arguments after the last flag for a procedure in order
   for those to be handled with an auxiliary "after flags" procedure
} -type integer -value 5

} -uninstall {

### Delete the package variables
::Simple::Variable::delete ::Simple::Script::MarkerStart
::Simple::Variable::delete ::Simple::Script::MarkerFormat

### Delete the package options
::Simple::Variable::delete ::Simple::Proc::%Options%(-checktype)
::Simple::Variable::delete ::Simple::Proc::%Options%(-afterflags)

} -firsttimeinstall {

### Declare the package types

# Extended format argument
::Simple::Type::declare {
   extendedarg
} -description {
   Extended format argument as ||{name type ?value? description}||
} -matchingscript {
   ::Simple::Proc::Priv::parse-argument $value 1
   return 1
}

# Canonical argument
::Simple::Type::declare {
   canonicalarg
} -description {
   Canonical argument as ||{name ?value?}||
} -matchingscript {
   ::Simple::Proc::Priv::convert-extended $value
   return 1
}

### Declare the package errors
::Simple::Error::declare {
   ::Simple::Proc::NON-EXISTING-NAMESPACE 
} -message {
   can't create procedure "%s": unknown namespace
} -explanation {
   The extended procedure <procedure> could not be created because the
   namespace does not exist
} -corrective {
   Use K<namespace eval> to create the namespace
}

::Simple::Error::declare {
   ::Simple::Proc::TYPE-CHECKING-AND-NO-PARSE-FLAGS 
} -message {
   can not have "-parseflags false" and argument run-time type checking
} -explanation {
   A procedure declaration failed because argument run-time type checking was
   enabled and flag parsing disabled
} -corrective {
   Either disable argument run-time type checking or enable flag parsing
}

::Simple::Error::declare {
   ::Simple::Proc::ARGS-MUST-BE-LAST 
} -message {
   "args" must be the last argument
} -explanation {
   A procedure declaration failed because an argument named "args" was
   supplied not the last in the list
}

::Simple::Error::declare {
   ::Simple::Proc::NON-OPT-AFTER-OPT 
} -message {
   non-optional argument "%s" after at least one optional argument
} -explanation {
   A procedure declaration failed because the non-optional argument <argument>
   was supplied after an optional argument
} -corrective {
   All arguments following an optional one must be either explicitly optional
   or the special argument "args" which is implictly optional
}

::Simple::Error::declare {
   ::Simple::Proc::DEFAULT-NOT-ALLOWED 
} -message {
   can not have a default value for %s "%s" of type "%s"
} -explanation {
   A procedure or option declaration failed because a default value was
   supplied for an <option or argument> which can not have it
} -corrective {
   Default values must be supplied for procedure optional arguments and for
   flags or options whose type is not "boolflag", and only for those
} -explained {
   can not have a default value for <option or argument> "<option or argument
   name>" of type "<type>"
}

::Simple::Error::declare {
   ::Simple::Proc::ARGS-NOT-ALLOWED 
} -message {
   "args" is not allowed if flag parsing is disabled
} -explanation {
   A procedure declaration failed because an argument named "args" was
   supplied but flag parsing is disabled
} -corrective {
   Either remove the "args" argument or enable flags parsing
}

::Simple::Error::declare {
   ::Simple::Proc::ARGS-LIST-OR-ANY 
} -message {
   "args" must be of type "list" or "any"
} -explanation {
   A procedure declaration failed because an argument named "args" was
   supplied with a type other than "list" or "any"
}

::Simple::Error::declare {
   ::Simple::Proc::NO-FLAG-PARSING-AND-NO-FLAGS 
} -message {
   flag parsing disabled but no flags given
} -explanation {
   A procedure declaration failed because flag parsing was diabled
   but no flags were given in the argument list
} -corrective {
   Enable flag parsing
}

::Simple::Error::declare {
   ::Simple::Proc::NO-PARAM-FOR-FLAG 
} -message {
   no parameter given for flag "%s" to "%s"
} -explanation {
   A call to the procedure <procedure> failed because no parameter was
   supplied for the flag <flag>
} -corrective {
   Supply a parameter for the flag
} -translation {
   <flag> <procedure>
}

::Simple::Error::declare {
   ::Simple::Proc::CANT-BETWEEN-SUBCO-AND-NO-SUBCO 
} -message {
   can't %s %ssubcommand "%s" to %ssubcommand "%s"
} -explanation {
   It is not allowed to <copy or rename> a non-subcommand to a\
   subcommand or viceversa
} -explained {
   can't <copy or rename> [non-]subcommand "<name 1>" to
   [non-]subcommand "<name 2>"
}

::Simple::Error::declare {
   ::Simple::Proc::NON-EXISTING-PROCEDURE 
} -message {
   "%s" isn't %s
} -explanation {
   <procedure> is not <a procedure, an extended procedure or a declared\
   procedure>
} -corrective {
   Provide a valid procedure name
}

::Simple::Error::declare {
   ::Simple::Proc::NON-EXISTING-ARGUMENT 
} -message {
   procedure "%s" doesn't have an argument "%s"
} -explanation {
   Procedure <procedure> does not have an argument named <argument>
} -corrective {
   Provide an argument from the procedure argument list
}

::Simple::Error::declare {
   ::Simple::Proc::NO-TYPE-CHOICE 
} -message {
   argument "%s" from procedure "%s" is not of type "choice" nor "choice-list"
} -explanation {
   The extended procedure <procedure> argument <argument> type is not
   "choice" nor "choice-list"
} -translation {<argument> <procedure>}

::Simple::Error::declare {
   ::Simple::Proc::CANT-BE-CALLED-FROM-TOPLEVEL 
} -message {
   "%s" can't be called from the top level
} -explanation {
   It makes no sense to call procedure <procedure> from the top level
} -corrective {
   Call the procedure from a level other than the top one
}

::Simple::Error::declare {
   ::Simple::Script::DUPLICATED-TAG 
} -message {
   duplicated %s with tag "%s" in script "%s"
} -explanation {
   Script <script> already contains a <header or trailer> with tag <tag>
} -corrective {
   Use K<::Simple::Script::Header::delete> or
   K<::Simple::Script::Trailer::delete> to delete the header with that tag or
   use another tag
} -translation {<header or trailer> <tag> <script>}

::Simple::Error::declare {
   ::Simple::Proc::VALUE-REQUIRED 
} -message {
   a value is required for %s "%s"
} -explanation {
   <optional or flag argument> <optional or flag argument name> requires a
   value, but no value was supplied
} -corrective {
   Supply a value for the optional or flag argument.
}

::Simple::Error::declare {
   ::Simple::Proc::ARRAY-NOT-ALLOWED 
} -message {
   derived array type "%s" not allowed for %s "%s"
} -explanation {
   A derived array type <type> has been used for <argument or option> 
   <argument or option name> but that is not allowed
} -corrective {
   For arguments, a derived array type can be emulated using K<upvar>
}

::Simple::Error::declare {
   ::Simple::Proc::OPTIONAL-FLAG 
} -message {
   cannot have an explicit optional flag "%s"
} -explanation {
   Flag <flag> was explicitly stated as optional while all flags are
   implicitly optional
} -corrective {
   Remove the question marks around the flag name
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::reset
# -synopsis   : Deletes an extended procedure auxiliary data.
# -overview   :
#
#     This procedure deletes all auxiliary data (variables and procedures)
#  corresponding to a procedure with extended argument syntax. Notice that
#  the actual procedure is not deleted.
#
# -arguments  :
#  { procedure        qualifiedname  {Qualified procedure name}}
#
# -remarks    :
#
#  * All commands in this procedure are protecetd against failure via the
#    K<catch> command; this means that no error is issued even if there is no
#    auxiliary data to delete.
#
proc ::Simple::Proc::Priv::reset {
   procedure
} {
   ### Delete the "after flags" auxiliary procedure, if any
   catch {rename $procedure-%AFTER-FLAGS% {}}

   ### Split the procedure name
   regexp -- {^(.*::)(.*)$} $procedure -> namespace unqualifiedProcedure

   ### Delete all extended procedure auxiliary information
   catch {unset ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)}
   catch {unset ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)}
   catch {unset ${namespace}%ProcData%(%FlagsIndices%,$unqualifiedProcedure)}
   catch {unset ${namespace}%ProcData%(%ParseFlags%,$unqualifiedProcedure)}
   catch {unset ${namespace}%ProcData%(%Access%,$unqualifiedProcedure)}
   catch {unset ${namespace}%ProcData%(%Alias%,$unqualifiedProcedure)}
   catch {unset ${namespace}%ProcData%(%Metadata%,$unqualifiedProcedure)}
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::parse-qualify-split
# -synopsis   : Parses, qualifies and splits a name.
# -overview   :
#
#     This procedure parses a name which may be one or two words.
#
#     In case the name contains two words, it is assumed to be a base command
#  and subcommand pair, which are validated.  In case the name contains a
#  single word, the name is validated.  In case the name is unqualified, it is
#  qualified in the scope a given number of levels (two by default) up the
#  stack, unless the A<qualify> argument is zero.
#
# -arguments  :
#  { item             commandname    {Item name}}
#  {?kind?            string {procedure class method {class command}}
#                                    {Kind of item (procedure, class, ...)}}
#  {?levels?          integer 2      {Levels up the stack in which to operate}}
#  {?qualify?         boolean 1      {Whether to qualify unqualified names}}
#
# -returns    : The qualified name, namespace and tail.
#
proc ::Simple::Proc::Priv::parse-qualify-split {
   item
   {kind procedure}
   {levels 2}
   {qualify 1}
} {
   ### Invalid number of words in name or invalid name
   set nWords [llength $item]
   if {$nWords < 1 || $nWords > 2 ||\
      ($nWords == 1 && ![::Simple::Type::Priv::is-basic name $item])} {

      ### Throw error
      ::Simple::Error::throw ::Simple::BAD-ITEM "$kind name" $item
   }

   ### Subitem
   if {$nWords == 2} {

      ### Get the base item and subitem
      foreach {baseItem subitem} $item break

      ### Assert the base item and subitem names are valid
      if {![::Simple::Type::Priv::is-basic name $baseItem] ||\
         ![string compare $baseItem {}]} {

         ::Simple::Error::throw ::Simple::BAD-ITEM\
            {subcommand base command name} $baseItem
      }
      if {![::Simple::Type::Priv::is-basic unqualifiedname $subitem] ||\
         ![string compare $subitem {}]} {

         ::Simple::Error::throw ::Simple::BAD-ITEM {subcommand name}\
            $subitem
      }
   }

   ### Unqualified name and must not qualify it
   if {![string match ::* $item] && !$qualify} {

      ### Do not qualify the name
      set namespace {}
      set unqualifiedItem $item

   ### Qualified name or must qualify it
   } else {
      foreach {tmp namespace unqualifiedItem}\
         [::Simple::Priv::qualify-split $item [incr levels]] break
   }

   ### Return the qualified name, namespace and tail
   return [list $namespace$unqualifiedItem $namespace $unqualifiedItem]
}

### ===========================================================================
### -command  : ::Simple::Proc::create
# -synopsis   : Creates a procedure with extended argument syntax.
# -alias      : proc-ext
# -overview   :
#
#     This procedure acts as an interface to the K<proc> command to create
#  other procedures allowing a richer syntax for their arguments.  Upon
#  creation of a new procedure, its extended format argument list is parsed
#  and converted into a canonical argument list (in which each argument has
#  the format ||{name ?defaultValue?}||) which is used to actually create the
#  procedure via the K<proc> command.  If the argument list contains one or
#  more flags, or if run-time type checking is requested, the procedure body
#  is instrumented to add run-time handling of the argument list.
#
#     If the procedure name is two words, a subcommand procedure similar to
#  that created via the P<SimpleSubcommand> package K<proc-sub> procedure is
#  created, but with extended argument syntax.  It is thus not required to
#  use K<proc-sub> when using this package.
#
#  ~~~~~Extended format arguments
#
#     Extended syntax arguments are of the form ||{name type ?value? 
#  description}||
#
#  ~~~~~~Argument name
#
#     Arguments whose name starts by a hyphen are flags.  The presence of
#  flags in the procedure call actual argument list is optional but.
#
#     Flags whose type is A<boolflag> are boolean flags and do not require a
#  parameter.  For each boolean flag in the argument list, a variable is
#  created in the procedure body with the same name as the flag (excluding the
#  hyphen) and whose value is true or false (''1'' or ''0'') depending on
#  whether the flag is given or not in the procedure call actual argument
#  list.
#
#     Flags whose type is not A<boolflag> are regular flags and require a
#  parameter; the flag type is the expected parameter type.  For each regular
#  flag in the argument list, a variable is created in the procedure body with
#  the same name as the flag (excluding the hyphen) and whose value is the
#  parameter supplied after the flag, or its default value if the flag was not
#  given in the procedure call actual argument list.  The presence or absence
#  of a regular flag in the procedure call actual argument list can be
#  checked via the K<::Simple::Argument::information flaggiven> procedure. 
#
#     The list of all flags in the procedure call actual argument list
#  together with their parameters can be obtained via the
#  K<::Simple::Argument::information givenflags>.
#
#     Arguments whose name is enclosed between question marks are optional. 
#  They differ from the flags in that they behave exactly as optional
#  arguments to the K<proc> command, that is, they must appear at the end of
#  the argument list.
#
#     As with the K<proc> command, the A<args> special argument collects all
#  remaining arguments in the actual argument list and combines them into a
#  list which is available to the procedure body in the variable name A<args>.
#  If given, the A<args> special argument must be the last one.
#
#  ~~~~~~Argument type
#
#     The argument type must be a type known by The Simple Development
#  Library, that is, any Simple Development Library type or declared via the
#  K<::Simple::Type::declare> procedure.
#
#     An actual argument whose type is A<choice> can only hold one of the
#  values specified in the argument declaration default value field.
#
#  See also the A<boolflag> type in the argument name section above.
#
#  ~~~~~~Argument default value or choice list
#
#     An argument declaration can specify this field if and only if its type
#  is A<choice>, an optional argument, or a flag whose type is not
#  A<boolflag>.  For arguments of type A<choice>, this field specifies the
#  list of allowed choices; otherwise, it corresponds to the parameter default
#  value.  For optional arguments of type A<choice>, the default value is the
#  first choice.
#
#  ~~~~~~Argument description
#
#  This is simply a description of the argument.
#
#  ~~~~~Run-time handling
#
#     The A<-checktype> flag allows to override the package option of the same
#  name for the procedure being created.  This option controls whether a
#  run-time type check is performed to assess whether the arguments contents
#  match their type.  This has a performance penalty which should be
#  considered.  The overhead comes from two fronts.  First, the contents of
#  each argument must be checked.  Second, it is not possible to apply the
#  improving performance tricks described in the details section below.
#
#     The A<-parseflags> flag allows to disable the flag parsing which
#  is delegated to the user.  This results in better performance (the
#  procedure call is equivalent to a regular Tcl K<proc> procedure) at the
#  cost of some hassle and restrictions:
#
#  * The arguments from the first flag are replaced by the A<args> special
#    argument.  It is the user responsibility to handle the A<args> special
#    argument to attain the desired behaviour.  See the K<puts> clone in the
#    examples section below.
#
#  * No run-time type checking of the procedure arguments is performed.  As a
#    result, it is not allowed to supply both the A<-parseflags> and
#    A<-checktype> flags.  If the A<-checktype> package option is set to true,
#    it is not honored.
#
# -keywords   : procedure argument type flag
# -arguments  :
#  {-parseflags       extbool true   {Whether to enable flag parsing or
#                                     delegate it to the user}}
#  {-checktype        extbool false  {Whether to enable argument run-time
#                                     type checking}}
#  { procedure        commandname    {Procedure name}}
#  { arguments        extendedarg-list
#                                    {Extended argument list}}
#  { body             script         {Procedure body}}
#
# -returns    : The empty string.
# -details    :
#
#  * If A<-parseflags false> is given, the actual procedure which is created
#    has the first flag and following arguments replaced by the A<args>
#    special argument.  The procedure body is not instrumented.
#
#  * If run-time type checking is enabled, the actual procedure which is
#    created has no arguments.  A call to the
#    K<::Simple::Proc::Priv::parse-arguments-checktype> procedure is added at
#    the very beginning of the procedure body to perform the required
#    run-time parsing and checking.
#
#  * If run-time type checking is disabled, the arguments to the procedure are
#    divided into three sets:  arguments before the first flag, arguments from
#    the first to the last flag and arguments after the last flag.  The actual
#    procedure which is created contains the first set of arguments; the
#    second and third sets are handled by a call to the
#    K<::Simple::Proc::Priv::parse-arguments> procedure which is added at the
#    very beginning of the procedure body to perform the required run-time
#    parsing.  If the number of arguments in the third set exceeds the
#    A<-afterflags> package option value, those are handled via an auxiliary
#    "after flags" procedure instead.
#
# -remarks    :
#
#  * If the procedure namespace does not exists it is not created. This mimics
#    the behaviour of the K<proc> command.
#
#  * The "after flags" procedure is named K<''name''-%AFTER-FLAGS%>
#    where ''name'' is the qualified name of the procedure or method
#    being created.
#
#  * The list of flags given in the actual argument list is stored in a
#    variable named A<%GIVEN-FLAGS%> in the scope of the procedure being
#    created.
#
#  * The header added to the script body when argument run-time type checking
#    is enabled is tagged as ''%PRIV%-parse-arguments'' or
#    ''%PRIV%-parse-arguments-checktype'' with priority ''0''.
#
#  * The order of the A<-parseflags> and A<-checktype> flags is irrelevant,
#    but both of them must appear before the procedure name.
#
# -examples   :
#
#  # A regular procedure
#  proc-ext name {
#     {-boolflag         boolflag       {A boolean flag}}
#     { arg1             string         {First argument}}
#     { arg2             integer        {Second argument}}
#     { arg3             choice {foo bar}
#                                       {Choice between foo and bar}}
#     {-flag             string theflag {A flag (requires parameter)}}
#     {?arg4?            string default {Optional argument}}
#     { args             list           {Remaining arguments}}
#  } {
#     puts "-boolflag = <$boolflag>"
#     puts "arg1      = <$arg1>"
#     puts "arg2      = <$arg2>"
#     puts "arg3      = <$arg3>"
#     puts -nonewline "flag      = <$flag> "
#     if {[::Simple::Argument::information flaggiven -flag]} {
#        puts "(flag given)"
#     } else {
#        puts "(default value)"
#     }
#     puts "arg4      = <$arg4>"
#     puts "args      = <$args>"
#  }
#
#  # A subcommand procedure for which the contents of
#  # their arguments is run-time type checked
#  proc-ext -checktype true {command subcommand} {
#     {-boolflag         boolflag       {A boolean flag}}
#     { arg1             string         {First argument}}
#     { arg2             integer        {Second argument}}
#  } {
#     puts "-boolflag = <$boolflag>"
#     puts "arg1      = <$arg1>"
#     puts "arg2      = <$arg2>"
#  }
#
#  # A clone of puts (except in that it requires the channel)
#  # with flag parsing delegated to the user
#  proc-ext -parseflags false puts-clone {
#     {-nonewline        boolflag       {Do not append a newline}}
#     { channelId        channel        {Output channel}}
#     { string           string         {String to write}}
#  } {
#     foreach {arg1 arg2 arg3} $args break
#     if {![string compare $arg1 -nonewline]} {
#        puts -nonewline  $arg2 $arg3
#     } else {
#        puts             $arg1 $arg2
#     }
#  }
#
# -effects    :
#
#  * Creates a command named K<procedure>.
#
#  * May create an "after flags" command named K<''name''-%AFTER-FLAGS%>
#    where ''name'' is the qualified name of the procedure or method
#    being created.
#
proc ::Simple::Proc::create {
   args
} {
   ### Parse and verify the procedure declaration (phase 1)
   ::Simple::Error::Priv::push-location procedure {} 1
   if {[catch {

      ### Set defaults
      set parseflags 1
      set checktype [set ::Simple::Proc::%Options%(-checktype)]

      ### Parse flags
      set flags [::Simple::Priv::parse-flags\
         [list -parseflags trimmed\
         -checktype trimmed]\
         [list] $args 3 1]

      # Ensure A<-parseflags> and A<-checktype> are booleans
      ::Simple::Proc::Priv::uplevel-ensure-booleans

      ### Parse remaining arguments
      set remainingArgs [lrange $args [expr {2 * [llength $flags]}] end]
      set nRemainingArgs [llength $remainingArgs]
      if {$nRemainingArgs < 3} {
         ::Simple::Error::throw ::Simple::NO-VALUE-FOR-PARAM\
            [lindex {procedure arguments body} $nRemainingArgs]\
            [lindex [info level 0] 0]
      }
      foreach {procedure arguments body} $remainingArgs break

      ### Parse and verify the procedure declaration (phase 2)
      set procedure [string trim $procedure]
      ::Simple::Error::Priv::push-location procedure $procedure 1
      set declarationAndCreationScript\
         [::Simple::Proc::Priv::parse-and-verify-declaration $parseflags\
         $checktype $checktypeGiven $procedure $arguments $body {} {} [list] 0]
   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Declare and create the procedure
   eval $declarationAndCreationScript

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::parse-and-verify-declaration
# -synopsis   : Parses and verifies a procedure or method declaration.
# -arguments  :
#  { parseflags       boolean        {Whether to enable flag parsing or
#                                     delegate it to the user}}
#  { checktype        boolean        {Whether to enable argument run-time
#                                     type checking}}
#  { checktypeGiven   boolean        {Whether A<-checktype> was given in the
#                                     command creation commands}}
#  { item             commandname    {Procedure or method name}}
#  { arguments        extendedarg-list
#                                    {Extended argument list}}
#  { body             script         {Procedure body}}
#  { access           choice {public private exported}
#                                    {Procedure access mode}}
#  { alias            name           {Procedure alias}}
#  { metadata         any-pairs      {Procedure metadata}}
#  { hidden           boolean        {Whether to also consider hidden types}}
#  {?procCommand?     choice {::proc ::itcl::body}
#                                    {Name of the command used to create the
#                                     procedure (::proc) or method or class
#                                     command (::itcl::body)}}
#  {?isMethod?        boolean 0      {Whether the item is a method}}
#
# -returns    : The procedure declaration and creation script.
#
proc ::Simple::Proc::Priv::parse-and-verify-declaration {
   parseflags
   checktype
   checktypeGiven
   item
   arguments
   body
   access
   alias
   metadata
   hidden
   {procCommand ::proc}
   {isMethod 0}
} {
   ### "-parseflags false" and type checking
   if {$checktype && !$parseflags} {

      ### Assert both "-parseflags false" and "-checktype true" are not given
      if {$checktypeGiven} {
         ::Simple::Error::throw\
            ::Simple::Proc::TYPE-CHECKING-AND-NO-PARSE-FLAGS
      }

      ### Disable run-time type checking
      set checktype 0
   }

   ### Parse, qualify and split the procedure or method name
   #   Class command or method
   if {[string compare $procCommand ::proc]} {

      # Class command
      if {!$isMethod} {
         set kind {class command}

      # Method
      } else {
         set kind method
      }

   #   Non-class command
   } else {
      set kind procedure
   }
   foreach {qualifiedItem namespace unqualifiedItem}\
      [::Simple::Proc::Priv::parse-qualify-split $item $kind 3] break

   ### Loop over arguments
   set argsGiven 0
   set nArgs 0
   set firstFlag -1
   set lastFlag -1
   set optionalGiven 0
   set internal [list]
   set argNames [list]
   foreach argument $arguments {

      ### Assert "args" is the last argument
      if {$argsGiven} {
         ::Simple::Error::throw ::Simple::Proc::ARGS-MUST-BE-LAST
      }

      ### Parse the argument
      foreach {isOptional isFlag defaultValueGiven argName defaultValue\
         type choices description originalArgName canonical}\
         [::Simple::Proc::Priv::parse-argument $argument $hidden] break
      set argumentName($argument) $argName
      set argumentCanonical($argument) $canonical

      ### Handle optional arguments
      if {$isOptional} {
         set optionalGiven 1
      } elseif {$optionalGiven && [string compare $argName args]} {
         ::Simple::Error::throw ::Simple::Proc::NON-OPT-AFTER-OPT\
            $originalArgName
      }

      ### Handle flag arguments
      if {$isFlag} {
         if {$firstFlag == -1} {
            set firstFlag $nArgs
         }
         set lastFlag $nArgs
      }

      ### Assert the default value is required, if given
      if {$defaultValueGiven && !$isOptional &&\
         !($isFlag && [string compare $type boolflag])} {

         ::Simple::Error::throw ::Simple::Proc::DEFAULT-NOT-ALLOWED\
            argument $originalArgName $type
      }

      ### Handle special argument "args"
      if {![string compare $argName args]} {
         if {!$parseflags} {
            ::Simple::Error::throw ::Simple::Proc::ARGS-NOT-ALLOWED
         }
         if {[string compare $type list] && [string compare $type any]} {
            ::Simple::Error::throw ::Simple::Proc::ARGS-LIST-OR-ANY
         }
         set argsGiven 1
      }

      ### Argument is valid
      lappend internal [list $isFlag $defaultValueGiven $argName $defaultValue\
         $type $choices $description]
      if {$isFlag} {
         lappend argNames -$argName
      } else {
         lappend argNames $argName
      }

      ### Next argument
      incr nArgs
   }

   ### Assert there is at least one flag if flags parsing is disabled
   if {!$parseflags && $firstFlag == -1} {
      ::Simple::Error::throw ::Simple::Proc::NO-FLAG-PARSING-AND-NO-FLAGS
   }

   # The procedure metadata must not be stored
   if {![set ::Simple::%Options%(-storemetadata)]} {
      set metadata [list]
   }

   ### Run-time argument type checking or flag parsing disabled
   set createAfterFlags 0
   set afterFlagsArgs {}
   set afterFlagsBody {}
   if {$checktype || !$parseflags} {

      # Set the position of the last flag in the
      # argument list to -1 so that it is ignored
      set lastFlag -1

   ### No run-time argument type checking and flag parsing disabled
   } elseif {$lastFlag != -1} {

      ### "after flags" auxiliary procedure required
      #   This happens if the number of arguments after the last flag exceeds
      #   the after flags threshold, unless run-time type checking is enabled
      #   because then all arguments are parsed at run time.
      #      Notice that the "after flags" procedure does not need to be
      #   exported from its namespace, even if its originator procedure is
      #   exported, as imported commands simply point to the exported one in
      #   its original namespace.
      #      In the case of methods, the "after flags" procedure is really
      #   a class command.
      if {$nArgs - $lastFlag > [set ::Simple::Proc::%Options%(-afterflags)]} {

         # Bug 5: bad handling of arguments in "after flags" procedures
         # Solution: Handle V<argList1> and V<argList2> as strings
         set afterFlagsArgs [list]
         set argList1 {}
         set argList2 {}

         # Loop over arguments
         foreach argument [lrange $arguments [expr {$lastFlag + 1}] end] {

            # Append to argument list
            set argName $argumentName($argument)
            append argList1 " $argName"
            append argList2 " $$argName"

            # Append to "after flags" auxiliary procedure argument list
            lappend afterFlagsArgs $argumentCanonical($argument)
         }

         # Create the "after flags" auxiliary procedure
         set createAfterFlags 1
         set afterFlagsBody [format {uplevel 2\
            "foreach {[list%s]} {[list%s]} {}"} $argList1 $argList2]

      ### "after flags" auxiliary procedure not required
      } else {

         # Set the position of the last flag in the
         # argument list to -1 so that it is ignored
         set lastFlag -1
      }
   }

   ### Run-time type checking requested and required
   #   Run-time type checking is not required if the procedure or
   #   method has no arguments
   if {$checktype && $nArgs > 0} {

      ### Create the actual procedure or method
      if {$firstFlag == -1} {
         ::Simple::Script::Header::add body %PRIV%-parse-arguments-checktype 0\
            [format {::Simple::Proc::Priv::parse-arguments-checktype %s\
            $args} [list $namespace $unqualifiedItem]]
      } else {
         ::Simple::Script::Header::add body %PRIV%-parse-arguments-checktype 0\
            [format {set %%GIVEN-FLAGS%%\
            [::Simple::Proc::Priv::parse-arguments-checktype\
            %s $args]} [list $namespace $unqualifiedItem]]
      }
      set arguments args

   ### Run-time type checking not requested or not required
   } else {

      ### Get the argument list up to the first flag
      if {$firstFlag == -1} {
         set beforeFlagsExtendedArgs $arguments
      } else {
         set beforeFlagsExtendedArgs\
            [lrange $arguments 0 [expr {$firstFlag - 1}]]
      }
      set beforeFlagsArgs [list]
      foreach argument $beforeFlagsExtendedArgs {

         # Append to argument list
         lappend beforeFlagsArgs $argumentCanonical($argument)
      }

      ### Create the actual procedure or method
      if {$parseflags} {
         if {$firstFlag == -1} {

            # Run-time parsing of the actual argument list not required
            # "args" argument not required
            set arguments $beforeFlagsArgs

         } else {

            # Run-time parsing of the actual argument list required
            # "args" argument required
            ::Simple::Script::Header::add body %PRIV%-parse-arguments 0\
               [format {set %%GIVEN-FLAGS%%\
               [::Simple::Proc::Priv::parse-arguments %s $args]}\
               [list $namespace $unqualifiedItem]]
            set arguments [concat $beforeFlagsArgs args]
         }
      } else {

         # Run-time parsing of the actual argument list not required
         # "args" argument required
         set arguments [concat $beforeFlagsArgs args]
      }
   }

   ### Compose the procedure declaration and creation script
   set declarationAndCreationScript "::Simple::Proc::Priv::do-declare\
      [list $item] [list $unqualifiedItem] $namespace $createAfterFlags\
      [list $afterFlagsArgs] [list $afterFlagsBody] [list $internal]\
      [list $argNames] [list $access] [list $alias] [list $metadata]\
      [list [list $firstFlag $lastFlag]] $parseflags $procCommand $isMethod\
      [list $arguments] [list $body]"

   ### Return the procedure declaration and creation script
   return $declarationAndCreationScript
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::do-declare
# -synopsis   : Declares a procedure or method.
# -arguments  :
#
#  { item             commandname    {Procedure or method name}}
#  { unqualifiedItem  unqualifiedname
#                                    {Unqualified procedure name}}
#  { namespace        namespace      {Procedure namespace}}
#  { createAfterFlags boolean        {Whether the "after flags" procedure must
#                                     be created}}
#  { afterFlagsArgs   canonicalarg-list       
#                                    {"After flags" procedure arguments}}
#  { afterFlagsBody   script         {"After flags" procedure body}}
#  { internal         list           {Procedure argument list internal
#                                     representation}}
#  { argNames         canonicalarg-list
#                                    {Procedure canonical argument list}}
#  { access           choice {public private exported}
#                                    {Procedure access mode}}
#  { alias            name           {Procedure alias}}
#  { metadata         any-pairs      {Procedure metadata}}
#  { flagsIndices     integer-list   {Indices of the first and last flags in
#                                     the procedure argument list}}
#  { parseflags       boolean        {Whether the flag parsing is enabled}}
#  { procCommand      choice {::proc ::itcl::body}
#                                    {Name of the command used to create the
#                                     procedure (::proc) or method or class
#                                     command (::itcl::body)}}
#  { isMethod         boolean        {Whether the item is a method}}
#  { arguments        extendedarg-list
#                                    {Procedure extended argument list}}
#  { body             script         {Procedure body}}
#
# -returns    : The empty string.
#
proc ::Simple::Proc::Priv::do-declare {
   item
   unqualifiedItem
   namespace
   createAfterFlags
   afterFlagsArgs
   afterFlagsBody
   internal
   argNames
   access
   alias
   metadata
   flagsIndices
   parseflags
   procCommand
   isMethod
   arguments
   body
} {
   ### Assert the namespace exists
   if {[catch {namespace children $namespace}]} {
      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-NAMESPACE $item
   }

   ### Subcommand procedure
   if {[llength $unqualifiedItem] == 2} {

      ### Create the base command
      ::Simple::Priv::save-name [lindex $item 0] 3
      ::Simple::Subcommand::Priv::create-base-command\
         $namespace[lindex $unqualifiedItem 0] [lindex $unqualifiedItem 1]\
         $procCommand $isMethod
   }

   ### Delete the procedure and auxiliary data, in case it is being recreated
   #   Methods are not deleted as [incr Tcl] does not allow that
   set qualifiedItem $namespace$unqualifiedItem
   if {![string compare $procCommand ::proc]} {
      catch {rename $qualifiedItem {}}
   }
   ::Simple::Proc::Priv::reset $qualifiedItem

   ### The "after flags" auxiliary procedure must be created
   if {$createAfterFlags} {

      ### Create the "after flags" auxiliary procedure
      # With a name of the form <qualifiedItem>-%AFTER-FLAGS%
      proc $qualifiedItem-%AFTER-FLAGS% $afterFlagsArgs $afterFlagsBody
   }

   ### Store the procedure or method internal representation
   set ${namespace}%ProcData%(%FlagsIndices%,$unqualifiedItem) $flagsIndices
   set ${namespace}%ProcData%(%ParseFlags%,$unqualifiedItem) $parseflags
   set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedItem) $internal
   set ${namespace}%ProcData%(%ArgNames%,$unqualifiedItem) $argNames
   set ${namespace}%ProcData%(%Access%,$unqualifiedItem) $access
   set ${namespace}%ProcData%(%Alias%,$unqualifiedItem) $alias
   set ${namespace}%ProcData%(%Metadata%,$unqualifiedItem) $metadata

   ### Create the procedure or method
   $procCommand $qualifiedItem $arguments $body

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::uplevel-ensure-booleans
# -synopsis   : Ensures A<-parseflags> and A<-checktype> are booleans
# -remarks    :
#
#  * This is a full K<uplevel> procedure used to factor code out.
#
# -effects    :
#
#  * Modifies V<parseflags>, V<checktypeGiven> and V<checktype>
#
proc ::Simple::Proc::Priv::uplevel-ensure-booleans {
} {
   ### Up one level
   uplevel 1 {
      set parseflags [{::Simple::Type::convert boolean} $parseflags]
      set checktypeGiven [::Simple::Argument::information flaggiven -checktype]
      if {$checktypeGiven} {
         set checktype [{::Simple::Type::convert boolean} $checktype]
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::parse-argument
# -synopsis   : Parses an extended format argument.
# -overview   :
#
#     This procedure parses an extended format argument ensuring its format is
#  correct and returns all useful argument details.
#
# -arguments  :
#  { extended         extendedarg    {Extended format argument}}
#  { hidden           boolean        {Whether to also consider hidden types}}
#
# -returns    :
#
#  A list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0: whether the argument is optional
#     1: whether the argument is a flag
#     2: whether a value for the argument has been given
#     3: argument name
#     4: argument value
#     5: argument type
#     6: argument choices (empty if type is not "choice" nor derived from it)
#     7: argument description
#     8: argument original name, including the hyphen(s) for flags and question
#        marks for optional arguments
#     9: argument in canonical format ||{name ?value?}||
##end
#
proc ::Simple::Proc::Priv::parse-argument {
   extended
   hidden
} {
   ### Validate the argument format
   set nTokens [llength $extended]
   if {$nTokens < 3 || $nTokens > 4} {
      ::Simple::Error::throw ::Simple::BAD-ITEM {extended argument} $extended
   }
   set valueGiven [expr {$nTokens == 4}]

   ### Get tokens
   if {$valueGiven} {
      foreach {name type value description} $extended {}
   } else {
      foreach {name type description} $extended {}
      set value {}
   }

   ### Handle optional arguments
   set isOptional [::Simple::Type::Priv::is-basic optional $name]
   if {$isOptional} {
      set name [{::Simple::Type::convert optional} $name]
      if {!$valueGiven} {
         ::Simple::Error::throw ::Simple::Proc::VALUE-REQUIRED\
            {optional argument} $name
      }
   }

   ### Parse flags
   set originalName $name
   set isFlag [::Simple::Type::Priv::is-basic flag $name]
   if {$isFlag} {
      if {!$valueGiven && [string compare $type boolflag]} {
         ::Simple::Error::throw ::Simple::Proc::VALUE-REQUIRED\
            {flag argument} $originalName
      }
      if {$isOptional} {
         ::Simple::Error::throw ::Simple::Proc::OPTIONAL-FLAG ?$name?
      }
      set name [{::Simple::Type::convert flag} $name]
   } else {
      if {![string compare $type boolflag]} {
         ::Simple::Error::throw ::Simple::Variable::BAD-TYPE-BOOLFLAG\
            $originalName
      }
   }

   ### Validate the argument name
   if {![::Simple::Type::Priv::is-basic unqualifiedname $name]} {
      ::Simple::Error::throw ::Simple::BAD-ITEM {argument name} $originalName
   }

   ### Assert the type exists
   if {![::Simple::Type::Priv::exists-with-hidden $type $hidden]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING type $type
   }

   ### Assert it is not a derived array type
   if {[string match *-array $type]} {
      ::Simple::Error::throw ::Simple::Proc::ARRAY-NOT-ALLOWED\
         $type argument $name
   }

   ### Handle choices
   if {[string match choice* $type]} {
      if {!$valueGiven || [llength $value] < 2} {
         ::Simple::Error::throw ::Simple::Variable::AT-LEAST-TWO-CHOICES\
            argument $originalName
      }
      set choices $value

      # Default value is first choice
      set value [lindex $choices 0]

      # Not optional nor flag
      if {!$isOptional && !$isFlag} {

         # As if no value given
         set valueGiven 0
      }

   } else {
      set choices [list]
   }

   ### Assert the value is valid
   if {$valueGiven && ![string match choice* $type]} {
      ::Simple::Variable::Priv::assert-type-conformation\
         argument 0 $originalName $type $value $hidden
   }

   ### The argument descriptions must not be stored
   if {![set ::Simple::%Options%(-storemetadata)]} {
      set description {}
   } else {

      # Convert the description to a single line
      set description [::Simple::Priv::convert-to-line $description]
   }

   ### Canonical format
   # Argument name and value
   if {$valueGiven} {

      # {name value}
      set canonical [list $name $value]

   # Argument name only
   } else {

      # name
      set canonical $name
   }

   ### Argument is valid
   return [list $isOptional $isFlag $valueGiven\
      $name $value $type $choices $description $originalName $canonical]
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::parse-arguments
# -synopsis   : Parses an argument list in run-time.
# -arguments  :
#  { namespace        namespace      {Procedure namespace}}
#  { procedure        unqualifiedname
#                                    {Unqualified procedure name}}
#  { actualArgs       list           {Actual argument list}}
#
# -returns    : The list of flags given in the actual argument list.
# -effects    :
#
#  * Creates variables named after the command named K<procedure>
#    arguments in the calling scope.
#
# -references :
#
#  * The K<::Simple::Proc::Priv::parse-arguments-checktype> procedure.
#
#  * The K<::Simple::Error::Priv::parse-options> procedure.
#
proc ::Simple::Proc::Priv::parse-arguments {
   namespace
   procedure
   actualArgs
} {
   # Flags indices in the argument list
   foreach {firstFlag lastFlag}\
      [set ${namespace}%ProcData%(%FlagsIndices%,$procedure)] break

   # Argument list
   set arguments [set ${namespace}%ProcData%(%ArgInternal%,$procedure)]

   # First argument
   set argument [lindex $arguments [set argumentIndex $firstFlag]]

   # First actual argument
   set actualArg [lindex $actualArgs [set actualArgIndex 0]]

   ### Loop over actual arguments
   set givenFlags [list]
   set actualArgsLlength [llength $actualArgs]
   while {$actualArgIndex < $actualArgsLlength} {

      ### Loop over arguments
      while 1 {

         # If the argument name is empty we are beyond the last one
         set name [lindex $argument 2]
         set isFlag [lindex $argument 0]
         if {![string compare $name {}]} {
            ::Simple::Error::throw\
               ::Simple::TOO-MANY-ARGS [{::Simple::Proc::information caller}]
         }

         ### Parse flags
         if {$isFlag} {

            ### Handle match
            set requiresParameter [lindex $argument 1]
            if {![string compare $actualArg -$name]} {

               # Flag given
               lappend givenFlags -$name

               ### Parameter required
               if {$requiresParameter} {

                  # Take parameter from next actual argument
                  incr actualArgIndex
                  if {$actualArgIndex >= $actualArgsLlength} {
                     ::Simple::Error::throw\
                        ::Simple::Proc::NO-PARAM-FOR-FLAG -$name\
                        [{::Simple::Proc::information caller}]
                  }
                  set parameter [lindex $actualArgs $actualArgIndex]

                  # Set the flag in the calling scope
                  uplevel 1 [list ::set $name $parameter]

               ### Parameter not required
               } else {

                  # Set the boolflag in the calling scope: given
                  uplevel 1 [list ::set $name 1]
               }

               # If the last flag has already being processed and
               # there is an "after flags" auxiliary procedure
               if {$argumentIndex == $lastFlag} {

                  # Unset the "args" special argument in the calling scope
                  uplevel 1 [list ::unset args]

                  # Process the remaining arguments via the
                  # "after flags" auxiliary procedure

                  # Error evaluating the "after flags" auxiliary procedure
                  if {[catch {eval [list $namespace$procedure-%AFTER-FLAGS%]\
                     [lrange $actualArgs [incr actualArgIndex] end]} result]} {

                     # Replace the "after flags" auxiliary procedure name
                     # by the name of the caller for proper error reporting
                     regsub -all -- {"::[^"]*-%AFTER-FLAGS%} $result\
                        \"[{::Simple::Proc::information caller}] result

                     # Rethrow the error
                     return -code error -errorcode $::errorCode $result
                  }

                  # Return the list of flags given
                  # in the actual argument list
                  return $givenFlags
               }

               ### Eat argument
               set argument [lindex $arguments [incr argumentIndex]]

               ### Next actual argument
               break

            ### Handle non-match
            } else {

               ### Parameter required
               if {$requiresParameter} {

                  # Set the argument in the calling scope,
                  # taking the parameter from the default value
                  set defaultValue [lindex $argument 3]
                  uplevel 1 [list ::set $name $defaultValue]

               ### Parameter not required
               } else {

                  # Set the boolflag in the calling scope: not given
                  uplevel 1 [list ::set $name 0]
               }

               # If the last flag has already being processed and
               # there is an "after flags" auxiliary procedure
               if {$argumentIndex == $lastFlag} {

                  # Unset the "args" special argument in the calling scope
                  uplevel 1 [list ::unset args]

                  # Process the remaining arguments via the
                  # "after flags" auxiliary procedure

                  # Error evaluating the "after flags" auxiliary procedure
                  if {[catch {eval [list $namespace$procedure-%AFTER-FLAGS%]\
                     [lrange $actualArgs $actualArgIndex end]} result]} {

                     # Replace the "after flags" auxiliary procedure name
                     # by the name of the caller for proper error reporting
                     regsub -all -- {"::[^"]*-%AFTER-FLAGS%} $result\
                        \"[{::Simple::Proc::information caller}] result

                     # Rethrow the error
                     return -code error -errorcode $::errorCode $result
                  }

                  # Return the list of flags given
                  # in the actual argument list
                  return $givenFlags
               }

               ### Eat argument
               set argument [lindex $arguments [incr argumentIndex]]

               ### Next argument
               continue
            }

         ### Handle non-flags
         } else {

            ### Handle the special argument "args"
            if {![string compare $name args]} {

               # Get all remaining actual arguments
               uplevel 1 [list ::set args [lrange $actualArgs $actualArgIndex end]]

               # Return the list of flags given in the actual argument list
               return $givenFlags
            }

            # Set the argument in the calling scope
            uplevel 1 [list ::set $name $actualArg]

            ### Eat argument
            set argument [lindex $arguments [incr argumentIndex]]

            ### Next actual argument
            break
         }
      }

      ### Eat actual argument
      set actualArg [lindex $actualArgs [incr actualArgIndex]]
   }

   ### Loop over not used arguments
   set argumentsLlength [llength $arguments]
   while {$argumentIndex < $argumentsLlength} {

      set name [lindex $argument  2]
      if {![string compare $name args]} {

         # No remaining actual arguments
         uplevel 1 [list ::set args [list]]

         # Return the list of flags given in the actual argument list
         return $givenFlags
      }

      set defaultGiven [lindex $argument 1]
      if {$defaultGiven} {

         # Set the argument in the calling scope,
         # taking the parameter from the default value
         set defaultValue [lindex $argument 3]
         uplevel 1 [list ::set $name $defaultValue]

      } else {

         # boolflag
         if {[lindex $argument 0]} {

            # Set the boolflag in the calling scope: not given
            uplevel 1 [list ::set $name 0]
         } else {
            ::Simple::Error::throw ::Simple::NO-VALUE-FOR-PARAM $name\
               [{::Simple::Proc::information caller}]
         }
      }

      ### Eat non-used argument
      set argument [lindex $arguments [incr argumentIndex]]
   }

   # If this point is reached, this procedure does not use the
   # special argument "args", so unset it in the calling scope
   uplevel 1 [list ::unset args]

   # Return the list of flags given in the actual argument list
   set givenFlags
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::convert-extended
# -synopsis   : Converts an argument from canonical to extended format.
# -overview   :
#
#     This procedure converts canonical arguments (the format used by the
#  K<proc> command arguments, ||{name ?default?}||) to extended format.
#
# -arguments  :
#  { canonical        canonicalarg   {Canonical argument}}
#
# -returns    : The argument in extended format.
#
proc ::Simple::Proc::Priv::convert-extended {
   canonical
} {
   set length [llength $canonical]
   if {$length < 1 || $length > 2} {
      ::Simple::Error::throw ::Simple::BAD-ITEM\
         {canonical argument} $canonical
   }

   ### Validate the argument name
   set name [lindex $canonical 0]
   if {![::Simple::Type::Priv::is-basic name $name]} {
      ::Simple::Error::throw ::Simple::BAD-ITEM {argument name} $name
   }

   ### Argument name and value
   if {$length == 2} {

      # {?name? any value ""}
      set answer [list ?$name? any [lindex $canonical 1] {}]

   ### Argument name only
   } else {

      # {name any ""}
      set answer [list $name any {}]
   }

   ### Return the argument in extended format
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information exists
# -synopsis   : Returns whether a procedure exists.
# -overview   :
#
#  This command returns whether a procedure exists, being it extended or not.
#
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : Whether the procedure exists.
#
proc-sub {::Simple::Proc::information exists} {
   procedure
} {
   ### Parse the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### The extended procedure exists
      set answer 1

   ### Non existing or non-extended procedure
   } else {

      ### Try with a regular procedure
      set answer [::Simple::Priv::proc-information-exists\
         $namespace $unqualifiedProcedure]
   }

   ### Return the procedure exists
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information isextended
# -synopsis   : Returns whether a procedure is an extended procedure.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : Whether the extended procedure exists.
#
proc-sub {::Simple::Proc::information isextended} {
   procedure
} {
   ### Parse the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Return the extended procedure exists
   return [info exists ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]
}

### ===========================================================================
### -command  : ::Simple::Argument::information defaultgiven
# -synopsis   : Returns whether a procedure argument has a default value.
# -arguments  :
#
#  { procedure        commandname    {Procedure name}}
#  { argument         unqualifiedname
#                                    {Argument name}}
#
# -returns    : Whether the procedure argument has a default value.
#
proc-sub {::Simple::Argument::information defaultgiven} {
   procedure
   argument
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Parse the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Locate the argument
      set position [lsearch -exact [set\
         ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] $argument]

      ### Non-existing argument
      if {$position == -1} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-ARGUMENT\
            [::Simple::Priv::restore-name] $argument
      }

      ### Set whether the procedure argument has a default value
      set answer [lindex [lindex\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         $position] 1]

   ### Non existing or non-extended procedure
   } else {

      ### Try with a regular procedure
      set answer [uplevel 1 [list ::info default $procedure $argument tmp]]
   }

   ### Return whether the procedure argument has a default value
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Argument::information default
# -synopsis   : Returns a procedure argument default value.
# -arguments  :
#
#  { procedure        commandname    {Procedure name}}
#  { argument         unqualifiedname
#                                    {Argument name}}
#
# -returns    : Whether the procedure argument has a default value.
#
proc-sub {::Simple::Argument::information default} {
   procedure
   argument
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Parse the procedure name
   foreach {qualifiedProcedure namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Locate the argument
      set position [lsearch -exact [set\
         ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] $argument]

      ### Non-existing argument
      if {$position == -1} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-ARGUMENT\
            [::Simple::Priv::restore-name] $argument
      }

      ### Assert the argument has a default value
      if {![lindex [lindex\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         $position] 1]} {

         ::Simple::Error::throw ::Simple::Variable::NO-DEFAULT-VALUE\
            argument $argument
      }

      ### Set the procedure argument default value
      set answer [lindex [lindex\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         $position] 3]

   ### Non existing or non-extended procedure
   } else {

      ### Try with a regular procedure
      if {[::info default $qualifiedProcedure $argument answer] == 0} {
         ::Simple::Error::throw ::Simple::Variable::NO-DEFAULT-VALUE\
            argument $argument
      }
   }

   ### Return the procedure argument default value
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Argument::information flaggiven
# -synopsis   : Returns whether a flag was given in an extended procedure call.
# -arguments  :
#  { flag             flag           {Flag name}}
#
# -keywords   : procedure argument flag
# -returns    : Whether the flag was given in the actual argument list.
# -examples   :
#
#  # A procedure with a single flag argument which echoes its value
#  # informing on whether that value is the default or was given in the
#  # argument list
#  proc-ext name {
#     {-flag             string theflag {A flag (requires parameter)}}
#  } {
#     puts -nonewline "flag      = <$flag> "
#     if {[::Simple::Argument::information flaggiven -flag]} {
#        puts "(flag given)"
#     } else {
#        puts "(default value)"
#     }
#  }
#
# -references :
#
#  * The K<::Simple::Argument::information givenflags> procedure.
#
#  * The K<::Simple::Priv::flag-given> procedure.
#
#  * The K<::Simple::Option::information usage> procedure.
#
# -limitations:
#
#  * This procedure must be called from another procedure.
#
proc-sub {::Simple::Argument::information flaggiven} {
   flag
} {
   ### Return whether the flag was given in the actual argument list.
   upvar %GIVEN-FLAGS% flags
   return [expr {[lsearch -exact $flags $flag] != -1}]
}

### ===========================================================================
### -command  : ::Simple::Argument::information isdefault
# -synopsis   : Returns whether an argument value is the default one.
# -overview   :
#
#     For flags or optional arguments, this procedure returns whether its
#  actual value matches its corresponding default value. For other arguments
#  the procedure returns always false.
#
# -arguments  :
#  { argument         unqualifiedname
#                                    {Argument name}}
#
# -keywords   : procedure argument {default value}
# -returns    : 
#
#  Whether the argument has a default value and it matches its actual value.
#
# -examples   :
#
#  # A procedure with an optional argument whose default value
#  # is rather estrange, value which is used to check whether
#  # the argument was given in the actual argument list
#  proc-ext name {
#     {?argument?   string %DEFAULT%      {Optional argument}}
#  } {
#     if {[::Simple::Argument::information isdefault argument]} {
#        puts {argument not given}
#     } else {
#        puts "argument given = $argument"
#     }
#  }
#
# -limitations:
#
#  * This procedure must be called from another procedure.
#
# -references :
#
#  * The K<::Simple::Option::information isdefault> procedure.
#
proc-sub {::Simple::Argument::information isdefault} {
   argument
} {
   ### Get the caller name
   set caller [{::Simple::Proc::information caller}]

   ### Save the caller name
   ::Simple::Priv::save-name $caller

   ### Assert this procedure has been called from within another
   if {![string compare $caller {}]} {
      ::Simple::Error::throw ::Simple::Proc::CANT-BE-CALLED-FROM-TOPLEVEL\
         {::Simple::Argument::information isdefault}
   }

   ### Qualify the the name of the caller
   set caller [::Simple::Priv::qualify $caller]

   ### The caller is an extended procedure
   if {[{::Simple::Proc::information isextended} $caller]} {

      ### The argument has a default value
      if {[{::Simple::Argument::information defaultgiven}\
         $caller $argument]} {

         ### Fetch it
         set default\
            [{::Simple::Argument::information default} $caller $argument]

         ### It is a flag
         if {[::Simple::Type::Priv::is-basic flag $argument]} {

            ### Convert the flag to a name
            set argument [{::Simple::Type::convert flag} $argument]
         }

         ### Get whether the value of the argument
         ###    in the calling scope is the default one
         set answer [expr {![string compare $default [uplevel 1 ::set $argument]]}]

      ### The argument does not have a default value
      } else {

         ### Return false
         set answer 0
      }

   ### The caller is not an extended procedure
   } else {

      ### Error getting the argument default value
      if {[catch {info default $caller $argument default} result]} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-ARGUMENT\
            [::Simple::Priv::restore-name] $argument

      ### Argument default value got
      } else {

         ### The argument has a default value
         if {$result} {

            ### Get whether the value of the argument
            ###    in the calling scope is the default one
            set answer\
               [expr {![string compare $default [uplevel 1 ::set $argument]]}]

         ### The argument does not have a default value
         } else {

            ### Return false
            set answer 0
         }
      }
   }

   ### Return whether the value of the argument
   ###    in the calling scope is the default one
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Proc::information caller
# -synopsis   : Returns the caller procedure name.
# -overview   :
#
#     This procedure returns the name of a procedure up the calling stack.  By
#  default, one level up the stack is used.
#
# -returns    : The caller procedure name or the empty string if not found.
# -arguments  :
#  {?levels?          integer 1      {Levels up the stack in which to operate}}
#
proc-sub {::Simple::Proc::information caller} {
   {levels 1}
} {
   if {![catch {set answer [lindex [info level -[incr levels]] 0]}]} {
      return $answer
   }
}

### ===========================================================================
### -command  : ::Simple::Script::Header::add
# -synopsis   : Prepends a header to a script.
# -overview   :
#
#  This command prepends a header to a script.
#
#     The header is enclosed between a begin and an end of header markers
#  formatted through the V<MarkerFormat> package variable.  Both markers
#  contain the header tag and priority.  Successive headers within a script
#  are ordered by increasing priority which goes from ''0'' to ''9999''.  If a
#  header with the same priority as the one being prepended already exists,
#  the original has precedence, that is, the later gets added after the
#  former.
#
# -arguments  :
#  { scriptName       name           {Script name}}
#  { tag              string         {Header tag}}
#  { priority         integer        {Header priority}}
#  { header           script         {Header}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Prepends the header to the script called K<scriptName> in the calling
#    scope.
#
proc ::Simple::Script::Header::add {
   scriptName
   tag
   priority
   header
} {
   ### Assert the script exists
   upvar $scriptName script
   if {![info exists script]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING script $scriptName
   }

   # Header to prepend
   set header "[format $::Simple::Script::MarkerFormat %HEADER-BEGIN%\
      $priority $tag]\n$header\n[format $::Simple::Script::MarkerFormat\
      %HEADER-END% $priority $tag]"

   ### Headers present
   if {[regexp -- "^($::Simple::Script::MarkerStart\
      %HEADER-BEGIN%.*\n$::Simple::Script::MarkerStart\
      %HEADER-END%\[^\n\]*)(.*)$" $script -> originalHeaders afterHeaders]} {

      set originalHeaders\
         [::Simple::Script::Priv::split-headers-or-trailers $originalHeaders]
      # A<afterHeaders> contains the script after the headers
      # A<originalHeaders> contains the list of headers

      ### Loop over original headers
      set RE "^$::Simple::Script::MarkerStart %HEADER-BEGIN%\
         +(\[0-9\]+) (\[^\n\]+)\n.*$"
      set notFound 1
      set headers [list]
      set index 0
      set insertIndex end
      foreach originalHeader $originalHeaders {

         regexp -- $RE $originalHeader -> originalPriority originalTag
         # A<originalPriority> contains the priority of this header
         # A<originalTag> contains the tag of this header

         ### Assert the tag is not already used
         if {![string compare $originalTag $tag]} {
            ::Simple::Error::throw ::Simple::Script::DUPLICATED-TAG\
               header $tag $scriptName
         }

         ### Append the original header to the new list of headers
         lappend headers $originalHeader

         ### Header with higher priority found
         if {$notFound && ($originalPriority > $priority)} {
            set notFound 0
            set insertIndex $index
         }

         ### Next original header
         incr index
      }

      ### Insert the header in its place
      set headersString\
         [join [linsert $originalHeaders $insertIndex $header] \n]
      
      ### Rebuild the script
      set script $headersString$afterHeaders

   ### No original headers present
   } else {

      ### Prepend the header to the script
      set script $header\n$script
   }

   ### Return the empty script
   return {}
}

### ===========================================================================
### -command  : ::Simple::Script::Trailer::add
# -synopsis   : Appends a trailer to a script.
# -overview   :
#
#  This command appends a trailer to a script.
#
#     The trailer is enclosed between a begin and an end of trailer markers
#  formatted through the V<MarkerFormat> package variable.  Both markers
#  contain the trailer tag and priority.  Successive trailers within a script
#  are ordered by increasing priority which goes from ''0'' to ''9999''.  If a
#  trailer with the same priority as the one being appended already exists,
#  the original has precedence, that is, the later gets added after the
#  former.
#
# -arguments  :
#  { scriptName       name           {Script name}}
#  { tag              string         {Trailer tag}}
#  { priority         integer        {Trailer priority}}
#  { trailer          script         {Trailer}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Appends the trailer to the script called K<scriptName> in the calling
#    scope.
#
proc ::Simple::Script::Trailer::add {
   scriptName
   tag
   priority
   trailer
} {
   ### Assert the script exists
   upvar $scriptName script
   if {![info exists script]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING script $scriptName
   }

   # Trailer to append
   set trailer "[format $::Simple::Script::MarkerFormat %TRAILER-BEGIN%\
      $priority $tag]\n$trailer\n[format $::Simple::Script::MarkerFormat\
      %TRAILER-END% $priority $tag]"

   ### Trailers present
   set index\
      [string first "$::Simple::Script::MarkerStart %TRAILER-BEGIN" $script]
   if {$index != -1} {

      set originalTrailers [::Simple::Script::Priv::split-headers-or-trailers\
         [string range $script $index end]]
      set beforeTrailers [string range $script 0 [incr index -1]]
      # A<beforeTrailers> contains the script before the trailers
      # A<originalTrailers> contains the list of trailers

      ### Loop over original trailers
      set RE "^$::Simple::Script::MarkerStart %TRAILER-BEGIN%\
         +(\[0-9\]+) (\[^\n\]+)\n.*$"
      set notFound 1
      set trailers [list]
      set index 0
      set insertIndex end
      foreach originalTrailer $originalTrailers {

         ### Assert the tag is not already used
         regexp -- $RE $originalTrailer -> originalPriority originalTag
         if {![string compare $originalTag $tag]} {
            ::Simple::Error::throw ::Simple::Script::DUPLICATED-TAG\
               trailer $tag $scriptName
         }

         ### Append the original trailer to the new list of trailers
         lappend trailers $originalTrailer

         ### Trailer with higher priority found
         if {$notFound && ($originalPriority > $priority)} {
            set notFound 0
            set insertIndex $index
         }

         ### Next original trailer
         incr index
      }

      ### Insert the trailer in its place
      set trailersString\
         [join [linsert $originalTrailers $insertIndex $trailer] \n]
      
      ### Rebuild the script
      set script $beforeTrailers$trailersString

   ### No original trailers present
   } else {

      ### Append the trailer to the script
      set script $script\n$trailer
   }

   ### Return the empty script
   return {}
}

### ===========================================================================
### -command  : ::Simple::Script::append
# -synopsis   : Appends a non-empty script to another one.
# -arguments  :
#  { scriptName       name           {Script name}}
#  { scriptToAppend   script         {Script to append}}
#
# -returns    : The empty string.
# -effects    :
#
#  * Appends the K<scriptToAppend> to the script called K<scriptName> in the
#    calling scope.
#
proc ::Simple::Script::append {
   scriptName
   scriptToAppend
} {
   ### Assert the script exists
   upvar $scriptName script
   if {![info exists script]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING script $scriptName
   }

   ### Non-empty script to append
   if {[string compare $scriptToAppend {}]} {

      ### Append the script

      # Non-empty script
      if {[string compare $script {}]} {
         ::append script \n$scriptToAppend

      # Empty script
      } else {
         set script $scriptToAppend
      }
   }

   ### Return the empty script
   return {}
}

### ===========================================================================
### -command  : ::Simple::Proc::Priv::info-body
# -synopsis   : Returns an existing procedure or method body.
# -arguments  :
#  { namespaceOrClass namespace      {Namespace or class}}
#  { unqualifiedItem  unqualifiedname
#                                    {Unqualified item name}}
#
# -return     : The procedure or method body.
#
proc ::Simple::Proc::Priv::info-body {
   namespaceOrClass
   unqualifiedItem
} {
   ### Try as a procedure
   if {[catch {set answer [info body $namespaceOrClass$unqualifiedItem]}]} {

      ### Try as a method
      set answer [namespace eval $namespaceOrClass\
         {info function [uplevel 1 ::set unqualifiedItem] -body}]
   }

   ### Return the procedure or method body
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Script::Header::get
# -synopsis   : Gets headers from a script.
# -overview   :
#
#     This procedure returns the list of headers from a script added by
#  K<::Simple::Script::Header::add> matching the given pattern.
#
# -arguments  :
#  { script           script         {Script}}
#  {?pattern?         pattern {}     {Pattern}}
#
# -returns    : 
#
#     The list of script headers matching the given pattern in the format
#  ||{tag priority header}||.
#
# -remarks    :
#
#  * Use no pattern to get all headers.
#
proc ::Simple::Script::Header::get {
   script
   {pattern {}}
} {
   ### Headers present
   set answer [list]
   if {[regexp -- "^($::Simple::Script::MarkerStart\
      %HEADER-BEGIN%.*\n$::Simple::Script::MarkerStart\
      %HEADER-END%\[^\n\]*)" $script -> fullHeaders]} {

      set fullHeaders\
         [::Simple::Script::Priv::split-headers-or-trailers $fullHeaders]
      # A<fullHeaders> contains the list of full headers

      ### Loop over full headers
      set RE "^$::Simple::Script::MarkerStart %HEADER-BEGIN%\
         +(\[0-9\]+) (\[^\n\]+)\n(.*)\n$::Simple::Script::MarkerStart\
         %HEADER-END%"
      foreach fullHeader $fullHeaders {

         regexp -- $RE $fullHeader -> priority tag header
         # A<priority> contains the priority of this header
         # A<tag> contains the tag of this header

         ### Pattern given and the tag matches it
         if {![string compare $pattern {}] || [string match $pattern $tag]} {

            ### Add this header to the list of matching headers
            lappend answer [list $tag $priority $header]
         }
      }
   }

   ### Return the list of matching headers
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Argument::information type
# -synopsis   : Returns an extended procedure argument type.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#  { argument         unqualifiedname
#                                    {Argument name}}
#
# -returns    : The extended procedure argument type.
#
proc-sub {::Simple::Argument::information type} {
   procedure
   argument
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Locate the argument
      set position [lsearch -exact [set\
         ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] $argument]

      ### Non-existing argument
      if {$position == -1} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-ARGUMENT\
            [::Simple::Priv::restore-name] $argument
      }

      ### Return the argument type
      return [lindex [lindex\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         $position] 4]

   ### Non-extended procedure
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         [::Simple::Priv::restore-name] {an extended procedure}
   }
}

### ===========================================================================
### -command  : ::Simple::Proc::information checktype
# -synopsis   : Returns whether a procedure performs argument type checking.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : Whether the procedure performs argument type checking.
# -remarks    :
#
#  * This procedure returns false for non-existing or non-extended procedures.
#
#  * This procedure returns true for extended procedures with no arguments.
#
proc-sub {::Simple::Proc::information checktype} {
   procedure
} {
   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Extended procedure with no arguments
      if {[llength\
         [set ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)]] == 0} {

         ### Return true
         set answer 1

      ### Extended procedure with arguments
      } else {

         ### Return whether the procedure performs argument run-time type checking
         set answer [expr {[llength [::Simple::Script::Header::get\
            [::Simple::Proc::Priv::info-body $namespace $unqualifiedProcedure]\
            %PRIV%-parse-arguments-checktype]] > 0}]
      }

   ### Non existing or non-extended procedure
   } else {

      ### Return false
      set answer 0
   }

   ### Return whether the procedure performs argument run-time type checking
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Argument::check-type
# -synopsis   : Perform runtime type checking on named arguments.
# -overview   :
#
#     Performs runtime type checking of specific arguments for the calling
#  procedure when runtime type checking is turned off. If runtime type 
#  checking is turned on for the calling procedure, this procedure does 
#  nothing.
#
#     This procedure is useful to force run-time type checking for a 
#  subset of arguments even when run-time type checking is explicitly
#  turned off.
#
# -arguments  :
#  { args             list           {List of argument names}}
#
# -keywords   : argument type
# -returns    : Whether runtime type checking was performed
# -examples   :
#
#  # Create a procedure with runtime type checking turned off
#  # but that explicitly checks the type of A<arg2>
#  proc-ext -checktype false name {
#     { arg1             integer        {First argument}}
#     { arg2             choice {A B}   {Second argument}}
#  } {
#    ::Simple::Argument::check-type arg2
#  }
#  
#  # Invoke the command with non-conforming arguments. Notice that the
#  # error refers to A<arg2> because A<arg1> is not being type checked.
#  # This throws following error:
#  #   |invalid value "C" for argument "arg2" of type "choice": must be A or B
#  name foo C
#
# -remarks    :
#
#  * This procedure returns ''1'' even if there are no arguments to 
#    be checked.
#
proc ::Simple::Argument::check-type {
   args
} {
   ### Get the caller qualified name
   set caller [::Simple::Namespace::qualify [lindex [info level -1] 0]]

   ### Caller has already performed run time type checking
   if {[::Simple::Proc::information checktype $caller]} {

      ### Return false
      set answer 0

   ### Caller has not performed run time type checking
   } else {

      ### Return true
      set answer 1

      ### Loop over given arguments
      foreach arg $args {

         ### Get argument type, choices and value
         ::Simple::Priv::save-name $caller
         set type [::Simple::Argument::information type $caller $arg]
         if {[string match choice* $type]} {
            set choices [::Simple::Argument::information choices $caller $arg]
         } else {
            set choices {}
         }

         # Convert the flag to an unqualified name
         set argNoFlag [string trimleft $arg -]
         set value [uplevel 1 [list ::set $argNoFlag]]

         ### Assert the argument is of the proper type
         ::Simple::Variable::Priv::assert-type-conformation \
             argument 0 $arg $type $value 0 $choices
      }
   }

   ### Return whether runtime type checking was performed
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Argument::information choices
# -synopsis   : Returns an extended procedure argument choices list.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#  { argument         unqualifiedname
#                                    {Argument name}}
#
# -returns    : The extended procedure argument choices list.
#
proc-sub {::Simple::Argument::information choices} {
   procedure
   argument
} {
   ### Save the procedure name
   ::Simple::Priv::save-name $procedure

   ### Parse, qualify and split the procedure name
   foreach {tmp namespace unqualifiedProcedure}\
      [::Simple::Proc::Priv::parse-qualify-split $procedure] break

   ### Extended procedure
   if {[info exists\
      ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]} {

      ### Locate the argument
      set position [lsearch -exact [set\
         ${namespace}%ProcData%(%ArgNames%,$unqualifiedProcedure)] $argument]

      ### Non-existing argument
      if {$position == -1} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-ARGUMENT\
            [::Simple::Priv::restore-name] $argument
      }

      ### Get the argument type
      set type [lindex [lindex\
         [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
         $position] 4]

      ### The argument type is not "choice" nor derived from it
      if {![string match choice* $type]} {

         ### Throw error
         ::Simple::Error::throw ::Simple::Proc::NO-TYPE-CHOICE\
            $argument [::Simple::Priv::restore-name]

      ### The argument type is "choice" or "choice-list"
      } else {

         ### Return the extended procedure argument choices
         return [lindex [lindex\
            [set ${namespace}%ProcData%(%ArgInternal%,$unqualifiedProcedure)]\
            $position] 5]
      }

   ### Non-extended procedure
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::Proc::NON-EXISTING-PROCEDURE\
         [::Simple::Priv::restore-name] {an extended procedure}
   }
}

### ===========================================================================
### -command  : ::Simple::Proc::configure
# -synopsis   : Configures the package options.
# -arguments  :
#
#  {-checktype        extbool false  {Whether to perform run-time type checking
#                                     of the procedures arguments}}
#  {-afterflags       integer 5      {Minimum number of arguments after the
#                                     last flag for a procedure in order for
#                                     those to be handled with an auxiliary
#                                     "after flags" procedure}}
#
# -returns    : The empty string.
#
proc ::Simple::Proc::configure {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::AT-LEAST-ONE-FLAG\
         [::Simple::List::enumerate [::Simple::Proc::cget]]

   ### Arguments
   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      set arg2 [lindex $args 1]
      switch -exact -- $arg1 {

         -checktype {
         ### -checktype given

            ### Invalid argument
            if {![::Simple::Type::Priv::is-basic extbool $arg2]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -checktype extbool {}
            }

            ### Set the check type option
            set ::Simple::Proc::%Options%(-checktype)\
               [{::Simple::Type::convert boolean} $arg2]
         }

         -afterflags {
         ### -afterflags given

            ### Invalid argument
            if {![::Simple::Type::Priv::is-basic integer $arg2]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -afterflags integer {}
            }
            if {$arg2 <= 1} {
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -afterflags integer\
                  {: must be greater than one}
            }

            ### Set the after flags option
            set ::Simple::Proc::%Options%(-afterflags) $arg2
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1\
               [::Simple::List::enumerate [::Simple::Proc::cget]]
         }
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Proc::cget
# -synopsis   : Gets the package options.
# -arguments  :
#  {-checktype        boolflag       {Gets whether to perform run-time type
#                                     checking of the procedures arguments}}
#  {-afterflags       boolflag       {Gets the minimum number of arguments
#                                     after the last flag for a procedure in
#                                     order for those to be handled with an
#                                     auxiliary "after flags" procedure}}
#
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Proc::cget {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Return the list of options
      return {-checktype -afterflags}

   ### Arguments
   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      switch -exact -- $arg1 {

         -afterflags {
         ### -afterflags given
            set ::Simple::Proc::%Options%(-afterflags)
         }

         -checktype {
         ### -checktype given
            set ::Simple::Proc::%Options%(-checktype)
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1\
               [::Simple::List::enumerate [::Simple::Proc::cget]]
         }
      }
   }
}
}
