### ===========================================================================
### -package  : SimpleProgram
#
package require SimpleDevLibPackage 1.0
declare-package {

   SimpleProgram

} -version {

   1.0

} -project {

   The Simple Development Library.

} -synopsis {

   The Simple Development Library program declaration.

} -module {

   implementation

} -overview {

      This package provides a way to declare programs in an homogeneous,
   easily parseable, fully introspectable way.  It allows to substitute the
   comment block which is usually placed at the beginning of each program file
   by Tcl-handled text. 

      Programs are declared with a single command (K<declare-program>).  Every
   single component needed to either make the program work (such as the
   program options) or to document it (metadata, such as the program synopsis)
   is provided as arguments to that command.  Components can be fully
   introspected, but metadata is only available if The Simple Development
   Library A<-storemetadata> option is set when the program is declared.

      Procedures are provided to declare a program, get information about the
   program, parse the program command line and exit the program execution.
 
} -usage {

      Use K<declare-program> to declare a program.  The declared program
   metadata given to K<declare-program> (or composed from other data, if not
   given, as for A<--help>) can be obtained via the
   K<::Simple::Program::information metadata> procedure.  Other subcommands of
   this procedure provide information about a declared program:  K<name>,
   K<version>, K<requisites>, K<synopsis>, K<help>, K<examples>, K<options>,
   K<errors>, K<variables>, K<types> and K<commands>.

      Use K<::Simple::Program::parse-command-line> to parse the program
   command line arguments and K<::Simple::Program::die> to display an error
   and exit.

      There is a backdoor implemented which allows to display the error stack
   trace held in V<errorInfo> upon an uncatched error in the program
   execution:  give the A<--errorinfo> option as a command line argument.
 
} -keywords {

   program declaration {command line}

} -examples {

   # Install the package
   package require SimpleDevLibProgram 1.0

   # For this example we substitute the ||exit|| command by ||return|| so that
   # we can provide some hard-coded command lines below
   interp alias {} exit {} return

   # Declare the program
   declare-program {
      foo
   } -version {
      3.1
   } -synopsis {

      Example program

   } -option {
      {-p --processing} -type choice\
         -choices {A B C} -description {processing option}
   } -option {
      {-v --verbosity} -name verbosity -type integer\
         -range -1:4 -description {set the verbosity level}
   } -option {
      {-w --width} -name width -type integer\
         -default 80 -description {set the width of the output}
   } -option {
      {-a --all} -type boolflag -description {process all files}
   } -option {
      -- -name files -type string -values 3: -description {files to process}
   } -optionsusage {
      -p {O -v} {O -w} {X -a --}
   } -examples {

      * To process all files in mode A:
         foo -p A --all

      * To process files "bar" and "gee" in mode B in a 132 columns display:
         foo -p B -w 132 bar gee

      * To process files "bar" and "gee" in verbose mode C:
         foo -p C -v 4 bar

   } -requisites {
      {Tcl 8}
   } -command {
      {
         program-command

      } -synopsis {

         Example of program command

      } -body {
         puts {Program command}
      }

   } -body {

      # Parse the command line here using the following line:
      # ::Simple::Program::parse-command-line $argv

      # Here use ||::Simple::Option::information available||,
      # ||::Simple::Option::information given|| and
      # ||::Simple::Option::information value|| to access the command line
      # arguments
   }

   # This displays the following:
   #   |Usage: foo -p choice [-v verbosity] [-w width] (-a | files(3:)) 
   #   |       foo --help 
   #   |       foo --version
   #   |       foo --examples
   #   |
   #   |Example program
   #   |
   #   |   -p, --processing           processing option (A, B or C)
   #   |   -v, --verbosity VERBOSITY  set the verbosity level
   #   |   -w, --width WIDTH          set the width of the output (default 80)
   #   |   -a, --all                  process all files
   #   |   files                      files to process
   #   |       --help                 display this help and exit
   #   |       --version              output version information and exit
   #   |       --examples             display some usage examples and exit
   ::Simple::Program::parse-command-line --help

   # This displays the following:
   #   |foo 3.1
   ::Simple::Program::parse-command-line --version

   # This displays the following:
   #   |Example program
   #   |
   #   |      * To process all files in mode A:
   #   |         foo -p A --all
   #   |
   #   |      * To process files "bar" and "gee" in mode B in a 132 columns display:
   #   |         foo -p B -w 132 bar gee
   #   |
   #   |      * To process files "bar" and "gee" in verbose mode C:
   #   |         foo -p C -v 4 bar
   ::Simple::Program::parse-command-line --examples

   # This displays the following:
   #   |foo: "-p" (or "--processing") required
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-w 132 -a}

   # This displays the following:
   #   |foo: invalid value "D" for option "-p" of type "choice": must be
   #   |   A, B or C
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-p D -a}

   # This displays the following:
   #   |foo: invalid value "foo" for option "-w" of type "integer"
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-p A -w foo -a}

   # This displays the following:
   #   |foo: one of "-a" (or "--all") or "files" required
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-p A}

   # This displays the following:
   #   |foo: too few values, 3: files required, got 2
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-p A file1 file2}

   # This displays the following:
   #   |foo: too few values, 3: files required, got 2
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-p A file1 file2}

   # This displays the following:
   #   |foo: invalid value "7" for option "-v": value out of range "-1:4"
   #   |Try "foo --help" for more information.
   ::Simple::Program::parse-command-line {-p A -v 7 -a}

   # A valid command line
   ::Simple::Program::parse-command-line {-p A -v 3 -a}

} -api {

   * K<declare-program> program ?-version version? ?-project string?
        ?-synopsis string? ?-overview string? ?-usage string?
        ?-exitstatus string? ?-help string? ?-keywords list?
        ?-examples string? ?-error declaration? ?-type declaration?
        ?-variable declaration? ?-command declaration? ?-option declaration? 
        ?-optionsusage list? ?-requisites packagerequisite-list? 
        ?-assumptions string? ?-resources string? ?-limitations string? 
        ?-details string? ?-references string? ?-remarks string? 
        ?-todo string? ?-cmid string? ?-sprs string? ?-smrs string? 
        ?-history string? ?-copyright string? ?-body script?
     Declares a program.

   * K<::Simple::Program::parse-command-line> commandLine
     Parses a program command line.

   * K<::Simple::Program::die> ?-trymessage? message ?status?
     Displays an error message and exits unssuccessfully.

   * K<::Simple::Program::configure>
     Configures the package options.

   * K<::Simple::Program::cget>
     Gets the package options.

} -requisites {

   {SimpleOption 1.0}

} -namespaces {

   ::Simple::Program
   ::Simple::Program::Priv

} -aliases {

   {declare-program ::Simple::Program::declare}

} -history {

   23-apr-2000   First public release, version 0.2
   10-sep-2001   Second public release, version 0.4
   17-feb-2003   Third public release, version 0.5
   05-apr-2003   Added A<-body> to K<declare-program>, version 0.5.1
   27-apr-2003   Added A<-command> to K<declare-program> and added\
                 K<::Simple::Program::information metadata>, version 0.5.2
   01-may-2003   Extra package, version 0.5.3
   20-jan-2004   Comments allowed before embedded declarations in\
                 K<declare-program> version 0.5.4
   22-jun-2005   The Simple Development Library version 1.0

} -copyright {

   Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)

} -variable {

   {
      ::Simple::Program::Information

   } -description {

      Array containing for each program information item (name, version,
      synopsis, help, examples and requisites) its value

   } -type any-array 

} -variable {

   {
      ::Simple::Program::Metadata

   } -description {

      Program metadata

   } -type list -value {}

} -variable {

   {
      ::Simple::Program::Types

   } -description {

      Program types list

   } -type type-list

} -variable {

   {
      ::Simple::Program::Commands

   } -description {

      Program commands list

   } -type unqualifiedname-list

} -variable {

   {
      ::Simple::Program::ErrorInfo

   } -description {

      Whether to display the error stack trace held in V<errorInfo> upon an
      uncatched error in the program execution.

   } -type boolean -value 0

} -command {

### ===========================================================================
### -command  : ::Simple::Program::declare

{
   ::Simple::Program::declare

} -synopsis {

   Declares a program.

} -alias {

   declare-program

} -overview {

      This procedure provides a way to declare a program.  This procedure can
   be invoked just once.
      
   The A<-synopsis> flag is mandatory.  
      
      The A<-option> flag allows to declare the program command line options. 
   Its argument is an option declaration intended to be the arguments of a
   call to the K<::Simple::Option::declare> procedure.  This flag can be given
   multiple times.  The command line options actually accepted by the program
   include those plus A<--help>, A<--version> and A<--examples> (as well as
   the backdoor option A<--errorinfo>); see the
   K<::Simple::Program::parse-command-line> procedure.
   
      If A<-optionsusage> is given, the options usage is declared via the
   K<::Simple::Option::declare-usage> procedure. 
   
      If A<-requisites> is given, the program requisites are required and
   installed via the K<::Simple::Package::require-and-install> procedure.

      If A<-help> is given, its argument is the help message displayed when
   invoking the program with the A<--help> option, after substituting the
   %USAGE%, %SYNOPSIS% and %OPTIONS% strings by the program usage, the program
   synopsis or the detailed options explanation, respectively.  Otherwise the
   help message is composed from the synopsis and options information.

      The A<-command> flag allows to declare program commands.  Its argument
   is a command declaration intended to be the arguments of a call to
   K<::Simple::Declare::declare-proc>.  This flag can be given multiple times. 
   Program commands are intended as helper procedures for the program.  In
   principle, program commands should live in the global namespace, but this
   is currently not enforced.

      The A<-error> flag allows to declare program errors.  Its argument is an
   error declaration intended to be the arguments of a call to
   K<::Simple::Error::declare>.  This flag can be given multiple times.

      The A<-type> flag allows to declare program types.  Its argument is a
   type declaration intended to be the arguments of a call to
   K<::Simple::Type::declare>.  This flag can be given multiple times.

      The A<-variable> flag allows to create program variables.  Its argument
   is a variable declaration intended to be the arguments of a call to
   K<::Simple::Variable::declare>.  This flag can be given multiple times.

      It is not allowed to declare program classes.  Place the classes in a
   package and require that package instead.

      The mandatory A<-body> flag specifies the program body script.  This
   script is evaluated in the global scope.  If there is an error in the
   evaluation, K<::Simple::Program::die> is invoked with the error message as
   argument.  Otherwise, the program exits with exit status equal to zero.  If
   other exit status and/or other behaviour is needed, use either K<exit> or
   K<::Simple::Program::die> in the program body script.  Notice that
   A<-exitstatus> is just a documentation flag.
 
      All other arguments are metadata, and are stored or ignored depending on
   whether The Simple Development Library A<-storemetadata> option is set when
   the program is declared, but for the A<-body> flag which is never stored.
 
} -arguments {
   { program          unqualifiedname
                                     {Program name}}
   {-version          version 1.0    {Program version}}
   {-project          string {}      {Program project}}
   {-synopsis         string {}
                                     {Program synopsis}}
   {-overview         string {}      {Program overview}}
   {-usage            string {}      {Program usage message}}
   {-exitstatus       string {}      {Program exit status}}
   {-help             string %DEFAULT%
                                     {Program command line help message}}

   {-keywords         list {}        {Program keywords}}
   {-examples         string {}      {Program usage examples}}
   {-error            declaration {} {Program error declaration}}
   {-type             declaration {} {Program type declaration}}
   {-variable         declaration {} {Program variable declaration}}
   {-command          declaration {} {Program command declaration}}
   {-option           declaration {} {Program command line option declaration}}
   {-optionsusage     list {}        {Program command line options usage}}
   {-requisites       packagerequisite-list {}
                                     {Program requisites}}
   {-assumptions      string {}      {Program assumptions}}
   {-resources        string {}      {Program required resources}}
   {-limitations      string {}      {Program limitations}}
   {-details          string {}      {Program implementation details}}
   {-references       string {}      {Program references}}
   {-remarks          string {}      {Program remarks}}
   {-todo             string {}      {Program to do list}}
   {-cmid             string {}      {Program configuration management
                                      identification}}
   {-sprs             string {}      {Program software problem reports}}
   {-smrs             string {}      {Program software modification reports}}
   {-history          string {}      {Program history}}
   {-copyright        string {}      {Program copyright notice}}
   {-body             script {}      {Program body}}

} -parseflags {

   false

} -returns {

   The empty string.

} -effects {

   * Requires the required packages.

   * Those of the program body.

} -remarks {

   * The order of the flags is irrelevant but their parameters can not start
     by an hyphen.

} -body {

   ### Parse and verify the program declaration
   set program [string trim $program]
   set ::Simple::Program::Information(name) $program
   ::Simple::Error::Priv::push-location program $program 1
   if {[catch {

      ### Set defaults
      set version 1.0
      set help %DEFAULT%

      ### Parse flags
      ::Simple::Priv::parse-flags\
         [list -version trimmed\
         -project trimmed\
         -synopsis line\
         -overview paragraphs\
         -usage paragraphs\
         -exitstatus paragraphs\
         -help lines\
         -keywords list\
         -examples lines\
         -optionsusage untouched\
         -requisites list\
         -assumptions paragraphs\
         -resources paragraphs\
         -limitations paragraphs\
         -details paragraphs\
         -references paragraphs\
         -remarks paragraphs\
         -todo paragraphs\
         -cmid paragraphs\
         -sprs paragraphs\
         -smrs paragraphs\
         -history paragraphs\
         -copyright paragraphs\
         -body untouched]\
         [list -error untouched\
         -type untouched\
         -variable untouched\
         -command untouched\
         -option untouched] $args 0 1

      ### Handle A<-synopsis>
      ::Simple::Declare::Priv::assert-synopsis-given

      ### Handle A<-body>
      if {![::Simple::Argument::information flaggiven -body]} {
         ::Simple::Error::throw ::Simple::FLAG-REQUIRED\
            -body program $program
      }

      ### Handle A<-error>
      foreach errorDeclaration $error {

         ### Assert the error code is unqualified
         ::Simple::Declare::Priv::remove-leading-comments errorDeclaration
         set errorCode [string trim [lindex $errorDeclaration 0]]
         ::Simple::Type::assert-is\
            unqualifiedname $errorCode {} {unqualified error code}

         ### Parse and verify the error declaration
         ::Simple::Error::Priv::push-location error $errorCode
         set errorDeclarationScript\
            [::Simple::Error::Priv::parse-and-verify-declaration\
            ::$errorCode [lrange $errorDeclaration 1 end] 0]

         ### Declare the error
         eval $errorDeclarationScript
         ::Simple::Error::Priv::pop-location
      }

      ### Handle A<-type>
      # Types need to be created before other items in case those items
      # use the new types.

      ### Loop over types
      foreach typeDeclaration $type {

         ### Parse and verify the type declaration
         ::Simple::Declare::Priv::remove-leading-comments typeDeclaration
         set typeName [string trim [lindex $typeDeclaration 0]]
         ::Simple::Error::Priv::push-location type $typeName
         set typeDeclarationScript\
            [::Simple::Type::Priv::parse-and-verify-declaration\
            $typeName [lrange $typeDeclaration 1 end] 0]

         ### Declare the type
         eval $typeDeclarationScript
         ::Simple::Error::Priv::pop-location

         ### Add the type to the program types list
         lappend ::Simple::Program::Types $typeName
      }

      ### Handle A<-variable>

      ### Loop over variables
      foreach variableDeclaration $variable {

         ### Assert the variable name is unqualified
         ::Simple::Declare::Priv::remove-leading-comments variableDeclaration
         set variableName [string trim [lindex $variableDeclaration 0]]
         ::Simple::Type::assert-is\
            unqualifiedname $variableName {} {unqualified variable name}

         ### Parse and verify the variable declaration
         ::Simple::Error::Priv::push-location variable $variableName
         foreach {variableDeclarationScript variableCreationScript}\
            [::Simple::Variable::Priv::parse-and-verify-declaration\
            ::$variableName [lrange $variableDeclaration 1 end] 0] break

         ### Declare and create the variable
         eval $variableDeclarationScript\n$variableCreationScript
         ::Simple::Error::Priv::pop-location
      }

      ### Handle A<-command>

      ### Loop over commands
      foreach commandDeclaration $command {

         ### Assert the command name is unqualified
         ::Simple::Declare::Priv::remove-leading-comments commandDeclaration
         set commandName [string trim [lindex $commandDeclaration 0]]
         ::Simple::Type::assert-is unqualifiedname\
            [lindex $commandName 0] {} {unqualified command name}

         ### Parse and verify the command declaration
         #   The command is asserted not the be exported nor aliased
         #   as it lives in the global namespace
         ::Simple::Error::Priv::push-location command $commandName
         foreach {isExported aliasPair\
            commandDeclarationAndCreationScript commandInstallationScript}\
            [::Simple::Declare::Priv::parse-and-verify-procedure-declaration\
            $commandName [lrange $commandDeclaration 1 end] 1] break
         ::Simple::Error::Priv::pop-location

         ### Declare, create and install the procedure
         eval $commandDeclarationAndCreationScript\n$commandInstallationScript

         ### Add the command to the program command list
         lappend ::Simple::Program::Commands $commandName
      }

      ### Handle A<-option>
      foreach optionDeclaration $option {

         # Declare the program command line options
         ::Simple::Error::Priv::push-location option\
            [lindex $optionDeclaration 0]
         namespace eval :: ::Simple::Option::declare $optionDeclaration
         ::Simple::Error::Priv::pop-location
      }

      ### Handle A<-optionsusage>
      if {[::Simple::Argument::information flaggiven -optionsusage]} {

         # Specify the program command line options usage
         ::Simple::Option::declare-usage $optionsusage
      }

      ### Require and install the required packages
      foreach requisite $requisites {
         namespace eval :: ::Simple::Package::require-and-install $requisite
      }

   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Store the program information items
   foreach flag [list version synopsis help examples requisites] {
      set ::Simple::Program::Information($flag) [set $flag]
   }

   ### The program metadata must be stored
   if {[set ::Simple::%Options%(-storemetadata)]} {
      set ::Simple::Program::Metadata [list $project $overview $usage\
         $exitstatus $keywords $assumptions $resources $limitations $details\
         $references $remarks $todo $cmid $sprs $smrs $history $copyright]
   }

   ### Evaluate the program body
   set return [catch [list uplevel #0 $body] result]
   if {$return != 0} {
      ::Simple::Program::die $result
   } else {
      exit 0
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::parse-command-line

{
   ::Simple::Program::parse-command-line

} -synopsis {

   Parses a program command line.

} -overview {

      This procedure is similar to the K<::Simple::Option::parse> procedure
   in that it parses the program command line and assesses its compliance
   against the program options and their usage. If an error is found when
   parsing the command line, a suitable diagnostics is displayed in the
   standard error channel together with a message sugesting the use of the
   A<--help> option, and the program exits.

      On top of the program options given as parameter to the A<-option> flag
   in the call to K<declare-program>, this procedure also accepts the
   A<--help>, A<--version>, A<--examples> and A<--errorinfo> command line
   options:

   A<--help>::
   Displays the help message and exits.

   A<--version>::
   Displays the program version and exits.

   A<--examples>::
   Displays the program examples and exits.

   A<--errorinfo>::
   Displays the error stack trace held in V<errorInfo> upon an uncatched
   error and exists.

} -arguments {
   { commandLine      list           {Program command line arguments}}
} -effects {

   * May finish the program execution.

} -references {

   * The K<::Simple::Option::parse> procedure.

} -returns {

   The empty string.

} -body {

   ### Handle A<--help>, A<--version> and A<--examples>
   switch -exact -- [lindex $commandLine 0] {

      --help {
      ### Help message

         ### Ensure the help message is up to date
         ::Simple::Program::Priv::help-message

         ### Display the help message
         puts stderr $::Simple::Program::Information(help)

         ### Exit successfully
         exit 0
      }

      --version {
      ### Version message
         puts stderr "$::Simple::Program::Information(name)\
            $::Simple::Program::Information(version)"

         ### Exit successfully
         exit 0
      }

      --examples {
      ### Examples message
         puts stderr $::Simple::Program::Information(synopsis)\n
         if {[string compare $::Simple::Program::Information(examples) {}]} {
            puts stderr $::Simple::Program::Information(examples)
         } else {
            puts stderr {No examples available}
         }

         ### Exit successfully
         exit 0
      }

      --errorinfo {
      ### Backdoor: display the error information upon any uncatched error

         ### Set the provide error info on error flag
         set ::Simple::Program::ErrorInfo 1

         ### Eat the option
         set commandLine [lrange $commandLine 1 end]
      }
   }

   ### Parse the program command line with respect to the declared options
   if {[catch {::Simple::Option::parse $commandLine} result]} {
      ::Simple::Program::die -trymessage $result
   }

   ### Return the empty string
   return {}
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::die

{
   ::Simple::Program::die

} -synopsis {

   Displays an error message and exits unssuccessfully.

} -arguments {

   {-trymessage       boolflag       {Whether to display the try message}}
   { message          string         {Error message}}
   {?status?          integer 1      {Exit status}}

} -returns {

   None; the program is finished.

} -overview {

      This procedure displays an error message in the standard error channel
   and exists the program.  If the A<-trymessage> flag is given, the try
   message ("Try <program name --help> for more information") is also
   displayed.
   
} -effects {

   * Finishes the program execution.

} -body {

   ### Display error message
   puts stderr "$::Simple::Program::Information(name): $message"

   ### Must display the error information
   if {$::Simple::Program::ErrorInfo} {

      ### Backdoor: display the error information
      set callStack [lindex [::Simple::Error::Priv::split-errorinfo] 1]
      puts stderr $callStack
   }

   ### Must display the try message
   if {$trymessage} {

      ### Display the rey message
      puts stderr [format {Try "%s --help" for more information.}\
         $::Simple::Program::Information(name)]
   }

   ### Exit unsuccessfully
   exit $status
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::configure

{
   ::Simple::Program::configure

} -synopsis {

   Configures the package options.

} -returns {

   The "package with no options" error is thrown.

} -body {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}
} -command {

### ===========================================================================
### -command  : ::Simple::Program::cget

{
   ::Simple::Program::cget

} -synopsis {

   Gets the package options.

} -returns  {

   The requested option value or the whole list of options if none specified.

} -body {

   return [list]
}
}
