### ===========================================================================
### -package  : SimpleSubcommand-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library subcommand procedures handling.
# -module     : extra
# -api        :
#
#  * K<::Simple::Subcommand::delete> procedure
#    Deletes a subcommand procedure.
#
#  * K<::Simple::Subcommand::move> sourceProcedure targetProcedure
#    Renames a subcommand procedure.
#
#  * K<::Simple::Subcommand::information exists> command
#    Returns whether a procedure is a base command procedure.
#
#  * K<::Simple::Subcommand::information subcommands> command
#    Returns a base command procedure subcommand list.
#
#  * K<::Simple::Subcommand::configure>
#    Configures the package options.
#
#  * K<::Simple::Subcommand::cget>
#    Gets the package options.
#
# -history    :
#
#  24-oct-2004   Bug 8 correction, version 0.5.2
#  24-oct-2004   Copy functionality deleted, version 0.5.3
#  10-feb-2005   Bug 9 correction, version 0.5.4
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleSubcommand-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Subcommand::delete
# -synopsis   : Deletes a subcommand procedure.
# -overview   :
#
#     This command is similar to the K<rename> command when its second
#  argument is empty, but works for subcommand procedures only.
#
# -arguments  :
#  { procedure        commandname    {Base command and subcommand pair}}
#
# -returns    : The empty string.
# -details    :
#
#  *  The subcommand procedure is deleted and, should it be the last one, the
#     base command is deleted as well.
#
# -effects    :
#
#  * Deletes the procedure named K<procedure>.
#
proc ::Simple::Subcommand::delete {
   procedure
} {
   ### Save the command name
   ::Simple::Priv::save-name [lindex $procedure 0]

   ### Parse and qualify the procedure name
   foreach {qualifiedProcedure command subcommand}\
      [::Simple::Subcommand::Priv::parse-qualify $procedure] break

   ### Delete the subcommand procedure
   if {[catch {rename $qualifiedProcedure {}}]} {
      ::Simple::Error::throw ::Simple::Subcommand::CANT-DELETE-OR-RENAME\
         delete [list [::Simple::Priv::restore-name] $subcommand]
   }

   ### No subcommands remaining for the base command
   if {[catch {::Simple::Subcommand::Priv::subcommands $command}]} {

      ### Delete the base command
      rename $command {}
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Subcommand::Priv::subcommands
# -synopsis   : Returns a base command procedure subcommand list.
# -arguments  :
#  { command          qualifiedname  {Qualified command name}}
#
# -returns    : The base command procedure subcommand list.
# -remarks    :
#
#  * Procedures containing a percent sign in its name are assumed to be
#    internal to The Simple Development Library and are excluded from the
#    output.
#
proc ::Simple::Subcommand::Priv::subcommands {
   command
} {
   ### Get the base command procedure list
   set subcommandProcs\
      [namespace eval [::Simple::Namespace::qualifiers $command]\
      [list info commands [namespace tail $command]\ *]]

   ### Tidy up the subcommand list
   set answer [list]
   foreach fullCommand $subcommandProcs {

      ### Get the subcommand name
      set subcommand [lindex $fullCommand 1]

      ### The subcommand name is not internal
      if {[::Simple::Priv::is-not-internal $subcommand]} {

         ### Add the subcommand to the subcommand list
         lappend answer $subcommand
      }
   }

   ### Empty subcommand list
   if {[llength $answer] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Subcommand::NON-EXISTING-COMMAND\
         [::Simple::Priv::restore-name]

   ### Non-empty subcommand list
   } else {

      ### Return the subcommand list
      return $answer
   }
}

### ===========================================================================
### -command  : ::Simple::Subcommand::move
# -synopsis   : Renames a subcommand procedure.
# -overview   :
#
#     This command is similar to the K<rename> command when its second
#  argument is non-empty, but works for subcommand procedures only.
#
# -arguments  :
#  { sourceProcedure  name           {Source base command and subcommand pair}}
#  { targetProcedure  name           {Target base command and subcommand pair}}
#
# -returns    : The empty string.
# -details    :
#
#  * The source subcommand procedure is deleted and the target is created.
#
# -remarks    :
#
#  * If the target procedure namespace does not exists, it is created.
#    This mimics the behaviour of the K<rename> command.
#
# -todo       :
#
#  * Allow to rename from subcommand to regular procedure and viceversa.
#
# -effects    :
#
#  * Renames the procedure named K<sourceProcedure> to K<targetProcedure>.
#
proc ::Simple::Subcommand::move {
   sourceProcedure
   targetProcedure
} {
   ### Save the source command name
   ::Simple::Priv::save-name [lindex $sourceProcedure 0]

   ### Parse and qualify the procedure name
   foreach {qualifiedOldProcedure sourceCommand sourceSubcommand}\
      [::Simple::Subcommand::Priv::parse-qualify $sourceProcedure] break

   ### Parse the source command name
   foreach {tmp namespace tail}\
      [::Simple::Proc::Priv::parse-qualify-split $qualifiedOldProcedure] break

   ### Command procedure does not exists
   if {![::Simple::Priv::proc-information-exists $namespace $tail]} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Subcommand::CANT-DELETE-OR-RENAME\
         rename [list [::Simple::Priv::restore-name] $sourceSubcommand]

   ### Command procedure exists
   } else {

      ### Get the subcommand procedure necessary information
      set arguments\
         [{::Simple::Priv::proc-information-arguments} $qualifiedOldProcedure]
      set body [info body $qualifiedOldProcedure]

      ### Delete the subcommand procedure
      rename $qualifiedOldProcedure {}

      ### Parse and qualify the target procedure name
      foreach {qualifiedNewProcedure targetCommand targetSubcommand}\
         [::Simple::Subcommand::Priv::parse-qualify $targetProcedure] break

      ### Create the target command namespace if it does not exist
      namespace eval [::Simple::Namespace::qualifiers $targetCommand] {}

      ### Create the target subcommand procedure
      ::Simple::Subcommand::create\
         $targetCommand\ $targetSubcommand $arguments $body

      ### No subcommands remaining for the source command
      if {[catch {::Simple::Subcommand::Priv::subcommands $sourceCommand}]} {

         ### Delete the base command
         rename $sourceCommand {}
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Subcommand::information exists
# -synopsis   : Returns whether a procedure is a base command procedure.
# -arguments  :
#  { command          name           {Base command name}}
#
# -returns    : Whether the procedure is a base command procedure.
# -references :
#
#  * The K<::Simple::Proc::information issubcommand> procedure.
#
# -details    :
#
#  * A command is considered a base command procedure if its body contains the
#     base command given by V<::Simple::BaseCommandTag>.
#
::Simple::Subcommand::create {::Simple::Subcommand::information exists} {
   command
} {
   ### Qualify the base command name in the calling scope
   set command [::Simple::Priv::qualify $command]

   ### The procedure does not exist
   if {[catch {set body [info body $command]}]} {

      ### Return false
      set answer 0

   ### The procedure exists
   } else {

      ### Return whether the procedure is a base command procedure
      set answer [expr {[regexp -- $::Simple::BaseCommandTag $body]}]
   }

   ### Return whether the procedure is a base command procedure.
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Subcommand::information subcommands
# -synopsis   : Returns a base command procedure subcommand list.
# -arguments  :
#  { command          name           {Base command name}}
#
# -returns    : The base command procedure subcommand list.
# -references :
#
#  * The K<::Simple::Proc::information subcommands> procedure.
#
# -remarks    :
#
#  * The returned list does not contain subcommands with a percent sign in its
#    name are those are assumed to be internal to The Simple Development
#    Library.
#
::Simple::Subcommand::create {::Simple::Subcommand::information subcommands} {
   command
} {
   ### Save the command name
   ::Simple::Priv::save-name $command

   ### Qualify the command name in the calling scope
   set command [::Simple::Priv::qualify $command]

   ### Return the subcommand list
   ::Simple::Subcommand::Priv::subcommands $command
}

### ===========================================================================
### -command  : ::Simple::Subcommand::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Subcommand::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Subcommand::cget
# -synopsis   : Gets the package options.
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Subcommand::cget {
} {
   return [list]
}

}
