### ===========================================================================
### -package  : SimpleSubcommand
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library subcommand procedures handling.
# -module     : implementation
# -overview   :
#
#     This package allows to handle subcommand procedures.  Subcommands is The
#  Simple Development Library term for the concept named ensemble commands,
#  command families or composite commands in other contexts.  Basically one
#  can define the K<foo bar> and K<foo gee> commands, which are
#  ''subcommands'' K<bar> and K<gee> for the K<foo> ''base command''; each
#  subcommand has its own arguments.
#
#     Procedures are provided to create, delete, copy, rename and get
#  information about subcommand procedures.
#
# -usage      :
#
#     Subcommand procedures are create via the K<proc-sub> command.  Such
#  subcommand procedures behave similarly to the built-in Tcl commands with
#  subcommands (such as K<string> or K<array>):
#
#  * They are called simply by providing the command and subcommand name, such
#    as K<foo bar>.
#
#  * When invoking a base command with no subcommand, an error is thrown
#    informing of the allowed subcommands.
#
#  * When called with a bad number of arguments for a particular subcommand,
#    an error is thrown informing of the argument list for that subcommand.
#
#     Subcommand procedures can be deleted via K<::Simple::Subcommand::delete>
#  or renamed via K<::Simple::Subcommand::move>.  The
#  K<::Simple::Subcommand::information> procedure provides several subcommands
#  to query information about subcommand procedures:  K<exists> and
#  K<subcommands>.
#
# -keywords   : procedure command subcommand family composite ensemble
# -api        :
#
#  * K<proc-sub> procedure arguments body
#    Creates a subcommand procedure.
#
#  * K<::Simple::Subcommand::configure>
#    Configures the package options.
#
#  * K<::Simple::Subcommand::cget>
#    Gets the package options.
#
# -variable   :
#
#  {
#     ::Simple::Priv::SavedName
#
#  } -description {
#
#     Name saved by K<::Simple::Priv::save-name>
#
#  } -type name
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#  ::Simple::Package::require-and-install SimpleSubcommand
#
#  # Create some subcommands
#  proc-sub {foo bar} arg1 {}
#  proc-sub {foo gee} arg1 {}
#
#  # Assess whether foo is a subcommand base command
#  # This displays the following:
#  #   |Yes, foo is a subcommand base command
#  if {[::Simple::Subcommand::information exists foo]} {
#     puts {Yes, foo is a subcommand base command}
#  } else {
#     puts {No, foo is not a subcommand base command}
#  }
#
#  # Display the list of subcommands for foo
#  # This displays the following:
#  #   |gee bar
#  puts [::Simple::Subcommand::information subcommands foo]
#
#  # Delete the "foo bar" subcommand
#  ::Simple::Subcommand::delete {foo bar}
#
#  # Display the list of subcommands for foo
#  # This displays the following:
#  #   |gee
#  puts [::Simple::Subcommand::information subcommands foo]
#
#  # Rename the "foo gee" subcommand to "foo bar"
#  ::Simple::Subcommand::move {foo gee} {foo bar}
#
#  # Display the list of subcommands for foo
#  # This displays the following:
#  #   |bar
#  puts [::Simple::Subcommand::information subcommands foo]
#
#  # Rename the "foo bar" subcommand to "zoo fii"
#  ::Simple::Subcommand::move {foo bar} {zoo fii}
#
#  # Assess whether foo is a subcommand base command
#  # This displays the following:
#  #   |No, foo is not a subcommand base command
#  if {[::Simple::Subcommand::information exists foo]} {
#     puts {Yes, foo is a subcommand base command}
#  } else {
#     puts {No, foo is not a subcommand base command}
#  }
#
# -details    :
#
#  * Subcommand procedures are implemented with the aid of a base procedure
#    named as the command.  That is, if a subcommand procedure named K<foo
#    bar> is created, two procedures are actually created:  the subcommand
#    procedure itself (named K<foo bar>) and the base command (named K<foo>). 
#    The base command simply acts as a dispatcher for the actual subcommand
#    procedures.  It also catches the subcommand procedure errors and throws
#    explanation errors when appropriate.
#
#  * For this reason, it is slightly faster to invoke a subcommand procedure
#    as K<{foo bar}> or K<foo\ bar> instead of K<foo bar>, as the dispatch
#    from the base command is avoided.
#
# -history    :
#
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  29-mar-2003   Extra package, version 0.5.1
#  24-oct-2004   Bug 8 correction, version 0.5.2
#  24-oct-2004   Copy functionality deleted, version 0.5.3
#  10-feb-2005   Bug 9 correction, version 0.5.4
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleSubcommand 1.0 -requisites {

   {SimpleNamespace 1.0}

} -namespaces {

   ::Simple::Subcommand
   ::Simple::Subcommand::Priv

} -aliases {

   {proc-sub ::Simple::Subcommand::create}

} -install {

### Create the package variables
set ::Simple::Priv::SavedName {}

### Create the package options

} -uninstall {

### Delete the package variables
unset ::Simple::Priv::SavedName

### Delete the package options

} -firsttimeinstall {

### Declare the package types

### Declare the package errors
::Simple::Error::declare {
   ::Simple::Subcommand::CANT-CREATE 
} -message {
   can't create subcommand "%s %s": procedure "%s" already exists
} -explanation {
   The subcommand procedure "<command> <subcommand>" could not be created
   because a procedure named as the subcommand base command (<command>)
   already exists
} -corrective {
   Either delete that procedure or choose another name for the
   subcommand base command
}

::Simple::Error::declare {
   ::Simple::Subcommand::CANT-DELETE-OR-RENAME 
} -message {
   can't %s "%s": command doesn't exist
} -explanation {
   The procedure "<procedure>" could
   not be <delete or rename>d because it does not exist
} -translation {<delete or rename> <procedure>}

::Simple::Error::declare {
   ::Simple::Subcommand::NON-EXISTING-COMMAND 
} -message {
   "%s" isn't a subcommand base command procedure
} -explanation {
   The subcommand list could not be retrieved because <command> is not a
   subcommand base command procedure.  Either it is a regular procedure, a
   subcommand procedure or does not exist.
}

### ===========================================================================
### -command  : ::Simple::Subcommand::create
# -synopsis   : Creates a subcommand procedure.
# -alias      : proc-sub
# -overview   :
#
#     This procedure acts as an interface to the K<proc> command to create
#  subcommand procedures.  Such subcommand procedures behave similarly to Tcl
#  subcommands (such as K<string> or K<array>) in terms of error handling:  if
#  the base command is invoked with an incorrect subcommand, an error is
#  thrown informing of the allowed subcommands; also, when a subcommand is
#  called with a bad number of arguments, an error is thrown informing of the
#  argument list for that subcommand.
#
#     The syntax for the argument list is identical to that accepted by the
#  K<proc> command.
#
# -arguments  :
#  { procedure        commandname    {Base command and subcommand pair}}
#  { arguments        canonicalarg-list           
#                                    {Subcommand procedure argument list}}
#  { body             script         {Subcommand procedure body}}
#
# -returns    : The empty string.
# -keywords   : procedure command subcommand family composite ensemble
# -examples   :
#
#  # Create some subcommands
#  proc-sub {foo bar} arg1 {puts $arg1}
#  proc-sub {foo gee} {arg1 {arg2 99} args} {puts "$arg1 $arg2 $args"}
#
# -details    :
#
#     A base command is created with the given base command name.  The base
#  command body simply calls the K<::Simple::Subcommand::Priv::base-command>
#  procedure passing all arguments received.  The actual subcommand procedure
#  is then created with the name K<command subcommand> where K<command> is the
#  base command name and K<subcommand> is the subcommand name.
#
# -references :
#
#  * The K<proc-ext> procedure.
#
# -effects    :
#
#  * Creates a command named K<procedure>.
#
#  * Creates a base command procedure should this be the first subcommand.
#
proc ::Simple::Subcommand::create {
   procedure
   arguments
   body
} {
   ### Save the command name
   ::Simple::Priv::save-name [lindex $procedure 0]

   ### Parse and qualify the procedure name
   foreach {qualifiedProcedure command subcommand}\
      [::Simple::Subcommand::Priv::parse-qualify $procedure] break

   ### Create the base command
   ::Simple::Subcommand::Priv::create-base-command $command $subcommand

   ### Create the subcommand
   proc $qualifiedProcedure $arguments $body

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Subcommand::Priv::create-base-command
# -synopsis   : Creates a subcommand base command or base method.
# -overview   :
#
#     This procedure creates a subcommand base command or base method if it
#  does not exists, otherwise it asserts that it is a base command or base
#  method indeed.
#
# -arguments  :
#  { command          qualifiedname  {Qualified command or method name}}
#  { subcommand       unqualifiedname
#                                    {Subcommand name}}
#  {?procCommand?     choice {::proc ::itcl::body}
#                                    {Name of the command used to create the
#                                     procedure (::proc) or method or class
#                                     command (::itcl::body)}}
#  {?isMethod?        boolean 0      {Whether the item is a method}}
#
# -effects    :
#
#  * Creates a command or method named K<command>.
#
proc ::Simple::Subcommand::Priv::create-base-command {
   command
   subcommand
   {procCommand ::proc}
   {isMethod 0}
} {
   ### Compose the command body fetching script
   set script {

      # Assert the namespace exists
      namespace parent $namespace

      # Fetch the command body
      set baseBody [namespace eval $namespace [list info body $command]]
   }

   ### Class command or method
   if {[string compare $procCommand ::proc]} {

      # The fact that the base method or base class command body exists but
      # is the empty string means that we are in the middle of a class
      # declaration, where the base method or base class command has been
      # forward declared with an empty body.  We throw an error here so
      # that the base method or base class command is created.
      append script {
         if {![string compare $baseBody {}]} {
            error {}
         }
      }
   }

   ### The base command does not exist
   set namespace [::Simple::Namespace::qualifiers $command]
   if {$isMethod} {
      set this {$this}
   } else {
      set this {}
   }
   if {[catch $script]} {

      ### Create the base command
# fixme: instead of one base command per subcommand family, we could have
#        aliases to a single, general, base command.
      $procCommand $command args [format {

         #========= %%%s%%

         ### Evaluate the subcommand procedure
         ### Error from subcommand procedure
         #   The uplevel here ensures that the call to the base command
         #   is hidden to the user-supplied command procedure body
         if {[catch {uplevel 1 %s [lindex [info level 0] 0]\\ $args} result]} {

            ### Process the subcommand procedure error
            ::Simple::Error::Priv::subcommand-error %s $result

         ### No error from subcommand procedure
         } else {

            ### Return the user return value
            return $result
         }

      } $::Simple::BaseCommandTag $this $command]

   ### The base command exists
   } else {

      ###  Assert the base command procedure is a base command indeed
      set tmp [::Simple::Priv::restore-name]
      if {[string first $::Simple::BaseCommandTag $baseBody] == -1} {
         ::Simple::Error::throw ::Simple::Subcommand::CANT-CREATE\
            $tmp $subcommand $tmp
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Subcommand::Priv::parse-qualify
# -synopsis   : Parses and qualifies a subcommand procedure name.
# -overview   :
#
#     This procedure parses a subcommand procedure name which must be two
#  words, the base command and subcommand names.  In case the base command
#  name is unqualified, it is qualified in the scope two levels up the stack. 
#  The procedure then returns a three element list containing the qualified
#  full name, the qualified base command name and the subcommand name.
#
# -returns    : A list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0: full command
#     1: command
#     2: subcommand
##end
#
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
proc ::Simple::Subcommand::Priv::parse-qualify {
   procedure
} {
   ### Invalid number of words in name
   if {[llength $procedure] != 2} {

      ### Throw error
      ::Simple::Error::throw ::Simple::BAD-ITEM\
         {subcommand procedure name} $procedure
   }

   ### Get the base command and subcommand names
   foreach {command subcommand} $procedure break

   ### Assert the base command and subcommand names are not empty
   if {[llength $command] == 0} {
      ::Simple::Error::throw ::Simple::BAD-ITEM\
         {subcommand base command name} {}
   }
   if {[llength $subcommand] == 0} {
      ::Simple::Error::throw ::Simple::BAD-ITEM {subcommand name} {}
   }

   ### Qualify the command name in the calling scope
   set command [::Simple::Priv::qualify $command 2]

   ### Return the full command, the base command and the subcommand names
   list $command\ $subcommand $command $subcommand
}

### ===========================================================================
### -command  : ::Simple::Priv::proc-information-usage
# -synopsis   : Returns a regular procedure argument usage string.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : The procedure argument usage string.
#
# -references :
#
#  * The K<::Simple::Proc::information usage> procedure.
#
#  * The K<::Simple::Option::information usage> procedure.
#
proc {::Simple::Priv::proc-information-usage} {
   procedure
} {
   set answer [list]
   foreach argument [uplevel 1 [list ::info args $procedure]] {
      if {[uplevel 1 [list ::info default $procedure $argument default]]} {
         lappend answer ?$argument?
      } else {
         lappend answer $argument
      }
   }
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Priv::proc-information-arguments
# -synopsis   : Returns a regular procedure argument list.
# -arguments  :
#  { procedure        commandname    {Procedure name}}
#
# -returns    : The procedure argument list.
#
# -references :
#
#  * The K<::Simple::Proc::information arguments> procedure.
#
proc {::Simple::Priv::proc-information-arguments} {
   procedure
} {
   set answer [list]
   foreach argument [uplevel 1 [list ::info args $procedure]] {
      if {[uplevel 1 [list ::info default $procedure $argument\
         ::Simple::Priv::%TMP%]]} {

         lappend answer [list $argument ${::Simple::Priv::%TMP%}]
      } else {
         lappend answer $argument
      }
   }
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Priv::proc-information-exists
# -synopsis   : Returns whether a regular procedure exists.
# -arguments  :
#  { namespace        namespace      {Procedure namespace}}
#  { procedure        unqualifiedname
#                                    {Unqualified procedure name}}
#
# -returns    : Whether the procedure exists.
#
proc {::Simple::Priv::proc-information-exists} {
   namespace
   procedure
} {
   ### Namespace exists
   #   The existance of the namespace is checked first so as to not
   #   create it with the "namespace eval" below
   if {[::Simple::Namespace::information exists $namespace]} {

      ### Get whether the procedure exists
      set answer [expr {[llength\
         [namespace eval $namespace [list info procs $procedure]]] == 1}]

   ### Namespace does not exist
   } else {

      ### Return false
      set answer 0
   }

   ### Return whether the procedure exists
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Priv::save-name
# -synopsis   : Saves a variable or procedure name to be accessed later.
# -overview   :
#
#     This procedure saves a variable or procedure name which can be later
#  obtained via the K<::Simple::Priv::restore-name> procedure.  The purpose of
#  this pair of functions is to save the original name of a variable or
#  procedure given by the user as an argument to a Simple Development Library
#  procedure in order to be used for error reporting.
#
# -arguments  :
#  { name             variableorcommandname 
#                                    {Variable or procedure name}}
#  {?levels?          integer 2      {Levels up the stack in which to operate}}
#
# -references :
#
#  * The K<::Simple::Priv::restore-name> procedure.
#
proc ::Simple::Priv::save-name {
   name
   {levels 2}
} {
   ### Save the name if the given number of levels up
   ###    the stack is outside The Simple Development Library
   if {![string match ::Simple* [uplevel $levels namespace current]]} {

      set ::Simple::Priv::SavedName $name
   }
}

### ===========================================================================
### -command  : ::Simple::Priv::restore-name
# -synopsis   : Restores a variable or procedure name previously saved.
# -overview   :
#
#     This procedure returns the previously saved variable or procedure name
#  via the K<::Simple::Priv::save-name> procedure
#
# -returns    : The previously saved variable or procedure name.
#
# -references :
#
#  * The K<::Simple::Priv::save-name> procedure.
#
proc ::Simple::Priv::restore-name {
} {
   return $::Simple::Priv::SavedName
}

### ===========================================================================
### -command  : ::Simple::Subcommand::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Subcommand::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Subcommand::cget
# -synopsis   : Gets the package options.
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Subcommand::cget {
} {
   return [list]
}

}
