### ===========================================================================
### -package  : SimpleTest
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library regression testing framework.
# -module     : implementation
# -overview   :
#
#     This package provides a complete framework for automated regression
#  testing.
#
#     Procedures are provided to initialize, reset and shutdown the test
#  environment, set criteria of which cases to test, run the test cases
#  as well as others to obtain the test results and statistics.
#
# -usage      :
#
#     The K<::Simple::Test::initialize> procedure sets up a test environment
#  where test cases can be run via the K<test-case> procedure.  By using
#  K<::Simple::Test::set-test-criteria>, one can set a criteria to specify
#  which test cases to test or ignore.
#
#     Upon testing a test case, its prerequisites are first assessed; if any
#  prerequisite is not fulfilled, the test case is skipped.  Next, the list of
#  dependencies are checked.  If any of the dependencies did not pass, the
#  test case is skipped.  Otherwise, the test case is run, and by comparing
#  its output, errors and return values with the corresponding expected
#  values, each test is flagged as passed or failed.
#
#     The detailed test results and the flags given to the corresponding call
#  to the P<test-case> procedures are available via the
#  K<::Simple::Test::results> and K<::Simple::Test::flags> procedures,
#  respectively.  Finally, the overall statistics of all the test cases can be
#  obtained via the K<::Simple::Test::statistics> procedure.
#
#     Use K<::Simple::Test::reset> to reset the test environment and
#  K<::Simple::Test::shutdown> to clean up after testing.
#
# -details    :
#
#  * The test environment is in fact a Tcl slave interpreter in which the
#    K<puts> command has been modified so that the test cases output can be
#    collected. 
#
#  * The test interpreter name is contained in the
#    V<::Simple::Test::TestInterp> variable.  It is possible to set it to the
#    empty string so as to use the toplevel interpreter.  This is useful to
#    run the tests within debugers which may have problems with child
#    interpreters.
#
# -keywords   : regression test package
# -variable   :
#
#  {
#     ::Simple::Test::TestInterp
#
#  } -description {
#
#     Test interpreter
#
#  } -type name -value test-interp
#
# -variable   :
#
#  {
#     ::Simple::Test::CaseOutput
#
#  } -description {
#
#     Current test case output
#
#  } -type string -value {}
#
# -variable   :
#
#  {
#     ::Simple::Test::CaseFlags
#
#  } -description {
#
#     Current test case flags
#
#  } -type flag-list -value {}
#
# -variable   :
#
#  {
#     ::Simple::Test::TestCaseFlags
#
#  } -description {
#
#     P<test-case> procedure flags
#
#  } -type flag-list -value {-regexp -prerequisites -dependencies\
#     -setup -script -cleanup -output -error -return}
#
# -variable   :
#
#  {
#     ::Simple::Test::Results
#
#  } -description {
#
#     Tests cases tested list, each one containing the following elements:
#
##table;delimiter=":";ignore="true"
#        0: number
#        1: tag
#        2: title
#        3: whether it was ignored
#        4: whether it was skipped
#        5: whether it passed or failed
#        6: in case it failed, the diagnostics
##end
#
#  } -type list -value {}
#
# -variable   :
#
#  {
#     ::Simple::Test::TestCasesPassed
#
#  } -description {
#
#     Array containing the tags of the test cases run and
#     passed
#
#  } -type testtag-array -value {}
#
# -variable   :
#
#  {
#     ::Simple::Test::Flags
#
#  } -description {
#
#     Tests cases tested list, each one containing its stored flags and their
#     contents as given to the corresponding P<test-case> procedure.  Each
#     element is a list of two-element lists, being the first element an
#     option and the second its contents.
#
#  } -type list -value {}
#
# -variable   :
#
#  {
#     ::Simple::Test::NTests
#
#  } -description {
#
#     Number of tests cases tested
#
#  } -type integer -value 0
#
# -variable   :
#
#  {
#     ::Simple::Test::NTestsPassed
#
#  } -description {
#
#     Number of tests cases passed
#
#  } -type integer -value 0
#
# -variable   :
#
#  {
#     ::Simple::Test::NTestsFailed
#
#  } -description {
#
#     Number of tests cases failed
#
#  } -type integer -value 0
#
# -variable   :
#
#  {
#     ::Simple::Test::NTestsSkipped
#
#  } -description {
#
#     Number of tests cases skipped
#
#  } -type integer -value 0
#
# -variable   :
#
#  {
#     ::Simple::Test::NTestsIgnored
#
#  } -description {
#
#     Number of tests cases ignored
#
#  } -type integer -value 0
#
# -variable   :
#
#  {
#     ::Simple::Test::TestExpr
#
#  } -description {
#
#     Test cases to be tested expression
#
#  } -type expression -value 1
#
# -option     :
#
#  {
#     -store
#
#  } -description {
#
#        Flags to store for each test case tested, which can later be obtained
#     via the K<::Simple::Test::flags> procedure.
#
#  } -type list -value {}
#
# -api        :
#
#  * K<::Simple::Test::initialize>
#    Initializes the test environment.
#
#  * K<test-case> tag title ?-regexp? ?-prerequisites script-list?
#       ?-dependencies testtag-list? ?-setup script? ?-script script?
#       ?-cleanup script? ?-output string? ?-error string? ?-return string?
#    Runs a test case.
#
#  * K<::Simple::Test::set-test-criteria> testCriteria
#    Sets the criteria to specify which test cases to test.
#
#  * K<::Simple::Test::statistics>
#    Returns test cases statistics.
#
#  * K<::Simple::Test::results>
#    Returns the test cases results.
#
#  * K<::Simple::Test::flags>
#    Returns the test stored flags.
#
#  * K<::Simple::Test::reset>
#    Resets the test environment.
#
#  * K<::Simple::Test::shutdown>
#    Shutdowns the test environment.
#
#  * K<::Simple::Test::configure> ?-store list?
#    Configures the package options.
#
#  * K<::Simple::Test::cget> ?-store?
#    Gets the package options.
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#  ::Simple::Package::require-and-install SimpleTest
#
#  # Initialize the test environment
#  ::Simple::Test::initialize
#
#  # Set the test criteria: test cases number 1, 3 and following
#  ::Simple::Test::set-test-criteria 1,3:
#
#  # Execute several test cases
#
#  # This one passes
#  test-case test-case-1 {
#     First test case
#  } -setup {
#
#     # Create a variable in the test environment
#     set foo 1
#
#  } -script {
#        puts -nonewline $foo
#  } -output 1
#
#  # This one is ignored
#  test-case test-case-2 {
#     Second test case
#  } -script {
#  }
#
#  # This one fails
#  test-case test-case-3 {
#     Third test case
#  } -script {
#        return $foo
#  } -return 2
#
#  # This one is skipped
#  test-case test-case-4 {
#     Fourth test case
#  } -prerequisites {
#     {$foo == 2}
#  } -script {
#  }
#
#  # This one passes also
#  test-case test-case-5 {
#     Fifth test case
#  } -script {
#        error {ERROR MESSAGE}
#  } -cleanup {
#
#     # Delete the variable in the test environment
#     unset foo
#
#  } -error {ERROR MESSAGE}
#
#  # This one depends on test-case-1 and test-case-2
#  test-case test-case-6 {
#     Sixth test case
#  } -dependencies {
#     test-case-1 test-case-2
#  } -script {
#     puts -nonewline 1
#  } -output 1
#     
#  # This displays the following:
#  #   |1 test-case-1     First test case: passed
#  #   |2 test-case-2    Second test case: ignored
#  #   |3 test-case-3     Third test case: failed
#  #   |        -return  = {2}
#  #   |   Actual return = {1}
#  #   |4 test-case-4    Fourth test case: skipped
#  #   |5 test-case-5     Fifth test case: passed
#  #   |6 test-case-6     Sixth test case: skipped
#
#  foreach result [::Simple::Test::results]\ 
#     optionPairs [::Simple::Test::flags] {
#
#     # Get the test case information
#     foreach {number tag title ignored skipped passed diagnostics}\ 
#        $result break
#     if {$ignored} {
#        set status ignored
#     } elseif {$skipped} {
#        set status skipped
#     } elseif {$passed} {
#        set status passed
#     } else {
#        set status failed
#     }
#
#     # Display a short report for each test case
#     puts [format {%1d %10s %20s %s} $number $tag $title: $status]
#
#     # For failed tests, display the test case flags and diagnostics
#     if {![string compare $status failed]} {
#        foreach {option contents} $optionPairs {
#           puts [format {%16s = {%s}} $option $contents]
#        }
#        foreach {reason extra} $diagnostics {
#           puts [format {%16s = {%s}} $reason $extra]
#        }
#     }
#  }
#
#  # Final tests statistics
#  foreach {nTests passed failed skipped ignored}\ 
#     [::Simple::Test::statistics] break
#
#  # Display some test cases statistics
#  # This displays the following:
#  #   |Total test   cases 6
#  #   |Test cases  passed 2
#  #   |Test cases  failed 1
#  #   |Test cases skipped 2
#  #   |Test cases ignored 1
#  puts [format {Total test   cases %d} $nTests]
#  puts [format {Test cases  passed %d} $passed]
#  puts [format {Test cases  failed %d} $failed]
#  puts [format {Test cases skipped %d} $skipped]
#  puts [format {Test cases ignored %d} $ignored]
#
#  # Shutdown the test environment
#  ::Simple::Test::shutdown
#
# -todo       :
#
#  * Consider whether it is worth it migrating the test infrastructure to use
#    the facilities of the Tcl test harness package, P<tcltest>.
#
#  * Add support for criteria to specify which test cases to test based on the
#    test case tag.
#
#  * Keep track of the test cases tags. Ensure they are not duplicated.
#    Ensure tags of the form ''<tag>-<number>'' are correlative.
#
#  * Allow to specify a user defined procedure for handling each test case
#    results as soon as they are run.
#
# -history    :
#
#  19-feb-1999   Unreleased first version 0.1
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  25-sep-2003   Named test interpreter, version 0.5.1
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleTest 1.0 -requisites {

   {SimpleType 1.0}

} -namespaces {

   ::Simple::Test
   ::Simple::Test::Priv

} -exportedcommands {

   ::Simple::Test::test-case

} -install {

### Create the package variables
set ::Simple::Test::TestInterp test-interp
set ::Simple::Test::CaseOutput {}
set ::Simple::Test::CaseFlags [list]
set ::Simple::Test::TestCaseFlags [list -regexp -prerequisites -dependencies\
   -setup -script -cleanup -output -error -return]
set ::Simple::Test::Results {}
array set ::Simple::Test::TestCasesPassed {}
set ::Simple::Test::Flags [list]
set ::Simple::Test::NTests 0
set ::Simple::Test::NTestsPassed 0
set ::Simple::Test::NTestsFailed 0
set ::Simple::Test::NTestsSkipped 0
set ::Simple::Test::NTestsIgnored 0
set ::Simple::Test::TestExpr 1

### Create the package options

# A<-store>
set ::Simple::Test::%Options%(-store) [list]

} -uninstall {

### Delete the package variables
unset ::Simple::Test::TestInterp
unset ::Simple::Test::CaseOutput
unset ::Simple::Test::CaseFlags
unset ::Simple::Test::TestCaseFlags
unset ::Simple::Test::Results
unset ::Simple::Test::TestCasesPassed
unset ::Simple::Test::Flags
unset ::Simple::Test::NTests
unset ::Simple::Test::NTestsPassed
unset ::Simple::Test::NTestsFailed
unset ::Simple::Test::NTestsSkipped
unset ::Simple::Test::NTestsIgnored
unset ::Simple::Test::TestExpr

### Delete the package options
unset ::Simple::Test::%Options%(-store)

} -firsttimeinstall {

### Declare the package types
::Simple::Type::declare {
   testtag
} -description {
   Test case tag: ''public-api-1'', ...
} -matchingscript {
   return [regexp {^[a-zA-Z0-9]+(-[0-9]+)?$} $value]
}

::Simple::Type::declare {
   testcriteria
} -description {
   Criteria for test cases to be tested: '':'', ''3'', ''3,7'', '':7'', ..
} -matchingscript {
   ::Simple::Test::Priv::ranges-to-expr n $value
   return 1
}

### Declare the package errors
::Simple::Error::declare {
   ::Simple::Test::ENVIRONMENT-NOT-INITIALIZED
} -message {
   test environment not initialized
} -explanation {
   The attempted operation requires an initialized test environment
} -corrective {
   Use the K<::Simple::Test::initialize> procedure to initialize the test
   environment
}

::Simple::Error::declare {
   ::Simple::Test::NO-SCRIPT-FLAG 
} -message {
   no -script for test "%s"
} -explanation {
   Test case <test case tag> lacks the compulsory A<-script> flag
}

::Simple::Error::declare {
   ::Simple::Test::RETURN-AND-ERROR-FLAGS 
} -message {
   both -return and -error given for test case "%s"
} -explanation {
   Test case <test case tag> has both A<-return> and A<-error> flags, but
   these are mutually exclusive
}

::Simple::Error::declare {
   ::Simple::Test::BAD-SET-OF-RANGES 
} -message {
   bad set of ranges "%s", range "%s" is not well formed
} -explanation {
   The set of ranges ranges <ranges> has bad format; <range> is the bad
   range
} -corrective {
      A set of ranges is one or more ranges separated by commas.  A range is
   one unsigned integer or two unsigned integers separated by a colon where
   the first, the second or even both unsigned integers may not be present,
   meaning that the range is open by the left, by the right, or by both ends,
   respectively.
}

::Simple::Error::declare {
   ::Simple::Test::BAD-FLAG-TO-STORE 
} -message {
   bad flag to store "%s":  must be -cleanup, -dependencies, -error,\
   -output, -prerequisites, -regexp, -return, -script or -setup
} -explanation {
   Flag <flag> is not allowed 
} -corrective {
   Supply a flag among those in the list
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::test-puts
# -synopsis   : Replaces the K<puts> command in the test environment.
# -overview   :
#
#     Works as K<puts> except if the channel to write to is stdout, stderr or
#  not given.  In that case, the output is sent to the A<CaseOutput> namespace
#  variable, unless the argument A<ignoreOutput> is set, case in which the
#  output is ignored all together.
#
# -arguments  :
#  { ignoreOutput     boolean        {Whether to ignore the output to stdout
#                                     and stderr}}
#  {-nonewline        boolflag       {Do not append a newline}}
#  {?channel?         channel stdout {Channel to write to}}
#  { string           string         {String to write}}
#
# -limitations:
#
#  * To identify whether some arguments are given or not their default value
#    is set to ''%DEFAULT%''.
#
proc ::Simple::Test::Priv::test-puts {
   ignoreOutput
   arg1
   {arg2 %DEFAULT%}
   {arg3 %DEFAULT%}
} {
   ### Parse arguments
   set newline [string compare $arg1 -nonewline]
   if {!$newline} {
      set channel $arg2
      set string $arg3
   } else {
      set channel $arg1
      set string $arg2
   }
   set channelGiven [string compare $string %DEFAULT%]
   if {!$channelGiven} {
      set string $channel
   }

   ### Channel given but not stdout nor stderr
   if {$channelGiven &&\
      [string compare $channel stdout] && [string compare $channel stderr]} {

      ### Delete the alias for K<puts> in the test interpreter
      interp alias $::Simple::Test::TestInterp puts {}
      interp eval $::Simple::Test::TestInterp\
         {rename ::Simple::Test::Priv::real-puts ::puts}

      ### Write to the given channel
      if {$newline} {
         interp eval $::Simple::Test::TestInterp [list puts $channel $string]
      } else {
         interp eval $::Simple::Test::TestInterp\
            [list puts -nonewline $channel $string]
      }

      ### Recreate the alias for K<puts> in the test interpreter
      interp eval $::Simple::Test::TestInterp\
         {rename ::puts ::Simple::Test::Priv::real-puts}
      interp alias $::Simple::Test::TestInterp puts {}\
         ::Simple::Test::Priv::test-puts $ignoreOutput

   ### stdout or stderr channel or no channel given at all
   } else {

      ### Do not ignore output
      if {!$ignoreOutput} {

         ### Send output to the ::Simple::Test::CaseOutput variable
         append ::Simple::Test::CaseOutput $string
         if {$newline} {
            append ::Simple::Test::CaseOutput \n
         }
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::assert-test-environment
# -synopsis   : Throws an error if the test environment has not been initialized.
#
proc ::Simple::Test::Priv::assert-test-environment {
} {
   if {![interp exist $::Simple::Test::TestInterp]} {
      ::Simple::Error::throw ::Simple::Test::ENVIRONMENT-NOT-INITIALIZED
   }
}

### ===========================================================================
### -command  : ::Simple::Test::initialize
# -synopsis   : Initializes the test environment.
# -returns    : The empty string.
#
# -remarks    :
#
#  * It is compulsory to call this procedure before testing.
#
# -effects    :
#
#  * Creates the test interpreter.
#
proc ::Simple::Test::initialize {
} {
   ### Reset the test case number and test output
   ::Simple::Test::reset

   ### Create the test interpreter
   interp create $::Simple::Test::TestInterp

   ### Is Tk loaded into the master interpreter?
   if {![string compare [::Simple::Package::information state Tk] installed]} {

      ### Load Tk into tthee test interpreter
      load {} Tk $::Simple::Test::TestInterp
   }

   ### Give it the same V<auto_path> as the master interpreter
   interp eval $::Simple::Test::TestInterp [list set ::auto_path $::auto_path]

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::test-eval
# -synopsis   : Evaluates a script in the test environment.
# -overview   :
#
#     The script is evaluated in the test environment.  Its output to stdout
#  and stderr is collected into the A<CaseOutput> namespace variable, unless
#  the argument A<ignoreOutput> is set, case in which the output is ignored
#  all together.
#
# -arguments  :
#  { script           script         {Script to evaluate}}
#  {?ignoreOutput?    boolean 1      {Whether to ignore the script output to
#                                     stdout and stderr}}
#
# -remarks    :
#
#  * It is compulsory to have called K<::Simple::Test::initialize> before
#  calling this procedure.
#
# -effects    :
#
#  * Creates an alias for K<puts> in the test interpreter.
#
#  * Those of the evaluated script in the test environment.
#
proc ::Simple::Test::Priv::test-eval {
   script
   {ignoreOutput 1}
} {
   ### Create an alias for K<puts> in the test interpreter

   # Move the real puts in case it has not been moved before
   interp eval $::Simple::Test::TestInterp {
      if {[llength [info commands ::Simple::Test::Priv::real-puts]] == 0} {
         rename ::puts ::Simple::Test::Priv::real-puts
      }
   }

   # Create the actual alias
   interp alias $::Simple::Test::TestInterp puts {}\
      ::Simple::Test::Priv::test-puts $ignoreOutput

   ### Evaluate the script in the test environment
   interp eval $::Simple::Test::TestInterp $script
}

### ===========================================================================
### -command  : ::Simple::Test::test-case
# -synopsis   : Runs a test case.
# -access     : exported
# -keywords   : test
# -overview   :
#
#     This is the actual test case handler.  Works by running the test case
#  in the test environment where the output of K<puts> is available at the
#  end.  If that output matches the expected output and any error message and
#  return values match the expected ones, the test case is considered passed,
#  otherwise the test fails.  The output of this procedure is stored and
#  can be obtained via the K<::Simple::Test::results> procedure.
#
#     If the expected output, return value or error are empty, the test passes
#  only if it does not produce any output, return value or error,
#  respectively.  If the expected return value or error is not given, then its
#  actual value is not checked.
#
#     By specifying the A<-regexp> flag, the expected output and return value
#  or error are matched as regular expressions so that it is possible to
#  validate non-constant values.
#
#     Before running the test case, each prerequisite in the list of
#  prerequisites is evaluated.  If any of them fails, the test is skipped. 
#  Next, the list of dependencies is checked.  If any of the dependencies does
#  not pass, the test is skipped.
# 
#     Otherwise, the setup script is evaluated afterwards.  If there is an
#  error, the test is failed.  Otherwise, the actual test is run followed by
#  the cleanup script.  An error in the cleanup script also results in the
#  test case to be regarded as failed.
#
# -arguments  :
#  { tag              testtag        {Test case tag}}
#  { title            string         {Test case title}}
#  {-regexp           boolflag       {Whether the expected output, return
#                                     and error shall be matched as regular
#                                     expressions}}
#  {-prerequisites    script-list {} {Prerequisites scripts list}}
#  {-dependencies     testtag-list {}
#                                    {List of test cases tags this test depends
#                                     on}}
#  {-setup            script {}      {Setup script}}
#  {-script           script {}      {Test case script}}
#  {-cleanup          script {}      {Cleanup script}}
#  {-output           string {}      {Test case expected output}}
#  {-error            string {}      {Test case expected error}}
#  {-return           string {}      {Test case expected return}}
#
# -returns    : Whether the test passed or not.
# -examples   :
#
#  # A test which checks some output
#  test-case T1 {Test 1} -script {puts -nonewline "foo"} -output foo
#
#  # A test which checks an error message
#  test-case T2 {Test 2} -script {return -code error bar} -error bar
#
#  # A test which checks the output using a regular expression
#  test-case T3 {Test 3} -regexp -script {puts -nonewline foo} -output {oo$}
#
# -remarks    :
#
#  * The A<-script> argument is compulsory.
#
#  * The A<-error> and A<-return> arguments are mutually exclusive.
#
#  * It is compulsory to have called K<::Simple::Test::initialize> before
#    calling this procedure.
#
#  * The order of the flags is irrelevant but their parameters can not start
#    by an hyphen.
#
# -effects    :
#
#  * Those of the evaluated scripts in the test environment.
#
proc ::Simple::Test::test-case {
   tag
   title
   args
} {
   ### Initialization
   ::Simple::Test::Priv::reset-flags

   ### Parse arguments
   set title [string trim $title]

   ### Parse flags
   # Parse A<-regexp>
   set regexp 0
   set regexpIndex [lsearch -exact $args -regexp]
   if {$regexpIndex > -1} {
      set regexp 1
      ::Simple::Test::Priv::store-flag -regexp {}
      set args [lreplace $args $regexpIndex $regexpIndex]
   }
   set flags [::Simple::Priv::parse-flags\
      [list -regexp untouched\
      -prerequisites list\
      -dependencies list\
      -setup untouched\
      -script untouched\
      -cleanup untouched\
      -output untouched\
      -return untouched\
      -error untouched]\
      [list] $args 0 1]
   foreach flag $flags {
      ::Simple::Test::Priv::store-flag $flag [set [string trimleft $flag -]]
   }
   if {![::Simple::Priv::flag-given -script]} {
      ::Simple::Error::throw ::Simple::Test::NO-SCRIPT-FLAG $tag
   }
   set outputGiven [::Simple::Priv::flag-given -output]
   set returnGiven [::Simple::Priv::flag-given -return]
   set errorGiven  [::Simple::Priv::flag-given -error]
   if {$returnGiven && $errorGiven} {
      ::Simple::Error::throw ::Simple::Test::RETURN-AND-ERROR-FLAGS $tag
   }

   ### Assert the test environment has been initialized
   ::Simple::Test::Priv::assert-test-environment

   ### This test must be ignored
   set testCaseEvaluated 0
   set passed 1
   set skipped 0
   set ignored 0
   set diagnostics [list]
   incr ::Simple::Test::NTests
   if {![::Simple::Test::Priv::must-be-tested $::Simple::Test::NTests]} {

      ### Test ignored
      set ignored 1

      ### Test failed
      set passed 0
   }

   ### Check for dependencies
   if {$passed && [::Simple::Priv::flag-given -dependencies]} {

      ### Check that all required tests passed
      #   If we find some that didn't pass (ie:  either skipped, failed, or
      #   not yet run), add them to a temporary list
      set badDependencies [list]
      foreach dependency $dependencies {
         if {![info exists ::Simple::Test::TestCasesPassed($dependency)]} {
            lappend badDependencies $dependency
         }
      }

      ### Unfulfilled dependencies
      if {[llength $badDependencies] > 0} {

         ### Test skipped
         set skipped 1

         ### Fill in diagnostics
         set diagnostics [list {Dependencies failed} $badDependencies]

         ### Test failed
         set passed 0
      }
   }

   ### Test not failed yet
   if {$passed} {

      ### A<-prerequisites> given
      if {[::Simple::Priv::flag-given -prerequisites]} {

         ### Loop over prerequisites
         foreach prerequisite $prerequisites {

            ### Evaluate the prerequisite in the test environment

            ### Prerequisite failed
            if {[catch {set prerequisiteResult\
               [::Simple::Test::Priv::test-eval\
               [format {expr {%s}} $prerequisite]]}]} {

               ### Test skipped
               set skipped 1

               ### Fill in diagnostics
               set diagnostics [list {Prerequisite failed} $::errorInfo]

               ### Test failed
               set passed 0
               break

            ### Unfulfilled prerequisite
            } elseif {!$prerequisiteResult} {

               ### Test skipped
               set skipped 1

               ### Fill in diagnostics
               set diagnostics [list {Unfulfilled prerequisite} $prerequisite]

               ### Test failed
               set passed 0
               break
            }
         }
      }
   }

   ### Test not failed yet
   if {$passed} {

      ### A<-setup> given
      if {[::Simple::Priv::flag-given -setup]} {

         ### Evaluate the setup in the test environment
         if {[catch {::Simple::Test::Priv::test-eval $setup}]} {

            ### Fill in diagnostics
            set diagnostics [list {Setup failed} $::errorInfo]

            ### Test failed
            set passed 0
         }
      }
   }

   ### Test not failed yet
   if {$passed} {

      ### Evaluate the test case in the test environment
      #   collecting the output into the CaseOutput variable
      set ::Simple::Test::CaseOutput {}
      set testCaseResult [catch {set testCaseReturn\
         [::Simple::Test::Priv::test-eval $script 0]} testCaseMessage]
      set testCaseEvaluated 1

      ### A<-cleanup> given
      if {[::Simple::Priv::flag-given -cleanup]} {

         ### Evaluate the cleanup in the test environment
         if {[catch {::Simple::Test::Priv::test-eval $cleanup}]} {

            ### Fill in diagnostics
            set diagnostics [list {Cleanup failed} $::errorInfo]

            ### Test failed
            set passed 0
         }
      }

      ### Output does not match expectations
      #   If the expected output is empty ensure the test case output
      #   is also empty, otherwise try to match the expected output either
      #   exactly or as a regular expression as required
      if {(![string compare $output {}] &&\
         [string compare $::Simple::Test::CaseOutput {}]) || (!$regexp &&\
         [string compare $output $::Simple::Test::CaseOutput]) ||\
         ($regexp && ![regexp -- $output $::Simple::Test::CaseOutput])} {

         ### Fill in expected output and actual output diagnostics
         set diagnostics [concat $diagnostics\
            [list {Actual output} $::Simple::Test::CaseOutput]]

         ### Test failed
         set passed 0
      }

      ### A<-error> given
      if {$errorGiven} {

         ### Error during the test case
         if {$testCaseResult} {

            ### Error message does not match expectations
            #   If the expected error is empty ensure the test case
            #   error is also empty, otherwise try to match the expected
            #   error either exactly or as a regular expression as required
            if {(![string compare $error {}] &&\
               [string compare $testCaseMessage {}]) ||\
               (!$regexp && [string compare $error $testCaseMessage]) ||\
               ($regexp && ![regexp -- $error $testCaseMessage])} {

               ### Fill in expected error and actual error diagnostics
               set diagnostics [concat $diagnostics\
                  [list {Actual error} $testCaseMessage]\
                  [list {Error information} $::errorInfo]]

               ### Test failed
               set passed 0
            }

         ### No error during the test case
         } else {

            ### Fill in expected error and actual error diagnostics
            set diagnostics [concat $diagnostics [list {Actual error} NONE]]

            ### Test failed
            set passed 0
         }

      ### A<-error> not given
      } else {

         ### Error during the test case
         if {$testCaseResult} {

            ### Fill in expected error and actual error diagnostics
            ::Simple::Test::Priv::store-flag -error NONE 1
            set diagnostics [concat $diagnostics\
               [list {Actual error} $testCaseMessage]\
               [list {Error information} $::errorInfo]]

            ### Test failed
            set passed 0

         ### No error during the test case
         } else {

            ### A<-return> given
            if {$returnGiven} {

               ### Return value does not match expectations
               #   If the expected return value is empty ensure the test
               #   case return value is also empty, otherwise try to match
               #   the expected return value either exactly or as a regular
               #   expression as required
               if {(![string compare $return {}] &&\
                  [string compare $testCaseReturn {}]) ||\
                  (!$regexp && [string compare $return $testCaseReturn]) ||\
                  ($regexp && ![regexp -- $return $testCaseReturn])} {

                  ### Fill in expected return and actual return diagnostics
                  set diagnostics [concat $diagnostics\
                     [list {Actual return} $testCaseReturn]]

                  ### Test failed
                  set passed 0
               }
            }
         }
      }
   }

   ### Update the tests statistics
   if {$ignored} {
      incr ::Simple::Test::NTestsIgnored
   } elseif {$skipped} {
      incr ::Simple::Test::NTestsSkipped
   } else {
      if {$passed} {
         set ::Simple::Test::TestCasesPassed($tag) {}
         incr ::Simple::Test::NTestsPassed
      } else {
         incr ::Simple::Test::NTestsFailed

         # Store the appropriate flags
         if {$testCaseEvaluated} {
            if {$regexp} {
               ::Simple::Test::Priv::store-flag -regexp {} 1
            }
            if {$outputGiven} {
               ::Simple::Test::Priv::store-flag -output $output 1
            }
            if {$errorGiven} {
               ::Simple::Test::Priv::store-flag -error $error 1
            }
            if {$returnGiven} {
               ::Simple::Test::Priv::store-flag -return $return 1
            }
         }
      }
   }

   ### Set the test information
   lappend ::Simple::Test::Results [list $::Simple::Test::NTests\
      $tag $title $ignored $skipped $passed $diagnostics]
   lappend ::Simple::Test::Flags [::Simple::Test::Priv::get-flags]

   ### Return whether the test passed or not
   return $passed
}

### ===========================================================================
### -command  : ::Simple::Test::set-test-criteria
# -synopsis   : Sets the criteria to specify which test cases to test.
# -overview   :
#
#     This procedure sets the expression that specifies which test cases to
#  test from the given criteria.  Test cases matching the criteria are tested
#  while all others are ignored.
#
# -arguments  :
#
#  { testCriteria    testcriteria    {Criteria of which test cases to test}}
#
# -returns    : The empty string.
# -todo       :
#
#  * Allow criteria based on test case tags.
#
proc ::Simple::Test::set-test-criteria {
   testCriteria
} {
   ### Set the criteria to specify which test cases to test
   set ::Simple::Test::TestExpr [::Simple::Test::Priv::ranges-to-expr\
         ::Simple::Test::NTests $testCriteria]

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::ranges-to-expr
# -synopsis   : Transforms a set of ranges into an expression.
# -overview   :
#
#     This procedure takes a set of ranges of the form "1,2,5:10,20:"
#  and returns the equivalent expression using A<number> as the variable
#  holding the integer.
#
#     A set of ranges is one or more ranges separated by commas.  A range is
#  one unsigned integer or two unsigned integers separated by a colon where
#  the first, the second or even both unsigned integers may not be present
#  meaning that the range is open by the left, by the right, or by both ends,
#  respectively.  It is equivalent to the decimal-range type of the
#  P<SimpleType> package.
#
#  An error is thrown if the set of ranges has bad format.
#
# -arguments  :
#  { number           variablename   {Variable name}}
#  { ranges           string         {Set of ranges}}
#
# -examples   :
#
#  Using A<number> = "n".
#
#  ranges                  expression
#  ----------------------- ---------------------------------------------
#  {}                      1
#  :                       1
#  3                       $n == 3
#  3,7                     $n == 3 || $n == 7
#  3:                      $n >= 3
#  3:,:                    $n >= 3
#  :7                      $n <= 7
#  3:7                     $n >= 3 && $n <= 7
#  :7,10:15,20:            $n <= 7 || ($n >= 10 && $n <= 15) || $n >= 20
#
# -returns    : The expression equivalent to the set of ranges.
#
proc ::Simple::Test::Priv::ranges-to-expr {
   number
   ranges
} {
   ### Loop over ranges
   set rangesList [split $ranges ,]
   set nRanges [llength $rangesList]
   set answer {}
   foreach range $rangesList {

      ### Range is an integer
      if {[string first : $range] == -1} {

         ### Assert the integer is valid
         if {![regexp {^(0|[1-9][0-9]*)$} $range]} {
            ::Simple::Error::throw ::Simple::Test::BAD-SET-OF-RANGES\
               $ranges $range
         }

         ### Set the subexpression
         set subexpression [format {$%s == %s} $number $range]

      ### Range is a proper range
      } else {

         ### Assert the range is valid
         foreach {left right tmp} [split $range :] break
         set leftNotEmpty [string compare $left {}]
         set rightNotEmpty [string compare $right {}]
         if {[string compare $tmp {}] ||\
            ($leftNotEmpty && ![regexp {^[1-9][0-9]*$} $left]) ||\
            ($rightNotEmpty && ![regexp {^[1-9][0-9]*$} $right]) ||\
            ($leftNotEmpty && $rightNotEmpty && $left >= $right)} {

            ::Simple::Error::throw ::Simple::Test::BAD-SET-OF-RANGES\
               $ranges $range
         }

         ### Range closed by the left
         if {$leftNotEmpty} {

            ### Set the subexpression
            set subexpression [format {$%s >= %s} $number $left]

            ### Range closed by the right also
            if {$rightNotEmpty} {

               ### Append to the subexpression
               append subexpression [format { && $%s <= %s} $number $right]

               ### There are multiple ranges
               if {$nRanges > 1} {

                  ### Enclose the compound subexpression in parenthesis
                  set subexpression ($subexpression)
               }
            }

         ### Range open by the left
         } else {

            ### Range closed by the right only
            if {$rightNotEmpty} {

               ### Set the subexpression
               set subexpression [format {$%s <= %s} $number $right]

            ### Range open by both ends
            } else {

               ### Set the subexpression
               set subexpression {}
            }
         }
      }

      ### Add the subexpression
      if {[string compare $answer {}]} {
         if {[string compare $subexpression {}]} {
            append answer [format { || %s} $subexpression]
         }
      } else {
         set answer $subexpression
      }
   }

   ### Return the expression equivalent to the set of ranges
   if {![string compare $answer {}]} {
      set answer 1
   }
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::must-be-tested
# -synopsis   : Returns whether a test must be tested.
# -overview   :
#
#     This procedure returns whether the given test must be tested or not
#  by evaluating the test cases to be tested expression.
#
# -arguments  :
#  { testNumber       integer        {Test number}}
#
# -returns    : Whether the test must be tested or not.
#
proc ::Simple::Test::Priv::must-be-tested {
   testNumber
} {
   ### Evaluate the test expression
   set answer [expr $::Simple::Test::TestExpr]

   ### Return whether the test must be tested or not
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Test::statistics
# -synopsis   : Returns test cases statistics.
# -returns    : A list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0: number of test cases executed
#     1: number of test cases passed
#     2: number of test cases failed
#     3: number of test cases skipped
#     4: number of test cases ignored
##end
#
proc ::Simple::Test::statistics {
} {
   return [list $::Simple::Test::NTests $::Simple::Test::NTestsPassed\
      $::Simple::Test::NTestsFailed $::Simple::Test::NTestsSkipped\
      $::Simple::Test::NTestsIgnored]
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::reset-flags
# -synopsis   : Resets the current test case flags.
#
proc ::Simple::Test::Priv::reset-flags {
} {
   catch {unset ::Simple::Test::CaseFlags}
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::store-flag
# -synopsis   : Stores a current test case flag.
# -arguments  :
#  { flag             flag           {Flag}}
#  { contents         any            {Flag contents}}
#  {?force?           boolean 0      {Whether to store the flag irrespectively
#                                     of the A<-store> package option}}
#
proc ::Simple::Test::Priv::store-flag {
   flag
   contents
   {force 0}
} {
   if {$force ||\
      [lsearch -exact [set ::Simple::Test::%Options%(-store)] $flag] != -1} {

      set ::Simple::Test::CaseFlags($flag) $contents
   }
}

### ===========================================================================
### -command  : ::Simple::Test::Priv::get-flags
# -synopsis   : Gets all current test case flags.
# -overview   :
#
#     This procedure returns the current test case flags ordered as in the
#  input to the corresponding P<test-case> procedure.  The format of the
#  return value is that of the K<::Simple::Test::flags> procedure, i.e., a
#  single list containing for each test case tested its stored flags and their
#  contents as given to the corresponding P<test-case> procedure.  The format
#  of the returned list is suitable to be given as argument to the K<array
#  get> command:  the array indices would be the flags and the array values
#  their contents.
#
# -returns    : The current test case flags.
#
proc ::Simple::Test::Priv::get-flags {
} {
   set answer [list]
   foreach flag $::Simple::Test::TestCaseFlags {
      if {[info exists ::Simple::Test::CaseFlags($flag)]} {
         lappend answer $flag $::Simple::Test::CaseFlags($flag)
      }
   }
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Test::results
# -synopsis   : Returns the test cases results.
# -overview   :
#
#     This procedure returns a list containing for each test case tested the
#  following elements:
#
##table;delimiter=":";ignore="true"
#     0: number
#     1: tag
#     2: title
#     3: whether it was ignored
#     4: whether it was skipped
#     5: whether it was passed or failed
#     6: in case it failed, the diagnostics; this is a list of two-elements\
#        lists, being the first element a reason and the second extra\
#        information
##end
#
#     Possible reasons and their corresponding extra information are the
#  following:
#
##table;delimiter="!";ignore="true";headings="true"
#        Reason                   ! Extra information
#        Dependencies failed      ! List of failed dependencies
#        Prerequisite failed      ! Error information (A<::errorInfo> contents)
#        Unfulfilled prerequisite ! Prerequisite script failed
#        Setup failed             ! Error information (A<::errorInfo> contents)
#        Cleanup failed           ! Error information (A<::errorInfo> contents)
#        Actual output            ! Actual test case output
#        Actual error             ! Actual test case error
#        Actual return            ! Actual test case return
#        Error information        ! Error information (A<::errorInfo> contents)
##end
#
#     The actual output, error or return appears when they did not match the
#  expected output, error or return values, respectively.  Notice that the
#  actual output may appear together with the actual error or return, but not
#  both.  The error information appears when an error is thrown from the test
#  case script.
#
# -returns    : The test cases results.
#
proc ::Simple::Test::results {
} {
   return $::Simple::Test::Results
}

### ===========================================================================
### -command  : ::Simple::Test::flags
# -synopsis   : Returns the test stored flags.
# -overview   :
#
#     This procedure returns a list containing the stored flags and their
#  contents for a test case as given to the corresponding P<test-case>
#  procedure.  Each element in the list is a list of two-element lists, being
#  the first a flag and the second its contents; this format is suitable to be
#  used with the P<array get> command.  As A<-regexp> is a boolean flag, its
#  contents is set to the empty string.
#
#     The stored flag are those of the A<-store> package option.  By default,
#  no flags are stored.  Nevertheless, if a test case fails the given items
#  among those in the A<-store> package option are stored along with the
#  A<-regexp> flag.  If the test case throws an error when no error was
#  expected, this procedure return an A<-error> flag with "NONE" as contents.
#
# -returns    : The test cases flags.
#
proc ::Simple::Test::flags {
} {
   return $::Simple::Test::Flags
}

### ===========================================================================
### -command  : ::Simple::Test::reset
# -synopsis   : Resets the test environment.
# -returns    : The empty string.
#
proc ::Simple::Test::reset {
} {
   ### Reset the test cases statistics, results and flags
   set ::Simple::Test::Results [list]
   set ::Simple::Test::Flags [list]
   set ::Simple::Test::NTests 0
   set ::Simple::Test::NTestsPassed 0
   set ::Simple::Test::NTestsFailed 0
   set ::Simple::Test::NTestsSkipped 0
   set ::Simple::Test::NTestsIgnored 0
   set ::Simple::Test::TestExpr 1

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Test::shutdown
# -synopsis   : Shutdowns the test environment.
# -effects    :
#
#  * Deletes the test interpreter.
#
# -returns    : The empty string.
#
proc ::Simple::Test::shutdown {
} {
   ### Delete the alias for K<puts> in the test interpreter
   interp alias $::Simple::Test::TestInterp puts {}
   interp eval $::Simple::Test::TestInterp\
      {rename ::Simple::Test::Priv::real-puts ::puts}

   ### Delete the test interpreter
   catch {interp delete $::Simple::Test::TestInterp}

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Test::configure
# -synopsis   : Configures the package options.
# -arguments  :
#  {-store            list {}        {Flags to store for each test case
#                                     tested, which can be later be obtained
#                                     via the K<::Simple::Test::flags>
#                                     procedure. If the given argument is
#                                     "all", then all options are stored.}}
#
# -returns    : The empty string.
#
proc ::Simple::Test::configure {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::AT-LEAST-ONE-FLAG\
         [::Simple::Test::cget]

   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      set arg2 [lindex $args 1]
      switch -exact -- $arg1 {

         -store {
         ### A<-store> given

            ### Invalid argument
            if {[catch {llength $arg2}]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -store list {}
            }

            ### Set the flags to store option
            if {[llength $arg2] == 1 && ![string compare $arg2 all]} {
               set arg2 $::Simple::Test::TestCaseFlags
            } else {
               foreach flag $arg2 {
                  if {![regexp {^-[-_a-zA-Z0-9%]+$} $flag]} {
                     ::Simple::Error::throw ::Simple::BAD-ITEM flag $flag
                  }
                  if {[lsearch -exact\
                     $::Simple::Test::TestCaseFlags $flag] == -1} {

                     ::Simple::Error::throw ::Simple::Test::BAD-FLAG-TO-STORE\
                        $flag
                  }
               }
            }
            set ::Simple::Test::%Options%(-store) $arg2
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG $arg1\
               [::Simple::Test::cget]
         }
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Test::cget
# -synopsis   : Gets the package options.
# -arguments  :
#  {-store            boolflag       {Gets the list of options to store for
#                                     each test case tested which can be later
#                                     be obtained via the
#                                     K<::Simple::Test::flags> procedure.
#                                     The returned list of options options is
#                                     ordered as in the input to the
#                                     corresponding P<test-case> procedure.}}
#
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Test::cget {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Return the list of options
      return -store

   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      switch -exact -- $arg1 {

         -store {
         ### A<-store> given
            set answer [list]
            foreach flag $::Simple::Test::TestCaseFlags {
               if {[lsearch -exact\
                  [set ::Simple::Test::%Options%(-store)] $flag] !=-1} {

                  lappend answer $flag
               }
            }
            return $answer
         }


         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG\
               $arg1 [::Simple::Test::cget]
         }
      }
   }
}
}
