### ===========================================================================
### -package  : SimpleType-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library types handling.
# -module     : extra
# -api        :
#
#  * K<::Simple::Type::modify description> type description
#    Modifies a type description.
#
#  * K<::Simple::Type::modify matchingscript> type matchingScript
#    Modifies a type matching script.
#
#  * K<::Simple::Type::delete> type
#    Deletes a type.
#
#  * K<::Simple::Type::information declared> ?typePattern?
#    Returns the list of declared types.
#
#  * K<::Simple::Type::information description> type
#    Returns a type description.
#
#  * K<::Simple::Type::information matchingscript> type
#    Returns a type matching script.
#
#  * K<::Simple::Type::convert decimal> integer
#    Converts an integer in any base to decimal.
#
#  * K<::Simple::Type::convert hexadecimal> integer
#    Converts an integer in any base to hexadecimal.
#
#  * K<::Simple::Type::convert binary> integer
#    Converts an integer in any base to binary.
#
#  * K<::Simple::Type::convert octal> integer
#    Converts an integer in any base to octal.
#
# -history    :
#
#  18-nov-2004   Added K<::Simple::Type::assert-is>, version 0.5.3
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleType-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Type::modify description
# -synopsis   : Modifies a type description.
# -arguments  :
#  { type             type           {Type name}}
#  { description      string         {New description}}
#
# -returns    : The empty string.
#
proc-sub {::Simple::Type::modify description} {
   type
   description
} {
   ### Assert the type is not derived
   ::Simple::Type::Priv::assert-not-derived $type

   ### Type exists
   if {[info exists ::Simple::Type::Description($type)]} {

      ### Set the type description
      set ::Simple::Type::Description($type)\
         [::Simple::Priv::transform-flag-argument $description paragraphs]

   ### Type does not exist
   } else {

      ### Assert the type name is valid
      ::Simple::Type::Priv::assert-valid-type $type

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING type $type
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::modify matchingscript
# -synopsis   : Modifies a type matching script.
# -arguments  :
#  { type             type           {Type name}}
#  { matchingScript   script         {New matching script}}
#
# -returns    : The empty string.
#
# -effects    :
#
#  * Recreates the type matching procedure named
#    K<::Simple::Type::Priv::%MATCHING%-''type''>
#
proc-sub {::Simple::Type::modify matchingscript} {
   type
   matchingScript
} {
   ### Assert the type is not derived
   ::Simple::Type::Priv::assert-not-derived $type

   ### Type exists
   if {[info exists ::Simple::Type::Description($type)]} {

      ### Recreate the matching procedure
      proc ::Simple::Type::Priv::%MATCHING%-$type {value {extra {}}}\
         [string trim $matchingScript]

   ### Type does not exist
   } else {

      ### Assert the type name is valid
      ::Simple::Type::Priv::assert-valid-type $type

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING type $type
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::delete
# -synopsis   : Deletes a type.
# -arguments  :
#  { type             type           {Type name}}
#
# -returns    : The empty string.
#
# -effects    :
#
#  * Deletes the type matching procedure named
#    K<::Simple::Type::Priv::%MATCHING%-''type''>
#
proc ::Simple::Type::delete {
   type
} {
   ### Delete the type
   ::Simple::Type::Priv::delete-with-hidden $type 0

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::delete-with-hidden
# -synopsis   : Deletes a type, also considering hidden types.
# -arguments  :
#  { type             type           {Type name}}
#  { hidden           boolean        {Whether to keep the type hidden}}
#
proc ::Simple::Type::Priv::delete-with-hidden {
   type
   hidden
} {
   ### Assert the type is not derived
   ::Simple::Type::Priv::assert-not-derived $type

   ### Type exists
   if {[info exists ::Simple::Type::Description($type)]} {

      ### Delete the type description and matching procedure
      unset ::Simple::Type::Description($type)
      rename ::Simple::Type::Priv::%MATCHING%-$type {}

   ### Type does not exist
   # Try with the hidden types
   } elseif {$hidden &&\
      [info exists ::Simple::Type::HiddenDescription($type)]} {

      ### Delete the hidden type description and matching procedure
      unset ::Simple::Type::HiddenDescription($type)
      rename ::Simple::Type::Priv::%MATCHING%-$type {}

   } else {

      ### Assert the type name is valid
      ::Simple::Type::Priv::assert-valid-type $type

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING type $type
   }
}

### ===========================================================================
### -command  : ::Simple::Type::information declared
# -synopsis   : Returns the list of declared types.
# -overview   :
#
#     This procedure returns the list of declared types matching the given
#  pattern.  This includes native The Simple Development Library types as well
#  as others declared via the K<::Simple::Type::declare> procedure.
#
# -arguments  :
#  {?typePattern?     pattern {}     {Pattern}}
#
# -returns    : The list of types matching the given pattern.
# -remarks    :
#
#  * Use no pattern to obtain the complete list of types.
#
proc-sub {::Simple::Type::information declared} {
   {typePattern {}}
} {
   if {[string compare $typePattern {}]} {
      return [array names ::Simple::Type::Description $typePattern]
   } else {
      return [array names ::Simple::Type::Description]
   }
}

### ===========================================================================
### -command  : ::Simple::Type::information description
# -synopsis   : Returns a type description.
# -arguments  :
#  { type             type           {Type name}}
#
# -returns    : The type description.
#
proc-sub {::Simple::Type::information description} {
   type
} {
   ### Assert the type is not derived
   ::Simple::Type::Priv::assert-not-derived $type

   ### Type exists
   if {[info exists ::Simple::Type::Description($type)]} {

      ### Return the type description
      return $::Simple::Type::Description($type)

   ### Type does not exist
   } else {

      ### Assert the type name is valid
      ::Simple::Type::Priv::assert-valid-type $type

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING type $type
   }
}

### ===========================================================================
### -command  : ::Simple::Type::information matchingscript
# -synopsis   : Returns a type matching script.
# -arguments  :
#  { type             type           {Type name}}
#
# -returns    : The type matching script.
#
proc-sub {::Simple::Type::information matchingscript} {
   type
} {
   ### Assert the type is not derived
   ::Simple::Type::Priv::assert-not-derived $type

   ### Type exists
   if {[info exists ::Simple::Type::Description($type)]} {

      ### Return the type matching script
      return [info body ::Simple::Type::Priv::%MATCHING%-$type]

   ### Type does not exist
   } else {

      ### Assert the type name is valid
      ::Simple::Type::Priv::assert-valid-type $type

      ### Throw error
      ::Simple::Error::throw ::Simple::NON-EXISTING type $type
   }
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::array-is
# -synopsis   : Returns whether an array elements conform to a type.
# -arguments  :
#  { arrayName        name           {The array name}}
#  { type             type           {Type name}}
#  { value            any            {Value}}
#  { list             boolean        {Whether the value is a list of the given
#                                     type}}
#  { range            boolean        {Whether the value is a range of the given
#                                     type}}
#  { extra            string         {Extra data required to check the
#                                     conformation}}
#
# -returns    : Whether the array elements conform to the type.
# -remarks    :
#
#  * The value ''%DEFAULT%'' conforms to all types
#
proc ::Simple::Type::Priv::array-is {
   arrayName
   type
   value
   list
   range
   extra
} {
   ### Link to the array in the calling scope
   upvar $arrayName localArray

   ### Foreach array element
   set answer 1
   foreach element [array names localArray] {

      ### Get the element value
      set value $localArray($element)

      ### Get whether the value conforms to the type
      if {![::Simple::Type::Priv::is $type $value $list $range $extra]} {

         # Short-circuit on the first non-conforming value
         set answer 0
         break
      }
   }

   ### Return whether the array elements conform to the type.
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::binary-to-decimal
# -synopsis   : Converts a binary integer to decimal.
# -arguments  :
#  { binary           binary         {Binary integer}}
#
# -returns    : The converted integer.
#
proc ::Simple::Type::Priv::binary-to-decimal {
   binary
} {
   ### Get rid of the prefix
   set binary [string range $binary 2 end]

   ### Compute the base
   set base [expr {1 << [string length $binary]}]

   ### Loop over binary digits
   # fixme: this should be reimplemented using
   #        K<binary scan> as present in Tcl 8.3
   set answer 0
   foreach digit [split $binary {}] {

      ### Reduce the base
      set base [expr {$base / 2}]

      ### Binary digit set
      if {$digit == 1} {
         incr answer $base
      }
   }

   ### Return the converted integer
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::convert decimal
# -synopsis   : Converts an integer in any base to decimal.
# -overview   :
#
#  Converts a decimal, hexadecimal, binary or octal integer to decimal base
#
# -keywords   : decimal hexadecimal binary octal base convert
# -arguments  :
#  { integer          anybase        {Positive integer}}
#
# -returns    : The converted integer.
#
proc-sub {::Simple::Type::convert decimal} {
   integer
} {
   ### Binary integer
   if {[::Simple::Type::Priv::is-basic binary $integer]} {

      ### Convert the binary integer
      set answer [::Simple::Type::Priv::binary-to-decimal $integer]

   ### Decimal, hexadecimal or octal integer
   } else {

      ### Convert the integer
      set answer [format %d $integer]
   }

   ### Negative integer
   if {$answer < 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Type::NEGATIVE-INT $integer
   }

   ### Return the converted integer
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::convert hexadecimal
# -synopsis   : Converts an integer in any base to hexadecimal.
# -overview   :
#
#     Converts a decimal, hexadecimal, binary or octal integer to hexadecimal
#  base.  The resulting hexadecimal constant is of the form
#  0x''hexadecimal-digits'' where ''hexadecimal-digits'' are a sequence of
#  A<uppercase> hexadecimal digits, that is, the "0x" prefix contains a
#  A<lowercase> "x" but the actual digits are uppercase.  This is different
#  from the result of the "%#X" K<format> specifier which returns an uppercase
#  "X".
#
# -keywords   : decimal hexadecimal binary octal base convert
# -arguments  :
#  { integer          anybase        {Positive integer}}
#
# -returns    : The converted integer.
#
proc-sub {::Simple::Type::convert hexadecimal} {
   integer
} {
   ### Binary integer
   set originalInteger $integer
   if {[::Simple::Type::Priv::is-basic binary $integer]} {

      ### Convert the binary integer
      set integer [::Simple::Type::Priv::binary-to-decimal $integer]
   }

   ### Negative integer
   if {$integer < 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Type::NEGATIVE-INT $originalInteger
   }

   ### Return the converted integer
   return [format 0x%X $integer]
}

### ===========================================================================
### -command  : ::Simple::Type::convert binary
# -synopsis   : Converts an integer in any base to binary.
# -overview   :
#
#     Converts a decimal, hexadecimal, binary or octal integer to binary base.
#  The resulting binary constant is of the form 0b''binary-digits''.
#  where ''binary-digits'' are a sequence of ''0''s and ''1''s.
#
# -keywords   : decimal hexadecimal binary octal base convert
# -arguments  :
#  { integer          anybase        {Positive integer}}
#
# -returns    : The converted integer.
#
proc-sub {::Simple::Type::convert binary} {
   integer
} {
   ### Binary integer
   set originalInteger $integer
   if {[::Simple::Type::Priv::is-basic binary $integer]} {

      ### Convert the binary integer to decimal
      set integer [::Simple::Type::Priv::binary-to-decimal $integer]

   ### Decimal, hexadecimal or octal integer
   } else {

      ### Convert the integer to decimal
      set integer [format %d $integer]
   }

   ### Negative integer
   if {$integer < 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Type::NEGATIVE-INT $originalInteger
   }

   ### Zero integer
   if {$integer == 0} {

      ### Binary digit not set
      set answer 0

   ### Greater than zero integer
   } else {

      ### Loop over binary digits
      # fixme: this should be reimplemented using
      #        K<binary format> as present in Tcl 8.3
      set answer {}
      while {$integer > 0} {

         ### Binary digit set
         if {[expr {$integer % 2}]} {
            set answer 1$answer

         ### Binary digit not set
         } else {
            set answer 0$answer
         }

         ### Reduce the integer
         set integer [expr {$integer / 2}]
      }
   }

   ### Return the converted integer
   return 0b$answer
}

### ===========================================================================
### -command  : ::Simple::Type::convert octal
# -synopsis   : Converts an integer in any base to octal.
# -overview   :
#
#  Converts a decimal, hexadecimal, binary or octal integer to octal base.
#
# -keywords   : decimal hexadecimal binary octal base convert
# -arguments  :
#  { integer          anybase        {Positive integer}}
#
# -returns    : The converted integer.
#
proc-sub {::Simple::Type::convert octal} {
   integer
} {
   ### Binary integer
   set originalInteger $integer
   if {[::Simple::Type::Priv::is-basic binary $integer]} {

      ### Convert the binary integer to decimal
      set integer [::Simple::Type::Priv::binary-to-decimal $integer]
   }

   ### Negative integer
   if {$integer < 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::Type::NEGATIVE-INT $originalInteger
   }

   ### Return the converted integer
   return [format 0%o $integer]
}
}
