### ===========================================================================
### -package  : SimpleType
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library types handling.
# -module     : implementation
# -overview   :
#
#     This package allows to handle types.  A type associates a type name to a
#  ''matching script'' used to assess whether a value conforms to the type.
#
#     Procedures are provided to declare (although a wealth of predefined
#  types are supplied) delete and modify types, get information about a type
#  and obtain the list of types, as well as several others to convert between
#  different types.  K<::Simple::Type::is> allows to assess whether a value
#  conforms to a type.
#
# -usage      :
#
#     Types are declared via K<::Simple::Type::declare> or embedded within
#  programs (K<declare-program>) or packages (K<declare-package>).  Types can
#  be deleted via K<::Simple::Type::delete> or modified via
#  K<::Simple::Type::modify>.  The K<::Simple::Type::information> procedure
#  provides several subcommands to query information about types:  K<exists>,
#  K<declared>, K<description> and K<matchingscript>.
#
#     The K<::Simple::Type::convert> set of procedures convert a value between
#  different types.  The K<::Simple::Type::is> procedure returns whether a
#  value conforms to a type.
#
#     There are two kind of types:  basic and derived.  Derived types are
#  derived from basic types and are denoted as ''basic type''-''derivation''
#  where ''basic type'' is the basic type name (such as integer) and
#  ''derivation'' is the kind of derivation.  All basic types can be derived
#  but ''boolflag''. Four derivations are currently supported:  arrays, pairs,
#  lists and ranges:
#
#   * Derived array types (''basic type''-array) denote the name of an array
#     whose elements conform to the type.  Non-existing arrays conform to all
#     types.
#
#   * Derived pair types (''basic type''-pairs) denote a list with an even
#     number of elements suitable to be given as argument to the K<array get>
#     command.  The second element in each pair conforms to the basic type.
#
#   * Derived list types (''basic type''-list) denote a list whose elements
#     conform to the basic type.
#
#   * Derived range types (''numeric''-range) denote a range whose limits
#     conform to the type which must be numeric, i.e., one of integer, float,
#     decimal, hexadecimal, binary, octal, anybase or number.  The form of the
#     range is ''lower limit:upper limit'' where the lower, the upper or even
#     both limits may be missing.  The colon may also be missing if just one
#     limit is given:  the range is restricted then to that exact value; this
#     also happens if both limits are equal.
#
#     The double derivation ''numeric''-range-list is also valid and
#  corresponds to a list of numeric ranges, such as ''{1 -2:3 :100}'' for the
#  integer-range-list type.  As the double derivation ''type''-list-array is
#  always valid for any type but ''boolflag'', it follows that the triple
#  derivation ''numeric''-range-list-array is also valid.
#
#     Derived types are only accepted by the K<::Simple::Type::is> and
#  K<::Simple::Type::information exists> procedures.
#
# -keywords   : type decimal hexadecimal binary octal base convert range array
# -api        :
#
#  * K<::Simple::Type::declare> type ?-description string?
#       ?-matchingscript script?
#    Declares a type.
#
#  * K<::Simple::Type::information exists> type
#    Returns whether a type exists.
#
#  * K<::Simple::Type::is> type value ?extra?
#    Returns whether a value conforms to a type.
#
#  * K<::Simple::Type::convert boolean> extBoolean
#    Converts an extended boolean to boolean.
#
#  * K<::Simple::Type::convert flag> flag
#    Converts a flag to an unqualified name.
#
#  * K<::Simple::Type::convert optional> optional
#    Converts an optional value to an unqualified name.
#
#  * K<::Simple::Type::outside-range> range value
#    Returns whether a numeric value is outside a numeric range.
#
#  * K<::Simple::Type::assert-is> type value ?extra? ?description?
#    Throws an error if a value does not conform to a type.
#
#  * K<::Simple::Type::configure>
#    Configures the package options.
#
#  * K<::Simple::Type::cget>
#    Gets the package options.
#
# -variable   :
#
#  {
#     ::Simple::Type::Description
#
#  } -description {
#
#     Array containing for each type its description
#
#  } -type string-array
#
# -variable   :
#
#  {
#     ::Simple::Type::HiddenDescription
#
#  } -description {
#
#     Array containing for each hidden type its description
#
#  }  -type string-array
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#  ::Simple::Package::require-and-install SimpleType
#
#  # Configure The Simple Development Library to store the metadata
#  ::Simple::configure -storemetadata true
#
#  # Declare a type bar which can only hold either bar or BAR
#  ::Simple::Type::declare {
#     bar
#  } -description {
#     can only hold either bar or BAR
#  } -matchingscript {
#     return [expr\ 
#        {![string compare $value bar] || ![string compare $value BAR]}]
#  }
#
#  # Assert the type bar exists
#  # This displays the following:
#  #   |The type bar exists
#  if {[::Simple::Type::information exists bar]} {
#     puts {The type bar exists}
#  }
#
#  # Get some information about the type bar
#  # This displays the following:
#  #   |The type bar can only hold either bar or BAR
#  puts "The type bar [::Simple::Type::information description bar]"
#
#  # Check the allowed contents for a variable of type bar
#  # This displays the following:
#  #   |"bar" is of type bar
#  #   |"foo" is NOT of type bar
#  #   |"BAR" is of type bar
#  #   |"bAR" is NOT of type bar
#  foreach contents {bar foo BAR bAR} {
#     puts -nonewline "\"$contents\" is "
#     if {![::Simple::Type::is bar $contents]} {
#        puts -nonewline {NOT }
#     }
#     puts {of type bar}
#  }
#
#  # Get rid of the type bar
#  ::Simple::Type::delete bar
#
#  # Assert the type bar no longer exists
#  # This displays the following:
#  #   |The type bar no longer exists
#  if {![::Simple::Type::information exists bar]} {
#     puts {The type bar no longer exists}
#  }
#
#  # More fun. A type made of three elements:
#  # an integer, a boolean and an alphabetic
#  ::Simple::Type::declare {
#     foo
#  } -description {
#     integer + boolean + alphabetic
#  } -matchingscript {
#     foreach localType [list integer boolean alphabetic] localValue $value {
#        if {![::Simple::Type::is $localType $localValue]} {
#           return 0
#        }
#     }
#     return 1
#  }
#
#  # Check the allowed contents for a variable of type foo
#  # This displays the following:
#  #   |"2 1 bar" is of type foo
#  #   |"2 x bar" is NOT of type foo
#  foreach contents {{2 1 bar} {2 x bar}} {
#     puts -nonewline "\"$contents\" is "
#     if {![::Simple::Type::is foo $contents]} {
#        puts -nonewline {NOT }
#     }
#     puts {of type foo}
#  }
#
#  # Derived list type
#  # This displays the following:
#  #   |"-2:1" is a proper integer range
#  #   |"2:1" is NOT a proper integer range
#  foreach contents {{-2:1} {2:1}} {
#     puts -nonewline "\"$contents\" is "
#     if {![::Simple::Type::is integer-range $contents]} {
#        puts -nonewline {NOT }
#     }
#     puts {a proper integer range}
#  }
#
#  # Derived range type
#  # This displays the following:
#  #   |"1.0 3 -1.E-2" is a proper list of floats
#  #   |"1.0 foo -2.1" is NOT a proper list of floats
#  foreach contents {{1.0 3 -1.E-2} {1.0 foo -2.1}} {
#     puts -nonewline "\"$contents\" is "
#     if {![::Simple::Type::is float-list $contents]} {
#        puts -nonewline {NOT }
#     }
#     puts {a proper list of floats}
#  }
#
# -details    :
#
#  * This package supports the concept of hidden types for internal use by The
#    Simple Development Library.  Hidden types are similar to regular types,
#    but they remain hidden to the public API.  The purpose is that packages
#    that define new types may use those types for other entities (variables,
#    attributes, commands arguments, ...) but the types remain hidden until
#    package installation.
#
# -todo       :
#
#  * Extend the concept of derived types to support arbitrary nested
#    extensions of lists such as integer-list-list.  Not only that, but
#    derivations such as integer-pairs-list or word-pairs-array should be
#    handled correctly.  The K<::Simple::Type::is> and
#    K<::Simple::Type::information exists> procedures should be adapted to
#    invoke themselves recursively for such multiple derivations.
#
#  * It would be nice to have a procedure which guesses the type of a given
#    value.
#
#  * The basic name regular expression (''[_a-zA-Z0-9%][-_a-zA-Z0-9%]*'')
#    should be a package option.  All type matching scripts requiring it
#    should reference the package option.  Alternatively, for maximum
#    performance, upon changing this option, all types matching scripts using
#    it should be recreated.
#
# -history    :
#
#  19-feb-1999   Unreleased first version 0.1
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  29-mar-2003   Extra package, version 0.5.1
#  09-jan-2004   Derived type range accepts equal limits, version 0.5.2
#  18-nov-2004   Added K<::Simple::Type::assert-is>, version 0.5.3
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleType 1.0 -requisites {

   {SimpleSubcommand 1.0}

} -namespaces {

   ::Simple::Type
   ::Simple::Type::Priv

} -install {

### Create the package variables

### Create the package options

} -uninstall {

### Delete the package variables

### Delete the package options

} -firsttimeinstall {

### Declare the package types
#   Utility procedure to declare the package types
# fixme: many of the below matching scripts should be reimplemented using
#        K<string is> as present in Tcl 8.3
# fixme: types should include an A<-examples> flag including which may include
#        some examples of valid values for the type for documentation purposes
proc ::Simple::Type::Priv::declare-types {
} {

###    Numbers

# Integer number: ''+1'', ''0'', ''-192'', ...
::Simple::Type::declare {
   integer
} -description {
   Integer number: ''+1'', ''0'', ''-192'', ...
} -matchingscript {
   return [regexp {^[-+]?(0|[1-9][0-9]*)$} $value]
}

# Floating point number: ''-1.2e23'', ''+3.0'', ...
::Simple::Type::declare {
   float
} -description {
   Floating point number: ''-1.2e23'', ''+3.0'', ...
} -matchingscript {
   return [regexp\
      {^[-+]?(0|[1-9][0-9]*|[0-9]*\.[0-9]+|[0-9]+\.[0-9]*)([eE][-+]?[0-9]+)?$}\
      $value]
}

# Decimal number: ''1234'', ...
::Simple::Type::declare {
   decimal
} -description {
   Decimal number: ''1234'', ...
} -matchingscript {
   return [regexp {^(0|[1-9][0-9]*)$} $value]
}

# Hexadecimal number: ''xF12'', ...
::Simple::Type::declare {
   hexadecimal
} -description {
   Hexadecimal number: ''xF12'', ...
} -matchingscript {
   return [regexp {^(0[xX][0-9a-fA-F]+)$} $value]
}

# Binary number: ''b01011'', ...
::Simple::Type::declare {
   binary
} -description {
   Binary number: ''b01011'', ...
} -matchingscript {
   return [regexp {^(0[bB][01]+)$} $value]
}

# Octal number: ''0723'', ...
::Simple::Type::declare {
   octal
} -description {
   Octal number: ''0723'', ...
} -matchingscript {
   return [regexp {^0[0-7]*$} $value]
}

# Decimal, hexadecimal, binary or octal number
::Simple::Type::declare {
   anybase
} -description {
   Decimal, hexadecimal, binary or octal number
} -matchingscript {
   return [regexp\
      {^((0|[1-9][0-9]*)|(0[xX][0-9a-fA-F]+)|(0[bB][01]+)|0[0-7]*)$}\
      $value]
}

# Any number
::Simple::Type::declare {
   number
} -description {
   Any number
} -matchingscript {

   # Binary or number valid for K<expr>
   if {![regexp {^(0[bB][01]+)$} $value]} {
      expr {0 + $value}
   }
   return 1
}

###    Strings and characters

# Character
::Simple::Type::declare {
   character
} -description {
   Character
} -matchingscript {
   return [expr {[string length $value] == 1}]
}

# String
::Simple::Type::declare {
   string
} -description {
   String
}

# Alphabetic (letters only)
::Simple::Type::declare {
   alphabetic
} -description {
   Alphabetic (letters only)
} -matchingscript {
   return [regexp {^[a-zA-Z]+$} $value]
}

# Alphanumeric (letters and digits)
::Simple::Type::declare {
   alphanumeric
} -description {
   Alphanumeric (letters and digits)
} -matchingscript {
   return [regexp {^[a-zA-Z0-9]+$} $value]
}

###    Boolean

# Boolean: ''0'' or ''1''
::Simple::Type::declare {
   boolean
} -description {
   Boolean: ''0'' or ''1''
} -matchingscript {
   return [regexp ^(0|1)$ $value]
}

###    Tcl/Tk specific

# Tcl extended boolean: ''0'', ''1'', ''true'', ''false'', ...
::Simple::Type::declare {
   extbool
} -description {
   Tcl extended boolean: ''0'', ''1'', ''true'', ''false'', ...
} -matchingscript {
   set boolean [{::Simple::Type::convert boolean} $value]
   return [regexp ^(0|1)$ $boolean]
}

# Tcl word
::Simple::Type::declare {
   word
} -description {
   Tcl word
} -matchingscript {
   return [expr {[string wordend $value 0] == [string length $value]}]
}

# Tcl script
::Simple::Type::declare {
   script
} -description {
   Tcl script
} -matchingscript {
   llength $value
   return 1
}

# Tcl expression: ''2 + 3'', ...
::Simple::Type::declare {
   expression
} -description {
   Tcl expression: ''2 + 3'', ...
} -matchingscript {
   expr $value
   return 1
}

# Tcl list
::Simple::Type::declare {
   list
} -description {
   Tcl list
} -matchingscript {
   llength $value
   return 1
}

# Tcl channel: ''stdin'', ''file0'', ''sock2'', ...
::Simple::Type::declare {
   channel
} -description {
   Tcl channel: ''stdin'', ''file0'', ''sock2'', ...
} -matchingscript {
   return [regexp {^(std(in|out|err)|(file|sock|console)[0-9]+)} $value]
}

# Tcl namespace: ''::foo::bar'', ...
::Simple::Type::declare {
   namespace
} -description {
   Tcl namespace: ''::foo::bar'', ...
} -matchingscript {
   return [regexp {^(::|[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)+$} $value]
}

# Tcl qualified name: ''::foo::Bar'', ''::foo::bar'', ...
::Simple::Type::declare {
   qualifiedname
} -description {
   Tcl qualified name: ''::foo::Bar'', ''::foo::bar'', ...
} -matchingscript {
   return [regexp {^((::[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)+)$} $value]
}

# Tcl unqualified name: ''foo'', ''Bar'', ...
::Simple::Type::declare {
   unqualifiedname
} -description {
   Tcl unqualified name: ''foo'', ''Bar'', ...
} -matchingscript {
   return [regexp {^[_a-zA-Z0-9%][-_a-zA-Z0-9%]*$} $value]
}

# Tcl qualified or unqualified name: ''foo'', ''::Bar'', ...
::Simple::Type::declare {
   name
} -description {
   Tcl qualified or unqualified name: ''foo'', ''::Bar'', ...
} -matchingscript {
   return [regexp {^((::)?[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)+$} $value]
}

# Tcl command name with optional subcommand: ''foo bar'', ''::Gee zuu'', ...
::Simple::Type::declare {
   commandname
} -description {
   Tcl command name with optional subcommand: ''foo bar'', ''::Gee zuu'', ...
} -matchingscript {
   return [regexp {^((::)?[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)+( [_a-zA-Z0-9%][-_a-zA-Z0-9%]*)?$} $value]
}

# Tcl command name with optional subcommand: ''foo ba''r, ''Gee zu''u, ...
::Simple::Type::declare {
   unqualifiedcommandname
} -description {
   Tcl command name with optional subcommand: ''foo ba''r, ''Gee zu''u, ...
} -matchingscript {
   return [regexp {^[_a-zA-Z0-9%][-_a-zA-Z0-9%]*( [_a-zA-Z0-9%][-_a-zA-Z0-9%]*)?$} $value]
}

# Tcl qualified variable name: ''::foo::Bar'', ''::foo::bar(gee)'', ...
::Simple::Type::declare {
   qualifiedvariablename
} -description {
   Tcl qualified variable name: ''::foo::Bar'', ''::foo::bar(gee)'', ...
} -matchingscript {
   return [regexp {^((::[_a-zA-Z0-9%][-_a-zA-Z0-9%]*(\([_a-zA-Z0-9%][-_a-zA-Z0-9%]*\))?)+)$} $value]
}

# Tcl unqualified variable name: ''foo'', ''Bar(gee)'', ...
::Simple::Type::declare {
   unqualifiedvariablename
} -description {
   Tcl unqualified variable name: ''foo'', ''Bar(gee)'', ...
} -matchingscript {
   return [regexp {^[_a-zA-Z0-9%][-_a-zA-Z0-9%]*(\([_a-zA-Z0-9%][-_a-zA-Z0-9%]*\))?$} $value]
}

# Tcl qualified or unqualified variable name: ''foo'', ''::Bar(gee)'', ...
::Simple::Type::declare {
   variablename
} -description {
   Tcl qualified or unqualified variable name: ''foo'', ''::Bar(gee)'', ...
} -matchingscript {
   return [regexp {^((::)?[_a-zA-Z0-9%][-_a-zA-Z0-9%]*(\([_a-zA-Z0-9%][-_a-zA-Z0-9%]*\))?)+$} $value]
}

# Tcl qualified or unqualified variable or command name:
#    ''::Bar(gee)'', ''bar gee'', ...
::Simple::Type::declare {
   variableorcommandname
} -description {
   Tcl qualified or unqualified variable or command name: ''::Bar(gee)'', ''bar gee'', ...
} -matchingscript {
   return [regexp {^(((::)?[_a-zA-Z0-9%][-_a-zA-Z0-9%]*(\([_a-zA-Z0-9%][-_a-zA-Z0-9%]*\))?)+|((::)?[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)+( [_a-zA-Z0-9%][-_a-zA-Z0-9%]*)?)$} $value]
}

# Tcl pattern: ''foo*bar'', ...
::Simple::Type::declare {
   pattern
} -description {
   Tcl pattern: ''foo*bar'', ...
} -matchingscript {
   string match $value {}
   return 1
}

# Tcl regular expression: ''[A-Z]+'', ...
::Simple::Type::declare {
   regexp
} -description {
   Tcl regular expression: ''[A-Z]+'', ...
} -matchingscript {
   regexp $value {}
   return 1
}

# Tcl format string
::Simple::Type::declare {
   formatstring
} -description {
   Tcl format string
} -matchingscript {
   format [::Simple::Error::Priv::transform-format-string $value {}]
   return 1
}

# Tk widget: ''.foo.bar'', ...
::Simple::Type::declare {
   widget
} -description {
   Tk widget: ''.foo.bar'', ...
} -matchingscript {
   return [regexp {^(\.|(\.[_a-zA-Z0-9%][-_a-zA-Z0-9%]*)+)$} $value]
}

###    Simple Development Library types

# Type name: ''integer'', ...
::Simple::Type::declare {
   type
} -description {
   Type name: ''integer'', ...
} -matchingscript {
   ::Simple::Type::Priv::assert-valid-type $value
   return 1
}

# Flag: ''-name'', ...
::Simple::Type::declare {
   flag
} -description {
   Flag: ''-name'', ...
} -matchingscript {
   return [regexp {^-(-|(-)?[_a-zA-Z0-9%]+)$} $value]
}

# Optional variable: ''?name?'', ...
::Simple::Type::declare {
   optional
} -description {
   Optional variable: ''?name?'', ...
} -matchingscript {
   return [regexp {^\?[_a-zA-Z0-9%][-_a-zA-Z0-9%]*\?$} $value]
}

# Item declaration: a list containing a class, command, ... declaration
::Simple::Type::declare {
   declaration
} -description {
   Item declaration: a list containing a class, command, ... declaration
} -matchingscript {
   llength $value
   return 1
}

###    Other types

# Date: ''Fri Feb 03 19:09:35 WET 1984'', ...
::Simple::Type::declare {
   date
} -description {
   Date: ''Fri Feb 03 19:09:35 WET 1984'', ...
} -matchingscript {
   clock scan $value
   return 1
}

###    Special types

# Choice list: ''{one two three}'', ...
::Simple::Type::declare {
   choice
} -description {
   Choice list: ''{one two three}'', ...
} -matchingscript {
   return [expr {[lsearch -exact $extra $value] != -1}]
}

# Any value
::Simple::Type::declare {
   any
} -description {
   Any value
}

# Boolean flag (see the P<SimpleProc> package)
::Simple::Type::declare {
   boolflag
} -description {
   Boolean flag
} -matchingscript {
   return 0
}

# URL: ''http://wiki.tcl.tk''
::Simple::Type::declare {
   url
} -description {
   URL: ''http://wiki.tcl.tk''
} -matchingscript {
   return [regexp {^([-\w]+://([-\w]+\.)*[-\w]+(/([-.~?%/\w]+))?(#[-.\w]+|/)?|#[-.\w]+)$} $value]
}

# Declare previous packages types
::Simple::Error::Priv::declare-types
::Simple::Package::Priv::declare-types

# Delete this procedure
rename [info level 0] {}

}

### Declare the package errors
::Simple::Error::declare {
   ::Simple::Type::BAD-DERIVED-TYPE 
} -message {
   invalid derived type "%s"
} -explanation {
   The derived type <derived type> is invalid.  All basic types can be derived
   but "boolflag".  The only derived range types valid are those in which the
   basic type is one of integer, float, decimal, hexadecimal, binary, octal,
   anybase or number
}

::Simple::Error::declare {
   ::Simple::Type::DERIVED-TYPE-OUT-OF-CONTEXT 
} -message {
   a derived type such as "%s" is not valid in this context
} -explanation {
   <derived type> is a derived type, but these types are valid for the
   K<::Simple::Type::is> and K<::Simple::Type::information exists> procedures
   only
}

::Simple::Error::declare {
   ::Simple::Type::NEGATIVE-INT 
} -message {
   negative integer "%s"
} -explanation {
   The integer <integer> is negative but a positive integer was expected
}

::Simple::Error::declare {
   ::Simple::Type::ALREADY-EXISTS
} -message {
   type "%s" already exists
} -explanation {
   Type <type> could not be created because it already exists
} -corrective {
   Delete the type.
}

### ===========================================================================
### -command  : ::Simple::Type::declare
# -synopsis   : Declares a type.
# -overview   :
#
#  This procedure is used to declare a new type.
#
#  A type is made of three elements:  
#
#     1. the type name, 
#
#     1. the type description and
#
#     1. type matching script.
#
#     The type name is an all-lowercase string.  The type matching script is a
#  Tcl script used to check whether a value conforms to the type.  It shall
#  assume that the value to be checked is passed in the V<value> variable, and
#  shall return either ''1'' or ''0'' depending on whether the value conforms
#  to the type or not; throwing an error is also a valid action for
#  non-conforming values.  It gets evaluated as a procedure, so use K<upvar>
#  and K<uplevel> to access the current stack level if needed.
#
#  A type with no empty matching script may hold any value.
#
# -arguments  :
#  { type             type           {Type name}}
#  {-description      string {}      {Type description}}
#  {-matchingscript   script {return 1}
#                                    {Type matching script}}
#
# -returns    : The empty string.
#
# -examples   :
#
#  # Declare a type "foo" which may hold any value
#  ::Simple::Type::declare foo
#
#  # Declare a type bar which can only hold either bar or BAR
#  ::Simple::Type::declare -matchingscript {
#     return\
#        [expr {![string compare $value bar] || ![string compare $value BAR]}]
#  }
#
# -effects    :
#
#  * Creates the type matching procedure named
#    K<::Simple::Type::Priv::%MATCHING%-''type''>.
#
# -remarks    :
#
#  * The order of the flags is irrelevant but their parameters can not start
#    by an hyphen.
#
proc ::Simple::Type::declare {
   type
   args
} {
   ### Parse and verify the type declaration
   set type [string trim $type]
   ::Simple::Error::Priv::push-location type $type 1
   if {[catch {
      set declarationScript\
         [::Simple::Type::Priv::parse-and-verify-declaration $type $args 0]
   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Declare the type
   eval $declarationScript

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::parse-and-verify-declaration
# -synopsis   : Parses and verifies a type declaration.
# -arguments  :
#  { type             type           {Type name}}
#  { actualArgs       list           {Actual type declaration argument list}}
#  { hidden           boolean        {Whether to keep the type hidden}}
#
# -returns    : The type declaration script.
#
proc ::Simple::Type::Priv::parse-and-verify-declaration {
   type
   actualArgs
   hidden
} {
   ### Assert the type is not derived
   ::Simple::Type::Priv::assert-not-derived $type

   ### Assert the type does not exist
   if {[::Simple::Type::Priv::exists-with-hidden $type 0]} {
      ::Simple::Error::throw ::Simple::Type::ALREADY-EXISTS $type
   }

   ### Set defaults
   set matchingscript {return 1}

   ### Parse flags
   ::Simple::Priv::parse-flags\
      [list -description paragraphs\
      -matchingscript trimmed]\
      [list] $actualArgs 0 2

   ### Compose the type declaration script
   set declarationScript [list ::Simple::Type::Priv::do-declare\
      $type $description $matchingscript $hidden]

   ### Return the type declaration script
   return $declarationScript
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::do-declare
# -synopsis   : Declares a type.
# -arguments  :
#  { type             type           {Type name}}
#  { description      string         {Type description}}
#  { matchingScript   script         {Type matching script}}
#  { hidden           boolean        {Whether to keep the type hidden}}
#
# -returns    : The empty string.
#
# -effects    :
#
#  * Creates the type matching procedure named
#    K<::Simple::Type::Priv::%MATCHING%-''type''>.
#
proc ::Simple::Type::Priv::do-declare {
   type
   description
   matchingScript
   hidden
} {
   # The type descriptions must not be stored
   if {![set ::Simple::%Options%(-storemetadata)]} {

      # Do not store the type description
      set description {}
   }

   ### Store the type internal representation
   if {$hidden} {
      set ::Simple::Type::HiddenDescription($type) $description
   } else  {
      set ::Simple::Type::Description($type) $description
   }

   ### Create the matching procedure
   proc ::Simple::Type::Priv::%MATCHING%-$type\
      {value {extra {}}} $matchingScript

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::assert-not-derived
# -synopsis   : Throws an error if a type name is derived.
# -arguments  :
#  { type             type           {Type name}}
#
proc ::Simple::Type::Priv::assert-not-derived {
   type
} {
   if {[regexp {^(.*)-(list|array|range)$} $type]} {
      ::Simple::Error::throw ::Simple::Type::DERIVED-TYPE-OUT-OF-CONTEXT $type
   }
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::assert-valid-type
# -synopsis   : Throws an error if a type name is not valid.
# -arguments  :
#  { type             type           {Type name}}
#
proc ::Simple::Type::Priv::assert-valid-type {
   type
} {
   if {![regexp {^[a-z0-9]+$} $type]} {
      ::Simple::Error::throw ::Simple::BAD-ITEM {type name} $type
   }
}

### ===========================================================================
### -command  : ::Simple::Type::information exists
# -synopsis   : Returns whether a type exists.
# -overview   :
#
#     This procedure returns whether a type exists.  This includes native The
#  Simple Development Library types as well as others declared via the
#  K<::Simple::Type::declare> procedure.  For derived types, the procedure
#  returns whether the basic type exists.
#
# -arguments  :
#  { type             type           {Type name}}
#
# -returns    : Whether the type exists.
#
proc-sub {::Simple::Type::information exists} {
   type
} {
   return [::Simple::Type::Priv::exists-with-hidden $type 0]
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::exists-with-hidden
# -synopsis   : Returns whether a type exists, also considering hidden types.
#
# -arguments  :
#  { type             type           {Type name}}
#  { hidden           boolean        {Whether to also consider hidden types}}
#
# -returns    : Whether the type exists.
#
proc ::Simple::Type::Priv::exists-with-hidden {
   type
   hidden
} {
   ### Type not derived from boolflag
   if {![string match boolflag-* $type]} {

      ### Remove a possible trailing "-array" in the type name
      regexp {^(.*)-array$} $type -> type

      ### Remove a possible trailing "-pairs" in the type name
      regexp {^(.*)-pairs$} $type -> type

      ### Remove a possible trailing "-list" in the type name
      regexp {^(.*)-list$} $type -> type

      ### Derived range type
      if {[regexp {^(.*)-range$} $type -> type]} {

         ### Return whether the basic type accepts a range derivation
         set answer [expr {[lsearch -exact [list integer float decimal\
            hexadecimal binary octal anybase number] $type] != -1}]

      ### Not derived range type
      } else {

         ### Assert the type name is valid
         ::Simple::Type::Priv::assert-valid-type $type

         ### Return whether the type exists
         set answer [info exists ::Simple::Type::Description($type)]

         # Try with the hidden types
         if {$hidden & !$answer} {
            set answer [info exists ::Simple::Type::HiddenDescription($type)]
         }
      }

   ### Type derived from boolflag
   } else {

      ### boolflag can not be derived
      set answer 0
   }

   ### Return whether the type exists
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::is
# -synopsis   : Returns whether a value conforms to a type.
# -overview   :
#
#     This procedure returns whether a value conforms to a type.
#
#     If the type is a derived list type, the value is considered a list of
#  values, and the procedure returns whether all of them conform to the basic
#  type.  Empty lists conform to all types.
#
#     If the type is a derived array type, the value is considered the name of
#  an array of values in the calling scope, and the procedure returns whether
#  all of them conform to the basic type.  Non-existing arrays conform to
#  all types.
#
#     If the type is a derived range type, the value is considered a range
#  whose limits are of the basic type.
#
# -keywords   : type
# -arguments  :
#  { type             type           {Type name}}
#  { value            any            {Value or list of values}}
#  {?extra?           string {}      {Extra data required to check the
#                                     conformation}}
#
# -returns    : Whether the value conforms to the type.
# -remarks    :
#
#  * The value ''%DEFAULT%'' conforms to all types
#
proc ::Simple::Type::is {
   type
   value
   {extra {}}
} {
   return [::Simple::Type::Priv::is-with-hidden $type $value 0 $extra 2]
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::is-with-hidden
# -synopsis   : Returns whether a value conforms to a type, considering hidden types.
# -arguments  :
#  { type             type           {Type name}}
#  { value            any            {Value or list of values}}
#  { hidden           boolean        {Whether to also consider hidden types}}
#  { extra            string         {Extra data required to check the
#                                     conformation}}
#  {?levels?          integer 1      {Levels up the stack in which to operate}}
#
# -returns    : Whether the value values conforms to the type.
#
proc ::Simple::Type::Priv::is-with-hidden {
   type
   value
   hidden
   extra
   {levels 1}
} {
   ### Assert the type is not derived from boolflag
   if {[string match boolflag-* $type]} {
      ::Simple::Error::throw ::Simple::Type::BAD-DERIVED-TYPE $type
   }

   ### Asses whether the type is a derived array type
   set array [regexp {^(.*)-array$} $type -> type]

   ### Asses whether the type is a derived pairs type
   set pairs [regexp {^(.*)-pairs$} $type -> type]

   ### Asses whether the type is a derived list type
   set list [regexp {^(.*)-list$} $type -> type]

   ### Assess whether the type is a derived range
   ###    type and, if so, assert it is valid
   set range [regexp {^(.*)-range$} $type -> type]
   if {$range && [lsearch -exact [list integer float decimal hexadecimal\
      binary octal anybase number] $type] == -1} {

      ::Simple::Error::throw ::Simple::Type::BAD-DERIVED-TYPE $type-range
   }

   ### Type does not exist
   if {![info exists ::Simple::Type::Description($type)]} {

      ### Type is not hidden
      # Try with the hidden types
      if {!$hidden || ![info exists ::Simple::Type::HiddenDescription($type)]} {

         ### Assert the type name is valid
         ::Simple::Type::Priv::assert-valid-type $type

         ### Throw error
         ::Simple::Error::throw ::Simple::NON-EXISTING type $type
      }
   }

   ### The value is ''%DEFAULT%''
   if {![string compare $value %DEFAULT%]} {

      ### The value ''%DEFAULT%'' conforms to all types
      set answer 1

   ### The value is not ''%DEFAULT%''
   } else {

      ### Derived array type
      if {$array} {

         ### Link to the array in the calling scope
         upvar $levels $value localArray

         ### The array exists 
         if {[array exists localArray]} {

            ### Get whether the array elements conform to the type
            set answer [::Simple::Type::Priv::array-is\
               localArray $type $value $list $range $extra]

         ### The array does not exist
         } else {

            ### It is a regular variable
            if {[info exists localArray]} {

               ### No conformation
               set answer 0

            ### It is not a regular variable
            } else {

               ### Non-existing arrays conform to all types
               set answer 1
            }
         }

      ### Derived pairs type
      } elseif {$pairs} {

         ### Error upon etting the array
         if {[catch {array set localArray $value}]} {

            ### No conformation
            set answer 0

         ### No error upon setting the array
         } else {

            ### Get whether the array elements conform to the type
            set answer [::Simple::Type::Priv::array-is\
               localArray $type $value $list $range $extra]
         }

      ### Non-derived array nor pairs type
      } else {

         ### Get whether the value conforms to the type
         set answer [::Simple::Type::Priv::is $type $value $list $range $extra]
      }
   }

   ### Return whether the value conforms to the type
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::is-basic
# -synopsis   : Returns whether a value conforms to a basic type.
# -arguments  :
#  { type             type           {Type name}}
#  { value            any            {Value}}
#  {?range?           boolean 0      {Whether the value is a range of the given
#                                     type}}
#  {?extra?           string {}      {Extra data required to check the
#                                     conformation}}
#
# -returns    : Whether the value conforms to the type.
# -remarks    :
#
#  * The value ''%DEFAULT%'' conforms to all types
#
proc ::Simple::Type::Priv::is-basic {
   type
   value
   {range 0}
   {extra {}}
} {
   ### Derived range type
   if {$range} {

      ### Switch on range
      switch -exact -- $value {

         {} {
         ### Empty range

            ### Return invalid range
            set answer 0
         }

         : {
         ### Range is ":"

            ### Return valid range
            set answer 1
         }

         default {
         ### default

            ### Get the lower and upper range limits
            foreach {lower upper tmp} [split $value :] break
            set lowerNotEmpty [string compare $lower {}]
            set upperNotEmpty [string compare $upper {}]

            ### The range has no more than two limits and all
            ###    limits conform to the range type
            if {!([string compare $tmp {}] ||\
               ($lowerNotEmpty && ![::Simple::Type::Priv::is-basic $type $lower]) ||\
               ($upperNotEmpty && ![::Simple::Type::Priv::is-basic $type $upper]))} {

               ### Both limits given
               if {$lowerNotEmpty && $upperNotEmpty} {

                  ### The range type is not float nor integer
                  if {[string compare $type float] &&\
                     [string compare $type integer]} {

                     ### The range type is not number
                     if {[string compare $type number]} {

                        ### Convert the limits to decimal
                        set lower [{::Simple::Type::convert decimal} $lower]
                        set upper [{::Simple::Type::convert decimal} $upper]

                     ### The range type is number
                     } {

                        ### Convert the limits to decimal if they are not float
                        if {![::Simple::Type::Priv::is-basic float $lower]} {
                           set lower [{::Simple::Type::convert decimal} $lower]
                        }
                        if {![::Simple::Type::Priv::is-basic float $upper]} {
                           set upper [{::Simple::Type::convert decimal} $upper]
                        }
                     }
                  }

                  ### Return whether the lower limit is actually lower
                  ###    than the upper one or equal to it
                  set answer [expr {$lower <= $upper}]

               ### Both limits not given
               } else {

                  ### Return valid range
                  set answer 1
               }

            ### The range has more than two limits or at least one
            ###    limit does not conform to the range type
            } else {

               ### Return invalid range
               set answer 0
            }
         }
      }

   ### Not derived range type
   } else {

      ### Evaluate the type matching script
      #   The K<uplevel 2> here ensures the matching
      #   script is evaluated in the proper stack level
      set script [list ::Simple::Type::Priv::%MATCHING%-$type $value $extra]
      if {[catch {set answer [uplevel 2 $script]}]} {

         # Error evaluating the matching script, return no conformation
         set answer 0
      }
   }

   ### Return whether the value conforms to the type
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::is
# -synopsis   : Returns whether a value conforms to a type.
# -arguments  :
#  { type             type           {Type name}}
#  { value            any            {Value}}
#  { list             boolean        {Whether the value is a list of the given
#                                     type}}
#  { range            boolean        {Whether the value is a range of the given
#                                     type}}
#  { extra            string         {Extra data required to check the
#                                     conformation}}
#
# -returns    : Whether the value conforms to the type.
# -remarks    :
#
#  * The value ''%DEFAULT%'' conforms to all types
#
proc ::Simple::Type::Priv::is {
   type
   value
   list
   range
   extra
} {
   ### Derived list type
   if {$list} {

      ### Non-empty list
      set answer 1
      if {[llength $value] > 0} {

         ### Loop over values
         foreach elementValue $value {

            ### Return whether all values conform to the type
            if {![::Simple::Type::Priv::is-basic\
               $type $elementValue $range $extra]} {

               # Short-circuit on the first non-conforming value
               set answer 0
               break
            }
         }
      }

   ### Single value
   } else {

      ### Get whether the value conforms to the type
      set answer [::Simple::Type::Priv::is-basic $type $value $range $extra]
   }

   ### Return whether the value conforms to the type
   return $answer
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::unhide
# -synopsis   : Unhides types.
# -arguments  :
#  { types            type-list      {Types}}
#
# -returns    : The empty string.
#
proc ::Simple::Type::Priv::unhide {
   types
} {
   foreach type $types {
      set ::Simple::Type::Description($type)\
         $::Simple::Type::HiddenDescription($type)
      unset ::Simple::Type::HiddenDescription($type)
   }
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::Priv::hide
# -synopsis   : Hides types.
# -arguments  :
#  { types            type-list      {Types}}
#
# -returns    : The empty string.
#
proc ::Simple::Type::Priv::hide {
   types
} {
   foreach type $types {
      set ::Simple::Type::HiddenDescription($type)\
         $::Simple::Type::Description($type)
      unset ::Simple::Type::Description($type)
   }
   return {}
}

### ===========================================================================
### -command  : ::Simple::Type::convert boolean
# -synopsis   : Converts an extended boolean to boolean.
# -overview   :
#
#     This procedure converts from extended booleans (''true''/''false'',
#  ''yes''/''no'' or ''on''/''off'') to ''1''/''0'' because the Tcl K<expr>
#  and K<if> commands ''can't use non-numeric string as operand of "!"''
#  (sic).
#
# -keywords   : boolean convert
# -arguments  :
#  { extBoolean       extbool        {Extended boolean}}
#
# -returns    : The converted boolean.
# -examples   :
#
#  # A robust negation test
#  if {![::Simple::Type::convert boolean $extbool]} {
#     puts "$extbool is false"
#  }
#
proc-sub {::Simple::Type::convert boolean} {
   extBoolean
} {
   if {$extBoolean} {
      return 1
   } else {
      return 0
   }
}

### ===========================================================================
### -command  : ::Simple::Type::convert flag
# -synopsis   : Converts a flag to an unqualified name.
# -arguments  :
#  { flag             flag           {Flag}}
#
# -returns    : The flag name.
#
proc-sub {::Simple::Type::convert flag} {
   flag
} {
   ### Assert the flag name is valid
   ::Simple::Type::assert-is flag $flag {} {flag name}

   ### Return the flag name
   return [string trimleft $flag -]
}

### ===========================================================================
### -command  : ::Simple::Type::convert optional
# -synopsis   : Converts an optional value to an unqualified name.
# -arguments  :
#  { optional         optional       {Optional value}}
#
# -returns    : The optional value name.
#
proc-sub {::Simple::Type::convert optional} {
   optional
} {
   ### Assert the optional value name is valid
   ::Simple::Type::assert-is optional $optional {} {optional value name}

   ### Return the optional value name
   return [string trim $optional ?]
}

### ===========================================================================
### -command  : ::Simple::Type::outside-range
# -synopsis   : Returns whether a numeric value is outside a numeric range.
# -arguments  :
#  { range            number-range   {Numeric range}}
#  { value            number         {Numeric value}}
#
# -returns    :  
#
#     ''-1'', ''0'' or ''1'', depending on whether the value is lower than the
#  lower range limit, within the range or greater than the upper range limit.
#
proc ::Simple::Type::outside-range {
   range
   value
} {
   ### Get the lower and upper range limits
   foreach {lower upper tmp} [split $range :] break
   set lowerNotEmpty [string compare $lower {}]
   set upperNotEmpty [string compare $upper {}]

   ### The range has no more than two limits and all limits are numbers
   if {!([string compare $tmp {}] ||\
      ($lowerNotEmpty && ![::Simple::Type::Priv::is-basic number $lower]) ||\
      ($upperNotEmpty && ![::Simple::Type::Priv::is-basic number $upper]))} {

      ### Assert the value is a number
      ::Simple::Type::assert-is number $value {} {numeric value}

      ### Convert binary integers to decimal
      foreach numberName {lower upper value} {
         set number [set $numberName]
         if {[::Simple::Type::Priv::is-basic binary $number]} {
            set $numberName [::Simple::Type::Priv::binary-to-decimal $number]
         }
      }

      ### No limits given
      if {!$lowerNotEmpty && !$upperNotEmpty} {

         ### Return value outside the range
         set answer 0

      ### Both limits given
      } elseif {$lowerNotEmpty && $upperNotEmpty} {

         ### Return whether the value is at the left, within or
         ###    at the right of both limits
         if {$value < $lower}  {
            set answer -1
         } elseif {$value > $upper} {
            set answer 1
         } else  {
            set answer 0
         }

      ### Lower limit given
      } elseif {$lowerNotEmpty} {

         ### Check whether the range is a single number
         if {[::Simple::Type::Priv::is-basic number $range]} {

            ### Return whether the value is lower, equal or
            ###    greater than the single number
            if {$value < $range}  {
               set answer -1
            } elseif {$value > $range} {
               set answer 1
            } else  {
               set answer 0
            }

         } else {

            ### Return whether the value is lower that the lower limit
            if {$value < $lower} {
               set answer -1
            } else  {
               set answer 0
            }
         }

      ### Rigth limit given
      } else {

         ### Return whether the value is greater that the upper limit
         if {$value > $upper} {
            set answer 1
         } else  {
            set answer 0
         }
      }

      ### Return whether the value is outside the range
      return $answer

   ### The range has more than two limits or a limit is not a number
   } else {

      ### Throw error
      ::Simple::Error::throw ::Simple::BAD-ITEM {numeric range} $range
   }
}

### ===========================================================================
### -command  : ::Simple::Type::assert-is
# -synopsis   : Throws an error if a value does not conform to a type.
# -overview   :
#
#     This procedure throws an error if a value does not conform to a type. 
#  The message uses the A<description>, if given; otherwise, the type
#  description is used, if it was stored when the type was created; otherwise,
#  the type name is used.
#
# -arguments  :
#  { type             type           {Type name}}
#  { value            any            {Value or list of values}}
#  {?extra?           string {}      {Extra data required to check the
#                                     conformation}}
#  {?description?     string {}      {Type description}}
#
# -returns    : Error if the value does not conform to the type.
# -remarks    :
#
#  * The value ''%DEFAULT%'' conforms to all types
#
proc ::Simple::Type::assert-is {
   type
   value
   {extra {}}
   {description {}}
} {
   ### The value does not conform to the type
   if {![::Simple::Type::Priv::is-with-hidden $type $value 0 $extra 2]} {

      ### Description not given
      if {![string compare $description {}]} {

         ### Type description stored
         if {[string compare $::Simple::Type::Description($type) {}]} {

            ### Use it
            set description\
               [string tolower $::Simple::Type::Description($type) 0 0]

         ### Type description not stored
         } else {

            ### Use the type name
            set description $type
         }
      }

      ### Throw error
      ::Simple::Error::throw ::Simple::BAD-ITEM $description $value
   }
}

### ===========================================================================
### -command  : ::Simple::Type::configure
# -synopsis   : Configures the package options.
# -returns    : The "package with no options" error is thrown.
#
proc ::Simple::Type::configure {
} {
   ::Simple::Error::throw ::Simple::NO-OPTIONS
}

### ===========================================================================
### -command  : ::Simple::Type::cget
# -synopsis   : Gets the package options.
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Type::cget {
} {
   return [list]
}

### ===========================================================================

# Declare the package types
::Simple::Type::Priv::declare-types

}
