### ===========================================================================
### -package  : SimpleVariable-extra
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library typed variables handling.
# -module     : extra
# -api        :
#
#  * K<::Simple::Variable::delete> variable
#    Deletes a typed variable.
#
#  * K<::Simple::Variable::information exists> variable
#    Returns whether a typed variable exists.
#
#  * K<::Simple::Variable::information defaultgiven> variable
#    Returns whether a typed variable has a default value.
#
#  * K<::Simple::Variable::information default> variable
#    Returns a typed variable default value.
#
#  * K<::Simple::Variable::information type> variable
#    Returns a typed variable type.
#
#  * K<::Simple::Variable::information choices> variable
#    Returns a typed variable choices.
#
#  * K<::Simple::Variable::information description> variable
#    Returns a typed variable description.
#
#  * K<::Simple::Variable::information access> variable
#    Returns a typed variable access mode.
#
#  * K<::Simple::Variable::information monitortype> variable
#    Returns whether a typed variable is being monitored.
#
# -history    :
#
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleVariable-extra 1.0 -firsttimeinstall {

### ===========================================================================
### -command  : ::Simple::Variable::delete
# -synopsis   : Deletes a typed variable.
# -overview   :
#
#     This procedure deletes a typed variable declared via the
#  K<::Simple::Variable::declare> procedure.
#
# -arguments  :
#  { variable         variablename   {Variable name}}
# -returns    : The empty string.
#
proc ::Simple::Variable::delete {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### The typed variable internal representation exists
   if {[info exists ${namespace}%Variables%($unqualifiedName)]}  {

      ### Delete the typed variable
      # The variable has been initialized
      if {[info exists $qualifiedName]} {

         # Delete the variable
         unset $qualifiedName
      }

      # Delete its write trace, if any
      if {[lsearch -exact [trace vinfo $variable]\
         {w ::Simple::Variable::Priv::trace-monitor}] != -1} {

         trace vdelete $qualifiedName w ::Simple::Variable::Priv::trace-monitor
      }

      # Delete its internal representation
      unset ${namespace}%Variables%($unqualifiedName)

   ### The typed variable internal representation does not exit
   } else  {

      ### Throw error
      ::Simple::Error::throw ::Simple::Variable::CANT-DELETE $variable
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::throw-bad-value
# -synopsis   : Throws a "bad value" error
# -arguments  :
#  { kind             word           {Kind of variable (variable, argument...)}}
#  { variable         variablename   {Variable name}}
#  { type             type           {Variable type}}
#  { value            any            {Variable value}}
#  { choices          list           {Variable choices}}
#
# -returns    : The error is thrown.
#
proc ::Simple::Variable::Priv::throw-bad-value {
   kind
   variable
   type
   value
   choices
} {
   ### Choices given
   if {[llength $choices] != 0} {
      set tmp {: }
      if {[string compare $type choice]} {
         append tmp {each element }
      }
      append tmp "must be [::Simple::List::enumerate $choices]"
      set choices $tmp
   } else {
      set choices [list]
   }

   ### Throw error
   ::Simple::Error::throw ::Simple::BAD-VALUE\
      $value $kind $variable $type $choices
}

### ===========================================================================
### -command  : ::Simple::Variable::information exists
# -synopsis   : Returns whether a typed variable exists.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : Whether the typed variable exists.
#
proc-sub {::Simple::Variable::information exists} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Return the typed variable exists
   return [info exists ${namespace}%Variables%($unqualifiedName)]
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::assert-variable-exists
# -synopsis   : Throws an error if a typed variable does not exist.
# -arguments  :
#  { name             name           {Variable name}}
#  { namespace        namespace      {Variable namespace}}
#  { unqualifiedName  unqualifiedvariablename
#                                    {Variable unqualified name}}
#
# -returns    : Error if the typed variable does not exist.
#
proc ::Simple::Variable::Priv::assert-variable-exists {
   name
   namespace
   unqualifiedName
} {
   ### Assert the variable exists
   if {![info exists ${namespace}%Variables%($unqualifiedName)]} {
      ::Simple::Error::throw ::Simple::NON-EXISTING {typed variable} $name
   }
}

### ===========================================================================
### -command  : ::Simple::Variable::information defaultgiven
# -synopsis   : Returns whether a typed variable has a default value.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : Whether the typed variable has a default value.
# -references :
#
#  * The K<::Simple::Option::information defaultgiven> procedure.
#
#  * The K<::Simple::ItkOption::information defaultgiven> procedure.
#
proc-sub {::Simple::Variable::information defaultgiven} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Return whether the typed variable has a default value
   return [lindex [set ${namespace}%Variables%($unqualifiedName)] 4]
}

### ===========================================================================
### -command  : ::Simple::Variable::information default
# -synopsis   : Returns a typed variable default value.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : The typed variable default value.
# -references :
#
#  * The K<::Simple::Option::information default> procedure.
#
#  * The K<::Simple::ItkOption::information default> procedure.
#
proc-sub {::Simple::Variable::information default} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Assert the variable has a default value
   if {![lindex [set ${namespace}%Variables%($unqualifiedName)] 4]} {
      ::Simple::Error::throw ::Simple::Variable::NO-DEFAULT-VALUE\
         variable $variable
   }

   ### Return the variable value
   return [lindex [set ${namespace}%Variables%($unqualifiedName)] 5]
}

### ===========================================================================
### -command  : ::Simple::Variable::information type
# -synopsis   : Returns a typed variable type.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : The typed variable type.
#
proc-sub {::Simple::Variable::information type} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Return the variable type
   return [lindex [set ${namespace}%Variables%($unqualifiedName)] 0]
}

### ===========================================================================
### -command  : ::Simple::Variable::information choices
# -synopsis   : Returns a typed variable choices.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : The typed variable choices.
#
proc-sub {::Simple::Variable::information choices} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Assert the variable type is "choice" or derived from it
   set type [lindex [set ${namespace}%Variables%($unqualifiedName)] 0]
   if {![string match choice* $type]} {
      if {[string match *-array $type]} {
         ::Simple::Error::throw\
            ::Simple::Variable::NO-TYPE-CHOICE-ARRAY $variable
      } else {
         ::Simple::Error::throw\
            ::Simple::Variable::NO-TYPE-CHOICE-VAR $variable
      }
   }

   ### Return the variable list of choices
   return [lindex [set ${namespace}%Variables%($unqualifiedName)] 1]
}

### ===========================================================================
### -command  : ::Simple::Variable::information description
# -synopsis   : Returns a typed variable description.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : The typed variable description.
#
proc-sub {::Simple::Variable::information description} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Return the variable description
   return [lindex [set ${namespace}%Variables%($unqualifiedName)] 2]
}

### ===========================================================================
### -command  : ::Simple::Variable::information access
# -synopsis   : Returns a typed variable access mode.
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : The typed variable access mode.
#
proc-sub {::Simple::Variable::information access} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Return the variable access mode
   return [lindex [set ${namespace}%Variables%($unqualifiedName)] 3]
}

### ===========================================================================
### -command  : ::Simple::Variable::information monitortype
# -synopsis   : Returns whether a typed variable is being monitored.
# -overview   :
#
#     This procedure returns whether a typed variable is being monitored to
#  ensure its values conform to its type.
#
# -arguments  :
#  { variable         variablename   {Variable name}}
#
# -returns    : Whether the typed variable is being monitored.
# -references :
#
#  * The K<::Simple::ItkOption::information monitortype> procedure.
#
proc-sub {::Simple::Variable::information monitortype} {
   variable
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $variable] break

   ### Assert the variable exists
   ::Simple::Variable::Priv::assert-variable-exists\
      $variable $namespace $unqualifiedName

   ### Return whether the variable is being monitored
   #   to ensure its values conform to its type
   #   that is, whether there is a write trace on the variable
   #   directed to the monitoring procedure
   return [expr {[lsearch -exact [trace vinfo $qualifiedName]\
      {w ::Simple::Variable::Priv::trace-monitor}] != -1}]
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::trace-monitor
# -synopsis   : Monitors typed variables.
# -overview   :
#
#     This procedure is called whenever a typed variable declared through
#  K<::Simple::Variable::declare> and being monitored is set (written) to
#  ensure that its new value conforms to the variable type, an error being
#  thrown otherwise.
#
# -arguments  :
#  { name             variablename   {Variable name}}
#  { index            unqualifiedvariablename
#                                    {Array index name (empty if not an array)}}
#  { operation        string         {Operation performed on the variable}}
#
# -details    :
#
#     This procedure gets called through the K<trace> command set by the
#  K<::Simple::Variable::declare> procedure.
#
#     If there is a trace on an array (because the array was created via
#  K<::Simple::Variable::declare> with a derived array type) and another trace
#  on an element of that array (because the element was created via
#  K<::Simple::Variable::declare> as well), this procedures gets called
#  A<twice> with exactly the same arguments: each time both types are checked.
#
proc ::Simple::Variable::Priv::trace-monitor {
   name
   index
   operation
} {
   ### Qualify and split the variable name
   foreach {qualifiedName namespace unqualifiedName}\
      [::Simple::Priv::qualify-split $name] break

   ### The variable is accessible to K<info exists> through its qualified name
   if {[info exists $qualifiedName]} {

      ### It is a variable
      set kind variable

      ### The namespace or classes list is the variable namespace
      set namespaceOrClassesList $namespace

   ### The variable is not accessible to K<info exists> through its qualified name
   } else {

      ### Get the class inheritance list
      if {[catch {set namespaceOrClassesList [namespace eval $namespace\
         {[uplevel 1 set namespace]::info heritage}]}]} {

         # An error has occured, so that the variable is not accessible to
         # K<info exists> through its qualified name but it is not an
         # attribute either.  This happens when a typed variable has been
         # aliased through the use of the K<upvar> or K<global> commands.  Due
         # to a Tcl limitation it is then impossible to access the name of the
         # original typed variable, and so its type, so the monitorng
         # procedure can not proceed
         ::Simple::Error::throw ::Simple::Variable::MONITORING-ALIAS
      }

      ### It is a class attribute
      set kind attribute
   }

   ### Locate the variable internal representation
   set internalRepLocation\
      [::Simple::Variable::Priv::locate-internal-representation\
      $namespaceOrClassesList %Variables%(${unqualifiedName})]

   ### The variable is an array
   if {[string compare $index {}]} {

      ### Get the array element value
      upvar ${name}($index) value

      ### The whole array is typed
      if {[string compare $internalRepLocation {}]} {

         ### Remove the trailing "-array" from the type name
         foreach {type choices} [set $internalRepLocation] break
         regexp {^(.*)-array$} $type -> type

         ### Assert the element value conforms to the array type
         #   The choices only apply if the type is choice or derived from it
         ::Simple::Variable::Priv::assert-type-conformation\
            $kind 1 $name $type $value 0 $choices
      }

      ### Locate the internal representation for the array element
      set internalRepLocation\
         [::Simple::Variable::Priv::locate-internal-representation\
         $namespaceOrClassesList %Variables%(${unqualifiedName}($index))]

      ### The array element is specifically typed
      if {[string compare $internalRepLocation {}]} {

         ### Assert the element value conforms to the proceed element type
         #   The choices only apply if the type is choice or derived from it
         foreach {type choices} [set $internalRepLocation] break
         ::Simple::Variable::Priv::assert-type-conformation\
            $kind 0 ${name}($index) $type $value 0 $choices
      }

   ### The variable is not an array
   } else {

      ### The typed variable exists
      if {[string compare $internalRepLocation {}]} {

         ### Get the variable value
         upvar $name value

         ### Assert the element value conforms to the element type
         #   The choices only apply if the type is choice or derived from it
         foreach {type choices} [set $internalRepLocation] break
         ::Simple::Variable::Priv::assert-type-conformation\
            $kind 0 $name $type $value 0 $choices
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::locate-internal-representation
# -synopsis   : Locates a variable internal representation.
# -overview   :
#
#     This procedure locates a variable internal representation.  It is
#  necessary because the internal representation for attributes of derived
#  classes is stored in the base class.
#
# -arguments  :
#  { namespaceOrClassesList
#                     namespace-list {Namespace or classes list}}
#  { unqualifiedInternalRepLocation
#                     unqualifiedvariablename
#                                    {Array index name (empty if not an array)}}
#
# -returns    : 
#
#     The qualified internal representation or the empty string if it could
#  not be found.
#
proc ::Simple::Variable::Priv::locate-internal-representation {
   namespaceOrClassesList
   unqualifiedInternalRepLocation
} {
   ### Loop over namespaces or classes
   set answer {}
   foreach namespaceOrClass $namespaceOrClassesList {

      ### Internal representation found
      if {[info exists ${namespaceOrClass}::$unqualifiedInternalRepLocation]} {

         ### Short-circuit
         set answer ${namespaceOrClass}::$unqualifiedInternalRepLocation
         break
      }
   }

   ### Return the qualified internal representation location
   return $answer
}
}
