### ===========================================================================
### -package  : SimpleVariable
# -version    : 1.0
# -project    : The Simple Development Library.
# -synopsis   : The Simple Development Library typed variables handling.
# -module     : implementation
# -overview   :
#
#     This package allows to handle typed variables.  Typed variables not only
#  have a name and a value if initialized, as regular Tcl variables do, but
#  also a type and a description.  Typed variables can be used as non-local
#  (namespace or global) Tcl variables.  Typed variables can be monitored to
#  ensure their values conform to their type.
#
#     Procedures are provided to declare, delete and get information about
#  typed variables.
#
# -usage      :
#
#     Typed variables are declared via K<::Simple::Variable::declare> or
#  embedded within programs (K<declare-program>), packages
#  (K<declare-package>) or classes (K<declare-class>).  Upon declaration,
#  typed variables can be optionally initialized to a value.  The declaration
#  can also specify whether a variable shall be monitored to ensure its values
#  conform to its type.  Typed variables can be deleted via
#  K<::Simple::Variable::delete>.  The K<::Simple::Variable::information>
#  procedure provides several subcommands to query information about typed
#  variables:  K<exists>, K<defaultgiven>, K<default>, K<type>, K<choices>,
#  K<description>, K<access> and K<monitortype>.
#
# -keywords   : variable type
# -option     :
#
#  {
#     -monitortype
#
#  } -description {
#
#        Whether to monitor the typed variables to ensure their values conform
#     to their types.  Notice that only those variables declared via the
#     K<::Simple::Variable::declare> procedure A<after> modifying this option
#     are affected by the change.
#
#  } -type boolean -value 0
#
# -api        :
#
#  * K<::Simple::Variable::declare> variable ?-description string?
#       ?-monitortype extbool? ?-type type? ?-choices any-list? ?-value any?
#    Declares a typed variable.
#
#  * K<::Simple::Variable::configure> ?-monitortype extbool?
#    Configures the package options.
#
#  * K<::Simple::Variable::cget> ?-monitortype?
#    Gets the package options.
#
# -examples   :
#
#  # Install the package
#  package require SimplePackage
#  ::Simple::Package::require-and-install SimpleVariable
#
#  # Configure The Simple Development Library to store the metadata
#  ::Simple::configure -storemetadata true
#
#  # Declare a typed variable and monitor its value
#  ::Simple::Variable::declare foobar\ 
#     -description {may contain foo or bar only}\ 
#     -monitortype true -type choice -choices {foo bar}
#
#  # Get some information about the just declared variable
#  # This displays the following:
#  #   |Type: "choice"
#  #   |Description: "may contain foo or bar only"
#  puts "Type: \"[::Simple::Variable::information type foobar]\""
#  puts "Description: \"[::Simple::Variable::information description foobar]\""
#
#  # Set the variable to a valid value
#  # This displays the following:
#  #   |bar
#  set foobar bar
#  puts $foobar
#
#  # Set the variable to an invalid value
#  # This throws an error:
#  #    can't set "foobar": invalid value "gee" for variable
#  #    "::foobar" of type "choice": must be foo or bar
#  set foobar gee
#
# -details    :
#
#  * Each typed variable internal representation is stored in one
#    element of an array named V<%Variables%> in the variable namespace.  The
#    name of the array element is the unqualified variable name.  The array
#    element is a list with the following elements:
#
##table;delimiter=":";ignore="true"
#     0: variable type
#     1: variable choices (empty if type is not "choice" or derived from it)
#     2: variable description
#     3: variable access mode
#     4: whether the variable has a default (initial) value
#     5: variable default (initial) value
##end
#
# -remarks    :
#
#  * It is possible to declare typed arrays using derived array types as in
#    K<::Simple::Variable::declare> ''foo -type number-array''.  It is even
#    possible to have an specific type for array elements by declaring them as
#    in K<::Simple::Variable::declare> ''foo(bar) -type float''.  If both
#    things are done together, all the elements in the array ''foo'' must be
#    numbers, but the values of ''foo(bar)'' must be float numbers.
#
# -todo       :
#
#  * It might be useful to be able to handle variables with more than one type
#    so that its values must conform to one of those.
#
# -history    :
#
#  23-apr-2000   First public release, version 0.2
#  10-sep-2001   Second public release, version 0.4
#  17-feb-2003   Third public release, version 0.5
#  30-mar-2003   Extra package, version 0.5.1
#  09-may-2004   Package renamed to P<SimpleVariable>, version 0.5.2
#  22-jun-2005   The Simple Development Library version 1.0
#
# -copyright  :
#
#  Copyright (C) 1999-2005, Juan C. Gil (jgil@gmv.es)
#
### ===========================================================================

### Declare the package
package require SimplePackage
::Simple::Package::declare SimpleVariable 1.0 -requisites {

   {SimpleType 1.0}

} -namespaces {

   ::Simple::Variable
   ::Simple::Variable::Priv

} -install {

### Create the package variables

### Create the package options

# A<-monitortype>
set ::Simple::Variable::%Options%(-monitortype) 0

} -uninstall {

### Delete the package variables

### Delete the package options
unset ::Simple::Variable::%Options%(-monitortype)

} -firsttimeinstall {

### Declare the package types

### Declare the package errors
::Simple::Error::declare {
   ::Simple::Variable::IS-ARRAY
} -message {
   variable "%s" is array
} -explanation {
   Non-array variable <name> could not be declared because it already exists
   and is an array
} -corrective {
   Delete the existing array.
}

::Simple::Error::declare {
   ::Simple::Variable::NOT-AN-ARRAY
} -message {
   variable "%s" isn't array
} -explanation {
   Array <name> could not be declared because it already exists
   and is not an array
} -corrective {
   Delete the existing non-array variable.
}

::Simple::Error::declare {
   ::Simple::Variable::CANT-DELETE 
} -message {
   can't delete "%s": no such variable
} -explanation {
   Variable <name> could not be deleted because it does not exist
}

::Simple::Error::declare {
   ::Simple::Variable::BAD-TYPE-BOOLFLAG 
} -message {
   invalid type "boolflag" for non-flag "%s"
} -explanation {
   Non-flag <name> type is "boolflag", but this type is only valid for flags
} -corrective {
   Either convert the variable or argument to a flag or modify its type
}

::Simple::Error::declare {
   ::Simple::Variable::AT-LEAST-TWO-CHOICES 
} -message {
   %s "%s" requires at least two choices
} -explanation {
   <variable, array element or argument> <name> is of type "choice" or
   derived from it, but less than two choices were supplied
} -corrective {
   Supply at least two choices
}

::Simple::Error::declare {
   ::Simple::Variable::NO-CHOICES-ALLOWED 
} -message {
   no choices allowed for %s "%s" of type "%s"
} -explanation {
   <variable, array element or argument> <name> is of type <type>, which is
   not "choice" nor derived from it, and so it does not allow the
   specification of choices
}

::Simple::Error::declare {
   ::Simple::Variable::NO-TYPE-CHOICE-VAR 
} -message {
   variable "%s" is not of type "choice" nor "choice-list"
} -explanation {
   The typed variable <name> type is not "choice" nor "choice-list"
}

::Simple::Error::declare {
   ::Simple::Variable::NO-DEFAULT-VALUE
} -message {
   %s "%s" has no default value
} -explanation {
   The typed <variable, [incr Tk] option, argument or option> <name> has no
   default value as it was not given in its declaration
}

::Simple::Error::declare {
   ::Simple::Variable::NO-TYPE-CHOICE-ARRAY 
} -message {
   array "%s" is not of type "choice-array" nor "choice-list-array"
} -explanation {
   The typed array <name> type is not "choice-array" nor "choice-list-array"
}

::Simple::Error::declare {
   ::Simple::Variable::MONITORING-ALIAS 
} -message {
   type monitoring for aliased variable
} -explanation {
   The procedure which monitors the typed variables to ensure their values
   conform to their types was called for a posibly correct typed variable
   but aliased through the use of the K<upvar> or K<global> commands.  Due to
   a Tcl limitation it is then impossible to access the name of the original
   typed variable, and so its type, so the monitorng procedure can not
   proceed
} -corrective {
   Either deactivate the variable type monitoring or use no alias
}

### ===========================================================================
### -command  : ::Simple::Variable::declare
# -synopsis   : Declares a typed variable.
# -overview   :
#
#     This procedure declares a non-local (namespace or global) typed variable
#  or array and optionally assign a value to it.
#
#     If A<-type> is given, the variable type is set to its parameter,
#  otherwise the variable type is set to A<any>. 
#
#     When the variable type is A<choice> or derived from it, A<-choices> is
#  required.
#
#     Use A<-value> to initialize the typed variable to a value. This actually
#  creates the variable.
#
#     The A<-monitortype> flag allows to override the package option of the
#  same name for the varible being declared.  This option controls whether the
#  variable will be monitored to ensure its values conform to its type.  An
#  error is thrown if a variable being monitored is set to a value which does
#  not conform to its type, but the variable still contains the non-conforming
#  value.
#
# -arguments  :
#  { variable         variablename   {Variable name}}
#  {-description      string {}      {Variable description}}
#  {-monitortype      extbool false  {Whether to enable run-time type
#                                     monitoring}}
#  {-type             type any       {Variable type}}
#  {-choices          any-list {}    {List of choices for choice type
#                                     and its derivations}}
#  {-value            any %DEFAULT%  {Variable value}}
#
# -returns    : The empty string.
# -details    :
#
#     The type monitoring is achieved by setting a write trace on the variable
#  via the K<trace> procedure with the
#  K<::Simple::Variable::Priv::trace-monitor> as handler procedure.  This
#  write trace is only set for monitored variables.
#
# -remarks    :
#
#  * The order of the flags is irrelevant but their parameters can not start
#    by an hyphen.
#
proc ::Simple::Variable::declare {
   variable
   args
} {
   ### Parse and verify the variable declaration
   set variable [string trim $variable]
   ::Simple::Error::Priv::push-location variable $variable 1
   if {[catch {
      foreach {declarationScript creationScript}\
         [::Simple::Variable::Priv::parse-and-verify-declaration\
         $variable $args 0] break
   }]} {

      # Rethrow the error after rationalizing the stack call
      eval [::Simple::Error::Priv::rationalize-call-stack]
   } 

   ### Declare and create the variable
   eval $declarationScript\n$creationScript

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::parse-and-verify-declaration
# -synopsis   : Parses and verifies a typed variable declaration.
# -arguments  :
#  { variable         variablename   {Variable name}}
#  { actualArgs       list           {Actual variable declaration argument
#                                     list}}
#  { hidden           boolean        {Whether to also consider hidden types}}
#
# -returns    :
#
#  A list with the following elements:
#
##table;delimiter=":";ignore="true"
#   0: the variable declaration script.
#   1: the variable creation script.
##end
#
proc ::Simple::Variable::Priv::parse-and-verify-declaration {
   variable
   actualArgs
   hidden
} {
   ### Set defaults
   set monitortype [set ::Simple::Variable::%Options%(-monitortype)]
   set type any

   ### Parse flags
   set flags [::Simple::Priv::parse-flags\
      [list -description paragraphs\
      -monitortype untouched\
      -type trimmed\
      -choices list\
      -value untouched]\
      [list] $actualArgs 0 2]

   ### First section factored out
   foreach {variableQualifiedName variableNamespace variableUnqualifiedName}\
      [::Simple::Priv::qualify-split $variable 3] break
   set kind variable
   set item $variable
   ::Simple::Variable::Priv::uplevel-declare-1

   ### Assert the variable, if existing, is an
   ###    array or not as the one being declared

   # The variable exists
   if {[info exists $variableQualifiedName]} {

      # Derived array type
      set isArray [array exists $variableQualifiedName]
      if {$isArrayType} {
         if {!$isArray} {
            ::Simple::Error::throw ::Simple::Variable::NOT-AN-ARRAY $item
         }

      # Non-derived array
      } elseif {$isArray} {
         ::Simple::Error::throw ::Simple::Variable::IS-ARRAY $item
      }
   }

   ### Compose the variable declaration script
   #   The variable access mode is public
   set declarationScript [list ::Simple::Variable::Priv::do-declare variable\
      $variableNamespace $variableUnqualifiedName $description\
      [list $type $choices public $valueGiven $value]]

   ### Compose the variable creation script
   set creationScript [list ::Simple::Variable::Priv::do-create\
      $variableQualifiedName $monitortype $valueGiven $type $value]

   ### Return the variable declaration and creation scripts
   return [list $declarationScript $creationScript]
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::do-declare
# -synopsis   : Declares a typed variable or similar.
# -overview   :
#
#     This procedure is used to declare typed variables but also other similar
#  items such as [incr Tcl] attributes or [incr Tk] options (as given by
#  K<SimpleClass>); even [incr Tk] components are handled by this procedure,
#  although they are not similar to variables.
#
# -arguments  :
#  { kind             choice {variable attribute itkoption itkcomponent}
#                                    {Kind of item}}
#  { itemNamespace    namespace      {Item namespace}}
#  { itemUnqualifiedName     
#                     unqualifiedname
#                                    {Item unqualified name}}
#  { description      string         {Item description}}
#  { internalRepWithoutDescription
#                     any-list       {Internal item representation without the
#                                     description}}
#
# -returns    : The empty string.
#
proc ::Simple::Variable::Priv::do-declare {
   kind
   itemNamespace
   itemUnqualifiedName
   description
   internalRepWithoutDescription
} {
   ### Compose the item internal representation

   # Switch on kind
   switch -exact -- $kind {

      variable -
      attribute {
      # variable or attribute

         set internalRepArray %Variables%
         set internalRepIndex $itemUnqualifiedName
         set descriptionIndex 2
      }

      itkoption {
      # [incr Tk] option 

         set internalRepArray %ItkOptions%
         set internalRepIndex -$itemUnqualifiedName
         set descriptionIndex 2
      }

      itkcomponent {
      # [incr Tk] component 

         set internalRepArray %ItkComponents%
         set internalRepIndex $itemUnqualifiedName
         set descriptionIndex 1
      }
   }

   # The item descriptions must not be stored
   if {![set ::Simple::%Options%(-storemetadata)]} {

      # Do not store the item description
      set description {}
   }

   ### Store the item internal representation
   set ${itemNamespace}${internalRepArray}($internalRepIndex)\
      [linsert $internalRepWithoutDescription $descriptionIndex $description]

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::do-create
# -synopsis   : Creates a typed variable or similar.
# -arguments  :
#  { variableQualifiedName     
#                     qualifiedname
#                                    {Variable qualified name}}
#  { monitorType      extbool        {Whether to monitor the variable being
#                                     declared to ensure its values conform to
#                                     its type}}
#  { valueGiven       boolean        {Whether A<-value> was given in the
#                                     variable declaration}}
#  { type             type           {Variable type}}
#  { value            any            {Variable value}}
#
# -effects    :
#
#  * Creates the typed variable if A<valueGiven> is true.
#  
#  * Sets a write trace on the variable to monitor its values to ensure they
#    conform to its type if A<monitorType> is true.
#
# -details    :
#
#  * The actual creation code is evaluated one level up the stack even though
#    the variable is qualified. This is so in order for [incr Tcl]
#    attributes to be created in the proper context within the class
#    constructor.
#
# -returns    : The empty string.
#
proc ::Simple::Variable::Priv::do-create {
   variableQualifiedName
   monitorType
   valueGiven
   type
   value
} {
   ### A<-value> given
   if {$valueGiven} {

      ### Derived array type
      if {[string match *-array $type]} {

         ### Set the array
         uplevel 1 [list array set $variableQualifiedName $value]

      ### No derived array type
      } else {

         ### Set the variable value
         uplevel 1 [list set $variableQualifiedName $value]
      }
   }

   ### Monitor type
   if {$monitorType} {

      ### Set a write trace on the variable to monitor
      ###    its values to ensure they conform to its type
      uplevel 1 [list trace variable $variableQualifiedName w\
         ::Simple::Variable::Priv::trace-monitor]
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::uplevel-declare-1
# -synopsis   : Declares a typed variable (first section factored out).
# -remarks    :
#
#  * This is a full K<uplevel> procedure used to factor code out.
#
# -effects    :
#
#  * Modifies V<monitortype>.
#
proc ::Simple::Variable::Priv::uplevel-declare-1 {
} {
   ### Up one level
   uplevel 1 {

      # Ensure A<-monitortype> is boolean
      if {[::Simple::Priv::flag-given -monitortype]} {
         set monitortype [{::Simple::Type::convert boolean} $monitortype]
      }

      ### Assert the type exists
      if {![::Simple::Type::Priv::exists-with-hidden $type $hidden]} {
         ::Simple::Error::throw ::Simple::NON-EXISTING type $type
      }

      ### Assert the type is not "boolflag"
      if {![string compare $type boolflag]} {
         ::Simple::Error::throw ::Simple::Variable::BAD-TYPE-BOOLFLAG $item
      }

      ### Handle A<-choices>
      if {[string match *-array $type]} {
         set isArrayType 1
         set container {array }
      } else {
         set isArrayType 0
         set container {}
      }
      set choicesGiven [::Simple::Priv::flag-given -choices]
      if {[string match choice* $type]} {
         if {!$choicesGiven || [llength $choices] < 2} {
            ::Simple::Error::throw ::Simple::Variable::AT-LEAST-TWO-CHOICES\
               "${container}$kind" $item
         }
      } else {
         if {$choicesGiven} {
            ::Simple::Error::throw ::Simple::Variable::NO-CHOICES-ALLOWED\
               "${container}$kind" $item $type
         } else {
            set choices [list]
         }
      }

      ### A<-value> given
      set valueGiven [::Simple::Priv::flag-given -value]
      if {$valueGiven} {

         ### Derived array type
         if {$isArrayType} {

            ### Assert the value is valid
            #   Assert the value is accepted by K<array set>
            if {[catch {array set localArray $value} result]} {
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $value {array variable} $item $type ": $result"
            }

            # Assert the actual array values are valid
            if {![::Simple::Type::Priv::is-with-hidden\
               $type localArray $hidden $choices]} {

               ::Simple::Variable::Priv::throw-bad-value\
                  {array variable} $item $type $value $choices
            }

         ### No derived array type
         } else {

            ### Assert the value is valid
            #   The variable choices only apply if the
            #   type is choice or derived from it
            ::Simple::Variable::Priv::assert-type-conformation\
               $kind 0 $item $type $value $hidden $choices
         }
      }
   }
}

### ===========================================================================
### -command  : ::Simple::Variable::Priv::assert-type-conformation
# -synopsis   : Throws an error if an item value does not conform to a type.
# -arguments  :
#  { kind             word           {Kind of item (variable, argument...)}}
#  { array            boolean        {Whether the item is an array}}
#  { item             variablename   {Item name}}
#  { type             type           {Item type}}
#  { value            any            {Item value}}
#  { hidden           boolean        {Whether to also consider hidden types}}
#  {?choices?         list {}        {Item choices}}
#
# -returns    : Error if the item value does not conform to the type.
#
proc ::Simple::Variable::Priv::assert-type-conformation {
   kind
   array
   item
   type
   value
   hidden
   {choices {}}
} {
   ### The value does not match the type
   if {![::Simple::Type::Priv::is-with-hidden $type $value $hidden $choices]} {

      ### Array
      if {$array} {

         ### Add "array" suffix to the kind
         set kind "array $kind"

         ### Add the "-array" suffix to the type name
         set type $type-array
      }

      ### Throw bad value error
      ::Simple::Variable::Priv::throw-bad-value\
         $kind $item $type $value $choices
   }
}

### ===========================================================================
### -command  : ::Simple::Variable::configure
# -synopsis   : Configures the package options.
#
# -arguments  :
#  {-monitortype      extbool false  {Whether to monitor the typed variables
#                                     to ensure their values conform to their
#                                     types}}
#
# -returns    : The empty string.
#
proc ::Simple::Variable::configure {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Throw error
      ::Simple::Error::throw ::Simple::AT-LEAST-ONE-FLAG\
         [::Simple::Variable::cget]

   ### Arguments
   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      set arg2 [lindex $args 1]
      switch -exact -- $arg1 {

         -monitortype {
         ### -monitortype given

            ### Invalid argument
            if {![::Simple::Type::Priv::is-basic extbool $arg2]} {

               ### Throw error
               ::Simple::Error::throw ::Simple::BAD-VALUE\
                  $arg2 argument -monitortype extbool {}
            }

            ### Set the monitor type option
            set ::Simple::Variable::%Options%(-monitortype)\
               [{::Simple::Type::convert boolean} $arg2]
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG\
               $arg1 [::Simple::Variable::cget]
         }
      }
   }

   ### Return the empty string
   return {}
}

### ===========================================================================
### -command  : ::Simple::Variable::cget
# -synopsis   : Gets the package options.
# -arguments  :
#  {-monitortype      boolflag       {Gets whether to monitor the typed
#                                     variables to ensure their values conform
#                                     to their types}}
#
# -returns    : 
#
#  The requested option value or the whole list of options if none specified.
#
proc ::Simple::Variable::cget {
   args
} {
   ### No arguments
   if {[llength $args] == 0} {

      ### Return the list of options
      return -monitortype

   ### Arguments
   } else {

      ### Switch on flag
      set arg1 [lindex $args 0]
      switch -exact -- $arg1 {

         -monitortype {
         ### -monitortype given
            set ::Simple::Variable::%Options%(-monitortype)
         }

         default {
         ### default, throw error
            ::Simple::Error::throw ::Simple::BAD-FLAG\
               $arg1 [::Simple::Variable::cget]
         }
      }
   }
}
}
